@@ -1,6 +1,7 @@
-/* $Id: dns.c 1239 2007-05-01 12:25:01Z bennylp $ */
+/* $Id$ */
 /* 
- * Copyright (C) 2003-2007 Benny Prijono <benny@prijono.org>
+ * Copyright (C) 2008-2011 Teluu Inc. (http://www.teluu.com)
+ * Copyright (C) 2003-2008 Benny Prijono <benny@prijono.org>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -29,6 +30,7 @@ PJ_DEF(const char *) pj_dns_get_type_name(int type)
 {
     switch (type) {
     case PJ_DNS_TYPE_A:	    return "A";
+    case PJ_DNS_TYPE_AAAA:  return "AAAA";
     case PJ_DNS_TYPE_SRV:   return "SRV";
     case PJ_DNS_TYPE_NS:    return "NS";
     case PJ_DNS_TYPE_CNAME: return "CNAME";
@@ -41,6 +43,13 @@ PJ_DEF(const char *) pj_dns_get_type_name(int type)
 }
 
 
+static void write16(pj_uint8_t *p, pj_uint16_t val)
+{
+    p[0] = (pj_uint8_t)(val >> 8);
+    p[1] = (pj_uint8_t)(val & 0xFF);
+}
+
+
 /**
  * Initialize a DNS query transaction.
  */
@@ -50,11 +59,9 @@ PJ_DEF(pj_status_t) pj_dns_make_query( void *packet,
 				       int qtype,
 				       const pj_str_t *name)
 {
-    pj_dns_hdr *hdr;
-    char *query, *p;
+    pj_uint8_t *p = (pj_uint8_t*)packet;
     const char *startlabel, *endlabel, *endname;
-    pj_uint16_t tmp;
-    unsigned d;
+    pj_size_t d;
 
     /* Sanity check */
     PJ_ASSERT_RETURN(packet && size && qtype && name, PJ_EINVAL);
@@ -66,14 +73,14 @@ PJ_DEF(pj_status_t) pj_dns_make_query( void *packet,
     PJ_ASSERT_RETURN(*size >= d, PJLIB_UTIL_EDNSQRYTOOSMALL);
 
     /* Initialize header */
-    hdr = (pj_dns_hdr*) packet;
-    pj_bzero(hdr, sizeof(struct pj_dns_hdr));
-    hdr->id = pj_htons(id);
-    hdr->flags = pj_htons(PJ_DNS_SET_RD(1));
-    hdr->qdcount = pj_htons(1);
+    pj_assert(sizeof(pj_dns_hdr)==12);
+    pj_bzero(p, sizeof(struct pj_dns_hdr));
+    write16(p+0, id);
+    write16(p+2, (pj_uint16_t)PJ_DNS_SET_RD(1));
+    write16(p+4, (pj_uint16_t)1);
 
     /* Initialize query */
-    query = p = (char*)(hdr+1);
+    p = ((pj_uint8_t*)packet)+sizeof(pj_dns_hdr);
 
     /* Tokenize name */
     startlabel = endlabel = name->ptr;
@@ -81,7 +88,7 @@ PJ_DEF(pj_status_t) pj_dns_make_query( void *packet,
     while (endlabel != endname) {
 	while (endlabel != endname && *endlabel != '.')
 	    ++endlabel;
-	*p++ = (char)(endlabel - startlabel);
+	*p++ = (pj_uint8_t)(endlabel - startlabel);
 	pj_memcpy(p, startlabel, endlabel-startlabel);
 	p += (endlabel-startlabel);
 	if (endlabel != endname && *endlabel == '.')
@@ -91,17 +98,15 @@ PJ_DEF(pj_status_t) pj_dns_make_query( void *packet,
     *p++ = '\0';
 
     /* Set type */
-    tmp = pj_htons((pj_uint16_t)(qtype));
-    pj_memcpy(p, &tmp, 2);
+    write16(p, (pj_uint16_t)qtype);
     p += 2;
 
     /* Set class (IN=1) */
-    tmp = pj_htons(1);
-    pj_memcpy(p, &tmp, 2);
+    write16(p, 1);
     p += 2;
 
     /* Done, calculate length */
-    *size = p - (char*)packet;
+    *size = (unsigned)(p - (pj_uint8_t*)packet);
 
     return 0;
 }
@@ -111,7 +116,7 @@ PJ_DEF(pj_status_t) pj_dns_make_query( void *packet,
  * it may contain pointers when name compression is applied) 
  */
 static pj_status_t get_name_len(int rec_counter, const pj_uint8_t *pkt, 
-				const pj_uint8_t *start, const pj_uint8_t *max,
+				const pj_uint8_t *start, const pj_uint8_t *max, 
 				int *parsed_len, int *name_len)
 {
     const pj_uint8_t *p;
@@ -239,7 +244,7 @@ static pj_status_t get_name(int rec_counter, const pj_uint8_t *pkt,
 }
 
 
-/* Skip query records. */
+/* Parse query records. */
 static pj_status_t parse_query(pj_dns_parsed_query *q, pj_pool_t *pool,
 			       const pj_uint8_t *pkt, const pj_uint8_t *start,
 			       const pj_uint8_t *max, int *parsed_len)
@@ -326,12 +331,12 @@ static pj_status_t parse_rr(pj_dns_parsed_rr *rr, pj_pool_t *pool,
 
     /* Get TTL */
     pj_memcpy(&rr->ttl, p, 4);
-    rr->ttl = pj_htonl(rr->ttl);
+    rr->ttl = pj_ntohl(rr->ttl);
     p += 4;
 
     /* Get rdlength */
     pj_memcpy(&rr->rdlength, p, 2);
-    rr->rdlength = pj_htons(rr->rdlength);
+    rr->rdlength = pj_ntohs(rr->rdlength);
     p += 2;
 
     /* Check that length is valid */
@@ -343,6 +348,10 @@ static pj_status_t parse_rr(pj_dns_parsed_rr *rr, pj_pool_t *pool,
 	pj_memcpy(&rr->rdata.a.ip_addr, p, 4);
 	p += 4;
 
+    } else if (rr->type == PJ_DNS_TYPE_AAAA) {
+	pj_memcpy(&rr->rdata.aaaa.ip_addr, p, 16);
+	p += 16;
+
     } else if (rr->type == PJ_DNS_TYPE_CNAME ||
 	       rr->type == PJ_DNS_TYPE_NS ||
 	       rr->type == PJ_DNS_TYPE_PTR) 
@@ -588,6 +597,9 @@ static void copy_rr(pj_pool_t *pool, pj_dns_parsed_rr *dst,
 			 pool, &dst->rdata.srv.target);
     } else if (src->type == PJ_DNS_TYPE_A) {
 	dst->rdata.a.ip_addr.s_addr =  src->rdata.a.ip_addr.s_addr;
+    } else if (src->type == PJ_DNS_TYPE_AAAA) {
+	pj_memcpy(&dst->rdata.aaaa.ip_addr, &src->rdata.aaaa.ip_addr,
+		  sizeof(pj_in6_addr));
     } else if (src->type == PJ_DNS_TYPE_CNAME) {
 	pj_strdup(pool, &dst->rdata.cname.name, &src->rdata.cname.name);
     } else if (src->type == PJ_DNS_TYPE_NS) {
@@ -679,3 +691,54 @@ PJ_DEF(void) pj_dns_packet_dup(pj_pool_t *pool,
     }
 }
 
+
+PJ_DEF(void) pj_dns_init_srv_rr( pj_dns_parsed_rr *rec,
+				 const pj_str_t *res_name,
+				 unsigned dnsclass,
+				 unsigned ttl,
+				 unsigned prio,
+				 unsigned weight,
+				 unsigned port,
+				 const pj_str_t *target)
+{
+    pj_bzero(rec, sizeof(*rec));
+    rec->name = *res_name;
+    rec->type = PJ_DNS_TYPE_SRV;
+    rec->dnsclass = (pj_uint16_t) dnsclass;
+    rec->ttl = ttl;
+    rec->rdata.srv.prio = (pj_uint16_t) prio;
+    rec->rdata.srv.weight = (pj_uint16_t) weight;
+    rec->rdata.srv.port = (pj_uint16_t) port;
+    rec->rdata.srv.target = *target;
+}
+
+
+PJ_DEF(void) pj_dns_init_cname_rr( pj_dns_parsed_rr *rec,
+				   const pj_str_t *res_name,
+				   unsigned dnsclass,
+				   unsigned ttl,
+				   const pj_str_t *name)
+{
+    pj_bzero(rec, sizeof(*rec));
+    rec->name = *res_name;
+    rec->type = PJ_DNS_TYPE_CNAME;
+    rec->dnsclass = (pj_uint16_t) dnsclass;
+    rec->ttl = ttl;
+    rec->rdata.cname.name = *name;
+}
+
+
+PJ_DEF(void) pj_dns_init_a_rr( pj_dns_parsed_rr *rec,
+			       const pj_str_t *res_name,
+			       unsigned dnsclass,
+			       unsigned ttl,
+			       const pj_in_addr *ip_addr)
+{
+    pj_bzero(rec, sizeof(*rec));
+    rec->name = *res_name;
+    rec->type = PJ_DNS_TYPE_A;
+    rec->dnsclass = (pj_uint16_t) dnsclass;
+    rec->ttl = ttl;
+    rec->rdata.a.ip_addr = *ip_addr;
+}
+
