@@ -1,6 +1,7 @@
-/* $Id: resolver.c 1323 2007-05-30 15:20:56Z bennylp $ */
+/* $Id$ */
 /* 
- * Copyright (C) 2003-2007 Benny Prijono <benny@prijono.org>
+ * Copyright (C) 2008-2011 Teluu Inc. (http://www.teluu.com)
+ * Copyright (C) 2003-2008 Benny Prijono <benny@prijono.org>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -27,6 +28,7 @@
 #include <pj/os.h>
 #include <pj/pool.h>
 #include <pj/pool_buf.h>
+#include <pj/rand.h>
 #include <pj/string.h>
 #include <pj/sock.h>
 #include <pj/timer.h>
@@ -62,6 +64,14 @@ enum ns_state
     STATE_BAD,
 };
 
+static const char *state_names[3] =
+{
+    "Probing",
+    "Active",
+    "Bad"
+};
+
+
 /* 
  * Each nameserver entry.
  * A name server is identified by its socket address (IP and port).
@@ -142,11 +152,12 @@ struct cached_res
 {
     PJ_DECL_LIST_MEMBER(struct cached_res);
 
+    pj_pool_t		    *pool;	    /**< Cache's pool.		    */
     struct res_key	     key;	    /**< Resource key.		    */
-    char		     buf[RES_BUF_SZ];/**< Resource buffer.	    */
     pj_hash_entry_buf	     hbuf;	    /**< Hash buffer		    */
     pj_time_val		     expiry_time;   /**< Expiration time.	    */
     pj_dns_parsed_packet    *pkt;	    /**< The response packet.	    */
+    unsigned		     ref_cnt;	    /**< Reference counter.	    */
 };
 
 
@@ -170,7 +181,8 @@ struct pj_dns_resolver
     unsigned char	 udp_rx_pkt[UDPSZ];/**< UDP receive buffer.	    */
     unsigned char	 udp_tx_pkt[UDPSZ];/**< UDP receive buffer.	    */
     pj_ssize_t		 udp_len;	/**< Length of received packet.	    */
-    pj_ioqueue_op_key_t	 udp_op_key;	/**< UDP read operation key.	    */
+    pj_ioqueue_op_key_t	 udp_op_rx_key;	/**< UDP read operation key.	    */
+    pj_ioqueue_op_key_t	 udp_op_tx_key;	/**< UDP write operation key.	    */
     pj_sockaddr_in	 udp_src_addr;	/**< Source address of packet	    */
     int			 udp_addr_len;	/**< Source address length.	    */
 
@@ -187,9 +199,6 @@ struct pj_dns_resolver
     /* Hash table for cached response */
     pj_hash_table_t	*hrescache;	/**< Cached response in hash table  */
 
-    /* Cached response free list */
-    struct cached_res	 res_free_nodes;
-
     /* Pending asynchronous query, hashed by transaction ID. */
     pj_hash_table_t	*hquerybyid;
 
@@ -216,6 +225,75 @@ static pj_status_t select_nameservers(pj_dns_resolver *resolver,
 				      unsigned servers[]);
 
 
+/* Close UDP socket */
+static void close_sock(pj_dns_resolver *resv)
+{
+    /* Close existing socket */
+    if (resv->udp_key != NULL) {
+	pj_ioqueue_unregister(resv->udp_key);
+	resv->udp_key = NULL;
+	resv->udp_sock = PJ_INVALID_SOCKET;
+    } else if (resv->udp_sock != PJ_INVALID_SOCKET) {
+	pj_sock_close(resv->udp_sock);
+	resv->udp_sock = PJ_INVALID_SOCKET;
+    }
+}
+
+
+/* Initialize UDP socket */
+static pj_status_t init_sock(pj_dns_resolver *resv)
+{
+    pj_ioqueue_callback socket_cb;
+    pj_status_t status;
+
+    /* Create the UDP socket */
+    status = pj_sock_socket(pj_AF_INET(), pj_SOCK_DGRAM(), 0, &resv->udp_sock);
+    if (status != PJ_SUCCESS)
+	return status;
+
+    /* Bind to any address/port */
+    status = pj_sock_bind_in(resv->udp_sock, 0, 0);
+    if (status != PJ_SUCCESS)
+	return status;
+
+    /* Register to ioqueue */
+    pj_bzero(&socket_cb, sizeof(socket_cb));
+    socket_cb.on_read_complete = &on_read_complete;
+    status = pj_ioqueue_register_sock(resv->pool, resv->ioqueue,
+				      resv->udp_sock, resv, &socket_cb,
+				      &resv->udp_key);
+    if (status != PJ_SUCCESS)
+	return status;
+
+    pj_ioqueue_op_key_init(&resv->udp_op_rx_key, sizeof(resv->udp_op_rx_key));
+    pj_ioqueue_op_key_init(&resv->udp_op_tx_key, sizeof(resv->udp_op_tx_key));
+
+    /* Start asynchronous read to the UDP socket */
+    resv->udp_len = sizeof(resv->udp_rx_pkt);
+    resv->udp_addr_len = sizeof(resv->udp_src_addr);
+    status = pj_ioqueue_recvfrom(resv->udp_key, &resv->udp_op_rx_key,
+				 resv->udp_rx_pkt, &resv->udp_len,
+				 PJ_IOQUEUE_ALWAYS_ASYNC,
+				 &resv->udp_src_addr, &resv->udp_addr_len);
+    if (status != PJ_EPENDING)
+	return status;
+
+    return PJ_SUCCESS;
+}
+
+
+/* Initialize DNS settings with default values */
+PJ_DEF(void) pj_dns_settings_default(pj_dns_settings *s)
+{
+    pj_bzero(s, sizeof(pj_dns_settings));
+    s->qretr_delay = PJ_DNS_RESOLVER_QUERY_RETRANSMIT_DELAY;
+    s->qretr_count = PJ_DNS_RESOLVER_QUERY_RETRANSMIT_COUNT;
+    s->cache_max_ttl = PJ_DNS_RESOLVER_MAX_TTL;
+    s->good_ns_ttl = PJ_DNS_RESOLVER_GOOD_NS_TTL;
+    s->bad_ns_ttl = PJ_DNS_RESOLVER_BAD_NS_TTL;
+}
+
+
 /*
  * Create the resolver.
  */
@@ -228,7 +306,6 @@ PJ_DEF(pj_status_t) pj_dns_resolver_create( pj_pool_factory *pf,
 {
     pj_pool_t *pool;
     pj_dns_resolver *resv;
-    pj_ioqueue_callback socket_cb;
     pj_status_t status;
 
     /* Sanity check */
@@ -257,10 +334,9 @@ PJ_DEF(pj_status_t) pj_dns_resolver_create( pj_pool_factory *pf,
     resv->timer = timer;
     resv->ioqueue = ioqueue;
     resv->last_id = 1;
+
+    pj_dns_settings_default(&resv->settings);
     resv->settings.options = options;
-    resv->settings.qretr_delay = PJ_DNS_RESOLVER_QUERY_RETRANSMIT_DELAY;
-    resv->settings.qretr_count = PJ_DNS_RESOLVER_QUERY_RETRANSMIT_COUNT;
-    resv->settings.cache_max_ttl = PJ_DNS_RESOLVER_MAX_TTL;
 
     /* Create the timer heap if one is not specified */
     if (resv->timer == NULL) {
@@ -276,46 +352,19 @@ PJ_DEF(pj_status_t) pj_dns_resolver_create( pj_pool_factory *pf,
 	    goto on_error;
     }
 
-    /* Response cache hash table and item list */
+    /* Response cache hash table */
     resv->hrescache = pj_hash_create(pool, RES_HASH_TABLE_SIZE);
-    pj_list_init(&resv->res_free_nodes);
 
     /* Query hash table and free list. */
     resv->hquerybyid = pj_hash_create(pool, Q_HASH_TABLE_SIZE);
     resv->hquerybyres = pj_hash_create(pool, Q_HASH_TABLE_SIZE);
     pj_list_init(&resv->query_free_nodes);
 
-    /* Create the UDP socket */
-    status = pj_sock_socket(PJ_AF_INET, PJ_SOCK_DGRAM, 0, &resv->udp_sock);
+    /* Initialize the UDP socket */
+    status = init_sock(resv);
     if (status != PJ_SUCCESS)
 	goto on_error;
 
-    /* Bind to any address/port */
-    status = pj_sock_bind_in(resv->udp_sock, 0, 0);
-    if (status != PJ_SUCCESS)
-	goto on_error;
-
-    /* Register to ioqueue */
-    pj_bzero(&socket_cb, sizeof(socket_cb));
-    socket_cb.on_read_complete = &on_read_complete;
-    status = pj_ioqueue_register_sock(pool, resv->ioqueue, resv->udp_sock,
-				      resv, &socket_cb, &resv->udp_key);
-    if (status != PJ_SUCCESS)
-	goto on_error;
-
-    pj_ioqueue_op_key_init(&resv->udp_op_key, sizeof(resv->udp_op_key));
-
-    /* Start asynchronous read to the UDP socket */
-    resv->udp_len = sizeof(resv->udp_rx_pkt);
-    resv->udp_addr_len = sizeof(resv->udp_src_addr);
-    status = pj_ioqueue_recvfrom(resv->udp_key, &resv->udp_op_key, 
-				 resv->udp_rx_pkt, &resv->udp_len, 
-				 PJ_IOQUEUE_ALWAYS_ASYNC,
-				 &resv->udp_src_addr, &resv->udp_addr_len);
-    if (status != PJ_EPENDING)
-	goto on_error;
-
-
     /* Looks like everything is okay */
     *p_resolver = resv;
     return PJ_SUCCESS;
@@ -332,14 +381,13 @@ on_error:
 PJ_DEF(pj_status_t) pj_dns_resolver_destroy( pj_dns_resolver *resolver,
 					     pj_bool_t notify)
 {
+    pj_hash_iterator_t it_buf, *it;
     PJ_ASSERT_RETURN(resolver, PJ_EINVAL);
 
     if (notify) {
 	/*
 	 * Notify pending queries if requested.
 	 */
-	pj_hash_iterator_t it_buf, *it;
-
 	it = pj_hash_first(resolver->hquerybyid, &it_buf);
 	while (it) {
 	    pj_dns_async_query *q = (pj_dns_async_query *)
@@ -358,19 +406,25 @@ PJ_DEF(pj_status_t) pj_dns_resolver_destroy( pj_dns_resolver *resolver,
 	}
     }
 
+    /* Destroy cached entries */
+    it = pj_hash_first(resolver->hrescache, &it_buf);
+    while (it) {
+	struct cached_res *cache;
+
+	cache = (struct cached_res*) pj_hash_this(resolver->hrescache, it);
+	pj_hash_set(NULL, resolver->hrescache, &cache->key, 
+		    sizeof(cache->key), 0, NULL);
+	pj_pool_release(cache->pool);
+
+	it = pj_hash_first(resolver->hrescache, &it_buf);
+    }
+
     if (resolver->own_timer && resolver->timer) {
 	pj_timer_heap_destroy(resolver->timer);
 	resolver->timer = NULL;
     }
 
-    if (resolver->udp_key != NULL) {
-	pj_ioqueue_unregister(resolver->udp_key);
-	resolver->udp_key = NULL;
-	resolver->udp_sock = PJ_INVALID_SOCKET;
-    } else if (resolver->udp_sock != PJ_INVALID_SOCKET) {
-	pj_sock_close(resolver->udp_sock);
-	resolver->udp_sock = PJ_INVALID_SOCKET;
-    }
+    close_sock(resolver);
 
     if (resolver->own_ioqueue && resolver->ioqueue) {
 	pj_ioqueue_destroy(resolver->ioqueue);
@@ -405,6 +459,7 @@ PJ_DEF(pj_status_t) pj_dns_resolver_set_ns( pj_dns_resolver *resolver,
     pj_status_t status;
 
     PJ_ASSERT_RETURN(resolver && count && servers, PJ_EINVAL);
+    PJ_ASSERT_RETURN(count < PJ_DNS_RESOLVER_MAX_NS, PJ_EINVAL);
 
     pj_mutex_lock(resolver->mutex);
 
@@ -531,15 +586,6 @@ static pj_status_t transmit_query(pj_dns_resolver *resolver,
     pj_time_val delay;
     pj_status_t status;
 
-    /* Create DNS query packet */
-    pkt_size = sizeof(resolver->udp_tx_pkt);
-    name = pj_str(q->key.name);
-    status = pj_dns_make_query(resolver->udp_tx_pkt, &pkt_size, 
-			       q->id, q->key.qtype, &name);
-    if (status != PJ_SUCCESS) {
-	return status;
-    }
-
     /* Select which nameserver(s) to send requests to. */
     server_cnt = PJ_ARRAY_SIZE(servers);
     status = select_nameservers(resolver, &server_cnt, servers);
@@ -565,6 +611,28 @@ static pj_status_t transmit_query(pj_dns_resolver *resolver,
 	return status;
     }
 
+    /* Check if the socket is available for sending */
+    if (pj_ioqueue_is_pending(resolver->udp_key, &resolver->udp_op_tx_key)) {
+	++q->transmit_cnt;
+	PJ_LOG(4,(resolver->name.ptr,
+		  "Socket busy in transmitting DNS %s query for %s%s",
+		  pj_dns_get_type_name(q->key.qtype),
+		  q->key.name,
+		  (q->transmit_cnt < resolver->settings.qretr_count?
+		   ", will try again later":"")));
+	return PJ_SUCCESS;
+    }
+
+    /* Create DNS query packet */
+    pkt_size = sizeof(resolver->udp_tx_pkt);
+    name = pj_str(q->key.name);
+    status = pj_dns_make_query(resolver->udp_tx_pkt, &pkt_size,
+			       q->id, q->key.qtype, &name);
+    if (status != PJ_SUCCESS) {
+	pj_timer_heap_cancel(resolver->timer, &q->timer_entry);
+	return status;
+    }
+
     /* Get current time. */
     pj_gettimeofday(&now);
 
@@ -573,13 +641,16 @@ static pj_status_t transmit_query(pj_dns_resolver *resolver,
 	pj_ssize_t sent  = (pj_ssize_t) pkt_size;
 	struct nameserver *ns = &resolver->ns[servers[i]];
 
-	pj_sock_sendto(resolver->udp_sock, resolver->udp_tx_pkt, &sent, 0,
-		       &resolver->ns[servers[i]].addr, sizeof(pj_sockaddr_in));
+	status = pj_ioqueue_sendto(resolver->udp_key,
+				   &resolver->udp_op_tx_key,
+				   resolver->udp_tx_pkt, &sent, 0,
+				   &resolver->ns[servers[i]].addr,
+				   sizeof(pj_sockaddr_in));
 
-	PJ_LOG(4,(resolver->name.ptr, 
+	PJ_PERROR(4,(resolver->name.ptr, status,
 		  "%s %d bytes to NS %d (%s:%d): DNS %s query for %s",
 		  (q->transmit_cnt==0? "Transmitting":"Re-transmitting"),
-		  (int)sent, servers[i],
+		  (int)pkt_size, servers[i],
 		  pj_inet_ntoa(ns->addr.sin_addr), 
 		  (int)pj_ntohs(ns->addr.sin_port),
 		  pj_dns_get_type_name(q->key.qtype), 
@@ -602,7 +673,8 @@ static pj_status_t transmit_query(pj_dns_resolver *resolver,
  */
 static void init_res_key(struct res_key *key, int type, const pj_str_t *name)
 {
-    unsigned i, len;
+    unsigned i;
+    pj_size_t len;
     char *dst = key->name;
     const char *src = name->ptr;
 
@@ -619,6 +691,47 @@ static void init_res_key(struct res_key *key, int type, const pj_str_t *name)
 }
 
 
+/* Allocate new cache entry */
+static struct cached_res *alloc_entry(pj_dns_resolver *resolver)
+{
+    pj_pool_t *pool;
+    struct cached_res *cache;
+
+    pool = pj_pool_create(resolver->pool->factory, "dnscache",
+			  RES_BUF_SZ, 256, NULL);
+    cache = PJ_POOL_ZALLOC_T(pool, struct cached_res);
+    cache->pool = pool;
+    cache->ref_cnt = 1;
+
+    return cache;
+}
+
+/* Re-allocate cache entry, to free cached packet */
+static void reset_entry(struct cached_res **p_cached)
+{
+    pj_pool_t *pool;
+    struct cached_res *cache = *p_cached;
+    unsigned ref_cnt;
+
+    pool = cache->pool;
+    ref_cnt = cache->ref_cnt;
+
+    pj_pool_reset(pool);
+
+    cache = PJ_POOL_ZALLOC_T(pool, struct cached_res);
+    cache->pool = pool;
+    cache->ref_cnt = ref_cnt;
+    *p_cached = cache;
+}
+
+/* Put unused/expired cached entry to the free list */
+static void free_entry(pj_dns_resolver *resolver, struct cached_res *cache)
+{
+    PJ_UNUSED_ARG(resolver);
+    pj_pool_release(cache->pool);
+}
+
+
 /*
  * Create and start asynchronous DNS query for a single resource.
  */
@@ -682,6 +795,13 @@ PJ_DEF(pj_status_t) pj_dns_resolver_start_query( pj_dns_resolver *resolver,
 	    status = PJ_DNS_GET_RCODE(cache->pkt->hdr.flags);
 	    status = PJ_STATUS_FROM_DNS_RCODE(status);
 
+	    /* Workaround for deadlock problem. Need to increment the cache's
+	     * ref counter first before releasing mutex, so the cache won't be
+	     * destroyed by other thread while in callback.
+	     */
+	    cache->ref_cnt++;
+	    pj_mutex_unlock(resolver->mutex);
+
 	    /* This cached response is still valid. Just return this
 	     * response to caller.
 	     */
@@ -690,6 +810,14 @@ PJ_DEF(pj_status_t) pj_dns_resolver_start_query( pj_dns_resolver *resolver,
 	    }
 
 	    /* Done. No host resolution is necessary */
+	    pj_mutex_lock(resolver->mutex);
+
+	    /* Decrement the ref counter. Also check if it is time to free
+	     * the cache (as it has been expired).
+	     */
+	    cache->ref_cnt--;
+	    if (cache->ref_cnt <= 0)
+		free_entry(resolver, cache);
 
 	    /* Must return PJ_SUCCESS */
 	    status = PJ_SUCCESS;
@@ -702,8 +830,10 @@ PJ_DEF(pj_status_t) pj_dns_resolver_start_query( pj_dns_resolver *resolver,
 	 */
 	pj_hash_set(NULL, resolver->hrescache, &key, sizeof(key), 0, NULL);
 
-	/* Store the entry into free nodes */
-	pj_list_push_back(&resolver->res_free_nodes, cache);
+	/* Also free the cache, if it is not being used (by callback). */
+	cache->ref_cnt--;
+	if (cache->ref_cnt <= 0)
+	    free_entry(resolver, cache);
 
 	/* Must continue with creating a query now */
     }
@@ -732,6 +862,7 @@ PJ_DEF(pj_status_t) pj_dns_resolver_start_query( pj_dns_resolver *resolver,
     q = alloc_qnode(resolver, options, user_data, cb);
 
     /* Save the ID and key */
+    /* TODO: dnsext-forgery-resilient: randomize id for security */
     q->id = resolver->last_id++;
     if (resolver->last_id == 0)
 	resolver->last_id = 1;
@@ -782,6 +913,118 @@ PJ_DEF(pj_status_t) pj_dns_resolver_cancel_query(pj_dns_async_query *query,
 }
 
 
+/* 
+ * DNS response containing A packet. 
+ */
+PJ_DEF(pj_status_t) pj_dns_parse_a_response(const pj_dns_parsed_packet *pkt,
+					    pj_dns_a_record *rec)
+{
+    enum { MAX_SEARCH = 20 };
+    pj_str_t hostname, alias = {NULL, 0}, *resname;
+    pj_size_t bufstart = 0;
+    pj_size_t bufleft = sizeof(rec->buf_);
+    unsigned i, ansidx, search_cnt=0;
+
+    PJ_ASSERT_RETURN(pkt && rec, PJ_EINVAL);
+
+    /* Init the record */
+    pj_bzero(rec, sizeof(pj_dns_a_record));
+
+    /* Return error if there's error in the packet. */
+    if (PJ_DNS_GET_RCODE(pkt->hdr.flags))
+	return PJ_STATUS_FROM_DNS_RCODE(PJ_DNS_GET_RCODE(pkt->hdr.flags));
+
+    /* Return error if there's no query section */
+    if (pkt->hdr.qdcount == 0)
+	return PJLIB_UTIL_EDNSINANSWER;
+
+    /* Return error if there's no answer */
+    if (pkt->hdr.anscount == 0)
+	return PJLIB_UTIL_EDNSNOANSWERREC;
+
+    /* Get the hostname from the query. */
+    hostname = pkt->q[0].name;
+
+    /* Copy hostname to the record */
+    if (hostname.slen > (int)bufleft) {
+	return PJ_ENAMETOOLONG;
+    }
+
+    pj_memcpy(&rec->buf_[bufstart], hostname.ptr, hostname.slen);
+    rec->name.ptr = &rec->buf_[bufstart];
+    rec->name.slen = hostname.slen;
+
+    bufstart += hostname.slen;
+    bufleft -= hostname.slen;
+
+    /* Find the first RR which name matches the hostname */
+    for (ansidx=0; ansidx < pkt->hdr.anscount; ++ansidx) {
+	if (pj_stricmp(&pkt->ans[ansidx].name, &hostname)==0)
+	    break;
+    }
+
+    if (ansidx == pkt->hdr.anscount)
+	return PJLIB_UTIL_EDNSNOANSWERREC;
+
+    resname = &hostname;
+
+    /* Keep following CNAME records. */
+    while (pkt->ans[ansidx].type == PJ_DNS_TYPE_CNAME &&
+	   search_cnt++ < MAX_SEARCH)
+    {
+	resname = &pkt->ans[ansidx].rdata.cname.name;
+
+	if (!alias.slen)
+	    alias = *resname;
+
+	for (i=0; i < pkt->hdr.anscount; ++i) {
+	    if (pj_stricmp(resname, &pkt->ans[i].name)==0) {
+		break;
+	    }
+	}
+
+	if (i==pkt->hdr.anscount)
+	    return PJLIB_UTIL_EDNSNOANSWERREC;
+
+	ansidx = i;
+    }
+
+    if (search_cnt >= MAX_SEARCH)
+	return PJLIB_UTIL_EDNSINANSWER;
+
+    if (pkt->ans[ansidx].type != PJ_DNS_TYPE_A)
+	return PJLIB_UTIL_EDNSINANSWER;
+
+    /* Copy alias to the record, if present. */
+    if (alias.slen) {
+	if (alias.slen > (int)bufleft)
+	    return PJ_ENAMETOOLONG;
+
+	pj_memcpy(&rec->buf_[bufstart], alias.ptr, alias.slen);
+	rec->alias.ptr = &rec->buf_[bufstart];
+	rec->alias.slen = alias.slen;
+
+	bufstart += alias.slen;
+	bufleft -= alias.slen;
+    }
+
+    /* Get the IP addresses. */
+    for (i=0; i < pkt->hdr.anscount; ++i) {
+	if (pkt->ans[i].type == PJ_DNS_TYPE_A &&
+	    pj_stricmp(&pkt->ans[i].name, resname)==0 &&
+	    rec->addr_count < PJ_DNS_MAX_IP_IN_A_REC)
+	{
+	    rec->addr[rec->addr_count++].s_addr =
+		pkt->ans[i].rdata.a.ip_addr.s_addr;
+	}
+    }
+
+    if (rec->addr_count == 0)
+	return PJLIB_UTIL_EDNSNOANSWERREC;
+
+    return PJ_SUCCESS;
+}
+
 
 /* Set nameserver state */
 static void set_nameserver_state(pj_dns_resolver *resolver,
@@ -790,18 +1033,23 @@ static void set_nameserver_state(pj_dns_resolver *resolver,
 				 const pj_time_val *now)
 {
     struct nameserver *ns = &resolver->ns[index];
+    enum ns_state old_state = ns->state;
 
     ns->state = state;
     ns->state_expiry = *now;
 
     if (state == STATE_PROBING)
-	ns->state_expiry.sec += ((PJ_DNS_RESOLVER_QUERY_RETRANSMIT_COUNT + 2) *
-				 PJ_DNS_RESOLVER_QUERY_RETRANSMIT_DELAY) / 1000;
+	ns->state_expiry.sec += ((resolver->settings.qretr_count + 2) *
+				 resolver->settings.qretr_delay) / 1000;
     else if (state == STATE_ACTIVE)
-	ns->state_expiry.sec += PJ_DNS_RESOLVER_GOOD_NS_TTL;
+	ns->state_expiry.sec += resolver->settings.good_ns_ttl;
     else
-	ns->state_expiry.sec += PJ_DNS_RESOLVER_BAD_NS_TTL;
+	ns->state_expiry.sec += resolver->settings.bad_ns_ttl;
 
+    PJ_LOG(5, (resolver->name.ptr, "Nameserver %s:%d state changed %s --> %s",
+	       pj_inet_ntoa(ns->addr.sin_addr),
+	       (int)pj_ntohs(ns->addr.sin_port),
+	       state_names[old_state], state_names[state]));
 }
 
 
@@ -943,17 +1191,18 @@ static void update_res_cache(pj_dns_resolver *resolver,
 			     const pj_dns_parsed_packet *pkt)
 {
     struct cached_res *cache;
-    pj_pool_t *res_pool;
     pj_uint32_t hval=0, ttl;
-    PJ_USE_EXCEPTION;
 
     /* If status is unsuccessful, clear the same entry from the cache */
     if (status != PJ_SUCCESS) {
 	cache = (struct cached_res *) pj_hash_get(resolver->hrescache, key, 
 						  sizeof(*key), &hval);
-	if (cache)
-	    pj_list_push_back(&resolver->res_free_nodes, cache);
+	/* Remove the entry before releasing its pool (see ticket #1710) */
 	pj_hash_set(NULL, resolver->hrescache, key, sizeof(*key), hval, NULL);
+	
+	/* Free the entry */
+	if (cache && --cache->ref_cnt <= 0)
+	    free_entry(resolver, cache);
     }
 
 
@@ -987,9 +1236,12 @@ static void update_res_cache(pj_dns_resolver *resolver,
     if (ttl == 0) {
 	cache = (struct cached_res *) pj_hash_get(resolver->hrescache, key, 
 						  sizeof(*key), &hval);
-	if (cache)
-	    pj_list_push_back(&resolver->res_free_nodes, cache);
+	/* Remove the entry before releasing its pool (see ticket #1710) */
 	pj_hash_set(NULL, resolver->hrescache, key, sizeof(*key), hval, NULL);
+
+	/* Free the entry */
+	if (cache && --cache->ref_cnt <= 0)
+	    free_entry(resolver, cache);
 	return;
     }
 
@@ -997,31 +1249,31 @@ static void update_res_cache(pj_dns_resolver *resolver,
     cache = (struct cached_res *) pj_hash_get(resolver->hrescache, key, 
     					      sizeof(*key), &hval);
     if (cache == NULL) {
-	if (!pj_list_empty(&resolver->res_free_nodes)) {
-	    cache = resolver->res_free_nodes.next;
-	    pj_list_erase(cache);
-	} else {
-	    cache = PJ_POOL_ZALLOC_T(resolver->pool, struct cached_res);
-	}
+	cache = alloc_entry(resolver);
+    } else if (cache->ref_cnt > 1) {
+	/* When cache entry is being used by callback (to app), just decrement
+	 * ref_cnt so it will be freed after the callback returns and allocate
+	 * new entry.
+	 */
+	cache->ref_cnt--;
+	cache = alloc_entry(resolver);
+    } else {
+	/* Remove the entry before resetting its pool (see ticket #1710) */
+	pj_hash_set(NULL, resolver->hrescache, key, sizeof(*key), hval, NULL);
+
+	/* Reset cache to avoid bloated cache pool */
+	reset_entry(&cache);
     }
 
     /* Duplicate the packet.
-     * We don't need to keep the query, NS, and AR sections from the packet,
-     * so exclude from duplication.
+     * We don't need to keep the NS and AR sections from the packet,
+     * so exclude from duplication. We do need to keep the Query
+     * section since DNS A parser needs the query section to know
+     * the name being requested.
      */
-    res_pool = pj_pool_create_on_buf("respool", cache->buf, sizeof(cache->buf));
-    PJ_TRY {
-	cache->pkt = NULL;
-	pj_dns_packet_dup(res_pool, pkt, 
-			  PJ_DNS_NO_QD | PJ_DNS_NO_NS | PJ_DNS_NO_AR,
-			  &cache->pkt);
-    }
-    PJ_CATCH_ANY {
-	PJ_LOG(1,(THIS_FILE, 
-		  "Not enough memory to duplicate DNS response. Response was "
-		  "truncated."));
-    }
-    PJ_END;
+    pj_dns_packet_dup(cache->pool, pkt, 
+		      PJ_DNS_NO_NS | PJ_DNS_NO_AR,
+		      &cache->pkt);
 
     /* Calculate expiration time */
     if (set_expiry) {
@@ -1038,6 +1290,7 @@ static void update_res_cache(pj_dns_resolver *resolver,
     /* Update the hash table */
     pj_hash_set_np(resolver->hrescache, &cache->key, sizeof(*key), hval,
 		   cache->hbuf, cache);
+
 }
 
 
@@ -1070,7 +1323,7 @@ static void on_timeout( pj_timer_heap_t *timer_heap,
     q->timer_entry.id = 0;
 
     /* Check to see if we should retransmit instead of time out */
-    if (q->transmit_cnt < PJ_DNS_RESOLVER_QUERY_RETRANSMIT_COUNT) {
+    if (q->transmit_cnt < resolver->settings.qretr_count) {
 	status = transmit_query(resolver, q);
 	if (status == PJ_SUCCESS) {
 	    pj_mutex_unlock(resolver->mutex);
@@ -1091,6 +1344,9 @@ static void on_timeout( pj_timer_heap_t *timer_heap,
     pj_hash_set(NULL, resolver->hquerybyid, &q->id, sizeof(q->id), 0, NULL);
     pj_hash_set(NULL, resolver->hquerybyres, &q->key, sizeof(q->key), 0, NULL);
 
+    /* Workaround for deadlock problem in #1565 (similar to #1108) */
+    pj_mutex_unlock(resolver->mutex);
+
     /* Call application callback, if any. */
     if (q->cb)
 	(*q->cb)(q->user_data, PJ_ETIMEDOUT, NULL);
@@ -1103,6 +1359,9 @@ static void on_timeout( pj_timer_heap_t *timer_heap,
 	cq = cq->next;
     }
 
+    /* Workaround for deadlock problem in #1565 (similar to #1108) */
+    pj_mutex_lock(resolver->mutex);
+
     /* Clear data */
     q->timer_entry.id = 0;
     q->user_data = NULL;
@@ -1128,7 +1387,7 @@ static void on_read_complete(pj_ioqueue_key_t *key,
                              pj_ssize_t bytes_read)
 {
     pj_dns_resolver *resolver;
-    pj_pool_t *pool;
+    pj_pool_t *pool = NULL;
     pj_dns_parsed_packet *dns_pkt;
     pj_dns_async_query *q;
     pj_status_t status;
@@ -1143,7 +1402,7 @@ static void on_read_complete(pj_ioqueue_key_t *key,
     if (bytes_read < 0) {
 	char errmsg[PJ_ERR_MSG_SIZE];
 
-	status = -bytes_read;
+	status = (pj_status_t)-bytes_read;
 	pj_strerror(status, errmsg, sizeof(errmsg));
 	PJ_LOG(4,(resolver->name.ptr, 
 		  "DNS resolver read error from %s:%d: %s", 
@@ -1203,11 +1462,10 @@ static void on_read_complete(pj_ioqueue_key_t *key,
 		    sizeof(dns_pkt->hdr.id), NULL);
     if (!q) {
 	PJ_LOG(5,(resolver->name.ptr, 
-		  "Unable to find query for DNS response id=%d from %s:%d "
-		  "(the query may had been answered by other name servers)",
-		  (unsigned)dns_pkt->hdr.id,
+		  "DNS response from %s:%d id=%d discarded",
 		  pj_inet_ntoa(resolver->udp_src_addr.sin_addr), 
-		  pj_ntohs(resolver->udp_src_addr.sin_port)));
+		  pj_ntohs(resolver->udp_src_addr.sin_port),
+		  (unsigned)dns_pkt->hdr.id));
 	goto read_next_packet;
     }
 
@@ -1223,6 +1481,9 @@ static void on_read_complete(pj_ioqueue_key_t *key,
     pj_hash_set(NULL, resolver->hquerybyid, &q->id, sizeof(q->id), 0, NULL);
     pj_hash_set(NULL, resolver->hquerybyres, &q->key, sizeof(q->key), 0, NULL);
 
+    /* Workaround for deadlock problem in #1108 */
+    pj_mutex_unlock(resolver->mutex);
+
     /* Notify applications first, to allow application to modify the 
      * record before it is saved to the hash table.
      */
@@ -1241,6 +1502,9 @@ static void on_read_complete(pj_ioqueue_key_t *key,
 	}
     }
 
+    /* Workaround for deadlock problem in #1108 */
+    pj_mutex_lock(resolver->mutex);
+
     /* Save/update response cache. */
     update_res_cache(resolver, &q->key, status, PJ_TRUE, dns_pkt);
     
@@ -1259,6 +1523,10 @@ static void on_read_complete(pj_ioqueue_key_t *key,
     pj_list_push_back(&resolver->query_free_nodes, q);
 
 read_next_packet:
+    if (pool) {
+	/* needed just in case PJ_HAS_POOL_ALT_API is set */
+	pj_pool_release(pool);
+    }
     bytes_read = sizeof(resolver->udp_rx_pkt);
     resolver->udp_addr_len = sizeof(resolver->udp_src_addr);
     status = pj_ioqueue_recvfrom(resolver->udp_key, op_key, 
@@ -1392,8 +1660,6 @@ PJ_DEF(void) pj_dns_resolver_dump(pj_dns_resolver *resolver,
 	    it = pj_hash_next(resolver->hrescache, it);
 	}
     }
-    PJ_LOG(3,(resolver->name.ptr, "  Nb. of cached response free nodes: %u",
-	      pj_list_size(&resolver->res_free_nodes)));
     PJ_LOG(3,(resolver->name.ptr, "  Nb. of pending queries: %u (%u)",
 	      pj_hash_count(resolver->hquerybyid),
 	      pj_hash_count(resolver->hquerybyres)));
