@@ -1,6 +1,7 @@
-/* $Id: srv_resolver.c 1239 2007-05-01 12:25:01Z bennylp $ */
+/* $Id$ */
 /* 
- * Copyright (C) 2003-2007 Benny Prijono <benny@prijono.org>
+ * Copyright (C) 2008-2011 Teluu Inc. (http://www.teluu.com)
+ * Copyright (C) 2003-2008 Benny Prijono <benny@prijono.org>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -21,6 +22,7 @@
 #include <pj/array.h>
 #include <pj/assert.h>
 #include <pj/log.h>
+#include <pj/os.h>
 #include <pj/pool.h>
 #include <pj/rand.h>
 #include <pj/string.h>
@@ -28,11 +30,19 @@
 
 #define THIS_FILE   "srv_resolver.c"
 
-#define ADDR_MAX_COUNT	    8
+#define ADDR_MAX_COUNT	    PJ_DNS_MAX_IP_IN_A_REC
+
+struct common
+{
+    pj_dns_type		     type;	    /**< Type of this structure.*/
+};
 
 struct srv_target
 {
+    struct common	    common;
+    pj_dns_srv_async_query *parent;
     pj_str_t		    target_name;
+    pj_dns_async_query	   *q_a;
     char		    target_buf[PJ_MAX_HOSTNAME];
     pj_str_t		    cname;
     char		    cname_buf[PJ_MAX_HOSTNAME];
@@ -44,19 +54,20 @@ struct srv_target
     pj_in_addr		    addr[ADDR_MAX_COUNT];
 };
 
-typedef struct pj_dns_srv_resolver_job
+struct pj_dns_srv_async_query
 {
+    struct common	     common;
     char		    *objname;
 
-    pj_dns_resolver	    *resolver;	    /**< Resolver SIP instance.	    */
     pj_dns_type		     dns_state;	    /**< DNS type being resolved.   */
+    pj_dns_resolver	    *resolver;	    /**< Resolver SIP instance.	    */
     void		    *token;
+    pj_dns_async_query	    *q_srv;
     pj_dns_srv_resolver_cb  *cb;
-    pj_dns_async_query	    *qobject;
     pj_status_t		     last_error;
 
     /* Original request: */
-    pj_bool_t		     fallback_a;
+    unsigned		     option;
     pj_str_t		     full_name;
     pj_str_t		     domain_part;
     pj_uint16_t		     def_port;
@@ -68,7 +79,7 @@ typedef struct pj_dns_srv_resolver_job
     /* Number of hosts in SRV records that the IP address has been resolved */
     unsigned		     host_resolved;
 
-} pj_dns_srv_resolver_job;
+};
 
 
 /* Async resolver callback, forward decl. */
@@ -86,13 +97,14 @@ PJ_DEF(pj_status_t) pj_dns_srv_resolve( const pj_str_t *domain_name,
 					unsigned def_port,
 					pj_pool_t *pool,
 					pj_dns_resolver *resolver,
-					pj_bool_t fallback_a,
+					unsigned option,
 					void *token,
-					pj_dns_srv_resolver_cb *cb)
+					pj_dns_srv_resolver_cb *cb,
+					pj_dns_srv_async_query **p_query)
 {
-    int len;
+    pj_size_t len;
     pj_str_t target_name;
-    pj_dns_srv_resolver_job *query_job;
+    pj_dns_srv_async_query *query_job;
     pj_status_t status;
 
     PJ_ASSERT_RETURN(domain_name && domain_name->slen &&
@@ -111,12 +123,13 @@ PJ_DEF(pj_status_t) pj_dns_srv_resolve( const pj_str_t *domain_name,
 
 
     /* Build the query_job state */
-    query_job = PJ_POOL_ZALLOC_T(pool, pj_dns_srv_resolver_job);
+    query_job = PJ_POOL_ZALLOC_T(pool, pj_dns_srv_async_query);
+    query_job->common.type = PJ_DNS_TYPE_SRV;
     query_job->objname = target_name.ptr;
     query_job->resolver = resolver;
     query_job->token = token;
     query_job->cb = cb;
-    query_job->fallback_a = fallback_a;
+    query_job->option = option;
     query_job->full_name = target_name;
     query_job->domain_part.ptr = target_name.ptr + len;
     query_job->domain_part.slen = target_name.slen - len;
@@ -127,18 +140,53 @@ PJ_DEF(pj_status_t) pj_dns_srv_resolve( const pj_str_t *domain_name,
     query_job->dns_state = PJ_DNS_TYPE_SRV;
 
     PJ_LOG(5, (query_job->objname, 
-	       "Starting async DNS %s query_job: target=%.*",
+	       "Starting async DNS %s query_job: target=%.*s:%d",
 	       pj_dns_get_type_name(query_job->dns_state),
-	       (int)target_name.slen, target_name.ptr));
+	       (int)target_name.slen, target_name.ptr,
+	       def_port));
 
     status = pj_dns_resolver_start_query(resolver, &target_name, 
 				         query_job->dns_state, 0, 
 					 &dns_callback,
-    					 query_job, &query_job->qobject);
+    					 query_job, &query_job->q_srv);
+    if (status==PJ_SUCCESS && p_query)
+	*p_query = query_job;
+
     return status;
 }
 
 
+/*
+ * Cancel pending query.
+ */
+PJ_DEF(pj_status_t) pj_dns_srv_cancel_query(pj_dns_srv_async_query *query,
+					    pj_bool_t notify)
+{
+    pj_bool_t has_pending = PJ_FALSE;
+    unsigned i;
+
+    if (query->q_srv) {
+	pj_dns_resolver_cancel_query(query->q_srv, PJ_FALSE);
+	query->q_srv = NULL;
+	has_pending = PJ_TRUE;
+    }
+
+    for (i=0; i<query->srv_cnt; ++i) {
+	struct srv_target *srv = &query->srv[i];
+	if (srv->q_a) {
+	    pj_dns_resolver_cancel_query(srv->q_a, PJ_FALSE);
+	    srv->q_a = NULL;
+	    has_pending = PJ_TRUE;
+	}
+    }
+
+    if (has_pending && notify && query->cb) {
+	(*query->cb)(query->token, PJ_ECANCELLED, NULL);
+    }
+
+    return has_pending? PJ_SUCCESS : PJ_EINVALIDOP;
+}
+
 
 #define SWAP(type,ptr1,ptr2) if (ptr1 != ptr2) { \
 				type tmp; \
@@ -151,7 +199,7 @@ PJ_DEF(pj_status_t) pj_dns_srv_resolve( const pj_str_t *domain_name,
 
 
 /* Build server entries in the query_job based on received SRV response */
-static void build_server_entries(pj_dns_srv_resolver_job *query_job, 
+static void build_server_entries(pj_dns_srv_async_query *query_job, 
 				 pj_dns_parsed_packet *response)
 {
     unsigned i;
@@ -187,6 +235,12 @@ static void build_server_entries(pj_dns_srv_resolver_job *query_job,
 	++query_job->srv_cnt;
     }
 
+    if (query_job->srv_cnt == 0) {
+	PJ_LOG(4,(query_job->objname, 
+		  "Could not find SRV record in DNS answer!"));
+	return;
+    }
+
     /* First pass: 
      *	order the entries based on priority.
      */
@@ -200,10 +254,7 @@ static void build_server_entries(pj_dns_srv_resolver_job *query_job,
     }
 
     /* Second pass:
-     *	pick one host among hosts with the same priority, according
-     *	to its weight. The idea is when one server fails, client should
-     *	contact the next server with higher priority rather than contacting
-     *	server with the same priority as the failed one.
+     *	Order the entry in a list.
      *
      *  The algorithm for selecting server among servers with the same
      *  priority is described in RFC 2782.
@@ -244,12 +295,15 @@ static void build_server_entries(pj_dns_srv_resolver_job *query_job,
 	    SWAP(struct srv_target, &query_job->srv[i], &query_job->srv[j]);
 
 	    /* Remove all other entries (of the same priority) */
+	    /* Don't need to do this.
+	     * See https://trac.pjsip.org/repos/ticket/1719
 	    while (count > 1) {
 		pj_array_erase(query_job->srv, sizeof(struct srv_target), 
 			       query_job->srv_cnt, i+1);
 		--count;
 		--query_job->srv_cnt;
 	    }
+	    */
 	}
     }
 
@@ -278,8 +332,10 @@ static void build_server_entries(pj_dns_srv_resolver_job *query_job,
 	    if (pj_stricmp(&rr->name, &query_job->srv[j].target_name)==0) {
 		unsigned cnt = query_job->srv[j].addr_cnt;
 		query_job->srv[j].addr[cnt].s_addr = rr->rdata.a.ip_addr.s_addr;
+		/* Only increment host_resolved once per SRV record */
+		if (query_job->srv[j].addr_cnt == 0)
+		    ++query_job->host_resolved;
 		++query_job->srv[j].addr_cnt;
-		++query_job->host_resolved;
 		break;
 	    }
 	}
@@ -347,30 +403,41 @@ static void build_server_entries(pj_dns_srv_resolver_job *query_job,
 
 
 /* Start DNS A record queries for all SRV records in the query_job structure */
-static pj_status_t resolve_hostnames(pj_dns_srv_resolver_job *query_job)
+static pj_status_t resolve_hostnames(pj_dns_srv_async_query *query_job)
 {
-    unsigned i;
+    unsigned i, err_cnt = 0;
     pj_status_t err=PJ_SUCCESS, status;
 
     query_job->dns_state = PJ_DNS_TYPE_A;
     for (i=0; i<query_job->srv_cnt; ++i) {
+	struct srv_target *srv = &query_job->srv[i];
+
 	PJ_LOG(5, (query_job->objname, 
 		   "Starting async DNS A query_job for %.*s",
-		   (int)query_job->srv[i].target_name.slen, 
-		   query_job->srv[i].target_name.ptr));
+		   (int)srv->target_name.slen, 
+		   srv->target_name.ptr));
 
+	srv->common.type = PJ_DNS_TYPE_A;
+	srv->parent = query_job;
+
+	/* See also #1809: dns_callback() will be invoked synchronously when response
+	 * is available in the cache, and var 'query_job->host_resolved' will get
+	 * incremented within the dns_callback(), which will cause this function
+	 * returning false error, so don't use that variable for counting errors.
+	 */
 	status = pj_dns_resolver_start_query(query_job->resolver,
-					     &query_job->srv[i].target_name,
+					     &srv->target_name,
 					     PJ_DNS_TYPE_A, 0,
 					     &dns_callback,
-					     query_job, NULL);
+					     srv, &srv->q_a);
 	if (status != PJ_SUCCESS) {
 	    query_job->host_resolved++;
+	    err_cnt++;
 	    err = status;
 	}
     }
     
-    return (query_job->host_resolved == query_job->srv_cnt) ? err : PJ_SUCCESS;
+    return (err_cnt == query_job->srv_cnt) ? err : PJ_SUCCESS;
 }
 
 /* 
@@ -381,15 +448,29 @@ static void dns_callback(void *user_data,
 			 pj_status_t status,
 			 pj_dns_parsed_packet *pkt)
 {
-    pj_dns_srv_resolver_job *query_job = (pj_dns_srv_resolver_job*) user_data;
+    struct common *common = (struct common*) user_data;
+    pj_dns_srv_async_query *query_job;
+    struct srv_target *srv = NULL;
     unsigned i;
 
-    /* Proceed to next stage */
+    if (common->type == PJ_DNS_TYPE_SRV) {
+	query_job = (pj_dns_srv_async_query*) common;
+	srv = NULL;
+    } else if (common->type == PJ_DNS_TYPE_A) {
+	srv = (struct srv_target*) common;
+	query_job = srv->parent;
+    } else {
+	pj_assert(!"Unexpected user data!");
+	return;
+    }
 
+    /* Proceed to next stage */
     if (query_job->dns_state == PJ_DNS_TYPE_SRV) {
 
 	/* We are getting SRV response */
 
+	query_job->q_srv = NULL;
+
 	if (status == PJ_SUCCESS && pkt->hdr.anscount != 0) {
 	    /* Got SRV response, build server entry. If A records are available
 	     * in additional records section of the DNS response, save them too.
@@ -410,7 +491,9 @@ static void dns_callback(void *user_data,
 		      errmsg));
 
 	    /* Trigger error when fallback is disabled */
-	    if (query_job->fallback_a == PJ_FALSE) {
+	    if ((query_job->option &
+		 (PJ_DNS_SRV_FALLBACK_A | PJ_DNS_SRV_FALLBACK_AAAA)) == 0) 
+	    {
 		goto on_error;
 	    }
 	}
@@ -457,95 +540,50 @@ static void dns_callback(void *user_data,
 
     } else if (query_job->dns_state == PJ_DNS_TYPE_A) {
 
+	/* Clear the outstanding job */
+	srv->q_a = NULL;
+
 	/* Check that we really have answer */
 	if (status==PJ_SUCCESS && pkt->hdr.anscount != 0) {
-	    int ans_idx = -1;
+	    pj_dns_a_record rec;
 
-	    /* Find the first DNS A record in the answer while processing
-	     * the CNAME info found in the response.
-	     */
-	    for (i=0; i < pkt->hdr.anscount; ++i) {
-
-		pj_dns_parsed_rr *rr = &pkt->ans[i];
-
-		if (rr->type == PJ_DNS_TYPE_A) {
-
-		    if (ans_idx == -1)
-			ans_idx = i;
-
-		} else if (rr->type == PJ_DNS_TYPE_CNAME) {
-		    /* Find which server entry to be updated with
-		     * the CNAME information.
-		     */
-		    unsigned j;
-		    pj_str_t cname = rr->rdata.cname.name;
-
-		    for (j=0; j<query_job->srv_cnt; ++j) 
-		    {
-			struct srv_target *srv = &query_job->srv[j];
-			if (pj_stricmp(&rr->name, &srv->target_name)==0)
-			{
-			    /* Update CNAME info for this server entry */
-			    srv->cname.ptr = srv->cname_buf;
-			    pj_strncpy(&srv->cname, &cname, 
-				       sizeof(srv->cname_buf));
-			    break;
-			}
-		    }
-		}
-	    }
-
-	    if (ans_idx == -1) {
-		/* There's no DNS A answer! */
-		PJ_LOG(5,(query_job->objname, 
-			  "No DNS A record in response!"));
-		status = PJLIB_UTIL_EDNSNOANSWERREC;
+	    /* Parse response */
+	    status = pj_dns_parse_a_response(pkt, &rec);
+	    if (status != PJ_SUCCESS)
 		goto on_error;
+
+	    pj_assert(rec.addr_count != 0);
+
+	    /* Update CNAME alias, if present. */
+	    if (rec.alias.slen) {
+		pj_assert(rec.alias.slen <= (int)sizeof(srv->cname_buf));
+		srv->cname.ptr = srv->cname_buf;
+		pj_strcpy(&srv->cname, &rec.alias);
+	    } else {
+		srv->cname.slen = 0;
 	    }
 
 	    /* Update IP address of the corresponding hostname or CNAME */
-	    for (i=0; i<query_job->srv_cnt; ++i) {
-		pj_dns_parsed_rr *rr = &pkt->ans[ans_idx];
-		struct srv_target *srv = &query_job->srv[i];
-
-		if (pj_stricmp(&rr->name, &srv->target_name)==0 ||
-		    pj_stricmp(&rr->name, &srv->cname)==0) 
-		{
-		    break;
-		}
-	    }
-
-	    if (i == query_job->srv_cnt) {
-		PJ_LOG(4,(query_job->objname, 
-			  "Received answer to DNS A request with no matching "
-			  "SRV record! The unknown name is %.*s",
-			  (int)pkt->ans[ans_idx].name.slen, 
-			  pkt->ans[ans_idx].name.ptr));
-	    } else {
-		unsigned j;
-
-		query_job->srv[i].addr[query_job->srv[i].addr_cnt++].s_addr =
-		    pkt->ans[ans_idx].rdata.a.ip_addr.s_addr;
+	    if (srv->addr_cnt < ADDR_MAX_COUNT) {
+		srv->addr[srv->addr_cnt++].s_addr = rec.addr[0].s_addr;
 
 		PJ_LOG(5,(query_job->objname, 
 			  "DNS A for %.*s: %s",
-			  (int)query_job->srv[i].target_name.slen, 
-			  query_job->srv[i].target_name.ptr,
-			  pj_inet_ntoa(pkt->ans[ans_idx].rdata.a.ip_addr)));
+			  (int)srv->target_name.slen, 
+			  srv->target_name.ptr,
+			  pj_inet_ntoa(rec.addr[0])));
+	    }
 
-		/* Check for multiple IP addresses */
-		for (j=ans_idx+1; j<pkt->hdr.anscount && 
-			    query_job->srv[i].addr_cnt < ADDR_MAX_COUNT; ++j)
-		{
-		    query_job->srv[i].addr[query_job->srv[i].addr_cnt++].s_addr = 
-			pkt->ans[j].rdata.a.ip_addr.s_addr;
+	    /* Check for multiple IP addresses */
+	    for (i=1; i<rec.addr_count && srv->addr_cnt < ADDR_MAX_COUNT; ++i)
+	    {
+		srv->addr[srv->addr_cnt++].s_addr = rec.addr[i].s_addr;
 
-		    PJ_LOG(5,(query_job->objname, 
-			      "Additional DNS A for %.*s: %s",
-			      (int)query_job->srv[i].target_name.slen, 
-			      query_job->srv[i].target_name.ptr,
-			      pj_inet_ntoa(pkt->ans[j].rdata.a.ip_addr)));
-		}
+		PJ_LOG(5,(query_job->objname, 
+			  "Additional DNS A for %.*s: %s",
+			  (int)srv->target_name.slen, 
+			  srv->target_name.ptr,
+			  pj_inet_ntoa(rec.addr[i])));
 	    }
 
 	} else if (status != PJ_SUCCESS) {
@@ -571,48 +609,42 @@ static void dns_callback(void *user_data,
     /* Check if all hosts have been resolved */
     if (query_job->host_resolved == query_job->srv_cnt) {
 	/* Got all answers, build server addresses */
-	pj_dns_srv_record svr_addr;
+	pj_dns_srv_record srv_rec;
 
-	svr_addr.count = 0;
+	srv_rec.count = 0;
 	for (i=0; i<query_job->srv_cnt; ++i) {
 	    unsigned j;
+	    struct srv_target *srv = &query_job->srv[i];
 
-	    /* Do we have IP address for this server? */
-	    /* This log is redundant really.
-	    if (query_job->srv[i].addr_cnt == 0) {
-		PJ_LOG(5,(query_job->objname, 
-			  " SRV target %.*s:%d does not have IP address!",
-			  (int)query_job->srv[i].target_name.slen,
-			  query_job->srv[i].target_name.ptr,
-			  query_job->srv[i].port));
-		continue;
+	    srv_rec.entry[srv_rec.count].priority = srv->priority;
+	    srv_rec.entry[srv_rec.count].weight = srv->weight;
+	    srv_rec.entry[srv_rec.count].port = (pj_uint16_t)srv->port ;
+
+	    srv_rec.entry[srv_rec.count].server.name = srv->target_name;
+	    srv_rec.entry[srv_rec.count].server.alias = srv->cname;
+	    srv_rec.entry[srv_rec.count].server.addr_count = 0;
+
+	    pj_assert(srv->addr_cnt <= PJ_DNS_MAX_IP_IN_A_REC);
+
+	    for (j=0; j<srv->addr_cnt; ++j) {
+		srv_rec.entry[srv_rec.count].server.addr[j].s_addr = 
+		    srv->addr[j].s_addr;
+		++srv_rec.entry[srv_rec.count].server.addr_count;
 	    }
-	    */
 
-	    for (j=0; j<query_job->srv[i].addr_cnt; ++j) {
-		unsigned idx = svr_addr.count;
-		pj_sockaddr_in *addr;
-
-		svr_addr.entry[idx].priority = query_job->srv[i].priority;
-		svr_addr.entry[idx].weight = query_job->srv[i].weight;
-		svr_addr.entry[idx].addr_len = sizeof(pj_sockaddr_in);
-	     
-		addr = (pj_sockaddr_in*)&svr_addr.entry[idx].addr;
-		pj_bzero(addr, sizeof(pj_sockaddr_in));
-		addr->sin_family = PJ_AF_INET;
-		addr->sin_addr = query_job->srv[i].addr[j];
-		addr->sin_port = pj_htons((pj_uint16_t)query_job->srv[i].port);
-
-		++svr_addr.count;
+	    if (srv->addr_cnt > 0) {
+		++srv_rec.count;
+		if (srv_rec.count == PJ_DNS_SRV_MAX_ADDR)
+		    break;
 	    }
 	}
 
 	PJ_LOG(5,(query_job->objname, 
 		  "Server resolution complete, %d server entry(s) found",
-		  svr_addr.count));
+		  srv_rec.count));
 
 
-	if (svr_addr.count > 0)
+	if (srv_rec.count > 0)
 	    status = PJ_SUCCESS;
 	else {
 	    status = query_job->last_error;
@@ -621,7 +653,7 @@ static void dns_callback(void *user_data,
 	}
 
 	/* Call the callback */
-	(*query_job->cb)(query_job->token, status, &svr_addr);
+	(*query_job->cb)(query_job->token, status, &srv_rec);
     }
 
 
