@@ -1,6 +1,7 @@
-/* $Id: pool.h 1040 2007-03-05 00:58:24Z bennylp $ */
+/* $Id$ */
 /* 
- * Copyright (C)2003-2007 Benny Prijono <benny@prijono.org>
+ * Copyright (C) 2008-2011 Teluu Inc. (http://www.teluu.com)
+ * Copyright (C) 2003-2008 Benny Prijono <benny@prijono.org>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -39,7 +40,6 @@ PJ_BEGIN_DECL
 
 /**
  * @defgroup PJ_POOL_GROUP Fast Memory Pool
- * @ingroup PJ
  * @brief
  * Memory pools allow dynamic memory allocation comparable to malloc or the 
  * new in operator C++. Those implementations are not desirable for very
@@ -227,9 +227,9 @@ PJ_BEGIN_DECL
 	    return 1;
 	}
 
-	// Create the pool factory, in this case, a caching pool.
-	pj_caching_pool_init(&cp, &pj_pool_factory_default_policy, 
-			     1024*1024 );
+	// Create the pool factory, in this case, a caching pool,
+	// using default pool policy.
+	pj_caching_pool_init(&cp, NULL, 1024*1024 );
 
 	// Do a demo
 	pool_demo_1(&cp.factory);
@@ -505,8 +505,14 @@ PJ_INLINE(void*) pj_pool_zalloc(pj_pool_t *pool, pj_size_t size)
 #define PJ_POOL_ZALLOC_T(pool,type) \
 	    ((type*)pj_pool_zalloc(pool, sizeof(type)))
 
+/*
+ * Internal functions
+ */
+PJ_IDECL(void*) pj_pool_alloc_from_block(pj_pool_block *block, pj_size_t size);
+PJ_DECL(void*) pj_pool_allocate_find(pj_pool_t *pool, pj_size_t size);
 
 
+	
 /**
  * @}	// PJ_POOL
  */
@@ -606,10 +612,18 @@ typedef struct pj_pool_factory_policy
 } pj_pool_factory_policy;
 
 /**
+ * \def PJ_NO_MEMORY_EXCEPTION
  * This constant denotes the exception number that will be thrown by default
  * memory factory policy when memory allocation fails.
+ *
+ * @see pj_NO_MEMORY_EXCEPTION()
  */
-extern int PJ_NO_MEMORY_EXCEPTION;
+PJ_DECL_DATA(int) PJ_NO_MEMORY_EXCEPTION;
+
+/**
+ * Get #PJ_NO_MEMORY_EXCEPTION constant.
+ */ 
+PJ_DECL(int) pj_NO_MEMORY_EXCEPTION(void);
 
 /**
  * This global variable points to default memory pool factory policy.
@@ -617,8 +631,19 @@ extern int PJ_NO_MEMORY_EXCEPTION;
  *  - block allocation and deallocation use malloc() and free().
  *  - callback will raise PJ_NO_MEMORY_EXCEPTION exception.
  *  - access to pool factory is not serialized (i.e. not thread safe).
+ *
+ * @see pj_pool_factory_get_default_policy
  */
-extern pj_pool_factory_policy pj_pool_factory_default_policy;
+PJ_DECL_DATA(pj_pool_factory_policy) pj_pool_factory_default_policy;
+
+
+/**
+ * Get the default pool factory policy.
+ *
+ * @return the pool policy.
+ */
+PJ_DECL(const pj_pool_factory_policy*) pj_pool_factory_get_default_policy(void);
+
 
 /**
  * This structure contains the declaration for pool factory interface.
@@ -831,7 +856,7 @@ struct pj_caching_pool
     /**
      * Internal pool.
      */
-    char	    pool_buf[256 * (sizeof(long) / 4)];
+    char	    pool_buf[256 * (sizeof(size_t) / 4)];
 
     /**
      * Mutex.
