@@ -1,6 +1,7 @@
-/* $Id: string.h 1210 2007-04-22 12:48:30Z bennylp $ */
+/* $Id$ */
 /* 
- * Copyright (C)2003-2007 Benny Prijono <benny@prijono.org>
+ * Copyright (C) 2008-2011 Teluu Inc. (http://www.teluu.com)
+ * Copyright (C) 2003-2008 Benny Prijono <benny@prijono.org>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -341,7 +342,7 @@ PJ_IDECL(int) pj_strncmp2( const pj_str_t *str1, const char *str2,
 			   pj_size_t len);
 
 /**
- * Perform lowercase comparison to the strings.
+ * Perform case-insensitive comparison to the strings.
  *
  * @param str1	    The string to compare.
  * @param str2	    The string to compare.
@@ -398,7 +399,7 @@ PJ_IDECL(int) pj_stricmp_alnum(const pj_str_t *str1, const pj_str_t *str2);
 #endif
 
 /**
- * Perform lowercase comparison to the strings.
+ * Perform case-insensitive comparison to the strings.
  *
  * @param str1	    The string to compare.
  * @param str2	    The string to compare.
@@ -411,7 +412,7 @@ PJ_IDECL(int) pj_stricmp_alnum(const pj_str_t *str1, const pj_str_t *str2);
 PJ_IDECL(int) pj_stricmp2( const pj_str_t *str1, const char *str2);
 
 /**
- * Perform lowercase comparison to the strings.
+ * Perform case-insensitive comparison to the strings.
  *
  * @param str1	    The string to compare.
  * @param str2	    The string to compare.
@@ -426,7 +427,7 @@ PJ_IDECL(int) pj_strnicmp( const pj_str_t *str1, const pj_str_t *str2,
 			   pj_size_t len);
 
 /**
- * Perform lowercase comparison to the strings.
+ * Perform case-insensitive comparison to the strings.
  *
  * @param str1	    The string to compare.
  * @param str2	    The string to compare.
@@ -471,6 +472,31 @@ PJ_INLINE(char*) pj_strchr( const pj_str_t *str, int chr)
     return (char*) memchr((char*)str->ptr, chr, str->slen);
 }
 
+/**
+ * Find the occurence of a substring substr in string str.
+ *
+ * @param str	    The string to search.
+ * @param substr    The string to search fo.
+ *
+ * @return the pointer to the position of substr in str, or NULL. Note
+ *         that if str is not NULL terminated, the returned pointer
+ *         is pointing to non-NULL terminated string.
+ */
+PJ_DECL(char*) pj_strstr(const pj_str_t *str, const pj_str_t *substr);
+
+/**
+ * Performs substring lookup like pj_strstr() but ignores the case of
+ * both strings.
+ *
+ * @param str	    The string to search.
+ * @param substr    The string to search fo.
+ *
+ * @return the pointer to the position of substr in str, or NULL. Note
+ *         that if str is not NULL terminated, the returned pointer
+ *         is pointing to non-NULL terminated string.
+ */
+PJ_DECL(char*) pj_stristr(const pj_str_t *str, const pj_str_t *substr);
+
 /**
  * Remove (trim) leading whitespaces from the string.
  *
@@ -499,7 +525,8 @@ PJ_DECL(pj_str_t*) pj_strrtrim( pj_str_t *str );
 PJ_IDECL(pj_str_t*) pj_strtrim( pj_str_t *str );
 
 /**
- * Initialize the buffer with some random string.
+ * Initialize the buffer with some random string. Note that the 
+ * generated string is not NULL terminated.
  *
  * @param str	    the string to store the result.
  * @param length    the length of the random string to generate.
@@ -509,7 +536,20 @@ PJ_IDECL(pj_str_t*) pj_strtrim( pj_str_t *str );
 PJ_DECL(char*) pj_create_random_string(char *str, pj_size_t length);
 
 /**
- * Convert string to unsigned integer.
+ * Convert string to signed integer. The conversion will stop as
+ * soon as non-digit character is found or all the characters have
+ * been processed.
+ *
+ * @param str	the string.
+ *
+ * @return the integer.
+ */
+PJ_DECL(long) pj_strtol(const pj_str_t *str);
+
+/**
+ * Convert string to unsigned integer. The conversion will stop as
+ * soon as non-digit character is found or all the characters have
+ * been processed.
  *
  * @param str	the string.
  *
@@ -534,6 +574,15 @@ PJ_DECL(unsigned long) pj_strtoul(const pj_str_t *str);
 PJ_DECL(unsigned long) pj_strtoul2(const pj_str_t *str, pj_str_t *endptr,
 				   unsigned base);
 
+/**
+ * Convert string to float.
+ *
+ * @param str	the string.
+ *
+ * @return the value.
+ */
+PJ_DECL(float) pj_strtof(const pj_str_t *str);
+
 /**
  * Utility to convert unsigned integer to string. Note that the
  * string will be NULL terminated.
