@@ -1,6 +1,7 @@
-/* $Id: os_core_win32.c 1046 2007-03-06 23:35:54Z bennylp $ */
+/* $Id$ */
 /* 
- * Copyright (C)2003-2007 Benny Prijono <benny@prijono.org>
+ * Copyright (C) 2008-2011 Teluu Inc. (http://www.teluu.com)
+ * Copyright (C) 2003-2008 Benny Prijono <benny@prijono.org>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -37,6 +38,17 @@
 #  include <winsock2.h>
 #endif
 
+/* Activate mutex related logging if PJ_DEBUG_MUTEX is set, otherwise
+ * use default level 6 logging.
+ */
+#if defined(PJ_DEBUG_MUTEX) && PJ_DEBUG_MUTEX
+#   undef PJ_DEBUG
+#   define PJ_DEBUG	    1
+#   define LOG_MUTEX(expr)  PJ_LOG(5,expr)
+#else
+#   define LOG_MUTEX(expr)  PJ_LOG(6,expr)
+#endif
+
 #define THIS_FILE	"os_core_win32.c"
 
 /*
@@ -104,6 +116,11 @@ struct pj_atomic_t
     long value;
 };
 
+/*
+ * Flag and reference counter for PJLIB instance.
+ */
+static int initialized;
+
 /*
  * Static global variables.
  */
@@ -130,6 +147,12 @@ PJ_DEF(pj_status_t) pj_init(void)
     pj_str_t guid;
     pj_status_t rc;
 
+    /* Check if PJLIB have been initialized */
+    if (initialized) {
+	++initialized;
+	return PJ_SUCCESS;
+    }
+
     /* Init Winsock.. */
     if (WSAStartup(MAKEWORD(2,0), &wsa) != 0) {
 	return PJ_RETURN_OS_ERROR(WSAGetLastError());
@@ -140,17 +163,21 @@ PJ_DEF(pj_status_t) pj_init(void)
 	return rc;
     }
     
-    /* Init random seed. */
-    pj_srand( GetCurrentProcessId() );
+    /* Init logging */
+    pj_log_init();
 
-    /* Startup GUID. */
-    guid.ptr = dummy_guid;
-    pj_generate_unique_string( &guid );
+    /* Init random seed. */
+    /* Or probably not. Let application in charge of this */
+    /* pj_srand( GetCurrentProcessId() ); */
 
     /* Initialize critical section. */
     if ((rc=init_mutex(&critical_section_mutex, "pj%p")) != PJ_SUCCESS)
 	return rc;
 
+    /* Startup GUID. */
+    guid.ptr = dummy_guid;
+    pj_generate_unique_string( &guid );
+
     /* Initialize exception ID for the pool. 
      * Must do so after critical section is configured.
      */
@@ -171,6 +198,10 @@ PJ_DEF(pj_status_t) pj_init(void)
     }
 #endif   
 
+    /* Flag PJLIB as initialized */
+    ++initialized;
+    pj_assert(initialized == 1);
+
     PJ_LOG(4,(THIS_FILE, "pjlib %s for win32 initialized",
 	      PJ_VERSION));
 
@@ -197,6 +228,20 @@ PJ_DEF(void) pj_shutdown()
 {
     int i;
 
+    /* Only perform shutdown operation when 'initialized' reaches zero */
+    pj_assert(initialized > 0);
+    if (--initialized != 0)
+	return;
+
+    /* Display stack usage */
+#if defined(PJ_OS_HAS_CHECK_STACK) && PJ_OS_HAS_CHECK_STACK!=0
+    {
+	pj_thread_t *rec = (pj_thread_t*)main_thread;
+	PJ_LOG(5,(rec->obj_name, "Main thread stack max usage=%u by %s:%d", 
+		  rec->stk_max_usage, rec->caller_file, rec->caller_line));
+    }
+#endif
+
     /* Call atexit() functions */
     for (i=atexit_count-1; i>=0; --i) {
 	(*atexit_func[i])();
@@ -221,8 +266,11 @@ PJ_DEF(void) pj_shutdown()
     /* Clear static variables */
     pj_errno_clear_handlers();
 
+    /* Ticket #1132: Assertion when (re)starting PJLIB on different thread */
+    pj_bzero(main_thread, sizeof(main_thread));
+
     /* Shutdown Winsock */
-    //WSACleanup();
+    WSACleanup();
 }
 
 
@@ -243,6 +291,76 @@ PJ_DEF(pj_bool_t) pj_thread_is_registered(void)
     return pj_thread_local_get(thread_tls_id) != 0;
 }
 
+
+/*
+ * Get thread priority value for the thread.
+ */
+PJ_DEF(int) pj_thread_get_prio(pj_thread_t *thread)
+{
+    return GetThreadPriority(thread->hthread);
+}
+
+
+/*
+ * Set the thread priority.
+ */
+PJ_DEF(pj_status_t) pj_thread_set_prio(pj_thread_t *thread,  int prio)
+{
+#if PJ_HAS_THREADS
+    PJ_ASSERT_RETURN(thread, PJ_EINVAL);
+    PJ_ASSERT_RETURN(prio>=THREAD_PRIORITY_IDLE && 
+			prio<=THREAD_PRIORITY_TIME_CRITICAL,
+		     PJ_EINVAL);
+
+    if (SetThreadPriority(thread->hthread, prio) == FALSE)
+	return PJ_RETURN_OS_ERROR(GetLastError());
+
+    return PJ_SUCCESS;
+
+#else
+    PJ_UNUSED_ARG(thread);
+    PJ_UNUSED_ARG(prio);
+    pj_assert("pj_thread_set_prio() called in non-threading mode!");
+    return PJ_EINVALIDOP;
+#endif
+}
+
+
+/*
+ * Get the lowest priority value available on this system.
+ */
+PJ_DEF(int) pj_thread_get_prio_min(pj_thread_t *thread)
+{
+    PJ_UNUSED_ARG(thread);
+    return THREAD_PRIORITY_IDLE;
+}
+
+
+/*
+ * Get the highest priority value available on this system.
+ */
+PJ_DEF(int) pj_thread_get_prio_max(pj_thread_t *thread)
+{
+    PJ_UNUSED_ARG(thread);
+    return THREAD_PRIORITY_TIME_CRITICAL;
+}
+
+
+/*
+ * Get native thread handle
+ */
+PJ_DEF(void*) pj_thread_get_os_handle(pj_thread_t *thread) 
+{
+    PJ_ASSERT_RETURN(thread, NULL);
+
+#if PJ_HAS_THREADS
+    return thread->hthread;
+#else
+    pj_assert("pj_thread_is_registered() called in non-threading mode!");
+    return NULL;
+#endif
+}
+
 /*
  * pj_thread_register(..)
  */
@@ -290,7 +408,7 @@ PJ_DEF(pj_status_t) pj_thread_register ( const char *cstr_thread_name,
 			 cstr_thread_name, thread->idthread);
     else
 	pj_ansi_snprintf(thread->obj_name, sizeof(thread->obj_name), 
-		         "thr%p", (void*)thread->idthread);
+		         "thr%p", (void*)(pj_ssize_t)thread->idthread);
     
     rc = pj_thread_local_set(thread_tls_id, thread);
     if (rc != PJ_SUCCESS)
@@ -333,6 +451,11 @@ static DWORD WINAPI thread_main(void *param)
     result = (*rec->proc)(rec->arg);
 
     PJ_LOG(6,(rec->obj_name, "Thread quitting"));
+#if defined(PJ_OS_HAS_CHECK_STACK) && PJ_OS_HAS_CHECK_STACK!=0
+    PJ_LOG(5,(rec->obj_name, "Thread stack max usage=%u by %s:%d", 
+	      rec->stk_max_usage, rec->caller_file, rec->caller_line));
+#endif
+
     return (DWORD)result;
 }
 
@@ -376,7 +499,7 @@ PJ_DEF(pj_status_t) pj_thread_create( pj_pool_t *pool,
     PJ_LOG(6, (rec->obj_name, "Thread created"));
 
 #if defined(PJ_OS_HAS_CHECK_STACK) && PJ_OS_HAS_CHECK_STACK!=0
-    rec->stk_size = stack_size ? stack_size : 0xFFFFFFFFUL;
+    rec->stk_size = stack_size ? (pj_uint32_t)stack_size : 0xFFFFFFFFUL;
     rec->stk_max_usage = 0;
 #endif
 
@@ -456,6 +579,9 @@ PJ_DEF(pj_status_t) pj_thread_join(pj_thread_t *p)
     PJ_CHECK_STACK();
     PJ_ASSERT_RETURN(p, PJ_EINVAL);
 
+    if (p == pj_thread_this())
+	return PJ_ECANCELLED;
+
     PJ_LOG(6, (pj_thread_this()->obj_name, "Joining thread %s", p->obj_name));
 
     rc = WaitForSingleObject(rec->hthread, INFINITE);
@@ -508,8 +634,9 @@ PJ_DEF(void) pj_thread_check_stack(const char *file, int line)
     pj_assert(thread);
 
     /* Calculate current usage. */
-    usage = (&stk_ptr > thread->stk_start) ? &stk_ptr - thread->stk_start :
-		thread->stk_start - &stk_ptr;
+    usage = (&stk_ptr > thread->stk_start) ? 
+		(pj_uint32_t)(&stk_ptr - thread->stk_start) :
+		(pj_uint32_t)(thread->stk_start - &stk_ptr);
 
     /* Assert if stack usage is dangerously high. */
     pj_assert("STACK OVERFLOW!! " && (usage <= thread->stk_size - 128));
@@ -818,7 +945,7 @@ PJ_DEF(pj_status_t) pj_mutex_lock(pj_mutex_t *mutex)
     PJ_CHECK_STACK();
     PJ_ASSERT_RETURN(mutex, PJ_EINVAL);
 
-    PJ_LOG(6,(mutex->obj_name, "Mutex: thread %s is waiting", 
+    LOG_MUTEX((mutex->obj_name, "Mutex: thread %s is waiting", 
 				pj_thread_this()->obj_name));
 
 #if PJ_WIN32_WINNT >= 0x0400
@@ -831,7 +958,7 @@ PJ_DEF(pj_status_t) pj_mutex_lock(pj_mutex_t *mutex)
         status = PJ_STATUS_FROM_OS(GetLastError());
 
 #endif
-    PJ_LOG(6,(mutex->obj_name, 
+    LOG_MUTEX((mutex->obj_name, 
 	      (status==PJ_SUCCESS ? "Mutex acquired by thread %s" : "FAILED by %s"),
 	      pj_thread_this()->obj_name));
 
@@ -862,7 +989,7 @@ PJ_DEF(pj_status_t) pj_mutex_unlock(pj_mutex_t *mutex)
     }
 #endif
 
-    PJ_LOG(6,(mutex->obj_name, "Mutex released by thread %s", 
+    LOG_MUTEX((mutex->obj_name, "Mutex released by thread %s", 
 				pj_thread_this()->obj_name));
 
 #if PJ_WIN32_WINNT >= 0x0400
@@ -885,7 +1012,7 @@ PJ_DEF(pj_status_t) pj_mutex_trylock(pj_mutex_t *mutex)
     PJ_CHECK_STACK();
     PJ_ASSERT_RETURN(mutex, PJ_EINVAL);
 
-    PJ_LOG(6,(mutex->obj_name, "Mutex: thread %s is trying", 
+    LOG_MUTEX((mutex->obj_name, "Mutex: thread %s is trying", 
 				pj_thread_this()->obj_name));
 
 #if PJ_WIN32_WINNT >= 0x0400
@@ -895,7 +1022,7 @@ PJ_DEF(pj_status_t) pj_mutex_trylock(pj_mutex_t *mutex)
                 PJ_SUCCESS : PJ_ETIMEDOUT;
 #endif
     if (status==PJ_SUCCESS) {
-	PJ_LOG(6,(mutex->obj_name, "Mutex acquired by thread %s", 
+	LOG_MUTEX((mutex->obj_name, "Mutex acquired by thread %s", 
 				  pj_thread_this()->obj_name));
 
 #if PJ_DEBUG
@@ -903,7 +1030,7 @@ PJ_DEF(pj_status_t) pj_mutex_trylock(pj_mutex_t *mutex)
 	++mutex->nesting_level;
 #endif
     } else {
-	PJ_LOG(6,(mutex->obj_name, "Mutex: thread %s's trylock() failed", 
+	LOG_MUTEX((mutex->obj_name, "Mutex: thread %s's trylock() failed", 
 				    pj_thread_this()->obj_name));
     }
 
@@ -918,7 +1045,7 @@ PJ_DEF(pj_status_t) pj_mutex_destroy(pj_mutex_t *mutex)
     PJ_CHECK_STACK();
     PJ_ASSERT_RETURN(mutex, PJ_EINVAL);
 
-    PJ_LOG(6,(mutex->obj_name, "Mutex destroyed"));
+    LOG_MUTEX((mutex->obj_name, "Mutex destroyed"));
 
 #if PJ_WIN32_WINNT >= 0x0400
     DeleteCriticalSection(&mutex->crit);
@@ -929,15 +1056,19 @@ PJ_DEF(pj_status_t) pj_mutex_destroy(pj_mutex_t *mutex)
 #endif
 }
 
-#if PJ_DEBUG
 /*
  * pj_mutex_is_locked()
  */
 PJ_DEF(pj_bool_t) pj_mutex_is_locked(pj_mutex_t *mutex)
 {
+#if PJ_DEBUG
     return mutex->owner == pj_thread_this();
-}
+#else
+    PJ_UNUSED_ARG(mutex);
+    pj_assert(!"PJ_DEBUG is not set!");
+    return 1;
 #endif
+}
 
 ///////////////////////////////////////////////////////////////////////////////
 /*
@@ -996,7 +1127,7 @@ PJ_DEF(pj_status_t) pj_sem_create( pj_pool_t *pool,
 	sem->obj_name[PJ_MAX_OBJ_NAME-1] = '\0';
     }
 
-    PJ_LOG(6, (sem->obj_name, "Semaphore created"));
+    LOG_MUTEX((sem->obj_name, "Semaphore created"));
 
     *sem_ptr = sem;
     return PJ_SUCCESS;
@@ -1009,15 +1140,15 @@ static pj_status_t pj_sem_wait_for(pj_sem_t *sem, unsigned timeout)
     PJ_CHECK_STACK();
     PJ_ASSERT_RETURN(sem, PJ_EINVAL);
 
-    PJ_LOG(6, (sem->obj_name, "Semaphore: thread %s is waiting", 
+    LOG_MUTEX((sem->obj_name, "Semaphore: thread %s is waiting", 
 			      pj_thread_this()->obj_name));
 
     result = WaitForSingleObject(sem->hSemaphore, timeout);
     if (result == WAIT_OBJECT_0) {
-	PJ_LOG(6, (sem->obj_name, "Semaphore acquired by thread %s", 
+	LOG_MUTEX((sem->obj_name, "Semaphore acquired by thread %s", 
 				  pj_thread_this()->obj_name));
     } else {
-	PJ_LOG(6, (sem->obj_name, "Semaphore: thread %s FAILED to acquire", 
+	LOG_MUTEX((sem->obj_name, "Semaphore: thread %s FAILED to acquire", 
 				  pj_thread_this()->obj_name));
     }
 
@@ -1059,7 +1190,7 @@ PJ_DEF(pj_status_t) pj_sem_post(pj_sem_t *sem)
     PJ_CHECK_STACK();
     PJ_ASSERT_RETURN(sem, PJ_EINVAL);
 
-    PJ_LOG(6, (sem->obj_name, "Semaphore released by thread %s",
+    LOG_MUTEX((sem->obj_name, "Semaphore released by thread %s",
 			      pj_thread_this()->obj_name));
 
     if (ReleaseSemaphore(sem->hSemaphore, 1, NULL))
@@ -1076,7 +1207,7 @@ PJ_DEF(pj_status_t) pj_sem_destroy(pj_sem_t *sem)
     PJ_CHECK_STACK();
     PJ_ASSERT_RETURN(sem, PJ_EINVAL);
 
-    PJ_LOG(6, (sem->obj_name, "Semaphore destroyed by thread %s",
+    LOG_MUTEX((sem->obj_name, "Semaphore destroyed by thread %s",
 			      pj_thread_this()->obj_name));
 
     if (CloseHandle(sem->hSemaphore))
@@ -1314,3 +1445,13 @@ PJ_DEF(pj_color_t) pj_term_get_color(void)
 }
 
 #endif	/* PJ_TERM_HAS_COLOR */
+
+/*
+ * pj_run_app()
+ */
+PJ_DEF(int) pj_run_app(pj_main_func_ptr main_func, int argc, char *argv[],
+                       unsigned flags)
+{
+    PJ_UNUSED_ARG(flags);
+    return (*main_func)(argc, argv);
+}
