@@ -1,6 +1,7 @@
-/* $Id: sock_linux_kernel.c 974 2007-02-19 01:13:53Z bennylp $ */
+/* $Id$ */
 /* 
- * Copyright (C)2003-2007 Benny Prijono <benny@prijono.org>
+ * Copyright (C) 2008-2011 Teluu Inc. (http://www.teluu.com)
+ * Copyright (C) 2003-2008 Benny Prijono <benny@prijono.org>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -35,6 +36,8 @@
 #include <asm/ioctls.h>		/* FIONBIO	*/
 #include <linux/utsname.h>	/* for pj_gethostname() */
 
+#define THIS_FILE   "sock_linux_kernel.c"
+
 /*
  * Address families conversion.
  * The values here are indexed based on pj_addr_family-0xFF00.
@@ -624,6 +627,34 @@ PJ_DEF(pj_status_t) pj_sock_setsockopt( pj_sock_t sockfd,
 	return PJ_SUCCESS;
 }
 
+/*
+ * Set socket option.
+ */
+PJ_DEF(pj_status_t) pj_sock_setsockopt_params( pj_sock_t sockfd,
+					       const pj_sockopt_params *params)
+{
+    unsigned int i = 0;
+    pj_status_t retval = PJ_SUCCESS;
+    PJ_CHECK_STACK();
+    PJ_ASSERT_RETURN(params, PJ_EINVAL);
+    
+    for (;i<params->cnt && i<PJ_MAX_SOCKOPT_PARAMS;++i) {
+	pj_status_t status = pj_sock_setsockopt(sockfd, 
+						params->options[i].level,
+						params->options[i].optname,
+						params->options[i].optval, 
+						params->options[i].optlen);
+	if (status != PJ_SUCCESS) {
+	    retval = status;
+	    PJ_PERROR(4,(THIS_FILE, status,
+			 "Warning: error applying sock opt %d",
+			 params->options[i].optname));
+	}
+    }
+
+    return retval;
+}
+
 /*
  * Shutdown socket.
  */
