@@ -1,6 +1,7 @@
-/* $Id: ioq_udp.c 1235 2007-04-30 21:03:32Z bennylp $ */
+/* $Id$ */
 /* 
- * Copyright (C)2003-2007 Benny Prijono <benny@prijono.org>
+ * Copyright (C) 2008-2011 Teluu Inc. (http://www.teluu.com)
+ * Copyright (C) 2003-2008 Benny Prijono <benny@prijono.org>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -114,7 +115,7 @@ static pj_ioqueue_callback test_cb =
     &on_ioqueue_connect,
 };
 
-#ifdef PJ_WIN32
+#if defined(PJ_WIN32) || defined(PJ_WIN64)
 #  define S_ADDR S_un.S_addr
 #else
 #  define S_ADDR s_addr
@@ -125,7 +126,7 @@ static pj_ioqueue_callback test_cb =
  * To test that the basic IOQueue functionality works. It will just exchange
  * data between two sockets.
  */ 
-static int compliance_test(void)
+static int compliance_test(pj_bool_t allow_concur)
 {
     pj_sock_t ssock=-1, csock=-1;
     pj_sockaddr_in addr, dst_addr;
@@ -133,10 +134,11 @@ static int compliance_test(void)
     pj_pool_t *pool = NULL;
     char *send_buf, *recv_buf;
     pj_ioqueue_t *ioque = NULL;
-    pj_ioqueue_key_t *skey, *ckey;
+    pj_ioqueue_key_t *skey = NULL, *ckey = NULL;
     pj_ioqueue_op_key_t read_op, write_op;
     int bufsize = BUF_MIN_SIZE;
-    pj_ssize_t bytes, status = -1;
+    pj_ssize_t bytes;
+    int status = -1;
     pj_str_t temp;
     pj_bool_t send_pending, recv_pending;
     pj_status_t rc;
@@ -152,9 +154,9 @@ static int compliance_test(void)
 
     // Allocate sockets for sending and receiving.
     TRACE_("creating sockets...");
-    rc = pj_sock_socket(PJ_AF_INET, PJ_SOCK_DGRAM, 0, &ssock);
+    rc = pj_sock_socket(pj_AF_INET(), pj_SOCK_DGRAM(), 0, &ssock);
     if (rc==PJ_SUCCESS)
-        rc = pj_sock_socket(PJ_AF_INET, PJ_SOCK_DGRAM, 0, &csock);
+        rc = pj_sock_socket(pj_AF_INET(), pj_SOCK_DGRAM(), 0, &csock);
     else
         csock = PJ_INVALID_SOCKET;
     if (rc != PJ_SUCCESS) {
@@ -165,7 +167,7 @@ static int compliance_test(void)
     // Bind server socket.
     TRACE_("bind socket...");
     pj_bzero(&addr, sizeof(addr));
-    addr.sin_family = PJ_AF_INET;
+    addr.sin_family = pj_AF_INET();
     addr.sin_port = pj_htons(PORT);
     if (pj_sock_bind(ssock, &addr, sizeof(addr))) {
 	status=-10; goto on_error;
@@ -178,6 +180,13 @@ static int compliance_test(void)
 	status=-20; goto on_error;
     }
 
+    // Set concurrency
+    TRACE_("set concurrency...");
+    rc = pj_ioqueue_set_default_concurrency(ioque, allow_concur);
+    if (rc != PJ_SUCCESS) {
+	status=-21; goto on_error;
+    }
+
     // Register server and client socket.
     // We put this after inactivity socket, hopefully this can represent the
     // worst waiting time.
@@ -258,7 +267,11 @@ static int compliance_test(void)
 	pj_time_val timeout = { 5, 0 };
 
 	TRACE_("poll...");
+#ifdef PJ_SYMBIAN
+	rc = pj_symbianos_poll(-1, PJ_TIME_VAL_MSEC(timeout));
+#else
 	rc = pj_ioqueue_poll(ioque, &timeout);
+#endif
 
 	if (rc == 0) {
 	    PJ_LOG(1,(THIS_FILE, "...ERROR: timed out..."));
@@ -285,7 +298,7 @@ static int compliance_test(void)
 	    if (addrlen != sizeof(pj_sockaddr_in)) {
 		status=-68; goto on_error;
 	    }
-	    if (addr.sin_family != PJ_AF_INET) {
+	    if (addr.sin_family != pj_AF_INET()) {
 		status=-69; goto on_error;
 	    }
 
@@ -312,10 +325,16 @@ static int compliance_test(void)
     status = 0;
 
 on_error:
-    if (ssock)
+    if (skey)
+    	pj_ioqueue_unregister(skey);
+    else if (ssock != -1)
 	pj_sock_close(ssock);
-    if (csock)
+    
+    if (ckey)
+    	pj_ioqueue_unregister(ckey);
+    else if (csock != -1)
 	pj_sock_close(csock);
+    
     if (ioque != NULL)
 	pj_ioqueue_destroy(ioque);
     pj_pool_release(pool);
@@ -341,7 +360,7 @@ static void on_read_complete(pj_ioqueue_key_t *key,
  * Check if callback is still called after socket has been unregistered or 
  * closed.
  */ 
-static int unregister_test(void)
+static int unregister_test(pj_bool_t allow_concur)
 {
     enum { RPORT = 50000, SPORT = 50001 };
     pj_pool_t *pool;
@@ -371,15 +390,22 @@ static int unregister_test(void)
 	return -110;
     }
 
+    // Set concurrency
+    TRACE_("set concurrency...");
+    status = pj_ioqueue_set_default_concurrency(ioqueue, allow_concur);
+    if (status != PJ_SUCCESS) {
+	return -112;
+    }
+
     /* Create sender socket */
-    status = app_socket(PJ_AF_INET, PJ_SOCK_DGRAM, 0, SPORT, &ssock);
+    status = app_socket(pj_AF_INET(), pj_SOCK_DGRAM(), 0, SPORT, &ssock);
     if (status != PJ_SUCCESS) {
 	app_perror("Error initializing socket", status);
 	return -120;
     }
 
     /* Create receiver socket. */
-    status = app_socket(PJ_AF_INET, PJ_SOCK_DGRAM, 0, RPORT, &rsock);
+    status = app_socket(pj_AF_INET(), pj_SOCK_DGRAM(), 0, RPORT, &rsock);
     if (status != PJ_SUCCESS) {
 	app_perror("Error initializing socket", status);
 	return -130;
@@ -435,7 +461,11 @@ static int unregister_test(void)
 
     /* Check if packet is received. */
     timeout.sec = 1; timeout.msec = 0;
+#ifdef PJ_SYMBIAN
+    pj_symbianos_poll(-1, 1000);
+#else
     pj_ioqueue_poll(ioqueue, &timeout);
+#endif
 
     if (packet_cnt != 1) {
 	return -180;
@@ -469,8 +499,12 @@ static int unregister_test(void)
     pj_ioqueue_unregister(key);
 
     /* Poll ioqueue. */
+#ifdef PJ_SYMBIAN
+    pj_symbianos_poll(-1, 1000);
+#else
     timeout.sec = 1; timeout.msec = 0;
     pj_ioqueue_poll(ioqueue, &timeout);
+#endif
 
     /* Must NOT receive any packets after socket is closed! */
     if (packet_cnt > 0) {
@@ -494,7 +528,7 @@ static int unregister_test(void)
  * This will just test registering PJ_IOQUEUE_MAX_HANDLES count
  * of sockets to the ioqueue.
  */
-static int many_handles_test(void)
+static int many_handles_test(pj_bool_t allow_concur)
 {
     enum { MAX = PJ_IOQUEUE_MAX_HANDLES };
     pj_pool_t *pool;
@@ -521,10 +555,16 @@ static int many_handles_test(void)
 	return -10;
     }
 
+    // Set concurrency
+    rc = pj_ioqueue_set_default_concurrency(ioqueue, allow_concur);
+    if (rc != PJ_SUCCESS) {
+	return -11;
+    }
+
     /* Register as many sockets. */
     for (count=0; count<MAX; ++count) {
 	sock[count] = PJ_INVALID_SOCKET;
-	rc = pj_sock_socket(PJ_AF_INET, PJ_SOCK_DGRAM, 0, &sock[count]);
+	rc = pj_sock_socket(pj_AF_INET(), pj_SOCK_DGRAM(), 0, &sock[count]);
 	if (rc != PJ_SUCCESS || sock[count] == PJ_INVALID_SOCKET) {
 	    PJ_LOG(3,(THIS_FILE, "....unable to create %d-th socket, rc=%d", 
 				 count, rc));
@@ -582,7 +622,8 @@ static int many_handles_test(void)
 /*
  * Benchmarking IOQueue
  */
-static int bench_test(int bufsize, int inactive_sock_count)
+static int bench_test(pj_bool_t allow_concur, int bufsize, 
+		      int inactive_sock_count)
 {
     pj_sock_t ssock=-1, csock=-1;
     pj_sockaddr_in addr;
@@ -591,7 +632,7 @@ static int bench_test(int bufsize, int inactive_sock_count)
     pj_ioqueue_op_key_t *inactive_read_op;
     char *send_buf, *recv_buf;
     pj_ioqueue_t *ioque = NULL;
-    pj_ioqueue_key_t *skey, *ckey, *key;
+    pj_ioqueue_key_t *skey, *ckey, *keys[SOCK_INACTIVE_MAX+2];
     pj_timestamp t1, t2, t_elapsed;
     int rc=0, i;    /* i must be signed */
     pj_str_t temp;
@@ -607,9 +648,9 @@ static int bench_test(int bufsize, int inactive_sock_count)
     recv_buf = (char*)pj_pool_alloc(pool, bufsize);
 
     // Allocate sockets for sending and receiving.
-    rc = pj_sock_socket(PJ_AF_INET, PJ_SOCK_DGRAM, 0, &ssock);
+    rc = pj_sock_socket(pj_AF_INET(), pj_SOCK_DGRAM(), 0, &ssock);
     if (rc == PJ_SUCCESS) {
-        rc = pj_sock_socket(PJ_AF_INET, PJ_SOCK_DGRAM, 0, &csock);
+        rc = pj_sock_socket(pj_AF_INET(), pj_SOCK_DGRAM(), 0, &csock);
     } else
         csock = PJ_INVALID_SOCKET;
     if (rc != PJ_SUCCESS) {
@@ -619,7 +660,7 @@ static int bench_test(int bufsize, int inactive_sock_count)
 
     // Bind server socket.
     pj_bzero(&addr, sizeof(addr));
-    addr.sin_family = PJ_AF_INET;
+    addr.sin_family = pj_AF_INET();
     addr.sin_port = pj_htons(PORT);
     if (pj_sock_bind(ssock, &addr, sizeof(addr)))
 	goto on_error;
@@ -633,6 +674,13 @@ static int bench_test(int bufsize, int inactive_sock_count)
 	goto on_error;
     }
 
+    // Set concurrency
+    rc = pj_ioqueue_set_default_concurrency(ioque, allow_concur);
+    if (rc != PJ_SUCCESS) {
+	app_perror("...error: pj_ioqueue_set_default_concurrency()", rc);
+	goto on_error;
+    }
+
     // Allocate inactive sockets, and bind them to some arbitrary address.
     // Then register them to the I/O queue, and start a read operation.
     inactive_sock = (pj_sock_t*)pj_pool_alloc(pool, 
@@ -640,11 +688,11 @@ static int bench_test(int bufsize, int inactive_sock_count)
     inactive_read_op = (pj_ioqueue_op_key_t*)pj_pool_alloc(pool,
                               inactive_sock_count*sizeof(pj_ioqueue_op_key_t));
     pj_bzero(&addr, sizeof(addr));
-    addr.sin_family = PJ_AF_INET;
+    addr.sin_family = pj_AF_INET();
     for (i=0; i<inactive_sock_count; ++i) {
         pj_ssize_t bytes;
 
-	rc = pj_sock_socket(PJ_AF_INET, PJ_SOCK_DGRAM, 0, &inactive_sock[i]);
+	rc = pj_sock_socket(pj_AF_INET(), pj_SOCK_DGRAM(), 0, &inactive_sock[i]);
 	if (rc != PJ_SUCCESS || inactive_sock[i] < 0) {
 	    app_perror("...error: pj_sock_socket()", rc);
 	    goto on_error;
@@ -656,7 +704,7 @@ static int bench_test(int bufsize, int inactive_sock_count)
 	    goto on_error;
 	}
 	rc = pj_ioqueue_register_sock(pool, ioque, inactive_sock[i], 
-			              NULL, &test_cb, &key);
+			              NULL, &test_cb, &keys[i]);
 	if (rc != PJ_SUCCESS) {
 	    pj_sock_close(inactive_sock[i]);
 	    inactive_sock[i] = PJ_INVALID_SOCKET;
@@ -665,7 +713,7 @@ static int bench_test(int bufsize, int inactive_sock_count)
 	    goto on_error;
 	}
         bytes = bufsize;
-	rc = pj_ioqueue_recv(key, &inactive_read_op[i], recv_buf, &bytes, 0);
+	rc = pj_ioqueue_recv(keys[i], &inactive_read_op[i], recv_buf, &bytes, 0);
 	if (rc != PJ_EPENDING) {
 	    pj_sock_close(inactive_sock[i]);
 	    inactive_sock[i] = PJ_INVALID_SOCKET;
@@ -721,7 +769,7 @@ static int bench_test(int bufsize, int inactive_sock_count)
 	}
 	if (rc == PJ_SUCCESS) {
 	    if (bytes < 0) {
-		app_perror("...error: pj_ioqueue_sendto()", -bytes);
+		app_perror("...error: pj_ioqueue_sendto()",(pj_status_t)-bytes);
 		break;
 	    }
 	}
@@ -735,7 +783,11 @@ static int bench_test(int bufsize, int inactive_sock_count)
 	TRACE__((THIS_FILE, "     waiting for key = %p", skey));
 	do {
 	    pj_time_val timeout = { 1, 0 };
+#ifdef PJ_SYMBIAN
+	    rc = pj_symbianos_poll(-1, PJ_TIME_VAL_MSEC(timeout));
+#else
 	    rc = pj_ioqueue_poll(ioque, &timeout);
+#endif
 	    TRACE__((THIS_FILE, "     poll rc=%d", rc));
 	} while (rc >= 0 && callback_read_key != skey);
 
@@ -760,7 +812,12 @@ static int bench_test(int bufsize, int inactive_sock_count)
 	// Poll until all events are exhausted, before we start the next loop.
 	do {
 	    pj_time_val timeout = { 0, 10 };
+#ifdef PJ_SYMBIAN
+	    PJ_UNUSED_ARG(timeout);
+	    rc = pj_symbianos_poll(-1, 100);
+#else	    
 	    rc = pj_ioqueue_poll(ioque, &timeout);
+#endif
 	} while (rc>0);
 
 	rc = 0;
@@ -784,11 +841,11 @@ static int bench_test(int bufsize, int inactive_sock_count)
 
     // Cleaning up.
     for (i=inactive_sock_count-1; i>=0; --i) {
-	pj_sock_close(inactive_sock[i]);
+	pj_ioqueue_unregister(keys[i]);
     }
 
-    pj_sock_close(ssock);
-    pj_sock_close(csock);
+    pj_ioqueue_unregister(skey);
+    pj_ioqueue_unregister(ckey);
 
 
     pj_ioqueue_destroy(ioque);
@@ -813,27 +870,29 @@ on_error:
     return -1;
 }
 
-int udp_ioqueue_test()
+static int udp_ioqueue_test_imp(pj_bool_t allow_concur)
 {
     int status;
     int bufsize, sock_count;
 
+    PJ_LOG(3,(THIS_FILE, "..testing with concurency=%d", allow_concur));
+
     //goto pass1;
 
     PJ_LOG(3, (THIS_FILE, "...compliance test (%s)", pj_ioqueue_name()));
-    if ((status=compliance_test()) != 0) {
+    if ((status=compliance_test(allow_concur)) != 0) {
 	return status;
     }
     PJ_LOG(3, (THIS_FILE, "....compliance test ok"));
 
 
     PJ_LOG(3, (THIS_FILE, "...unregister test (%s)", pj_ioqueue_name()));
-    if ((status=unregister_test()) != 0) {
+    if ((status=unregister_test(allow_concur)) != 0) {
 	return status;
     }
     PJ_LOG(3, (THIS_FILE, "....unregister test ok"));
 
-    if ((status=many_handles_test()) != 0) {
+    if ((status=many_handles_test(allow_concur)) != 0) {
 	return status;
     }
     
@@ -853,7 +912,7 @@ int udp_ioqueue_test()
     //goto pass2;
 
     for (bufsize=BUF_MIN_SIZE; bufsize <= BUF_MAX_SIZE; bufsize *= 2) {
-	if ((status=bench_test(bufsize, SOCK_INACTIVE_MIN)) != 0)
+	if ((status=bench_test(allow_concur, bufsize, SOCK_INACTIVE_MIN)) != 0)
 	    return status;
     }
 //pass2:
@@ -863,12 +922,27 @@ int udp_ioqueue_test()
 	 sock_count *= 2) 
     {
 	//PJ_LOG(3,(THIS_FILE, "...testing with %d fds", sock_count));
-	if ((status=bench_test(bufsize, sock_count-2)) != 0)
+	if ((status=bench_test(allow_concur, bufsize, sock_count-2)) != 0)
 	    return status;
     }
     return 0;
 }
 
+int udp_ioqueue_test()
+{
+    int rc;
+
+    rc = udp_ioqueue_test_imp(PJ_TRUE);
+    if (rc != 0)
+	return rc;
+
+    rc = udp_ioqueue_test_imp(PJ_FALSE);
+    if (rc != 0)
+	return rc;
+
+    return 0;
+}
+
 #else
 /* To prevent warning about "translation unit is empty"
  * when this test is disabled. 
