@@ -1,6 +1,7 @@
-/* $Id: ilbc.c 1266 2007-05-11 15:14:34Z bennylp $ */
+/* $Id$ */
 /* 
- * Copyright (C)2003-2007 Benny Prijono <benny@prijono.org>
+ * Copyright (C) 2008-2011 Teluu Inc. (http://www.teluu.com)
+ * Copyright (C) 2003-2008 Benny Prijono <benny@prijono.org>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -29,9 +30,16 @@
 #include <pj/pool.h>
 #include <pj/string.h>
 #include <pj/os.h>
-#include "../../third_party/ilbc/iLBC_encode.h"
-#include "../../third_party/ilbc/iLBC_decode.h"
 
+#if defined(PJMEDIA_ILBC_CODEC_USE_COREAUDIO)&& PJMEDIA_ILBC_CODEC_USE_COREAUDIO
+    #include <AudioToolbox/AudioToolbox.h>
+    #define iLBC_Enc_Inst_t AudioConverterRef
+    #define iLBC_Dec_Inst_t AudioConverterRef
+    #define BLOCKL_MAX 		1
+#else
+    #include "../../third_party/ilbc/iLBC_encode.h"
+    #include "../../third_party/ilbc/iLBC_decode.h"
+#endif
 
 /*
  * Only build this file if PJMEDIA_HAS_ILBC_CODEC != 0
@@ -105,7 +113,8 @@ static pjmedia_codec_factory_op ilbc_factory_op =
     &ilbc_default_attr,
     &ilbc_enum_codecs,
     &ilbc_alloc_codec,
-    &ilbc_dealloc_codec
+    &ilbc_dealloc_codec,
+    &pjmedia_codec_ilbc_deinit
 };
 
 /* iLBC factory */
@@ -130,6 +139,7 @@ struct ilbc_codec
     pj_bool_t		 plc_enabled;
     pj_timestamp	 last_tx;
 
+
     pj_bool_t		 enc_ready;
     iLBC_Enc_Inst_t	 enc;
     unsigned		 enc_frame_size;
@@ -141,9 +151,19 @@ struct ilbc_codec
     unsigned		 dec_frame_size;
     unsigned		 dec_samples_per_frame;
     float		 dec_block[BLOCKL_MAX];
+
+#if defined(PJMEDIA_ILBC_CODEC_USE_COREAUDIO)&& PJMEDIA_ILBC_CODEC_USE_COREAUDIO
+    unsigned		 enc_total_packets;
+    char		 *enc_buffer;
+    unsigned		 enc_buffer_offset;
+
+    unsigned		 dec_total_packets;
+    char		 *dec_buffer;
+    unsigned		 dec_buffer_offset;
+#endif
 };
 
-
+static pj_str_t STR_MODE = {"mode", 4};
 
 /*
  * Initialize and register iLBC codec factory to pjmedia endpoint.
@@ -264,6 +284,7 @@ static pj_status_t ilbc_default_attr (pjmedia_codec_factory *factory,
     attr->info.clock_rate = CLOCK_RATE;
     attr->info.channel_cnt = 1;
     attr->info.avg_bps = ilbc_factory.bps;
+    attr->info.max_bps = 15200;
     attr->info.pcm_bits_per_sample = 16;
     attr->info.frm_ptime = (short)ilbc_factory.mode;
     attr->info.pt = PJMEDIA_RTP_PT_ILBC;
@@ -272,7 +293,12 @@ static pj_status_t ilbc_default_attr (pjmedia_codec_factory *factory,
     attr->setting.vad = 1;
     attr->setting.plc = 1;
     attr->setting.penh = 1;
-    attr->setting.dec_fmtp_mode = (pj_uint8_t)ilbc_factory.mode;
+    attr->setting.dec_fmtp.cnt = 1;
+    attr->setting.dec_fmtp.param[0].name = STR_MODE;
+    if (ilbc_factory.mode == 30)
+	attr->setting.dec_fmtp.param[0].val = pj_str("30");
+    else
+	attr->setting.dec_fmtp.param[0].val = pj_str("20");
 
     return PJ_SUCCESS;
 }
@@ -345,6 +371,18 @@ static pj_status_t ilbc_dealloc_codec( pjmedia_codec_factory *factory,
     PJ_ASSERT_RETURN(factory == &ilbc_factory.base, PJ_EINVAL);
 
     ilbc_codec = (struct ilbc_codec*) codec;
+
+#if defined(PJMEDIA_ILBC_CODEC_USE_COREAUDIO)&& PJMEDIA_ILBC_CODEC_USE_COREAUDIO
+    if (ilbc_codec->enc) {
+	AudioConverterDispose(ilbc_codec->enc);
+	ilbc_codec->enc = NULL;
+    }
+    if (ilbc_codec->dec) {
+	AudioConverterDispose(ilbc_codec->dec);
+	ilbc_codec->dec = NULL;
+    }
+#endif
+
     pj_pool_release(ilbc_codec->pool);
 
     return PJ_SUCCESS;
@@ -369,51 +407,108 @@ static pj_status_t ilbc_codec_open(pjmedia_codec *codec,
 {
     struct ilbc_codec *ilbc_codec = (struct ilbc_codec*)codec;
     pj_status_t status;
+    unsigned i;
+    pj_uint16_t dec_fmtp_mode = DEFAULT_MODE, 
+		enc_fmtp_mode = DEFAULT_MODE;
+
+#if defined(PJMEDIA_ILBC_CODEC_USE_COREAUDIO)&& PJMEDIA_ILBC_CODEC_USE_COREAUDIO
+    AudioStreamBasicDescription srcFormat, dstFormat;
+    UInt32 size;
+
+    srcFormat.mSampleRate       = attr->info.clock_rate;
+    srcFormat.mFormatID         = kAudioFormatLinearPCM;
+    srcFormat.mFormatFlags      = kLinearPCMFormatFlagIsSignedInteger
+				  | kLinearPCMFormatFlagIsPacked;
+    srcFormat.mBitsPerChannel   = attr->info.pcm_bits_per_sample;
+    srcFormat.mChannelsPerFrame = attr->info.channel_cnt;
+    srcFormat.mBytesPerFrame    = srcFormat.mChannelsPerFrame
+	                          * srcFormat.mBitsPerChannel >> 3;
+    srcFormat.mFramesPerPacket  = 1;
+    srcFormat.mBytesPerPacket   = srcFormat.mBytesPerFrame *
+				  srcFormat.mFramesPerPacket;
+
+    memset(&dstFormat, 0, sizeof(dstFormat));
+    dstFormat.mSampleRate 	= attr->info.clock_rate;
+    dstFormat.mFormatID 	= kAudioFormatiLBC;
+    dstFormat.mChannelsPerFrame = attr->info.channel_cnt;
+#endif
 
     pj_assert(ilbc_codec != NULL);
     pj_assert(ilbc_codec->enc_ready == PJ_FALSE && 
 	      ilbc_codec->dec_ready == PJ_FALSE);
 
-    /* Decoder mode must be set */
-    PJ_ASSERT_RETURN(attr->setting.dec_fmtp_mode==20 ||
-		     attr->setting.dec_fmtp_mode==30, PJMEDIA_CODEC_EINMODE);
-
-    /* The enc mode must be set in the attribute 
-     * (from the mode parameter in fmtp attribute in the SDP
-     * received from remote) 
-     */
-    if (attr->setting.enc_fmtp_mode == 0)
-	attr->setting.enc_fmtp_mode = attr->setting.dec_fmtp_mode;
-
-    PJ_ASSERT_RETURN(attr->setting.enc_fmtp_mode==20 ||
-		     attr->setting.enc_fmtp_mode==30, PJMEDIA_CODEC_EINMODE);
-
-    /* Update enc_ptime in the param */
-    if (attr->setting.enc_fmtp_mode != attr->setting.dec_fmtp_mode) {
-	attr->info.enc_ptime = attr->setting.enc_fmtp_mode;
-    } else {
-	attr->info.enc_ptime = 0;
+    /* Get decoder mode */
+    for (i = 0; i < attr->setting.dec_fmtp.cnt; ++i) {
+	if (pj_stricmp(&attr->setting.dec_fmtp.param[i].name, &STR_MODE) == 0)
+	{
+	    dec_fmtp_mode = (pj_uint16_t)
+			    pj_strtoul(&attr->setting.dec_fmtp.param[i].val);
+	    break;
+	}
     }
 
-    /* Create enc */
-    ilbc_codec->enc_frame_size = initEncode(&ilbc_codec->enc, 
-					    attr->setting.enc_fmtp_mode);
-    ilbc_codec->enc_samples_per_frame = CLOCK_RATE*attr->setting.enc_fmtp_mode/
-					1000;
+    /* Decoder mode must be set */
+    PJ_ASSERT_RETURN(dec_fmtp_mode == 20 || dec_fmtp_mode == 30, 
+		     PJMEDIA_CODEC_EINMODE);
+
+    /* Get encoder mode */
+    for (i = 0; i < attr->setting.enc_fmtp.cnt; ++i) {
+	if (pj_stricmp(&attr->setting.enc_fmtp.param[i].name, &STR_MODE) == 0)
+	{
+	    enc_fmtp_mode = (pj_uint16_t)
+			    pj_strtoul(&attr->setting.enc_fmtp.param[i].val);
+	    break;
+	}
+    }
+
+    PJ_ASSERT_RETURN(enc_fmtp_mode==20 || enc_fmtp_mode==30, 
+		     PJMEDIA_CODEC_EINMODE);
+
+    /* Both sides of a bi-directional session MUST use the same "mode" value.
+     * In this point, possible values are only 20 or 30, so when encoder and
+     * decoder modes are not same, just use the default mode, it is 30.
+     */
+    if (enc_fmtp_mode != dec_fmtp_mode) {
+	enc_fmtp_mode = dec_fmtp_mode = DEFAULT_MODE;
+	PJ_LOG(4,(ilbc_codec->obj_name, 
+		  "Normalized iLBC encoder and decoder modes to %d", 
+		  DEFAULT_MODE));
+    }
+
+    /* Update some attributes based on negotiated mode. */
+    attr->info.avg_bps = (dec_fmtp_mode == 30? 13333 : 15200);
+    attr->info.frm_ptime = dec_fmtp_mode;
+
+    /* Create encoder */
+#if defined(PJMEDIA_ILBC_CODEC_USE_COREAUDIO)&& PJMEDIA_ILBC_CODEC_USE_COREAUDIO
+    dstFormat.mFramesPerPacket  = CLOCK_RATE * enc_fmtp_mode / 1000;
+    dstFormat.mBytesPerPacket   = (enc_fmtp_mode == 20? 38 : 50);
+
+    /* Use AudioFormat API to fill out the rest of the description */
+    size = sizeof(dstFormat);
+    AudioFormatGetProperty(kAudioFormatProperty_FormatInfo,
+ 	                   0, NULL, &size, &dstFormat);
+
+    if (AudioConverterNew(&srcFormat, &dstFormat, &ilbc_codec->enc) != noErr)
+	return PJMEDIA_CODEC_EFAILED;
+    ilbc_codec->enc_frame_size = (enc_fmtp_mode == 20? 38 : 50);
+#else
+    ilbc_codec->enc_frame_size = initEncode(&ilbc_codec->enc, enc_fmtp_mode);
+#endif
+    ilbc_codec->enc_samples_per_frame = CLOCK_RATE * enc_fmtp_mode / 1000;
     ilbc_codec->enc_ready = PJ_TRUE;
 
     /* Create decoder */
+#if defined(PJMEDIA_ILBC_CODEC_USE_COREAUDIO)&& PJMEDIA_ILBC_CODEC_USE_COREAUDIO
+    if (AudioConverterNew(&dstFormat, &srcFormat, &ilbc_codec->dec) != noErr)
+	return PJMEDIA_CODEC_EFAILED;
+    ilbc_codec->dec_samples_per_frame = CLOCK_RATE * dec_fmtp_mode / 1000;
+#else
     ilbc_codec->dec_samples_per_frame = initDecode(&ilbc_codec->dec,
-						   attr->setting.dec_fmtp_mode,
+						   dec_fmtp_mode,
 						   attr->setting.penh);
-    if (attr->setting.dec_fmtp_mode == 20)
-	ilbc_codec->dec_frame_size = 38;
-    else if (attr->setting.dec_fmtp_mode == 30)
-	ilbc_codec->dec_frame_size = 50;
-    else {
-	pj_assert(!"Invalid iLBC mode");
-	ilbc_codec->dec_frame_size = ilbc_codec->enc_frame_size;
-    }
+#endif
+    ilbc_codec->dec_frame_size = (dec_fmtp_mode == 20? 38 : 50);
     ilbc_codec->dec_ready = PJ_TRUE;
 
     /* Save plc flags */
@@ -433,8 +528,7 @@ static pj_status_t ilbc_codec_open(pjmedia_codec *codec,
     pj_set_timestamp32(&ilbc_codec->last_tx, 0, 0);
 
     PJ_LOG(5,(ilbc_codec->obj_name, 
-	      "iLBC codec opened, encoder mode=%d, decoder mode=%d",
-	      attr->setting.enc_fmtp_mode, attr->setting.dec_fmtp_mode));
+	      "iLBC codec opened, mode=%d", dec_fmtp_mode));
 
     return PJ_SUCCESS;
 }
@@ -489,7 +583,7 @@ static pj_status_t  ilbc_codec_parse( pjmedia_codec *codec,
 	frames[count].buf = pkt;
 	frames[count].size = ilbc_codec->dec_frame_size;
 	frames[count].timestamp.u64 = ts->u64 + count * 
-					  ilbc_codec->dec_samples_per_frame;
+				      ilbc_codec->dec_samples_per_frame;
 
 	pkt = ((char*)pkt) + ilbc_codec->dec_frame_size;
 	pkt_size -= ilbc_codec->dec_frame_size;
@@ -501,6 +595,69 @@ static pj_status_t  ilbc_codec_parse( pjmedia_codec *codec,
     return PJ_SUCCESS;
 }
 
+#if defined(PJMEDIA_ILBC_CODEC_USE_COREAUDIO)&& PJMEDIA_ILBC_CODEC_USE_COREAUDIO
+static OSStatus encodeDataProc (
+    AudioConverterRef             inAudioConverter,
+    UInt32                        *ioNumberDataPackets,
+    AudioBufferList               *ioData,
+    AudioStreamPacketDescription  **outDataPacketDescription,
+    void                          *inUserData
+)
+{
+    struct ilbc_codec *ilbc_codec = (struct ilbc_codec*)inUserData;
+
+    /* Initialize in case of failure */
+    ioData->mBuffers[0].mData = NULL;
+    ioData->mBuffers[0].mDataByteSize = 0;
+
+    if (ilbc_codec->enc_total_packets < *ioNumberDataPackets) {
+	*ioNumberDataPackets = ilbc_codec->enc_total_packets;
+    }
+
+    if (*ioNumberDataPackets) {
+	ioData->mBuffers[0].mData = ilbc_codec->enc_buffer +
+				    ilbc_codec->enc_buffer_offset;
+	ioData->mBuffers[0].mDataByteSize = *ioNumberDataPackets *
+					    ilbc_codec->enc_samples_per_frame
+					    << 1;
+	ilbc_codec->enc_buffer_offset += ioData->mBuffers[0].mDataByteSize;
+    }
+
+    ilbc_codec->enc_total_packets -= *ioNumberDataPackets;
+    return noErr;
+}
+
+static OSStatus decodeDataProc (
+    AudioConverterRef             inAudioConverter,
+    UInt32                        *ioNumberDataPackets,
+    AudioBufferList               *ioData,
+    AudioStreamPacketDescription  **outDataPacketDescription,
+    void                          *inUserData
+)
+{
+    struct ilbc_codec *ilbc_codec = (struct ilbc_codec*)inUserData;
+
+    /* Initialize in case of failure */
+    ioData->mBuffers[0].mData = NULL;
+    ioData->mBuffers[0].mDataByteSize = 0;
+
+    if (ilbc_codec->dec_total_packets < *ioNumberDataPackets) {
+	*ioNumberDataPackets = ilbc_codec->dec_total_packets;
+    }
+
+    if (*ioNumberDataPackets) {
+	ioData->mBuffers[0].mData = ilbc_codec->dec_buffer +
+				    ilbc_codec->dec_buffer_offset;
+	ioData->mBuffers[0].mDataByteSize = *ioNumberDataPackets *
+					    ilbc_codec->dec_frame_size;
+	ilbc_codec->dec_buffer_offset += ioData->mBuffers[0].mDataByteSize;
+    }
+
+    ilbc_codec->dec_total_packets -= *ioNumberDataPackets;
+    return noErr;
+}
+#endif
+
 /*
  * Encode frame.
  */
@@ -510,16 +667,24 @@ static pj_status_t ilbc_codec_encode(pjmedia_codec *codec,
 				     struct pjmedia_frame *output)
 {
     struct ilbc_codec *ilbc_codec = (struct ilbc_codec*)codec;
-    unsigned i;
+    pj_int16_t *pcm_in;
+    pj_size_t nsamples;
+#if defined(PJMEDIA_ILBC_CODEC_USE_COREAUDIO)&& PJMEDIA_ILBC_CODEC_USE_COREAUDIO
+    UInt32 npackets;
+    OSStatus err;
+    AudioBufferList theABL;
+#endif
 
-    pj_assert(ilbc_codec != NULL);
-    PJ_ASSERT_RETURN(input && output, PJ_EINVAL);
+    pj_assert(ilbc_codec && input && output);
 
-    if (output_buf_len < ilbc_codec->enc_frame_size)
-	return PJMEDIA_CODEC_EFRMTOOSHORT;
+    pcm_in = (pj_int16_t*)input->buf;
+    nsamples = input->size >> 1;
 
-    if (input->size != (ilbc_codec->enc_samples_per_frame << 1))
-	return PJMEDIA_CODEC_EPCMFRMINLEN;
+    PJ_ASSERT_RETURN(nsamples % ilbc_codec->enc_samples_per_frame == 0, 
+		     PJMEDIA_CODEC_EPCMFRMINLEN);
+    PJ_ASSERT_RETURN(output_buf_len >= ilbc_codec->enc_frame_size * nsamples /
+		     ilbc_codec->enc_samples_per_frame,
+		     PJMEDIA_CODEC_EFRMTOOSHORT);
 
     /* Detect silence */
     if (ilbc_codec->vad_enabled) {
@@ -534,8 +699,8 @@ static pj_status_t ilbc_codec_encode(pjmedia_codec *codec,
 						(input->size >> 1),
 						NULL);
 	if (is_silence &&
-	    PJMEDIA_CODEC_MAX_SILENCE_PERIOD != -1 &&
-	    silence_period < PJMEDIA_CODEC_MAX_SILENCE_PERIOD)
+	    (PJMEDIA_CODEC_MAX_SILENCE_PERIOD == -1 ||
+	     silence_period < PJMEDIA_CODEC_MAX_SILENCE_PERIOD*8000/1000))
 	{
 	    output->type = PJMEDIA_FRAME_TYPE_NONE;
 	    output->buf = NULL;
@@ -547,18 +712,45 @@ static pj_status_t ilbc_codec_encode(pjmedia_codec *codec,
 	}
     }
 
-    /* Convert to float */
-    for (i=0; i<ilbc_codec->enc_samples_per_frame; ++i) {
-	ilbc_codec->enc_block[i] = (float) (((pj_int16_t*)input->buf)[i]);
-    }
-
     /* Encode */
-    iLBC_encode((unsigned char *)output->buf, 
-		ilbc_codec->enc_block, 
-		&ilbc_codec->enc);
+    output->size = 0;
+#if defined(PJMEDIA_ILBC_CODEC_USE_COREAUDIO)&& PJMEDIA_ILBC_CODEC_USE_COREAUDIO
+    npackets = nsamples / ilbc_codec->enc_samples_per_frame;
+
+    theABL.mNumberBuffers = 1;
+    theABL.mBuffers[0].mNumberChannels = 1;
+    theABL.mBuffers[0].mDataByteSize = output_buf_len;
+    theABL.mBuffers[0].mData = output->buf;
+
+    ilbc_codec->enc_total_packets = npackets;
+    ilbc_codec->enc_buffer = (char *)input->buf;
+    ilbc_codec->enc_buffer_offset = 0;
+
+    err = AudioConverterFillComplexBuffer(ilbc_codec->enc, encodeDataProc,
+					  ilbc_codec, &npackets,
+					  &theABL, NULL);
+    if (err == noErr) {
+	output->size = npackets * ilbc_codec->enc_frame_size;
+    }
+#else
+    while (nsamples >= ilbc_codec->enc_samples_per_frame) {
+	unsigned i;
+	
+	/* Convert to float */
+	for (i=0; i<ilbc_codec->enc_samples_per_frame; ++i) {
+	    ilbc_codec->enc_block[i] = (float) (*pcm_in++);
+	}
+
+	iLBC_encode((unsigned char *)output->buf + output->size, 
+		    ilbc_codec->enc_block, 
+		    &ilbc_codec->enc);
+
+	output->size += ilbc_codec->enc.no_of_bytes;
+	nsamples -= ilbc_codec->enc_samples_per_frame;
+    }
+#endif
 
     output->type = PJMEDIA_FRAME_TYPE_AUDIO;
-    output->size = ilbc_codec->enc.no_of_bytes;
     output->timestamp = input->timestamp;
 
     return PJ_SUCCESS;
@@ -573,7 +765,13 @@ static pj_status_t ilbc_codec_decode(pjmedia_codec *codec,
 				     struct pjmedia_frame *output)
 {
     struct ilbc_codec *ilbc_codec = (struct ilbc_codec*)codec;
+#if defined(PJMEDIA_ILBC_CODEC_USE_COREAUDIO)&& PJMEDIA_ILBC_CODEC_USE_COREAUDIO
+    UInt32 npackets;
+    OSStatus err;
+    AudioBufferList theABL;
+#else
     unsigned i;
+#endif
 
     pj_assert(ilbc_codec != NULL);
     PJ_ASSERT_RETURN(input && output, PJ_EINVAL);
@@ -585,6 +783,26 @@ static pj_status_t ilbc_codec_decode(pjmedia_codec *codec,
 	return PJMEDIA_CODEC_EFRMINLEN;
 
     /* Decode to temporary buffer */
+#if defined(PJMEDIA_ILBC_CODEC_USE_COREAUDIO)&& PJMEDIA_ILBC_CODEC_USE_COREAUDIO
+    npackets = input->size / ilbc_codec->dec_frame_size *
+	       ilbc_codec->dec_samples_per_frame;
+
+    theABL.mNumberBuffers = 1;
+    theABL.mBuffers[0].mNumberChannels = 1;
+    theABL.mBuffers[0].mDataByteSize = output_buf_len;
+    theABL.mBuffers[0].mData = output->buf;
+
+    ilbc_codec->dec_total_packets = npackets;
+    ilbc_codec->dec_buffer = (char *)input->buf;
+    ilbc_codec->dec_buffer_offset = 0;
+
+    err = AudioConverterFillComplexBuffer(ilbc_codec->dec, decodeDataProc,
+					  ilbc_codec, &npackets,
+					  &theABL, NULL);
+    if (err == noErr) {
+	output->size = npackets * (ilbc_codec->dec_samples_per_frame << 1);
+    }
+#else
     iLBC_decode(ilbc_codec->dec_block, (unsigned char*) input->buf,
 		&ilbc_codec->dec, 1);
 
@@ -593,6 +811,8 @@ static pj_status_t ilbc_codec_decode(pjmedia_codec *codec,
 	((short*)output->buf)[i] = (short)ilbc_codec->dec_block[i];
     }
     output->size = (ilbc_codec->dec_samples_per_frame << 1);
+#endif
+
     output->type = PJMEDIA_FRAME_TYPE_AUDIO;
     output->timestamp = input->timestamp;
 
@@ -608,7 +828,13 @@ static pj_status_t  ilbc_codec_recover(pjmedia_codec *codec,
 				      struct pjmedia_frame *output)
 {
     struct ilbc_codec *ilbc_codec = (struct ilbc_codec*)codec;
+#if defined(PJMEDIA_ILBC_CODEC_USE_COREAUDIO)&& PJMEDIA_ILBC_CODEC_USE_COREAUDIO
+    UInt32 npackets;
+    OSStatus err;
+    AudioBufferList theABL;
+#else
     unsigned i;
+#endif
 
     pj_assert(ilbc_codec != NULL);
     PJ_ASSERT_RETURN(output, PJ_EINVAL);
@@ -617,6 +843,29 @@ static pj_status_t  ilbc_codec_recover(pjmedia_codec *codec,
 	return PJMEDIA_CODEC_EPCMTOOSHORT;
 
     /* Decode to temporary buffer */
+#if defined(PJMEDIA_ILBC_CODEC_USE_COREAUDIO)&& PJMEDIA_ILBC_CODEC_USE_COREAUDIO
+    npackets = 1;
+
+    theABL.mNumberBuffers = 1;
+    theABL.mBuffers[0].mNumberChannels = 1;
+    theABL.mBuffers[0].mDataByteSize = output_buf_len;
+    theABL.mBuffers[0].mData = output->buf;
+
+    ilbc_codec->dec_total_packets = npackets;
+    ilbc_codec->dec_buffer_offset = 0;
+    if (ilbc_codec->dec_buffer) {
+	err = AudioConverterFillComplexBuffer(ilbc_codec->dec, decodeDataProc,
+					      ilbc_codec, &npackets,
+					      &theABL, NULL);
+	if (err == noErr) {
+	    output->size = npackets *
+		           (ilbc_codec->dec_samples_per_frame << 1);
+	}
+    } else {
+	output->size = npackets * (ilbc_codec->dec_samples_per_frame << 1);
+	pj_bzero(output->buf, output->size);
+    }
+#else
     iLBC_decode(ilbc_codec->dec_block, NULL, &ilbc_codec->dec, 0);
 
     /* Convert decodec samples from float to short */
@@ -624,6 +873,7 @@ static pj_status_t  ilbc_codec_recover(pjmedia_codec *codec,
 	((short*)output->buf)[i] = (short)ilbc_codec->dec_block[i];
     }
     output->size = (ilbc_codec->dec_samples_per_frame << 1);
+#endif
     output->type = PJMEDIA_FRAME_TYPE_AUDIO;
 
     return PJ_SUCCESS;
@@ -631,4 +881,3 @@ static pj_status_t  ilbc_codec_recover(pjmedia_codec *codec,
 
 
 #endif	/* PJMEDIA_HAS_ILBC_CODEC */
-
