@@ -1,6 +1,7 @@
-/* $Id: sdp.c 1266 2007-05-11 15:14:34Z bennylp $ */
+/* $Id$ */
 /* 
- * Copyright (C) 2003-2007 Benny Prijono <benny@prijono.org>
+ * Copyright (C) 2008-2011 Teluu Inc. (http://www.teluu.com)
+ * Copyright (C) 2003-2008 Benny Prijono <benny@prijono.org>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -33,7 +34,9 @@ enum {
     SKIP_WS = 0,
     SYNTAX_ERROR = 1,
 };
-#define TOKEN		"-.!%*_=`'~"
+// New token definition from RFC 4566 (SDP)
+#define TOKEN		"!#$%&'*+-.^_`{|}~"
+//#define TOKEN		"-.!%*_=`'~"
 //#define TOKEN		"'`-./:?\"#$&*;=@[]^_`{|}+~!"
 #define NTP_OFFSET	((pj_uint32_t)2208988800)
 #define THIS_FILE	"sdp.c"
@@ -56,6 +59,8 @@ static void parse_generic_line(pj_scanner *scanner, pj_str_t *str,
 			       parse_context *ctx);
 static void parse_connection_info(pj_scanner *scanner, pjmedia_sdp_conn *conn,
 				  parse_context *ctx);
+static void parse_bandwidth_info(pj_scanner *scanner, pjmedia_sdp_bandw *bandw,
+				  parse_context *ctx);
 static pjmedia_sdp_attr *parse_attr(pj_pool_t *pool, pj_scanner *scanner,
 				    parse_context *ctx);
 static void parse_media(pj_scanner *scanner, pjmedia_sdp_media *med,
@@ -131,11 +136,10 @@ PJ_DEF(pjmedia_sdp_attr*) pjmedia_sdp_attr_clone(pj_pool_t *pool,
     return attr;
 }
 
-PJ_DEF(pjmedia_sdp_attr*) 
-pjmedia_sdp_attr_find (unsigned count, 
-		       pjmedia_sdp_attr *const attr_array[],
-		       const pj_str_t *name,
-		       const pj_str_t *c_fmt)
+PJ_DEF(pjmedia_sdp_attr*) pjmedia_sdp_attr_find (unsigned count, 
+				       pjmedia_sdp_attr *const attr_array[],
+				       const pj_str_t *name,
+				       const pj_str_t *c_fmt)
 {
     unsigned i;
     unsigned c_pt = 0xFFFF;
@@ -158,11 +162,10 @@ pjmedia_sdp_attr_find (unsigned count,
     return NULL;
 }
 
-PJ_DEF(pjmedia_sdp_attr*) 
-pjmedia_sdp_attr_find2(unsigned count, 
-		       pjmedia_sdp_attr *const attr_array[],
-		       const char *c_name,
-		       const pj_str_t *c_fmt)
+PJ_DEF(pjmedia_sdp_attr*) pjmedia_sdp_attr_find2(unsigned count, 
+				       pjmedia_sdp_attr *const attr_array[],
+				       const char *c_name,
+				       const pj_str_t *c_fmt)
 {
     pj_str_t name;
 
@@ -173,6 +176,7 @@ pjmedia_sdp_attr_find2(unsigned count,
 }
 
 
+
 PJ_DEF(pj_status_t) pjmedia_sdp_attr_add(unsigned *count,
 					 pjmedia_sdp_attr *attr_array[],
 					 pjmedia_sdp_attr *attr)
@@ -250,6 +254,8 @@ PJ_DEF(pj_status_t) pjmedia_sdp_attr_get_rtpmap( const pjmedia_sdp_attr *attr,
 
     PJ_ASSERT_RETURN(attr->value.slen != 0, PJMEDIA_SDP_EINATTR);
 
+    init_sdp_parser();
+
     /* Check if input is null terminated, and null terminate if
      * necessary. Unfortunately this may crash the application if
      * attribute was allocated from a read-only memory location.
@@ -257,7 +263,8 @@ PJ_DEF(pj_status_t) pjmedia_sdp_attr_get_rtpmap( const pjmedia_sdp_attr *attr,
      * null terminated.
      */
     if (attr->value.ptr[attr->value.slen] != 0 &&
-	attr->value.ptr[attr->value.slen] != '\r')
+	attr->value.ptr[attr->value.slen] != '\r' &&
+	attr->value.ptr[attr->value.slen] != '\n')
     {
 	pj_assert(!"Shouldn't happen");
 	term = attr->value.ptr[attr->value.slen];
@@ -368,6 +375,8 @@ PJ_DEF(pj_status_t) pjmedia_sdp_attr_get_rtcp(const pjmedia_sdp_attr *attr,
 
     PJ_ASSERT_RETURN(pj_strcmp2(&attr->name, "rtcp")==0, PJ_EINVALIDOP);
 
+    init_sdp_parser();
+
     /* fmtp BNF:
      *	a=rtcp:<port> [nettype addrtype address]
      */
@@ -412,6 +421,40 @@ PJ_DEF(pj_status_t) pjmedia_sdp_attr_get_rtcp(const pjmedia_sdp_attr *attr,
 }
 
 
+PJ_DEF(pjmedia_sdp_attr*) pjmedia_sdp_attr_create_rtcp(pj_pool_t *pool,
+						       const pj_sockaddr *a)
+{
+    enum {
+	ATTR_LEN = PJ_INET6_ADDRSTRLEN+16
+    };
+    pjmedia_sdp_attr *attr;
+
+    attr = PJ_POOL_ALLOC_T(pool, pjmedia_sdp_attr);
+    attr->name = pj_str("rtcp");
+    attr->value.ptr = (char*) pj_pool_alloc(pool, ATTR_LEN);
+    if (a->addr.sa_family == pj_AF_INET()) {
+	attr->value.slen = 
+	    pj_ansi_snprintf(attr->value.ptr, ATTR_LEN,
+			    "%u IN IP4 %s",
+			    pj_ntohs(a->ipv4.sin_port),
+			    pj_inet_ntoa(a->ipv4.sin_addr));
+    } else if (a->addr.sa_family == pj_AF_INET6()) {
+	char tmp_addr[PJ_INET6_ADDRSTRLEN];
+	attr->value.slen = 
+	    pj_ansi_snprintf(attr->value.ptr, ATTR_LEN,
+			    "%u IN IP6 %s",
+			    pj_sockaddr_get_port(a),
+			    pj_sockaddr_print(a, tmp_addr, 
+					      sizeof(tmp_addr), 0));
+
+    } else {
+	pj_assert(!"Unsupported address family");
+	return NULL;
+    }
+
+    return attr;
+}
+
 
 PJ_DEF(pj_status_t) pjmedia_sdp_attr_to_rtpmap(pj_pool_t *pool,
 					       const pjmedia_sdp_attr *attr,
@@ -460,12 +503,12 @@ PJ_DEF(pj_status_t) pjmedia_sdp_rtpmap_to_attr(pj_pool_t *pool,
 			   (int)rtpmap->param.slen,
 			   rtpmap->param.ptr);
 
-    if (len < 1 || len > (int)sizeof(tempbuf))
+    if (len < 1 || len >= (int)sizeof(tempbuf))
 	return PJMEDIA_SDP_ERTPMAPTOOLONG;
 
     attr->value.slen = len;
-    attr->value.ptr = (char*) pj_pool_alloc(pool, attr->value.slen);
-    pj_memcpy(attr->value.ptr, tempbuf, attr->value.slen);
+    attr->value.ptr = (char*) pj_pool_alloc(pool, attr->value.slen+1);
+    pj_memcpy(attr->value.ptr, tempbuf, attr->value.slen+1);
 
     *p_attr = attr;
     return PJ_SUCCESS;
@@ -483,7 +526,7 @@ static int print_connection_info( pjmedia_sdp_conn *c, char *buf, int len)
 			       c->addr_type.ptr,
 			       (int)c->addr.slen,
 			       c->addr.ptr);
-    if (printed < 1 || printed > len)
+    if (printed < 1 || printed >= len)
 	return -1;
 
     return printed;
@@ -503,6 +546,39 @@ PJ_DEF(pjmedia_sdp_conn*) pjmedia_sdp_conn_clone (pj_pool_t *pool,
     return c;
 }
 
+PJ_DEF(pjmedia_sdp_bandw*)
+pjmedia_sdp_bandw_clone (pj_pool_t *pool, 
+			 const pjmedia_sdp_bandw *rhs)
+{
+    pjmedia_sdp_bandw *b = PJ_POOL_ALLOC_T(pool, pjmedia_sdp_bandw);
+    if (!b) return NULL;
+
+    if (!pj_strdup (pool, &b->modifier, &rhs->modifier)) return NULL;
+    b->value = rhs->value;
+
+    return b;
+}
+
+static pj_ssize_t print_bandw(const pjmedia_sdp_bandw *bandw,
+			      char *buf, pj_size_t len)
+{
+    char *p = buf;
+
+    if ((int)len < bandw->modifier.slen + 10 + 5)
+	return -1;
+
+    *p++ = 'b';
+    *p++ = '=';
+    pj_memcpy(p, bandw->modifier.ptr, bandw->modifier.slen);
+    p += bandw->modifier.slen;
+    *p++ = ':';
+    p += pj_utoa(bandw->value, p);
+
+    *p++ = '\r';
+    *p++ = '\n';
+    return p-buf;
+}
+
 static pj_ssize_t print_attr(const pjmedia_sdp_attr *attr, 
 			     char *buf, pj_size_t len)
 {
@@ -564,7 +640,16 @@ static int print_media_desc( pjmedia_sdp_media *m, char *buf, int len)
 
     /* print connection info, if present. */
     if (m->conn) {
-	printed = print_connection_info(m->conn, p, end-p);
+	printed = print_connection_info(m->conn, p, (int)(end-p));
+	if (printed < 0) {
+	    return -1;
+	}
+	p += printed;
+    }
+    
+    /* print optional bandwidth info. */
+    for (i=0; i<m->bandw_count; ++i) {
+	printed = (int)print_bandw(m->bandw[i], p, end-p);
 	if (printed < 0) {
 	    return -1;
 	}
@@ -573,14 +658,14 @@ static int print_media_desc( pjmedia_sdp_media *m, char *buf, int len)
 
     /* print attributes. */
     for (i=0; i<m->attr_count; ++i) {
-	printed = print_attr(m->attr[i], p, end-p);
+	printed = (int)print_attr(m->attr[i], p, end-p);
 	if (printed < 0) {
 	    return -1;
 	}
 	p += printed;
     }
 
-    return p-buf;
+    return (int)(p-buf);
 }
 
 PJ_DEF(pjmedia_sdp_media*) pjmedia_sdp_media_clone(
@@ -606,6 +691,12 @@ PJ_DEF(pjmedia_sdp_media*) pjmedia_sdp_media_clone(
 	m->conn = NULL;
     }
 
+    m->bandw_count = rhs->bandw_count;
+    for (i=0; i < rhs->bandw_count; ++i) {
+	m->bandw[i] = pjmedia_sdp_bandw_clone (pool, rhs->bandw[i]);
+	PJ_ASSERT_RETURN(m->bandw[i] != NULL, NULL);
+    }
+
     m->attr_count = rhs->attr_count;
     for (i=0; i < rhs->attr_count; ++i) {
 	m->attr[i] = pjmedia_sdp_attr_clone (pool, rhs->attr[i]);
@@ -615,9 +706,9 @@ PJ_DEF(pjmedia_sdp_media*) pjmedia_sdp_media_clone(
     return m;
 }
 
-PJ_DEF(pjmedia_sdp_attr*) 
-pjmedia_sdp_media_find_attr(const pjmedia_sdp_media *m,
-			    const pj_str_t *name, const pj_str_t *fmt)
+PJ_DEF(pjmedia_sdp_attr*) pjmedia_sdp_media_find_attr(
+				const pjmedia_sdp_media *m,
+				const pj_str_t *name, const pj_str_t *fmt)
 {
     PJ_ASSERT_RETURN(m && name, NULL);
     return pjmedia_sdp_attr_find(m->attr_count, m->attr, name, fmt);
@@ -625,9 +716,9 @@ pjmedia_sdp_media_find_attr(const pjmedia_sdp_media *m,
 
 
 
-PJ_DEF(pjmedia_sdp_attr*) 
-pjmedia_sdp_media_find_attr2(const pjmedia_sdp_media *m,
-			     const char *name, const pj_str_t *fmt)
+PJ_DEF(pjmedia_sdp_attr*) pjmedia_sdp_media_find_attr2(
+				const pjmedia_sdp_media *m,
+				const char *name, const pj_str_t *fmt)
 {
     PJ_ASSERT_RETURN(m && name, NULL);
     return pjmedia_sdp_attr_find2(m->attr_count, m->attr, name, fmt);
@@ -640,16 +731,20 @@ PJ_DEF(pj_status_t) pjmedia_sdp_media_add_attr( pjmedia_sdp_media *m,
     return pjmedia_sdp_attr_add(&m->attr_count, m->attr, attr);
 }
 
-PJ_DEF(unsigned) 
-pjmedia_sdp_media_remove_all_attr(pjmedia_sdp_media *m,
-				  const char *name)
+PJ_DEF(pj_status_t) pjmedia_sdp_session_add_attr(pjmedia_sdp_session *s,
+						 pjmedia_sdp_attr *attr)
+{
+    return pjmedia_sdp_attr_add(&s->attr_count, s->attr, attr);
+}
+
+PJ_DEF(unsigned) pjmedia_sdp_media_remove_all_attr(pjmedia_sdp_media *m,
+						   const char *name)
 {
     return pjmedia_sdp_attr_remove_all(&m->attr_count, m->attr, name);
 }
 
-PJ_DEF(pj_status_t)
-pjmedia_sdp_media_remove_attr(pjmedia_sdp_media *m,
-			      pjmedia_sdp_attr *attr)
+PJ_DEF(pj_status_t) pjmedia_sdp_media_remove_attr(pjmedia_sdp_media *m,
+			      			  pjmedia_sdp_attr *attr)
 {
     return pjmedia_sdp_attr_remove(&m->attr_count, m->attr, attr);
 }
@@ -710,13 +805,21 @@ static int print_session(const pjmedia_sdp_session *ses,
 
     /* Connection line (c=) if exist. */
     if (ses->conn) {
-	printed = print_connection_info(ses->conn, p, end-p);
+	printed = print_connection_info(ses->conn, p, (int)(end-p));
 	if (printed < 1) {
 	    return -1;
 	}
 	p += printed;
     }
 
+    /* print optional bandwidth info. */
+    for (i=0; i<ses->bandw_count; ++i) {
+	printed = (int)print_bandw(ses->bandw[i], p, end-p);
+	if (printed < 1) {
+	    return -1;
+	}
+	p += printed;
+    }
 
     /* Time */
     if ((end-p) < 24) {
@@ -734,7 +837,7 @@ static int print_session(const pjmedia_sdp_session *ses,
 
     /* Print all attribute (a=) lines. */
     for (i=0; i<ses->attr_count; ++i) {
-	printed = print_attr(ses->attr[i], p, end-p);
+	printed = (int)print_attr(ses->attr[i], p, end-p);
 	if (printed < 0) {
 	    return -1;
 	}
@@ -743,14 +846,14 @@ static int print_session(const pjmedia_sdp_session *ses,
 
     /* Print media (m=) lines. */
     for (i=0; i<ses->media_count; ++i) {
-	printed = print_media_desc(ses->media[i], p, end-p);
+	printed = print_media_desc(ses->media[i], p, (int)(end-p));
 	if (printed < 0) {
 	    return -1;
 	}
 	p += printed;
     }
 
-    return p-buf;
+    return (int)(p-buf);
 }
 
 /******************************************************************************
@@ -816,7 +919,7 @@ static void parse_origin(pj_scanner *scanner, pjmedia_sdp_session *ses,
     pj_scan_get_char(scanner);
 
     /* address */
-    pj_scan_get_until_chr(scanner, " \t\r", &ses->origin.addr);
+    pj_scan_get_until_chr(scanner, " \t\r\n", &ses->origin.addr);
 
     /* We've got what we're looking for, skip anything until newline */
     pj_scan_skip_line(scanner);
@@ -846,7 +949,7 @@ static void parse_time(pj_scanner *scanner, pjmedia_sdp_session *ses,
     pj_scan_get_char(scanner);
 
     /* stop time */
-    pj_scan_get_until_chr(scanner, " \t\r", &str);
+    pj_scan_get_until_chr(scanner, " \t\r\n", &str);
     ses->time.stop = pj_strtoul(&str);
 
     /* We've got what we're looking for, skip anything until newline */
@@ -868,7 +971,7 @@ static void parse_generic_line(pj_scanner *scanner, pj_str_t *str,
     pj_scan_advance_n(scanner, 2, SKIP_WS);
 
     /* get anything until newline (including whitespaces). */
-    pj_scan_get_until_ch(scanner, '\r', str);
+    pj_scan_get_until_chr(scanner, "\r\n", str);
 
     /* newline. */
     pj_scan_get_newline(scanner);
@@ -891,7 +994,30 @@ static void parse_connection_info(pj_scanner *scanner, pjmedia_sdp_conn *conn,
     pj_scan_get_char(scanner);
 
     /* address. */
-    pj_scan_get_until_chr(scanner, " \t\r", &conn->addr);
+    pj_scan_get_until_chr(scanner, "/ \t\r\n", &conn->addr);
+    PJ_TODO(PARSE_SDP_CONN_ADDRESS_SUBFIELDS);
+
+    /* We've got what we're looking for, skip anything until newline */
+    pj_scan_skip_line(scanner);
+}
+
+static void parse_bandwidth_info(pj_scanner *scanner, pjmedia_sdp_bandw *bandw,
+				  parse_context *ctx)
+{
+    pj_str_t str;
+
+    ctx->last_error = PJMEDIA_SDP_EINBANDW;
+
+    /* b= */
+    pj_scan_advance_n(scanner, 2, SKIP_WS);
+
+    /* modifier */
+    pj_scan_get_until_ch(scanner, ':', &bandw->modifier);
+    pj_scan_get_char(scanner);
+
+    /* value */
+    pj_scan_get_until_chr(scanner, " \t\r\n", &str);
+    bandw->value = pj_strtoul(&str);
 
     /* We've got what we're looking for, skip anything until newline */
     pj_scan_skip_line(scanner);
@@ -935,18 +1061,27 @@ static void parse_media(pj_scanner *scanner, pjmedia_sdp_media *med,
     }
 
     /* transport */
-    pj_scan_get_until_ch(scanner, ' ', &med->desc.transport);
+    pj_scan_get_until_chr(scanner, " \t\r\n", &med->desc.transport);
 
     /* format list */
     med->desc.fmt_count = 0;
     while (*scanner->curptr == ' ') {
+	pj_str_t fmt;
+
 	pj_scan_get_char(scanner);
 
 	/* Check again for the end of the line */
 	if ((*scanner->curptr == '\r') || (*scanner->curptr == '\n'))
 		break;
 
-	pj_scan_get(scanner, &cs_token, &med->desc.fmt[med->desc.fmt_count++]);
+	pj_scan_get(scanner, &cs_token, &fmt);
+	if (med->desc.fmt_count < PJMEDIA_MAX_SDP_FMT)
+	    med->desc.fmt[med->desc.fmt_count++] = fmt;
+	else
+	    PJ_PERROR(2,(THIS_FILE, PJ_ETOOMANY, 
+		         "Error adding SDP media format %.*s, "
+			 "format is ignored",
+			 (int)fmt.slen, fmt.ptr));
     }
 
     /* We've got what we're looking for, skip anything until newline */
@@ -981,14 +1116,16 @@ static pjmedia_sdp_attr *parse_attr( pj_pool_t *pool, pj_scanner *scanner,
     /* get attr name. */
     pj_scan_get(scanner, &cs_token, &attr->name);
 
-    if (*scanner->curptr != '\r' && *scanner->curptr != '\n') {
+    if (*scanner->curptr && *scanner->curptr != '\r' && 
+	*scanner->curptr != '\n') 
+    {
 	/* skip ':' if present. */
 	if (*scanner->curptr == ':')
 	    pj_scan_get_char(scanner);
 
 	/* get value */
-	if (*scanner->curptr != '\r') {
-	    pj_scan_get_until_ch(scanner, '\r', &attr->value);
+	if (*scanner->curptr != '\r' && *scanner->curptr != '\n') {
+	    pj_scan_get_until_chr(scanner, "\r\n", &attr->value);
 	} else {
 	    attr->value.ptr = NULL;
 	    attr->value.slen = 0;
@@ -1005,6 +1142,64 @@ static pjmedia_sdp_attr *parse_attr( pj_pool_t *pool, pj_scanner *scanner,
     return attr;
 }
 
+
+/*
+ * Apply direction attribute in session to all media.
+ */
+static void apply_media_direction(pjmedia_sdp_session *sdp)
+{
+    pjmedia_sdp_attr *dir_attr = NULL;
+    unsigned i;
+
+    const pj_str_t inactive = { "inactive", 8 };
+    const pj_str_t sendonly = { "sendonly", 8 };
+    const pj_str_t recvonly = { "recvonly", 8 };
+    const pj_str_t sendrecv = { "sendrecv", 8 };
+
+    /* Find direction attribute in session, don't need to find default 
+     * direction "sendrecv".
+     */
+    for (i = 0; i < sdp->attr_count && !dir_attr; ++i) {
+	if (!pj_strcmp(&sdp->attr[i]->name, &sendonly) ||
+	    !pj_strcmp(&sdp->attr[i]->name, &recvonly) ||
+	    !pj_strcmp(&sdp->attr[i]->name, &inactive)) 
+	{
+	    dir_attr = sdp->attr[i];
+	}
+    }
+
+    /* Found the direction attribute */
+    if (dir_attr) {
+	/* Remove the direction attribute in session */
+	pjmedia_sdp_attr_remove(&sdp->attr_count, sdp->attr, dir_attr);
+
+	/* Apply the direction attribute to all media, but not overriding it
+	 * if media already has direction attribute.
+	 */
+	for (i = 0; i < sdp->media_count; ++i) {
+	    pjmedia_sdp_media *m;
+	    unsigned j;
+
+	    /* Find direction attribute in this media */
+	    m = sdp->media[i];
+	    for (j = 0; j < m->attr_count; ++j) {
+		if (!pj_strcmp(&m->attr[j]->name, &sendrecv) ||
+		    !pj_strcmp(&m->attr[j]->name, &sendonly) ||
+		    !pj_strcmp(&m->attr[j]->name, &recvonly) ||
+		    !pj_strcmp(&m->attr[j]->name, &inactive)) 
+		{
+		    break;
+		}
+	    }
+
+	    /* Not found, apply direction attribute from session */
+	    if (j == m->attr_count)
+		pjmedia_sdp_media_add_attr(m, dir_attr);
+	}
+    }
+}
+
+
 /*
  * Parse SDP message.
  */
@@ -1017,6 +1212,7 @@ PJ_DEF(pj_status_t) pjmedia_sdp_parse( pj_pool_t *pool,
     pjmedia_sdp_media *media = NULL;
     pjmedia_sdp_attr *attr;
     pjmedia_sdp_conn *conn;
+    pjmedia_sdp_bandw *bandw;
     pj_str_t dummy;
     int cur_name = 254;
     parse_context ctx;
@@ -1030,6 +1226,10 @@ PJ_DEF(pj_status_t) pjmedia_sdp_parse( pj_pool_t *pool,
     session = PJ_POOL_ZALLOC_T(pool, pjmedia_sdp_session);
     PJ_ASSERT_RETURN(session != NULL, PJ_ENOMEM);
 
+    /* Ignore leading newlines */
+    while (*scanner.curptr=='\r' || *scanner.curptr=='\n')
+	pj_scan_get_char(&scanner);
+
     PJ_TRY {
 	while (!pj_scan_is_eof(&scanner)) {
 		cur_name = *scanner.curptr;
@@ -1038,9 +1238,9 @@ PJ_DEF(pj_status_t) pjmedia_sdp_parse( pj_pool_t *pool,
 		    attr = parse_attr(pool, &scanner, &ctx);
 		    if (attr) {
 			if (media) {
-			    media->attr[media->attr_count++] = attr;
+			    pjmedia_sdp_media_add_attr(media, attr);
 			} else {
-			    session->attr[session->attr_count++] = attr;
+			    pjmedia_sdp_session_add_attr(session, attr);
 			}
 		    }
 		    break;
@@ -1065,19 +1265,34 @@ PJ_DEF(pj_status_t) pjmedia_sdp_parse( pj_pool_t *pool,
 		case 'm':
 		    media = PJ_POOL_ZALLOC_T(pool, pjmedia_sdp_media);
 		    parse_media(&scanner, media, &ctx);
-		    session->media[ session->media_count++ ] = media;
+		    if (session->media_count < PJMEDIA_MAX_SDP_MEDIA)
+			session->media[ session->media_count++ ] = media;
+		    else
+			PJ_PERROR(2,(THIS_FILE, PJ_ETOOMANY,
+				     "Error adding media, media is ignored"));
 		    break;
 		case 'v':
 		    parse_version(&scanner, &ctx);
 		    break;
 		case 13:
-		    /* Allow empty newline at the end of the message */
-		    pj_scan_get_char(&scanner);
-		    /* Continue below */
 		case 10:
 		    pj_scan_get_char(&scanner);
-		    if (!pj_scan_is_eof(&scanner)) {
-			on_scanner_error(&scanner);
+		    /* Allow empty newlines at the end of the message */
+		    while (!pj_scan_is_eof(&scanner)) {
+			if (*scanner.curptr != 13 && *scanner.curptr != 10) {
+			    ctx.last_error = PJMEDIA_SDP_EINSDP;
+			    on_scanner_error(&scanner);
+			}
+			pj_scan_get_char(&scanner);
+		    }
+		    break;
+		case 'b':
+		    bandw = PJ_POOL_ZALLOC_T(pool, pjmedia_sdp_bandw);
+		    parse_bandwidth_info(&scanner, bandw, &ctx);
+		    if (media) {
+			media->bandw[media->bandw_count++] = bandw;
+		    } else {
+			session->bandw[session->bandw_count++] = bandw;
 		    }
 		    break;
 		default:
@@ -1111,6 +1326,9 @@ PJ_DEF(pj_status_t) pjmedia_sdp_parse( pj_pool_t *pool,
 
     pj_scan_fini(&scanner);
 
+    if (session)
+	apply_media_direction(session);
+
     *p_sdp = session;
     return ctx.last_error;
 }
@@ -1128,9 +1346,8 @@ PJ_DEF(int) pjmedia_sdp_print( const pjmedia_sdp_session *desc,
 /*
  * Clone session
  */
-PJ_DEF(pjmedia_sdp_session*) 
-pjmedia_sdp_session_clone( pj_pool_t *pool,
-			   const pjmedia_sdp_session *rhs)
+PJ_DEF(pjmedia_sdp_session*) pjmedia_sdp_session_clone( pj_pool_t *pool,
+			   			const pjmedia_sdp_session *rhs)
 {
     pjmedia_sdp_session *sess;
     unsigned i;
@@ -1157,6 +1374,12 @@ pjmedia_sdp_session_clone( pj_pool_t *pool,
 	PJ_ASSERT_RETURN(sess->conn != NULL, NULL);
     }
 
+    /* Duplicate bandwidth info */
+    sess->bandw_count = rhs->bandw_count;
+    for (i=0; i<rhs->bandw_count; ++i) {
+	sess->bandw[i] = pjmedia_sdp_bandw_clone(pool, rhs->bandw[i]);
+    }
+
     /* Clone time line. */
     sess->time.start = rhs->time.start;
     sess->time.stop = rhs->time.stop;
@@ -1199,6 +1422,14 @@ static pj_status_t validate_sdp_conn(const pjmedia_sdp_conn *c)
 
 /* Validate SDP session descriptor. */
 PJ_DEF(pj_status_t) pjmedia_sdp_validate(const pjmedia_sdp_session *sdp)
+{
+    return pjmedia_sdp_validate2(sdp, PJ_TRUE);
+}
+
+
+/* Validate SDP session descriptor. */
+PJ_DEF(pj_status_t) pjmedia_sdp_validate2(const pjmedia_sdp_session *sdp,
+					  pj_bool_t strict)
 {
     unsigned i;
     const pj_str_t STR_RTPMAP = { "rtpmap", 6 };
@@ -1248,7 +1479,8 @@ PJ_DEF(pj_status_t) pjmedia_sdp_validate(const pjmedia_sdp_session *sdp)
 	 */
 	if (m->conn == NULL) {
 	    if (sdp->conn == NULL)
-		return PJMEDIA_SDP_EMISSINGCONN;
+		if (strict || m->desc.port != 0)
+		    return PJMEDIA_SDP_EMISSINGCONN;
 	}
 
 	/* Verify payload type. */
@@ -1283,3 +1515,72 @@ PJ_DEF(pj_status_t) pjmedia_sdp_validate(const pjmedia_sdp_session *sdp)
 }
 
 
+PJ_DEF(pj_status_t) pjmedia_sdp_transport_cmp( const pj_str_t *t1,
+					       const pj_str_t *t2)
+{
+    static const pj_str_t ID_RTP_AVP  = { "RTP/AVP", 7 };
+    static const pj_str_t ID_RTP_SAVP = { "RTP/SAVP", 8 };
+
+    /* Exactly equal? */
+    if (pj_stricmp(t1, t2) == 0)
+	return PJ_SUCCESS;
+
+    /* Compatible? */
+    if ((!pj_stricmp(t1, &ID_RTP_AVP) || !pj_stricmp(t1, &ID_RTP_SAVP)) &&
+        (!pj_stricmp(t2, &ID_RTP_AVP) || !pj_stricmp(t2, &ID_RTP_SAVP)))
+	return PJ_SUCCESS;
+
+    return PJMEDIA_SDP_ETPORTNOTEQUAL;
+}
+
+
+PJ_DEF(pj_status_t) pjmedia_sdp_media_deactivate(pj_pool_t *pool,
+						 pjmedia_sdp_media *m)
+{
+    PJ_ASSERT_RETURN(m, PJ_EINVAL);
+    PJ_UNUSED_ARG(pool);
+
+    /* Set port to zero */
+    m->desc.port = 0;
+
+    /* And remove attributes */
+    m->attr_count = 0;
+
+    return PJ_SUCCESS;
+}
+
+
+PJ_DEF(pjmedia_sdp_media*) pjmedia_sdp_media_clone_deactivate(
+						pj_pool_t *pool,
+						const pjmedia_sdp_media *rhs)
+{
+    unsigned int i;
+    pjmedia_sdp_media *m;
+
+    PJ_ASSERT_RETURN(pool && rhs, NULL);
+
+    m = PJ_POOL_ZALLOC_T(pool, pjmedia_sdp_media);
+    pj_memcpy(m, rhs, sizeof(*m));
+
+    /* Clone the media line only */
+    pj_strdup (pool, &m->desc.media, &rhs->desc.media);
+    pj_strdup (pool, &m->desc.transport, &rhs->desc.transport);
+    for (i=0; i<rhs->desc.fmt_count; ++i)
+	pj_strdup(pool, &m->desc.fmt[i], &rhs->desc.fmt[i]);
+
+    if (rhs->conn) {
+	m->conn = pjmedia_sdp_conn_clone (pool, rhs->conn);
+	PJ_ASSERT_RETURN(m->conn != NULL, NULL);
+    }
+
+    m->bandw_count = rhs->bandw_count;
+    for (i=0; i < rhs->bandw_count; ++i) {
+	m->bandw[i] = pjmedia_sdp_bandw_clone (pool, rhs->bandw[i]);
+	PJ_ASSERT_RETURN(m->bandw[i] != NULL, NULL);
+    }
+
+    /* And deactivate it */
+    pjmedia_sdp_media_deactivate(pool, m);
+
+    return m;
+}
