@@ -1,6 +1,7 @@
-/* $Id: simpleua.c 1120 2007-03-30 18:57:01Z bennylp $ */
+/* $Id$ */
 /* 
- * Copyright (C) 2003-2007 Benny Prijono <benny@prijono.org>
+ * Copyright (C) 2008-2011 Teluu Inc. (http://www.teluu.com)
+ * Copyright (C) 2003-2008 Benny Prijono <benny@prijono.org>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -29,6 +30,7 @@
  *
  * Capabilities to be demonstrated here:
  *  - Basic call
+ *  - Should support IPv6 (not tested)
  *  - UDP transport at port 5060 (hard coded)
  *  - RTP socket at port 4000 (hard coded)
  *  - proper SDP negotiation
@@ -63,6 +65,22 @@
 
 #include "util.h"
 
+
+/* Settings */
+#define AF		pj_AF_INET() /* Change to pj_AF_INET6() for IPv6.
+				      * PJ_HAS_IPV6 must be enabled and
+				      * your system must support IPv6.  */
+#if 0
+#define SIP_PORT	5080	     /* Listening SIP port		*/
+#define RTP_PORT	5000	     /* RTP port			*/
+#else
+#define SIP_PORT	5060	     /* Listening SIP port		*/
+#define RTP_PORT	4000	     /* RTP port			*/
+#endif
+
+#define MAX_MEDIA_CNT	2	     /* Media count, set to 1 for audio
+				      * only or 2 for audio and video	*/
+
 /*
  * Static variables.
  */
@@ -72,15 +90,24 @@ static pjsip_endpoint	    *g_endpt;	    /* SIP endpoint.		*/
 static pj_caching_pool	     cp;	    /* Global pool factory.	*/
 
 static pjmedia_endpt	    *g_med_endpt;   /* Media endpoint.		*/
-static pjmedia_sock_info     g_med_skinfo;  /* Socket info for media	*/
-static pjmedia_transport    *g_med_transport;/* Media stream transport	*/
+
+static pjmedia_transport_info g_med_tpinfo[MAX_MEDIA_CNT]; 
+					    /* Socket info for media	*/
+static pjmedia_transport    *g_med_transport[MAX_MEDIA_CNT];
+					    /* Media stream transport	*/
+static pjmedia_sock_info     g_sock_info[MAX_MEDIA_CNT];  
+					    /* Socket info array	*/
 
 /* Call variables: */
 static pjsip_inv_session    *g_inv;	    /* Current invite session.	*/
-static pjmedia_session	    *g_med_session; /* Call's media session.	*/
-static pjmedia_snd_port	    *g_snd_player;  /* Call's sound player	*/
-static pjmedia_snd_port	    *g_snd_rec;	    /* Call's sound recorder.	*/
+static pjmedia_stream       *g_med_stream;  /* Call's audio stream.	*/
+static pjmedia_snd_port	    *g_snd_port;    /* Sound device.		*/
 
+#if defined(PJMEDIA_HAS_VIDEO) && (PJMEDIA_HAS_VIDEO != 0)
+static pjmedia_vid_stream   *g_med_vstream; /* Call's video stream.	*/
+static pjmedia_vid_port	    *g_vid_capturer;/* Call's video capturer.	*/
+static pjmedia_vid_port	    *g_vid_renderer;/* Call's video renderer.	*/
+#endif	/* PJMEDIA_HAS_VIDEO */
 
 /*
  * Prototypes:
@@ -126,6 +153,68 @@ static pjsip_module mod_simpleua =
 };
 
 
+/* Notification on incoming messages */
+static pj_bool_t logging_on_rx_msg(pjsip_rx_data *rdata)
+{
+    PJ_LOG(4,(THIS_FILE, "RX %d bytes %s from %s %s:%d:\n"
+			 "%.*s\n"
+			 "--end msg--",
+			 rdata->msg_info.len,
+			 pjsip_rx_data_get_info(rdata),
+			 rdata->tp_info.transport->type_name,
+			 rdata->pkt_info.src_name,
+			 rdata->pkt_info.src_port,
+			 (int)rdata->msg_info.len,
+			 rdata->msg_info.msg_buf));
+    
+    /* Always return false, otherwise messages will not get processed! */
+    return PJ_FALSE;
+}
+
+/* Notification on outgoing messages */
+static pj_status_t logging_on_tx_msg(pjsip_tx_data *tdata)
+{
+    
+    /* Important note:
+     *	tp_info field is only valid after outgoing messages has passed
+     *	transport layer. So don't try to access tp_info when the module
+     *	has lower priority than transport layer.
+     */
+
+    PJ_LOG(4,(THIS_FILE, "TX %d bytes %s to %s %s:%d:\n"
+			 "%.*s\n"
+			 "--end msg--",
+			 (tdata->buf.cur - tdata->buf.start),
+			 pjsip_tx_data_get_info(tdata),
+			 tdata->tp_info.transport->type_name,
+			 tdata->tp_info.dst_name,
+			 tdata->tp_info.dst_port,
+			 (int)(tdata->buf.cur - tdata->buf.start),
+			 tdata->buf.start));
+
+    /* Always return success, otherwise message will not get sent! */
+    return PJ_SUCCESS;
+}
+
+/* The module instance. */
+static pjsip_module msg_logger = 
+{
+    NULL, NULL,				/* prev, next.		*/
+    { "mod-msg-log", 13 },		/* Name.		*/
+    -1,					/* Id			*/
+    PJSIP_MOD_PRIORITY_TRANSPORT_LAYER-1,/* Priority	        */
+    NULL,				/* load()		*/
+    NULL,				/* start()		*/
+    NULL,				/* stop()		*/
+    NULL,				/* unload()		*/
+    &logging_on_rx_msg,			/* on_rx_request()	*/
+    &logging_on_rx_msg,			/* on_rx_response()	*/
+    &logging_on_tx_msg,			/* on_tx_request.	*/
+    &logging_on_tx_msg,			/* on_tx_response()	*/
+    NULL,				/* on_tsx_state()	*/
+
+};
+
 
 /*
  * main()
@@ -135,12 +224,15 @@ static pjsip_module mod_simpleua =
  */
 int main(int argc, char *argv[])
 {
+    pj_pool_t *pool = NULL;
     pj_status_t status;
+    unsigned i;
 
     /* Must init PJLIB first: */
     status = pj_init();
     PJ_ASSERT_RETURN(status == PJ_SUCCESS, 1);
 
+    pj_log_set_level(5);
 
     /* Then init PJLIB-UTIL: */
     status = pjlib_util_init();
@@ -179,13 +271,20 @@ int main(int argc, char *argv[])
      * resolves the address with STUN).
      */
     {
-	pj_sockaddr_in addr;
+	pj_sockaddr addr;
 
-	addr.sin_family = PJ_AF_INET;
-	addr.sin_addr.s_addr = 0;
-	addr.sin_port = pj_htons(5060);
+	pj_sockaddr_init(AF, &addr, NULL, (pj_uint16_t)SIP_PORT);
+	
+	if (AF == pj_AF_INET()) {
+	    status = pjsip_udp_transport_start( g_endpt, &addr.ipv4, NULL, 
+						1, NULL);
+	} else if (AF == pj_AF_INET6()) {
+	    status = pjsip_udp_transport_start6(g_endpt, &addr.ipv6, NULL,
+						1, NULL);
+	} else {
+	    status = PJ_EAFNOTSUP;
+	}
 
-	status = pjsip_udp_transport_start( g_endpt, &addr, NULL, 1, NULL);
 	if (status != PJ_SUCCESS) {
 	    app_perror(THIS_FILE, "Unable to start UDP transport", status);
 	    return 1;
@@ -235,6 +334,9 @@ int main(int argc, char *argv[])
 	PJ_ASSERT_RETURN(status == PJ_SUCCESS, 1);
     }
 
+    /* Initialize 100rel support */
+    status = pjsip_100rel_init_module(g_endpt);
+    PJ_ASSERT_RETURN(status == PJ_SUCCESS, status);
 
     /*
      * Register our module to receive incoming requests.
@@ -242,6 +344,12 @@ int main(int argc, char *argv[])
     status = pjsip_endpt_register_module( g_endpt, &mod_simpleua);
     PJ_ASSERT_RETURN(status == PJ_SUCCESS, 1);
 
+    /*
+     * Register message logger module.
+     */
+    status = pjsip_endpt_register_module( g_endpt, &msg_logger);
+    PJ_ASSERT_RETURN(status == PJ_SUCCESS, 1);
+
 
     /* 
      * Initialize media endpoint.
@@ -264,37 +372,64 @@ int main(int argc, char *argv[])
     PJ_ASSERT_RETURN(status == PJ_SUCCESS, 1);
 #endif
 
+
+#if defined(PJMEDIA_HAS_VIDEO) && (PJMEDIA_HAS_VIDEO != 0)
+    /* Init video subsystem */
+    pool = pjmedia_endpt_create_pool(g_med_endpt, "Video subsystem", 512, 512);
+    status = pjmedia_video_format_mgr_create(pool, 64, 0, NULL);
+    PJ_ASSERT_RETURN(status == PJ_SUCCESS, 1);
+    status = pjmedia_converter_mgr_create(pool, NULL);
+    PJ_ASSERT_RETURN(status == PJ_SUCCESS, 1);
+    status = pjmedia_vid_codec_mgr_create(pool, NULL);
+    PJ_ASSERT_RETURN(status == PJ_SUCCESS, 1);
+    status = pjmedia_vid_dev_subsys_init(&cp.factory);
+    PJ_ASSERT_RETURN(status == PJ_SUCCESS, 1);
+
+#   if defined(PJMEDIA_HAS_OPENH264_CODEC) && PJMEDIA_HAS_OPENH264_CODEC != 0
+    status = pjmedia_codec_openh264_vid_init(NULL, &cp.factory);
+    PJ_ASSERT_RETURN(status == PJ_SUCCESS, 1);
+#   endif
+
+#  if defined(PJMEDIA_HAS_FFMPEG_VID_CODEC) && PJMEDIA_HAS_FFMPEG_VID_CODEC!=0
+    /* Init ffmpeg video codecs */
+    status = pjmedia_codec_ffmpeg_vid_init(NULL, &cp.factory);
+    PJ_ASSERT_RETURN(status == PJ_SUCCESS, 1);
+#  endif  /* PJMEDIA_HAS_FFMPEG_VID_CODEC */
+
+#endif	/* PJMEDIA_HAS_VIDEO */
     
     /* 
      * Create media transport used to send/receive RTP/RTCP socket.
      * One media transport is needed for each call. Application may
      * opt to re-use the same media transport for subsequent calls.
      */
-    status = pjmedia_transport_udp_create(g_med_endpt, NULL, 4000, 0, 
-					  &g_med_transport);
-    if (status != PJ_SUCCESS) {
-	app_perror(THIS_FILE, "Unable to create media transport", status);
-	return 1;
-    }
+    for (i = 0; i < PJ_ARRAY_SIZE(g_med_transport); ++i) {
+	status = pjmedia_transport_udp_create3(g_med_endpt, AF, NULL, NULL, 
+					       RTP_PORT + i*2, 0, 
+					       &g_med_transport[i]);
+	if (status != PJ_SUCCESS) {
+	    app_perror(THIS_FILE, "Unable to create media transport", status);
+	    return 1;
+	}
 
-    /* 
-     * Get socket info (address, port) of the media transport. We will
-     * need this info to create SDP (i.e. the address and port info in
-     * the SDP).
-     */
-    {
-	pjmedia_transport_udp_info udp_info;
+	/* 
+	 * Get socket info (address, port) of the media transport. We will
+	 * need this info to create SDP (i.e. the address and port info in
+	 * the SDP).
+	 */
+	pjmedia_transport_info_init(&g_med_tpinfo[i]);
+	pjmedia_transport_get_info(g_med_transport[i], &g_med_tpinfo[i]);
 
-	pjmedia_transport_udp_get_info(g_med_transport, &udp_info);
-	pj_memcpy(&g_med_skinfo, &udp_info.skinfo, 
+	pj_memcpy(&g_sock_info[i], &g_med_tpinfo[i].sock_info,
 		  sizeof(pjmedia_sock_info));
     }
 
-
     /*
      * If URL is specified, then make call immediately.
      */
     if (argc > 1) {
+	pj_sockaddr hostaddr;
+	char hostip[PJ_INET6_ADDRSTRLEN+2];
 	char temp[80];
 	pj_str_t dst_uri = pj_str(argv[1]);
 	pj_str_t local_uri;
@@ -302,13 +437,20 @@ int main(int argc, char *argv[])
 	pjmedia_sdp_session *local_sdp;
 	pjsip_tx_data *tdata;
 
-	pj_ansi_sprintf(temp, "sip:simpleuac@%s", pjsip_endpt_name(g_endpt)->ptr);
+	if (pj_gethostip(AF, &hostaddr) != PJ_SUCCESS) {
+	    app_perror(THIS_FILE, "Unable to retrieve local host IP", status);
+	    return 1;
+	}
+	pj_sockaddr_print(&hostaddr, hostip, sizeof(hostip), 2);
+
+	pj_ansi_sprintf(temp, "<sip:simpleuac@%s:%d>", 
+			hostip, SIP_PORT);
 	local_uri = pj_str(temp);
 
 	/* Create UAC dialog */
 	status = pjsip_dlg_create_uac( pjsip_ua_instance(), 
 				       &local_uri,  /* local URI */
-				       NULL,	    /* local Contact */
+				       &local_uri,  /* local Contact */
 				       &dst_uri,    /* remote URI */
 				       &dst_uri,    /* remote target */
 				       &dlg);	    /* dialog */
@@ -324,6 +466,7 @@ int main(int argc, char *argv[])
 		pjsip_cred_info	cred[1];
 
 		cred[0].realm	  = pj_str("sip.server.realm");
+		cred[0].scheme    = pj_str("digest");
 		cred[0].username  = pj_str("theuser");
 		cred[0].data_type = PJSIP_CRED_DATA_PLAIN_PASSWD;
 		cred[0].data      = pj_str("thepassword");
@@ -334,6 +477,24 @@ int main(int argc, char *argv[])
 	 */
 
 
+	/* Get the SDP body to be put in the outgoing INVITE, by asking
+	 * media endpoint to create one for us.
+	 */
+	status = pjmedia_endpt_create_sdp( g_med_endpt,	    /* the media endpt	*/
+					   dlg->pool,	    /* pool.		*/
+					   MAX_MEDIA_CNT,   /* # of streams	*/
+					   g_sock_info,     /* RTP sock info	*/
+					   &local_sdp);	    /* the SDP result	*/
+	PJ_ASSERT_RETURN(status == PJ_SUCCESS, 1);
+
+
+
+	/* Create the INVITE session, and pass the SDP returned earlier
+	 * as the session's initial capability.
+	 */
+	status = pjsip_inv_create_uac( dlg, local_sdp, 0, &g_inv);
+	PJ_ASSERT_RETURN(status == PJ_SUCCESS, 1);
+
 	/* If we want the initial INVITE to travel to specific SIP proxies,
 	 * then we should put the initial dialog's route set here. The final
 	 * route set will be updated once a dialog has been established.
@@ -360,28 +521,6 @@ int main(int argc, char *argv[])
 	 * Note that Route URI SHOULD have an ";lr" parameter!
 	 */
 
-
-	/* Get the SDP body to be put in the outgoing INVITE, by asking
-	 * media endpoint to create one for us. The SDP will contain all
-	 * codecs that have been registered to it (in this case, only
-	 * PCMA and PCMU), plus telephony event.
-	 */
-	status = pjmedia_endpt_create_sdp( g_med_endpt,	    /* the media endpt	*/
-					   dlg->pool,	    /* pool.		*/
-					   1,		    /* # of streams	*/
-					   &g_med_skinfo,   /* RTP sock info	*/
-					   &local_sdp);	    /* the SDP result	*/
-	PJ_ASSERT_RETURN(status == PJ_SUCCESS, 1);
-
-
-
-	/* Create the INVITE session, and pass the SDP returned earlier
-	 * as the session's initial capability.
-	 */
-	status = pjsip_inv_create_uac( dlg, local_sdp, 0, &g_inv);
-	PJ_ASSERT_RETURN(status == PJ_SUCCESS, 1);
-
-
 	/* Create initial INVITE request.
 	 * This INVITE request will contain a perfectly good request and 
 	 * an SDP body as well.
@@ -416,6 +555,56 @@ int main(int argc, char *argv[])
     /* On exit, dump current memory usage: */
     dump_pool_usage(THIS_FILE, &cp);
 
+    /* Destroy audio ports. Destroy the audio port first
+     * before the stream since the audio port has threads
+     * that get/put frames to the stream.
+     */
+    if (g_snd_port)
+	pjmedia_snd_port_destroy(g_snd_port);
+
+#if defined(PJMEDIA_HAS_VIDEO) && (PJMEDIA_HAS_VIDEO != 0)
+    /* Destroy video ports */
+    if (g_vid_capturer)
+	pjmedia_vid_port_destroy(g_vid_capturer);
+    if (g_vid_renderer)
+	pjmedia_vid_port_destroy(g_vid_renderer);
+#endif
+
+    /* Destroy streams */
+    if (g_med_stream)
+	pjmedia_stream_destroy(g_med_stream);
+#if defined(PJMEDIA_HAS_VIDEO) && (PJMEDIA_HAS_VIDEO != 0)
+    if (g_med_vstream)
+	pjmedia_vid_stream_destroy(g_med_vstream);
+
+    /* Deinit ffmpeg codec */
+#   if defined(PJMEDIA_HAS_FFMPEG_VID_CODEC) && PJMEDIA_HAS_FFMPEG_VID_CODEC!=0
+    pjmedia_codec_ffmpeg_vid_deinit();
+#   endif
+#   if defined(PJMEDIA_HAS_OPENH264_CODEC) && PJMEDIA_HAS_OPENH264_CODEC != 0
+    pjmedia_codec_openh264_vid_deinit();
+#   endif
+
+#endif
+
+    /* Destroy media transports */
+    for (i = 0; i < MAX_MEDIA_CNT; ++i) {
+	if (g_med_transport[i])
+	    pjmedia_transport_close(g_med_transport[i]);
+    }
+
+    /* Deinit pjmedia endpoint */
+    if (g_med_endpt)
+	pjmedia_endpt_destroy(g_med_endpt);
+
+    /* Deinit pjsip endpoint */
+    if (g_endpt)
+	pjsip_endpt_destroy(g_endpt);
+
+    /* Release pool */
+    if (pool)
+	pj_pool_release(pool);
+
     return 0;
 }
 
@@ -466,6 +655,9 @@ static void call_on_forked(pjsip_inv_session *inv, pjsip_event *e)
  */
 static pj_bool_t on_rx_request( pjsip_rx_data *rdata )
 {
+    pj_sockaddr hostaddr;
+    char temp[80], hostip[PJ_INET6_ADDRSTRLEN];
+    pj_str_t local_uri;
     pjsip_dialog *dlg;
     pjmedia_sdp_session *local_sdp;
     pjsip_tx_data *tdata;
@@ -517,12 +709,25 @@ static pj_bool_t on_rx_request( pjsip_rx_data *rdata )
 	return PJ_TRUE;
     } 
 
+    /*
+     * Generate Contact URI
+     */
+    if (pj_gethostip(AF, &hostaddr) != PJ_SUCCESS) {
+	app_perror(THIS_FILE, "Unable to retrieve local host IP", status);
+	return PJ_TRUE;
+    }
+    pj_sockaddr_print(&hostaddr, hostip, sizeof(hostip), 2);
+
+    pj_ansi_sprintf(temp, "<sip:simpleuas@%s:%d>", 
+		    hostip, SIP_PORT);
+    local_uri = pj_str(temp);
+
     /*
      * Create UAS dialog.
      */
     status = pjsip_dlg_create_uas( pjsip_ua_instance(), 
 				   rdata,
-				   NULL, /* contact */
+				   &local_uri, /* contact */
 				   &dlg);
     if (status != PJ_SUCCESS) {
 	pjsip_endpt_respond_stateless(g_endpt, rdata, 500, NULL,
@@ -534,9 +739,8 @@ static pj_bool_t on_rx_request( pjsip_rx_data *rdata )
      * Get media capability from media endpoint: 
      */
 
-    status = pjmedia_endpt_create_sdp( g_med_endpt, rdata->tp_info.pool, 1,
-				       &g_med_skinfo, 
-				       &local_sdp);
+    status = pjmedia_endpt_create_sdp( g_med_endpt, rdata->tp_info.pool,
+				       MAX_MEDIA_CNT, g_sock_info, &local_sdp);
     PJ_ASSERT_RETURN(status == PJ_SUCCESS, PJ_TRUE);
 
 
@@ -599,7 +803,7 @@ static pj_bool_t on_rx_request( pjsip_rx_data *rdata )
 static void call_on_media_update( pjsip_inv_session *inv,
 				  pj_status_t status)
 {
-    pjmedia_session_info sess_info;
+    pjmedia_stream_info stream_info;
     const pjmedia_sdp_session *local_sdp;
     const pjmedia_sdp_session *remote_sdp;
     pjmedia_port *media_port;
@@ -622,88 +826,211 @@ static void call_on_media_update( pjsip_inv_session *inv,
     status = pjmedia_sdp_neg_get_active_remote(inv->neg, &remote_sdp);
 
 
-    /* Create session info based on the two SDPs. 
-     * We only support one stream per session for now.
-     */
-    status = pjmedia_session_info_from_sdp(inv->dlg->pool, g_med_endpt, 
-					   1, &sess_info, 
-					   local_sdp, remote_sdp);
+    /* Create stream info based on the media audio SDP. */
+    status = pjmedia_stream_info_from_sdp(&stream_info, inv->dlg->pool,
+					  g_med_endpt,
+					  local_sdp, remote_sdp, 0);
     if (status != PJ_SUCCESS) {
-	app_perror( THIS_FILE, "Unable to create media session", status);
+	app_perror(THIS_FILE,"Unable to create audio stream info",status);
 	return;
     }
 
-    /* If required, we can also change some settings in the session info,
+    /* If required, we can also change some settings in the stream info,
      * (such as jitter buffer settings, codec settings, etc) before we
-     * create the session.
+     * create the stream.
      */
 
-    /* Create new media session, passing the two SDPs, and also the
+    /* Create new audio media stream, passing the stream info, and also the
      * media socket that we created earlier.
-     * The media session is active immediately.
      */
-    status = pjmedia_session_create( g_med_endpt, &sess_info,
-				     &g_med_transport, NULL, &g_med_session );
+    status = pjmedia_stream_create(g_med_endpt, inv->dlg->pool, &stream_info,
+				   g_med_transport[0], NULL, &g_med_stream);
     if (status != PJ_SUCCESS) {
-	app_perror( THIS_FILE, "Unable to create media session", status);
+	app_perror( THIS_FILE, "Unable to create audio stream", status);
 	return;
     }
 
+    /* Start the audio stream */
+    status = pjmedia_stream_start(g_med_stream);
+    if (status != PJ_SUCCESS) {
+	app_perror( THIS_FILE, "Unable to start audio stream", status);
+	return;
+    }
 
-    /* Get the media port interface of the first stream in the session. 
+    /* Get the media port interface of the audio stream. 
      * Media port interface is basicly a struct containing get_frame() and
      * put_frame() function. With this media port interface, we can attach
      * the port interface to conference bridge, or directly to a sound
      * player/recorder device.
      */
-    pjmedia_session_get_port(g_med_session, 0, &media_port);
+    pjmedia_stream_get_port(g_med_stream, &media_port);
 
+    /* Create sound port */
+    pjmedia_snd_port_create(inv->pool,
+                            PJMEDIA_AUD_DEFAULT_CAPTURE_DEV,
+                            PJMEDIA_AUD_DEFAULT_PLAYBACK_DEV,
+                            PJMEDIA_PIA_SRATE(&media_port->info),/* clock rate	    */
+                            PJMEDIA_PIA_CCNT(&media_port->info),/* channel count    */
+                            PJMEDIA_PIA_SPF(&media_port->info), /* samples per frame*/
+                            PJMEDIA_PIA_BITS(&media_port->info),/* bits per sample  */
+                            0,
+                            &g_snd_port);
 
-
-    /* Create a sound Player device and connect the media port to the
-     * sound device.
-     */
-    status = pjmedia_snd_port_create_player( 
-		    inv->pool,				/* pool		    */
-		    -1,					/* sound dev id	    */
-		    media_port->info.clock_rate,	/* clock rate	    */
-		    media_port->info.channel_count,	/* channel count    */
-		    media_port->info.samples_per_frame, /* samples per frame*/
-		    media_port->info.bits_per_sample,   /* bits per sample  */
-		    0,					/* options	    */
-		    &g_snd_player);
     if (status != PJ_SUCCESS) {
-	app_perror( THIS_FILE, "Unable to create sound player", status);
+	app_perror( THIS_FILE, "Unable to create sound port", status);
 	PJ_LOG(3,(THIS_FILE, "%d %d %d %d",
-	    	    media_port->info.clock_rate,	/* clock rate	    */
-		    media_port->info.channel_count,	/* channel count    */
-		    media_port->info.samples_per_frame, /* samples per frame*/
-		    media_port->info.bits_per_sample    /* bits per sample  */
+		    PJMEDIA_PIA_SRATE(&media_port->info),/* clock rate	    */
+		    PJMEDIA_PIA_CCNT(&media_port->info),/* channel count    */
+		    PJMEDIA_PIA_SPF(&media_port->info), /* samples per frame*/
+		    PJMEDIA_PIA_BITS(&media_port->info) /* bits per sample  */
 	    ));
 	return;
     }
 
-    status = pjmedia_snd_port_connect(g_snd_player, media_port);
+    status = pjmedia_snd_port_connect(g_snd_port, media_port);
 
 
-    /* Create a sound recorder device and connect the media port to the
-     * sound device.
+    /* Get the media port interface of the second stream in the session,
+     * which is video stream. With this media port interface, we can attach
+     * the port directly to a renderer/capture video device.
      */
-    status = pjmedia_snd_port_create_rec( 
-		    inv->pool,				/* pool		    */
-		    -1,					/* sound dev id	    */
-		    media_port->info.clock_rate,	/* clock rate	    */
-		    media_port->info.channel_count,	/* channel count    */
-		    media_port->info.samples_per_frame, /* samples per frame*/
-		    media_port->info.bits_per_sample,   /* bits per sample  */
-		    0,					/* options	    */
-		    &g_snd_rec);
-    if (status != PJ_SUCCESS) {
-	app_perror( THIS_FILE, "Unable to create sound recorder", status);
-	return;
-    }
+#if defined(PJMEDIA_HAS_VIDEO) && (PJMEDIA_HAS_VIDEO != 0)
+    if (local_sdp->media_count > 1) {
+	pjmedia_vid_stream_info vstream_info;
+	pjmedia_vid_port_param vport_param;
 
-    status = pjmedia_snd_port_connect(g_snd_rec, media_port);
+	pjmedia_vid_port_param_default(&vport_param);
+
+	/* Create stream info based on the media video SDP. */
+	status = pjmedia_vid_stream_info_from_sdp(&vstream_info,
+						  inv->dlg->pool, g_med_endpt,
+						  local_sdp, remote_sdp, 1);
+	if (status != PJ_SUCCESS) {
+	    app_perror(THIS_FILE,"Unable to create video stream info",status);
+	    return;
+	}
+
+	/* If required, we can also change some settings in the stream info,
+	 * (such as jitter buffer settings, codec settings, etc) before we
+	 * create the video stream.
+	 */
+
+	/* Create new video media stream, passing the stream info, and also the
+	 * media socket that we created earlier.
+	 */
+	status = pjmedia_vid_stream_create(g_med_endpt, NULL, &vstream_info,
+	                                   g_med_transport[1], NULL,
+	                                   &g_med_vstream);
+	if (status != PJ_SUCCESS) {
+	    app_perror( THIS_FILE, "Unable to create video stream", status);
+	    return;
+	}
+
+	/* Start the video stream */
+	status = pjmedia_vid_stream_start(g_med_vstream);
+	if (status != PJ_SUCCESS) {
+	    app_perror( THIS_FILE, "Unable to start video stream", status);
+	    return;
+	}
+
+	if (vstream_info.dir & PJMEDIA_DIR_DECODING) {
+	    status = pjmedia_vid_dev_default_param(
+				inv->pool, PJMEDIA_VID_DEFAULT_RENDER_DEV,
+				&vport_param.vidparam);
+	    if (status != PJ_SUCCESS) {
+		app_perror(THIS_FILE, "Unable to get default param of video "
+			   "renderer device", status);
+		return;
+	    }
+
+	    /* Get video stream port for decoding direction */
+	    pjmedia_vid_stream_get_port(g_med_vstream, PJMEDIA_DIR_DECODING,
+					&media_port);
+
+	    /* Set format */
+	    pjmedia_format_copy(&vport_param.vidparam.fmt,
+				&media_port->info.fmt);
+	    vport_param.vidparam.dir = PJMEDIA_DIR_RENDER;
+	    vport_param.active = PJ_TRUE;
+
+	    /* Create renderer */
+	    status = pjmedia_vid_port_create(inv->pool, &vport_param, 
+					     &g_vid_renderer);
+	    if (status != PJ_SUCCESS) {
+		app_perror(THIS_FILE, "Unable to create video renderer device",
+			   status);
+		return;
+	    }
+
+	    /* Connect renderer to media_port */
+	    status = pjmedia_vid_port_connect(g_vid_renderer, media_port, 
+					      PJ_FALSE);
+	    if (status != PJ_SUCCESS) {
+		app_perror(THIS_FILE, "Unable to connect renderer to stream",
+			   status);
+		return;
+	    }
+	}
+
+	/* Create capturer */
+	if (vstream_info.dir & PJMEDIA_DIR_ENCODING) {
+	    status = pjmedia_vid_dev_default_param(
+				inv->pool, PJMEDIA_VID_DEFAULT_CAPTURE_DEV,
+				&vport_param.vidparam);
+	    if (status != PJ_SUCCESS) {
+		app_perror(THIS_FILE, "Unable to get default param of video "
+			   "capture device", status);
+		return;
+	    }
+
+	    /* Get video stream port for decoding direction */
+	    pjmedia_vid_stream_get_port(g_med_vstream, PJMEDIA_DIR_ENCODING,
+					&media_port);
+
+	    /* Get capturer format from stream info */
+	    pjmedia_format_copy(&vport_param.vidparam.fmt, 
+	                        &media_port->info.fmt);
+	    vport_param.vidparam.dir = PJMEDIA_DIR_CAPTURE;
+	    vport_param.active = PJ_TRUE;
+
+	    /* Create capturer */
+	    status = pjmedia_vid_port_create(inv->pool, &vport_param, 
+					     &g_vid_capturer);
+	    if (status != PJ_SUCCESS) {
+		app_perror(THIS_FILE, "Unable to create video capture device",
+			   status);
+		return;
+	    }
+
+	    /* Connect capturer to media_port */
+	    status = pjmedia_vid_port_connect(g_vid_capturer, media_port, 
+					      PJ_FALSE);
+	    if (status != PJ_SUCCESS) {
+		app_perror(THIS_FILE, "Unable to connect capturer to stream",
+			   status);
+		return;
+	    }
+	}
+
+	/* Start streaming */
+	if (g_vid_renderer) {
+	    status = pjmedia_vid_port_start(g_vid_renderer);
+	    if (status != PJ_SUCCESS) {
+		app_perror(THIS_FILE, "Unable to start video renderer",
+			   status);
+		return;
+	    }
+	}
+	if (g_vid_capturer) {
+	    status = pjmedia_vid_port_start(g_vid_capturer);
+	    if (status != PJ_SUCCESS) {
+		app_perror(THIS_FILE, "Unable to start video capturer",
+			   status);
+		return;
+	    }
+	}
+    }
+#endif	/* PJMEDIA_HAS_VIDEO */
 
     /* Done with media. */
 }
