@@ -1,6 +1,7 @@
-/* $Id: siprtp.c 974 2007-02-19 01:13:53Z bennylp $ */
+/* $Id$ */
 /* 
- * Copyright (C) 2003-2007 Benny Prijono <benny@prijono.org>
+ * Copyright (C) 2008-2011 Teluu Inc. (http://www.teluu.com)
+ * Copyright (C) 2003-2008 Benny Prijono <benny@prijono.org>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -35,8 +36,10 @@ static const char *USAGE =
 "\n"
 " Program options:\n"
 "   --count=N,        -c    Set number of calls to create (default:1) \n"
+"   --gap=N           -g    Set call gapping to N msec (default:0)\n"
 "   --duration=SEC,   -d    Set maximum call duration (default:unlimited) \n"
 "   --auto-quit,      -q    Quit when calls have been completed (default:no)\n"
+"   --call-report     -R    Display report on call termination (default:yes)\n"
 "\n"
 " Address and ports options:\n"
 "   --local-port=PORT,-p    Set local SIP port (default: 5060)\n"
@@ -74,6 +77,16 @@ static const char *USAGE =
 
 #include <stdlib.h>
 
+/* Uncomment these to disable threads.
+ * NOTE:
+ *   when threading is disabled, siprtp won't transmit any
+ *   RTP packets.
+ */
+/*
+#undef PJ_HAS_THREADS
+#define PJ_HAS_THREADS 0
+*/
+
 
 #if PJ_HAS_HIGH_RES_TIMER==0
 #   error "High resolution timer is needed for this sample"
@@ -149,6 +162,8 @@ struct call
 static struct app
 {
     unsigned		 max_calls;
+    unsigned		 call_gap;
+    pj_bool_t		 call_report;
     unsigned		 uac_calls;
     unsigned		 duration;
     pj_bool_t		 auto_quit;
@@ -217,10 +232,10 @@ static void destroy_call_media(unsigned call_index);
 static void destroy_media();
 
 /* This callback is called by media transport on receipt of RTP packet. */
-static void on_rx_rtp(void *user_data, const void *pkt, pj_ssize_t size);
+static void on_rx_rtp(void *user_data, void *pkt, pj_ssize_t size);
 
 /* This callback is called by media transport on receipt of RTCP packet. */
-static void on_rx_rtcp(void *user_data, const void *pkt, pj_ssize_t size);
+static void on_rx_rtcp(void *user_data, void *pkt, pj_ssize_t size);
 
 /* Display error */
 static void app_perror(const char *sender, const char *title, 
@@ -295,7 +310,7 @@ static pj_status_t init_sip()
 	pjsip_transport *tp;
 
 	pj_bzero(&addr, sizeof(addr));
-	addr.sin_family = PJ_AF_INET;
+	addr.sin_family = pj_AF_INET();
 	addr.sin_addr.s_addr = 0;
 	addr.sin_port = pj_htons((pj_uint16_t)app.sip_port);
 
@@ -336,6 +351,10 @@ static pj_status_t init_sip()
     status = pjsip_ua_init_module( app.sip_endpt, NULL );
     PJ_ASSERT_RETURN(status == PJ_SUCCESS, status);
 
+    /* Initialize 100rel support */
+    status = pjsip_100rel_init_module(app.sip_endpt);
+    PJ_ASSERT_RETURN(status == PJ_SUCCESS, status);
+
     /*  Init invite session module. */
     {
 	pjsip_inv_callback inv_cb;
@@ -401,7 +420,13 @@ static pj_status_t init_media()
     /* Initialize media endpoint so that at least error subsystem is properly
      * initialized.
      */
+#if PJ_HAS_THREADS
     status = pjmedia_endpt_create(&app.cp.factory, NULL, 1, &app.med_endpt);
+#else
+    status = pjmedia_endpt_create(&app.cp.factory, 
+				  pjsip_endpt_get_ioqueue(app.sip_endpt),
+				  0, &app.med_endpt);
+#endif
     PJ_ASSERT_RETURN(status == PJ_SUCCESS, status);
 
 
@@ -738,8 +763,11 @@ static void call_on_state_changed( pjsip_inv_session *inv,
 		  inv->cause,
 		  (int)inv->cause_text.slen,
 		  inv->cause_text.ptr));
-	PJ_LOG(3,(THIS_FILE, "Call #%d statistics:", call->index));
-	print_call(call->index);
+
+	if (app.call_report) {
+	    PJ_LOG(3,(THIS_FILE, "Call #%d statistics:", call->index));
+	    print_call(call->index);
+	}
 
 
 	call->inv = NULL;
@@ -826,6 +854,8 @@ static pj_status_t init_options(int argc, char *argv[])
 
     struct pj_getopt_option long_options[] = {
 	{ "count",	    1, 0, 'c' },
+	{ "gap",            1, 0, 'g' },
+	{ "call-report",    0, 0, 'R' },
 	{ "duration",	    1, 0, 'd' },
 	{ "auto-quit",	    0, 0, 'q' },
 	{ "local-port",	    1, 0, 'p' },
@@ -878,17 +908,24 @@ static pj_status_t init_options(int argc, char *argv[])
 
     /* Parse options */
     pj_optind = 0;
-    while((c=pj_getopt_long(argc,argv, "c:d:p:r:i:l:q", 
+    while((c=pj_getopt_long(argc,argv, "c:d:p:r:i:l:g:qR", 
 			    long_options, &option_index))!=-1) 
     {
 	switch (c) {
 	case 'c':
 	    app.max_calls = atoi(pj_optarg);
-	    if (app.max_calls < 0 || app.max_calls > MAX_CALLS) {
-		PJ_LOG(3,(THIS_FILE, "Invalid max calls value %s", pj_optarg));
+	    if (app.max_calls > MAX_CALLS) {
+		PJ_LOG(3,(THIS_FILE,"Invalid max calls value %s "
+				    "(must be <= %d)", pj_optarg, MAX_CALLS));
 		return 1;
 	    }
 	    break;
+	case 'g':
+	    app.call_gap = atoi(pj_optarg);
+	    break;
+	case 'R':
+	    app.call_report = PJ_TRUE;
+	    break;
 	case 'd':
 	    app.duration = atoi(pj_optarg);
 	    break;
@@ -970,14 +1007,15 @@ static pj_status_t create_sdp( pj_pool_t *pool,
     pjmedia_sdp_session *sdp;
     pjmedia_sdp_media *m;
     pjmedia_sdp_attr *attr;
-    pjmedia_transport_udp_info tpinfo;
+    pjmedia_transport_info tpinfo;
     struct media_stream *audio = &call->media[0];
 
     PJ_ASSERT_RETURN(pool && p_sdp, PJ_EINVAL);
 
 
     /* Get transport info */
-    pjmedia_transport_udp_get_info(audio->transport, &tpinfo);
+    pjmedia_transport_info_init(&tpinfo);
+    pjmedia_transport_get_info(audio->transport, &tpinfo);
 
     /* Create and initialize basic SDP session */
     sdp = pj_pool_zalloc (pool, sizeof(pjmedia_sdp_session));
@@ -1011,7 +1049,7 @@ static pj_status_t create_sdp( pj_pool_t *pool,
 
     /* Standard media info: */
     m->desc.media = pj_str("audio");
-    m->desc.port = pj_ntohs(tpinfo.skinfo.rtp_addr_name.sin_port);
+    m->desc.port = pj_ntohs(tpinfo.sock_info.rtp_addr_name.ipv4.sin_port);
     m->desc.port_count = 1;
     m->desc.transport = pj_str("RTP/AVP");
 
@@ -1064,7 +1102,7 @@ static pj_status_t create_sdp( pj_pool_t *pool,
 }
 
 
-#if defined(PJ_WIN32) && PJ_WIN32 != 0
+#if (defined(PJ_WIN32) && PJ_WIN32 != 0) || (defined(PJ_WIN64) && PJ_WIN64 != 0)
 #include <windows.h>
 static void boost_priority(void)
 {
@@ -1142,7 +1180,7 @@ static void boost_priority(void)
 /*
  * This callback is called by media transport on receipt of RTP packet.
  */
-static void on_rx_rtp(void *user_data, const void *pkt, pj_ssize_t size)
+static void on_rx_rtp(void *user_data, void *pkt, pj_ssize_t size)
 {
     struct media_stream *strm;
     pj_status_t status;
@@ -1158,13 +1196,13 @@ static void on_rx_rtp(void *user_data, const void *pkt, pj_ssize_t size)
 
     /* Check for errors */
     if (size < 0) {
-	app_perror(THIS_FILE, "RTP recv() error", -size);
+	app_perror(THIS_FILE, "RTP recv() error", (pj_status_t)-size);
 	return;
     }
 
     /* Decode RTP packet. */
     status = pjmedia_rtp_decode_rtp(&strm->in_sess, 
-				    pkt, size, 
+				    pkt, (int)size, 
 				    &hdr, &payload, &payload_len);
     if (status != PJ_SUCCESS) {
 	app_perror(THIS_FILE, "RTP decode error", status);
@@ -1185,7 +1223,7 @@ static void on_rx_rtp(void *user_data, const void *pkt, pj_ssize_t size)
 /*
  * This callback is called by media transport on receipt of RTCP packet.
  */
-static void on_rx_rtcp(void *user_data, const void *pkt, pj_ssize_t size)
+static void on_rx_rtcp(void *user_data, void *pkt, pj_ssize_t size)
 {
     struct media_stream *strm;
 
@@ -1197,7 +1235,7 @@ static void on_rx_rtcp(void *user_data, const void *pkt, pj_ssize_t size)
 
     /* Check for errors */
     if (size < 0) {
-	app_perror(THIS_FILE, "Error receiving RTCP packet", -size);
+	app_perror(THIS_FILE, "Error receiving RTCP packet",(pj_status_t)-size);
 	return;
     }
 
@@ -1326,7 +1364,7 @@ static int media_thread(void *arg)
 	    /*
 	     * Time to send RTCP packet.
 	     */
-	    pjmedia_rtcp_pkt *rtcp_pkt;
+	    void *rtcp_pkt;
 	    int rtcp_len;
 	    pj_ssize_t size;
 	    pj_status_t status;
@@ -1435,12 +1473,14 @@ static void call_on_media_update( pjsip_inv_session *inv,
 
     /* Start media thread. */
     audio->thread_quit_flag = 0;
+#if PJ_HAS_THREADS
     status = pj_thread_create( inv->pool, "media", &media_thread, audio,
 			       0, 0, &audio->thread);
     if (status != PJ_SUCCESS) {
 	app_perror(THIS_FILE, "Error creating media thread", status);
 	return;
     }
+#endif
 
     /* Set the media as active */
     audio->active = PJ_TRUE;
@@ -1621,7 +1661,7 @@ static void print_avg_stat(void)
 	/* Jitter  */
 	MIN_(min_stat.rx.jitter.min, audio->rtcp.stat.rx.jitter.min);
 	MAX_(max_stat.rx.jitter.max, audio->rtcp.stat.rx.jitter.max);
-	AVG_(avg_stat.rx.jitter.avg, audio->rtcp.stat.rx.jitter.avg);
+	AVG_(avg_stat.rx.jitter.mean, audio->rtcp.stat.rx.jitter.mean);
 
 
 	/* TX Statistisc: */
@@ -1654,13 +1694,13 @@ static void print_avg_stat(void)
 	/* Jitter  */
 	MIN_(min_stat.tx.jitter.min, audio->rtcp.stat.tx.jitter.min);
 	MAX_(max_stat.tx.jitter.max, audio->rtcp.stat.tx.jitter.max);
-	AVG_(avg_stat.tx.jitter.avg, audio->rtcp.stat.tx.jitter.avg);
+	AVG_(avg_stat.tx.jitter.mean, audio->rtcp.stat.tx.jitter.mean);
 
 
 	/* RTT */
 	MIN_(min_stat.rtt.min, audio->rtcp.stat.rtt.min);
 	MAX_(max_stat.rtt.max, audio->rtcp.stat.rtt.max);
-	AVG_(avg_stat.rtt.avg, audio->rtcp.stat.rtt.avg);
+	AVG_(avg_stat.rtt.mean, audio->rtcp.stat.rtt.mean);
 
 	++count;
     }
@@ -1729,7 +1769,7 @@ static void print_avg_stat(void)
 	   "packets",
 
 	   min_stat.rx.jitter.min/1000.0, 
-	   avg_stat.rx.jitter.avg/1000.0, 
+	   avg_stat.rx.jitter.mean/1000.0, 
 	   max_stat.rx.jitter.max/1000.0,
 	   "ms",
 	
@@ -1760,13 +1800,13 @@ static void print_avg_stat(void)
 	   "packets",
 
 	   min_stat.tx.jitter.min/1000.0, 
-	   avg_stat.tx.jitter.avg/1000.0, 
+	   avg_stat.tx.jitter.mean/1000.0, 
 	   max_stat.tx.jitter.max/1000.0,
 	   "ms",
 
 	   /* rtt */
 	   min_stat.rtt.min/1000.0, 
-	   avg_stat.rtt.avg/1000.0, 
+	   avg_stat.rtt.mean/1000.0, 
 	   max_stat.rtt.max/1000.0,
 	   "ms"
 	   );
@@ -1808,6 +1848,7 @@ static void hangup_all_calls()
 	if (!app.call[i].inv)
 	    continue;
 	hangup_call(i);
+	pj_thread_sleep(app.call_gap);
     }
     
     /* Wait until all calls are terminated */
@@ -1822,7 +1863,8 @@ static pj_bool_t simple_input(const char *title, char *buf, pj_size_t len)
     char *p;
 
     printf("%s (empty to cancel): ", title); fflush(stdout);
-    fgets(buf, len, stdin);
+    if (fgets(buf, (int)len, stdin) == NULL)
+	return PJ_FALSE;
 
     /* Remove trailing newlines. */
     for (p=buf; ; ++p) {
@@ -1858,7 +1900,10 @@ static void console_main()
 
     for (;;) {
 	printf(">>> "); fflush(stdout);
-	fgets(input1, sizeof(input1), stdin);
+	if (fgets(input1, sizeof(input1), stdin) == NULL) {
+	    puts("EOF while reading stdin, will quit now..");
+	    break;
+	}
 
 	switch (input1[0]) {
 
@@ -1980,7 +2025,8 @@ static void app_log_writer(int level, const char *buffer, int len)
 	pj_log_write(level, buffer, len);
 
     if (log_file) {
-	fwrite(buffer, len, 1, log_file);
+	pj_size_t count = fwrite(buffer, len, 1, log_file);
+	PJ_UNUSED_ARG(count);
 	fflush(log_file);
     }
 }
@@ -2070,10 +2116,12 @@ int main(int argc, char *argv[])
     }
 
     /* Start worker threads */
+#if PJ_HAS_THREADS
     for (i=0; i<app.thread_count; ++i) {
 	pj_thread_create( app.pool, "app", &sip_worker_thread, NULL,
 			  0, 0, &app.sip_thread[i]);
     }
+#endif
 
     /* If URL is specified, then make call immediately */
     if (app.uri_to_call.slen) {
@@ -2088,6 +2136,7 @@ int main(int argc, char *argv[])
 		app_perror(THIS_FILE, "Error making call", status);
 		break;
 	    }
+	    pj_thread_sleep(app.call_gap);
 	}
 
 	if (app.auto_quit) {
@@ -2096,8 +2145,10 @@ int main(int argc, char *argv[])
 		pj_thread_sleep(100);
 	    pj_thread_sleep(200);
 	} else {
+#if PJ_HAS_THREADS
 	    /* Start user interface loop */
 	    console_main();
+#endif
 	}
 
     } else {
@@ -2105,11 +2156,19 @@ int main(int argc, char *argv[])
 	PJ_LOG(3,(THIS_FILE, "Ready for incoming calls (max=%d)", 
 		  app.max_calls));
 
+#if PJ_HAS_THREADS
 	/* Start user interface loop */
 	console_main();
-
+#endif
     }
 
+#if !PJ_HAS_THREADS
+    PJ_LOG(3,(THIS_FILE, "Press Ctrl-C to quit"));
+    for (;;) {
+	pj_time_val t = {0, 10};
+	pjsip_endpt_handle_events(app.sip_endpt, &t);
+    }
+#endif
     
     /* Shutting down... */
     destroy_sip();
