@@ -1,6 +1,7 @@
-/* $Id: evsub.c 1241 2007-05-02 05:14:29Z bennylp $ */
+/* $Id$ */
 /* 
- * Copyright (C) 2003-2007 Benny Prijono <benny@prijono.org>
+ * Copyright (C) 2008-2011 Teluu Inc. (http://www.teluu.com)
+ * Copyright (C) 2003-2008 Benny Prijono <benny@prijono.org>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -31,6 +32,7 @@
 #include <pj/log.h>
 #include <pj/os.h>
 #include <pj/pool.h>
+#include <pj/rand.h>
 #include <pj/string.h>
 
 
@@ -49,18 +51,35 @@ enum
     PJSIP_NOTIFY_METHOD = PJSIP_OTHER_METHOD
 };
 
-const pjsip_method pjsip_subscribe_method = 
+PJ_DEF_DATA(const pjsip_method) pjsip_subscribe_method = 
 {
     (pjsip_method_e) PJSIP_SUBSCRIBE_METHOD,
     { "SUBSCRIBE", 9 }
 };
 
-const pjsip_method pjsip_notify_method = 
+PJ_DEF_DATA(const pjsip_method) pjsip_notify_method = 
 {
     (pjsip_method_e) PJSIP_NOTIFY_METHOD,
     { "NOTIFY", 6 }
 };
 
+/**
+ * SUBSCRIBE method constant.
+ */
+PJ_DEF(const pjsip_method*) pjsip_get_subscribe_method()
+{
+    return &pjsip_subscribe_method;
+}
+
+/**
+ * NOTIFY method constant.
+ */
+PJ_DEF(const pjsip_method*) pjsip_get_notify_method()
+{
+    return &pjsip_notify_method;
+}
+
+
 /*
  * Static prototypes.
  */
@@ -87,16 +106,16 @@ static pj_str_t evsub_state_names[] =
  */
 
 /* Number of seconds to send SUBSCRIBE before the actual expiration */
-#define TIME_UAC_REFRESH	5
+#define TIME_UAC_REFRESH	PJSIP_EVSUB_TIME_UAC_REFRESH
 
 /* Time to wait for the final NOTIFY after sending unsubscription */
-#define TIME_UAC_TERMINATE	5
+#define TIME_UAC_TERMINATE	PJSIP_EVSUB_TIME_UAC_TERMINATE
 
 /* If client responds NOTIFY with non-2xx final response (such as 401),
  * wait for this seconds for further NOTIFY, otherwise client will
  * unsubscribe
  */
-#define TIME_UAC_WAIT_NOTIFY	5
+#define TIME_UAC_WAIT_NOTIFY	PJSIP_EVSUB_TIME_UAC_WAIT_NOTIFY
 
 
 /*
@@ -204,15 +223,18 @@ struct pjsip_evsub
     pj_str_t		  state_str;	/**< String describing the state.   */
     pjsip_evsub_state	  dst_state;	/**< Pending state to be set.	    */
     pj_str_t		  dst_state_str;/**< Pending state to be set.	    */
+    pj_str_t		  term_reason;	/**< Termination reason.	    */
     pjsip_method	  method;	/**< Method that established subscr.*/
     pjsip_event_hdr	 *event;	/**< Event description.		    */
     pjsip_expires_hdr	 *expires;	/**< Expires header		    */
     pjsip_accept_hdr	 *accept;	/**< Local Accept header.	    */
+    pjsip_hdr             sub_hdr_list; /**< User-defined header.           */
 
     pj_time_val		  refresh_time;	/**< Time to refresh.		    */
     pj_timer_entry	  timer;	/**< Internal timer.		    */
     int			  pending_tsx;	/**< Number of pending transactions.*/
     pjsip_transaction	 *pending_sub;	/**< Pending UAC SUBSCRIBE tsx.	    */
+    pj_timer_entry	 *pending_sub_timer; /**< Stop pending sub timer.   */
 
     void		 *mod_data[PJSIP_MAX_MODULE];	/**< Module data.   */
 };
@@ -231,6 +253,7 @@ struct dlgsub
 
 /* Static vars. */
 static const pj_str_t STR_EVENT	     = { "Event", 5 };
+static const pj_str_t STR_EVENT_S    = { "o", 1 };
 static const pj_str_t STR_SUB_STATE  = { "Subscription-State", 18 };
 static const pj_str_t STR_TERMINATED = { "terminated", 10 };
 static const pj_str_t STR_ACTIVE     = { "active", 6 };
@@ -266,8 +289,10 @@ PJ_DEF(pj_status_t) pjsip_evsub_init_module(pjsip_endpoint *endpt)
 	{ "NOTIFY", 6}
     };
 
-    pj_register_strerror(PJSIP_SIMPLE_ERRNO_START, PJ_ERRNO_SPACE_SIZE,
-			 &pjsipsimple_strerror);
+    status = pj_register_strerror(PJSIP_SIMPLE_ERRNO_START,
+				  PJ_ERRNO_SPACE_SIZE,
+				  &pjsipsimple_strerror);
+    pj_assert(status == PJ_SUCCESS);
 
     PJ_ASSERT_RETURN(endpt != NULL, PJ_EINVAL);
     PJ_ASSERT_RETURN(mod_evsub.mod.id == -1, PJ_EINVALIDOP);
@@ -279,7 +304,7 @@ PJ_DEF(pj_status_t) pjsip_evsub_init_module(pjsip_endpoint *endpt)
     pj_list_init(&mod_evsub.pkg_list);
 
     /* Create pool: */
-    mod_evsub.pool = pjsip_endpt_create_pool(endpt, "evsub", 4000, 4000);
+    mod_evsub.pool = pjsip_endpt_create_pool(endpt, "evsub", 512, 512);
     if (!mod_evsub.pool)
 	return PJ_ENOMEM;
 
@@ -388,6 +413,9 @@ PJ_DEF(pj_status_t) pjsip_evsub_register_pkg( pjsip_module *pkg_mod,
     PJ_ASSERT_RETURN(accept_cnt < PJ_ARRAY_SIZE(pkg->pkg_accept->values), 
 		     PJ_ETOOMANY);
 
+    /* Make sure evsub module has been initialized */
+    PJ_ASSERT_RETURN(mod_evsub.mod.id != -1, PJ_EINVALIDOP);
+
     /* Make sure no module with the specified name already registered: */
 
     PJ_ASSERT_RETURN(find_pkg(event_name) == NULL, PJSIP_SIMPLE_EPKGEXISTS);
@@ -508,6 +536,13 @@ static void evsub_destroy( pjsip_evsub *sub )
     /* Kill timer */
     set_timer(sub, TIMER_TYPE_NONE, 0);
 
+    /* Kill timer for stopping pending sub (see ticket #1807) */
+    if (sub->pending_sub_timer && sub->pending_sub_timer->id == 1) {
+	pjsip_endpt_cancel_timer(sub->endpt, sub->pending_sub_timer);
+	sub->pending_sub_timer->id = 0;
+	sub->pending_sub_timer = NULL;
+    }
+
     /* Remove this session from dialog's list of subscription */
     dlgsub_head = (struct dlgsub *) sub->dlg->mod_data[mod_evsub.mod.id];
     dlgsub = dlgsub_head->next;
@@ -529,10 +564,12 @@ static void evsub_destroy( pjsip_evsub *sub )
  * Set subscription session state.
  */
 static void set_state( pjsip_evsub *sub, pjsip_evsub_state state,
-		       const pj_str_t *state_str, pjsip_event *event)
+		       const pj_str_t *state_str, pjsip_event *event,
+		       const pj_str_t *reason)
 {
     pjsip_evsub_state prev_state = sub->state;
     pj_str_t old_state_str = sub->state_str;
+    pjsip_event dummy_event;
 
     sub->state = state;
 
@@ -541,12 +578,22 @@ static void set_state( pjsip_evsub *sub, pjsip_evsub_state state,
     else
 	sub->state_str = evsub_state_names[state];
 
+    if (reason && sub->term_reason.slen==0)
+	pj_strdup(sub->pool, &sub->term_reason, reason);
+
     PJ_LOG(4,(sub->obj_name, 
 	      "Subscription state changed %.*s --> %.*s",
 	      (int)old_state_str.slen,
 	      old_state_str.ptr,
 	      (int)sub->state_str.slen,
 	      sub->state_str.ptr));
+    pj_log_push_indent();
+
+    /* don't call the callback with NULL event, it may crash the app! */
+    if (!event) {
+	PJSIP_EVENT_INIT_USER(dummy_event, 0, 0, 0, 0);
+	event = &dummy_event;
+    }
 
     if (sub->user.on_evsub_state && sub->call_cb)
 	(*sub->user.on_evsub_state)(sub, event);
@@ -554,10 +601,15 @@ static void set_state( pjsip_evsub *sub, pjsip_evsub_state state,
     if (state == PJSIP_EVSUB_STATE_TERMINATED &&
 	prev_state != PJSIP_EVSUB_STATE_TERMINATED) 
     {
+	/* Kill any timer. */
+	set_timer(sub, TIMER_TYPE_NONE, 0);
+
 	if (sub->pending_tsx == 0) {
 	    evsub_destroy(sub);
 	}
     }
+
+    pj_log_pop_indent();
 }
 
 
@@ -590,11 +642,14 @@ static void on_timer( pj_timer_heap_t *timer_heap,
 	    pj_status_t status;
 
 	    PJ_LOG(5,(sub->obj_name, "Refreshing subscription."));
+	    pj_log_push_indent();
 	    status = pjsip_evsub_initiate(sub, NULL, 
 					  sub->expires->ivalue,
 					  &tdata);
 	    if (status == PJ_SUCCESS)
 		pjsip_evsub_send_request(sub, tdata);
+
+	    pj_log_pop_indent();
 	}
 	break;
 
@@ -608,18 +663,26 @@ static void on_timer( pj_timer_heap_t *timer_heap,
 
 	    PJ_LOG(5,(sub->obj_name, "Timeout waiting for refresh. "
 				     "Sending NOTIFY to terminate."));
+	    pj_log_push_indent();
 	    status = pjsip_evsub_notify( sub, PJSIP_EVSUB_STATE_TERMINATED, 
 					 NULL, &STR_TIMEOUT, &tdata);
 	    if (status == PJ_SUCCESS)
 		pjsip_evsub_send_request(sub, tdata);
+
+	    pj_log_pop_indent();
 	}
 	break;
 
     case TIMER_TYPE_UAC_TERMINATE:
 	{
+	    pj_str_t timeout = {"timeout", 7};
+
 	    PJ_LOG(5,(sub->obj_name, "Timeout waiting for final NOTIFY. "
 				     "Terminating.."));
-	    set_state(sub, PJSIP_EVSUB_STATE_TERMINATED, NULL, NULL);
+	    pj_log_push_indent();
+	    set_state(sub, PJSIP_EVSUB_STATE_TERMINATED, NULL, NULL, 
+		      &timeout);
+	    pj_log_pop_indent();
 	}
 	break;
 
@@ -632,9 +695,12 @@ static void on_timer( pj_timer_heap_t *timer_heap,
 		     "Timeout waiting for subsequent NOTIFY (we did "
 		     "send non-2xx response for previous NOTIFY). "
 		     "Unsubscribing.."));
+	    pj_log_push_indent();
 	    status = pjsip_evsub_initiate( sub, NULL, 0, &tdata);
 	    if (status == PJ_SUCCESS)
 		pjsip_evsub_send_request(sub, tdata);
+
+	    pj_log_pop_indent();
 	}
 	break;
 
@@ -686,6 +752,7 @@ static pj_status_t evsub_create( pjsip_dialog *dlg,
     sub->expires = pjsip_expires_hdr_create(sub->pool, pkg->pkg_expires);
     sub->accept = (pjsip_accept_hdr*) 
     		  pjsip_hdr_clone(sub->pool, pkg->pkg_accept);
+    pj_list_init(&sub->sub_hdr_list);
 
     sub->timer.user_data = sub;
     sub->timer.cb = &on_timer;
@@ -705,25 +772,33 @@ static pj_status_t evsub_create( pjsip_dialog *dlg,
     pj_strdup(sub->pool, &sub->event->event_type, event);
 
 
-    /* Create subcription list: */
+    /* Check if another subscription has been registered to the dialog. In
+     * that case, just add ourselves to the subscription list, otherwise
+     * create and register a new subscription list.
+     */
+    if (pjsip_dlg_has_usage(dlg, &mod_evsub.mod)) {
+	dlgsub_head = (struct dlgsub*) dlg->mod_data[mod_evsub.mod.id];
+	dlgsub = PJ_POOL_ALLOC_T(sub->pool, struct dlgsub);
+	dlgsub->sub = sub;
+	pj_list_push_back(dlgsub_head, dlgsub);
+    } else {
+	dlgsub_head = PJ_POOL_ALLOC_T(sub->pool, struct dlgsub);
+	dlgsub = PJ_POOL_ALLOC_T(sub->pool, struct dlgsub);
+	dlgsub->sub = sub;
 
-    dlgsub_head = PJ_POOL_ALLOC_T(sub->pool, struct dlgsub);
-    dlgsub = PJ_POOL_ALLOC_T(sub->pool, struct dlgsub);
-    dlgsub->sub = sub;
-
-    pj_list_init(dlgsub_head);
-    pj_list_push_back(dlgsub_head, dlgsub);
+	pj_list_init(dlgsub_head);
+	pj_list_push_back(dlgsub_head, dlgsub);
 
 
-    /* Register as dialog usage: */
+	/* Register as dialog usage: */
 
-    status = pjsip_dlg_add_usage(dlg, &mod_evsub.mod, dlgsub_head);
-    if (status != PJ_SUCCESS) {
-	pjsip_dlg_dec_lock(dlg);
-	return status;
+	status = pjsip_dlg_add_usage(dlg, &mod_evsub.mod, dlgsub_head);
+	if (status != PJ_SUCCESS) {
+	    pjsip_dlg_dec_lock(dlg);
+	    return status;
+	}
     }
 
-
     PJ_LOG(5,(sub->obj_name, "%s subscription created, using dialog %s",
 	      (role==PJSIP_ROLE_UAC ? "UAC" : "UAS"),
 	      dlg->obj_name));
@@ -812,7 +887,8 @@ PJ_DEF(pj_status_t) pjsip_evsub_create_uas( pjsip_dialog *dlg,
      * the package name (don't want to add more arguments in the function).
      */
     event_hdr = (pjsip_event_hdr*) 
-	pjsip_msg_find_hdr_by_name(rdata->msg_info.msg, &STR_EVENT, NULL);
+		 pjsip_msg_find_hdr_by_names(rdata->msg_info.msg, &STR_EVENT,
+					     &STR_EVENT_S, NULL);
     if (event_hdr == NULL) {
 	return PJSIP_ERRNO_FROM_SIP_STATUS(PJSIP_SC_BAD_REQUEST);
     }
@@ -890,7 +966,7 @@ PJ_DEF(pj_status_t) pjsip_evsub_terminate( pjsip_evsub *sub,
     */
 
     sub->call_cb = notify;
-    set_state(sub, PJSIP_EVSUB_STATE_TERMINATED, NULL, NULL);
+    set_state(sub, PJSIP_EVSUB_STATE_TERMINATED, NULL, NULL, NULL);
 
     pjsip_dlg_dec_lock(sub->dlg);
     return PJ_SUCCESS;
@@ -912,6 +988,13 @@ PJ_DEF(const char*) pjsip_evsub_get_state_name(pjsip_evsub *sub)
     return sub->state_str.ptr;
 }
 
+/*
+ * Get termination reason.
+ */
+PJ_DEF(const pj_str_t*) pjsip_evsub_get_termination_reason(pjsip_evsub *sub)
+{
+    return &sub->term_reason;
+}
 
 /*
  * Initiate client subscription
@@ -951,6 +1034,19 @@ PJ_DEF(pj_status_t) pjsip_evsub_initiate( pjsip_evsub *sub,
     pjsip_msg_add_hdr( tdata->msg, (pjsip_hdr*)
 		       pjsip_hdr_shallow_clone(tdata->pool, sub->expires));
 
+    /* Add Supported header (it's optional in RFC 3265, but some event package
+     * RFC may bring this requirement to SHOULD strength - e.g. RFC 5373)
+     */
+    {
+       const pjsip_hdr *hdr = pjsip_endpt_get_capability(sub->endpt,
+						         PJSIP_H_SUPPORTED,
+						         NULL);
+       if (hdr) {
+	   pjsip_msg_add_hdr(tdata->msg, (pjsip_hdr*)
+			     pjsip_hdr_shallow_clone(tdata->pool, hdr));
+       }
+    }
+
     /* Add Accept header: */
     pjsip_msg_add_hdr( tdata->msg, (pjsip_hdr*)
 		       pjsip_hdr_shallow_clone(tdata->pool, sub->accept));
@@ -961,7 +1057,18 @@ PJ_DEF(pj_status_t) pjsip_evsub_initiate( pjsip_evsub *sub,
 		       pjsip_hdr_shallow_clone(tdata->pool, 
 					       mod_evsub.allow_events_hdr));
 
-   
+
+    /* Add custom headers */
+    {
+	const pjsip_hdr *hdr = sub->sub_hdr_list.next;
+	while (hdr != &sub->sub_hdr_list) {
+	    pjsip_msg_add_hdr( tdata->msg, (pjsip_hdr*)
+			       pjsip_hdr_shallow_clone(tdata->pool, hdr));
+	    hdr = hdr->next;
+	}
+    }
+
+
     *p_tdata = tdata;
 
 
@@ -972,6 +1079,27 @@ on_return:
 }
 
 
+/*
+ * Add custom headers.
+ */
+PJ_DEF(pj_status_t) pjsip_evsub_add_header( pjsip_evsub *sub,
+					    const pjsip_hdr *hdr_list )
+{
+    const pjsip_hdr *hdr;
+
+    PJ_ASSERT_RETURN(sub && hdr_list, PJ_EINVAL);
+
+    hdr = hdr_list->next;
+    while (hdr != hdr_list) {
+        pj_list_push_back(&sub->sub_hdr_list, (pjsip_hdr*)
+		          pjsip_hdr_clone(sub->pool, hdr));
+	hdr = hdr->next;
+    }
+
+    return PJ_SUCCESS;
+}
+
+
 /*
  * Accept incoming subscription request.
  */
@@ -1030,6 +1158,14 @@ PJ_DEF(pj_status_t) pjsip_evsub_accept( pjsip_evsub *sub,
     if (status != PJ_SUCCESS)
 	goto on_return;
 
+    /* Set UAS timeout timer, when status code is 2xx and state is not
+     * terminated.
+     */
+    if (st_code/100 == 2 && sub->state != PJSIP_EVSUB_STATE_TERMINATED) {
+	PJ_LOG(5,(sub->obj_name, "UAS timeout in %d seconds",
+		  sub->expires->ivalue));
+	set_timer(sub, TIMER_TYPE_UAS_TIMEOUT, sub->expires->ivalue);
+    }
 
 on_return:
 
@@ -1063,10 +1199,10 @@ static pjsip_sub_state_hdr* sub_state_create( pj_pool_t *pool,
     switch (state) {
     case PJSIP_EVSUB_STATE_NULL:
     case PJSIP_EVSUB_STATE_SENT:
-    case PJSIP_EVSUB_STATE_ACCEPTED:
 	pj_assert(!"Invalid state!");
 	/* Treat as pending */
 
+    case PJSIP_EVSUB_STATE_ACCEPTED:
     case PJSIP_EVSUB_STATE_PENDING:
 	sub_state->sub_state = STR_PENDING;
 	sub_state->expires_param = delay.sec;
@@ -1112,8 +1248,8 @@ PJ_DEF(pj_status_t) pjsip_evsub_notify( pjsip_evsub *sub,
     pjsip_dlg_inc_lock(sub->dlg);
 
     /* Create NOTIFY request */
-    status = pjsip_dlg_create_request( sub->dlg, &pjsip_notify_method, -1, 
-				       &tdata);
+    status = pjsip_dlg_create_request( sub->dlg, pjsip_get_notify_method(), 
+				       -1, &tdata);
     if (status != PJ_SUCCESS)
 	goto on_return;
 
@@ -1133,7 +1269,9 @@ PJ_DEF(pj_status_t) pjsip_evsub_notify( pjsip_evsub *sub,
     /* Add Authentication headers. */
     pjsip_auth_clt_init_req( &sub->dlg->auth_sess, tdata );
 
-    
+    /* Update reason */
+    if (reason)
+	pj_strdup(sub->dlg->pool, &sub->term_reason, reason);
 
     /* Save destination state. */
     sub->dst_state = state;
@@ -1196,7 +1334,7 @@ PJ_DEF(pj_status_t) pjsip_evsub_send_request( pjsip_evsub *sub,
 
 	set_state(sub, sub->dst_state, 
 		  (sub->dst_state_str.slen ? &sub->dst_state_str : NULL), 
-		  NULL);
+		  NULL, NULL);
 
 	sub->dst_state = PJSIP_EVSUB_STATE_NULL;
 	sub->dst_state_str.slen = 0;
@@ -1210,6 +1348,26 @@ on_return:
 }
 
 
+/* Callback to be called to terminate transaction. */
+static void terminate_timer_cb(pj_timer_heap_t *timer_heap,
+			       struct pj_timer_entry *entry)
+{
+    pj_str_t *key;
+    pjsip_transaction *tsx;
+
+    PJ_UNUSED_ARG(timer_heap);
+
+    /* Clear timer ID */
+    entry->id = 0;
+
+    key = (pj_str_t*)entry->user_data;
+    tsx = pjsip_tsx_layer_find_tsx(key, PJ_FALSE);
+    /* Chance of race condition here */
+    if (tsx) {
+	pjsip_tsx_terminate(tsx, PJSIP_SC_REQUEST_UPDATED);
+    }
+}
+
 
 /*
  * Attach subscription session to newly created transaction, if appropriate.
@@ -1260,7 +1418,8 @@ static pjsip_evsub *on_new_transaction( pjsip_transaction *tsx,
     }
 
     event_hdr = (pjsip_event_hdr*)
-    		pjsip_msg_find_hdr_by_name(msg, &STR_EVENT, NULL);
+    		pjsip_msg_find_hdr_by_names(msg, &STR_EVENT, 
+					    &STR_EVENT_S, NULL);
     if (!event_hdr) {
 	/* Not subscription related message */
 	return NULL;
@@ -1289,8 +1448,18 @@ static pjsip_evsub *on_new_transaction( pjsip_transaction *tsx,
 	    if (pj_strcmp(&dlgsub->sub->event->id_param, 
 			  &event_hdr->id_param)==0)
 	    {
-		
-		break;
+		/* Skip this subscription if it has no event ID and has been
+		 * terminated (see ticket #1647).
+		 */
+		if ((dlgsub->sub->option & PJSIP_EVSUB_NO_EVENT_ID) &&
+		    (pjsip_evsub_get_state(dlgsub->sub)==
+					PJSIP_EVSUB_STATE_TERMINATED))
+		{
+		    dlgsub = dlgsub->next;
+    		    continue;
+		} else {
+		    break;
+		}
 
 	    }
 	    /*
@@ -1320,10 +1489,21 @@ static pjsip_evsub *on_new_transaction( pjsip_transaction *tsx,
 	dlgsub = dlgsub->next;
     }
 
-    if (dlgsub == dlgsub_head) {
+    /* Note: 
+     *  the second condition is for http://trac.pjsip.org/repos/ticket/911 
+     */
+    if (dlgsub == dlgsub_head ||
+	(dlgsub->sub && 
+	    pjsip_evsub_get_state(dlgsub->sub)==PJSIP_EVSUB_STATE_TERMINATED))
+    {
+	const char *reason_msg = 
+	    (dlgsub == dlgsub_head ? "Subscription Does Not Exist" :
+	     "Subscription already terminated");
+
 	/* This could be incoming request to create new subscription */
 	PJ_LOG(4,(THIS_FILE, 
-		  "Subscription not found for %.*s, event=%.*s;id=%.*s",
+		  "%s for %.*s, event=%.*s;id=%.*s",
+		  reason_msg,
 		  (int)tsx->method.name.slen,
 		  tsx->method.name.ptr,
 		  (int)event_hdr->event_type.slen,
@@ -1335,10 +1515,11 @@ static pjsip_evsub *on_new_transaction( pjsip_transaction *tsx,
 	if (tsx->state == PJSIP_TSX_STATE_TRYING &&
 	    pjsip_method_cmp(&tsx->method, &pjsip_notify_method)==0)
 	{
-	    pj_str_t reason = pj_str("Subscription Does Not Exist");
+	    pj_str_t reason;
 	    pjsip_tx_data *tdata;
 	    pj_status_t status;
 
+	    pj_cstr(&reason, reason_msg);
 	    status = pjsip_dlg_create_response(dlg, 
 					       event->body.tsx_state.src.rdata, 
 					       481, &reason, 
@@ -1375,6 +1556,10 @@ static pjsip_evsub *on_new_transaction( pjsip_transaction *tsx,
 	if (sub->pending_sub && 
 	    sub->pending_sub->state < PJSIP_TSX_STATE_COMPLETED) 
 	{
+	    pj_timer_entry *timer;
+	    pj_str_t *key;
+	    pj_time_val timeout = {0, 0};
+
 	    PJ_LOG(4,(sub->obj_name, 
 		      "Cancelling pending subscription request"));
 
@@ -1382,7 +1567,24 @@ static pjsip_evsub *on_new_transaction( pjsip_transaction *tsx,
 	     * When transaction handler (below) see this status code, it
 	     * will ignore the transaction.
 	     */
-	    pjsip_tsx_terminate(sub->pending_sub, PJSIP_SC_REQUEST_UPDATED);
+	    /* This unfortunately may cause deadlock, because at the moment
+	     * we are holding dialog's mutex. If a response to this
+	     * transaction is in progress in another thread, that thread
+	     * will deadlock when trying to acquire dialog mutex, because
+	     * it is holding the transaction mutex.
+	     *
+	     * So the solution is to register timer to kill this transaction.
+	     */
+	    //pjsip_tsx_terminate(sub->pending_sub, PJSIP_SC_REQUEST_UPDATED);
+	    timer = PJ_POOL_ZALLOC_T(dlg->pool, pj_timer_entry);
+	    key = PJ_POOL_ALLOC_T(dlg->pool, pj_str_t);
+	    pj_strdup(dlg->pool, key, &sub->pending_sub->transaction_key);
+	    timer->cb = &terminate_timer_cb;
+	    timer->user_data = key;
+	    timer->id = 1;
+	    sub->pending_sub_timer = timer;
+
+	    pjsip_endpt_schedule_timer(dlg->endpt, timer, &timeout);
 	}
 
 	sub->pending_sub = tsx;
@@ -1487,7 +1689,7 @@ static void on_tsx_state_uac( pjsip_evsub *sub, pjsip_transaction *tsx,
 	if (sub->state == PJSIP_EVSUB_STATE_NULL &&
 	    tsx->state == PJSIP_TSX_STATE_CALLING)
 	{
-	    set_state(sub, PJSIP_EVSUB_STATE_SENT, NULL, event);
+	    set_state(sub, PJSIP_EVSUB_STATE_SENT, NULL, event, NULL);
 	    return;
 	}
 
@@ -1499,8 +1701,15 @@ static void on_tsx_state_uac( pjsip_evsub *sub, pjsip_transaction *tsx,
 	}
 
 	/* Clear pending subscription */
-	if (tsx == sub->pending_sub)
+	if (tsx == sub->pending_sub) {
 	    sub->pending_sub = NULL;
+	} else if (sub->pending_sub != NULL) {
+	    /* This SUBSCRIBE transaction has been "renewed" with another
+	     * SUBSCRIBE, so we can just ignore this. For example, user
+	     * sent SUBSCRIBE followed immediately with UN-SUBSCRIBE.
+	     */
+	    return;
+	}
 
 	/* Handle authentication. */
 	if (tsx->status_code==401 || tsx->status_code==407) {
@@ -1521,8 +1730,7 @@ static void on_tsx_state_uac( pjsip_evsub *sub, pjsip_transaction *tsx,
 	    if (status != PJ_SUCCESS) {
 		/* Authentication failed! */
 		set_state(sub, PJSIP_EVSUB_STATE_TERMINATED,
-			       NULL,
-			       event);
+			  NULL, event, &tsx->status_text);
 		return;
 	    }
 
@@ -1566,6 +1774,10 @@ static void on_tsx_state_uac( pjsip_evsub *sub, pjsip_transaction *tsx,
 		unsigned timeout = (sub->expires->ivalue > TIME_UAC_REFRESH) ?
 		    sub->expires->ivalue - TIME_UAC_REFRESH : sub->expires->ivalue;
 
+		/* Reduce timeout by about 1 - 10 secs (randomized) */
+		if (timeout > 10)
+		    timeout += -10 + (pj_rand() % 10);
+
 		PJ_LOG(5,(sub->obj_name, "Will refresh in %d seconds", 
 			  timeout));
 		set_timer(sub, TIMER_TYPE_UAC_REFRESH, timeout);
@@ -1580,7 +1792,7 @@ static void on_tsx_state_uac( pjsip_evsub *sub, pjsip_transaction *tsx,
 	    /* Set state, if necessary */
 	    pj_assert(sub->state != PJSIP_EVSUB_STATE_NULL);
 	    if (sub->state == PJSIP_EVSUB_STATE_SENT) {
-		set_state(sub, PJSIP_EVSUB_STATE_ACCEPTED, NULL, event);
+		set_state(sub, PJSIP_EVSUB_STATE_ACCEPTED, NULL, event, NULL);
 	    }
 
 	} else {
@@ -1614,12 +1826,9 @@ static void on_tsx_state_uac( pjsip_evsub *sub, pjsip_transaction *tsx,
 		return;
 	    }
 
-	    /* Kill any timer. */
-	    set_timer(sub, TIMER_TYPE_NONE, 0);
-
 	    /* Set state to TERMINATED */
 	    set_state(sub, PJSIP_EVSUB_STATE_TERMINATED, 
-		      NULL, event);
+		      NULL, event, &tsx->status_text);
 
 	}
 
@@ -1645,7 +1854,6 @@ static void on_tsx_state_uac( pjsip_evsub *sub, pjsip_transaction *tsx,
 
 	pjsip_tx_data *tdata;
 	pj_status_t status;
-	int next_refresh;
 
 	/* Only want to handle initial NOTIFY receive event. */
 	if (tsx->state != PJSIP_TSX_STATE_TRYING)
@@ -1705,7 +1913,12 @@ static void on_tsx_state_uac( pjsip_evsub *sub, pjsip_transaction *tsx,
 		unsigned timeout = TIME_UAC_WAIT_NOTIFY;
 		set_timer(sub, TIMER_TYPE_UAC_WAIT_NOTIFY, timeout);
 	    } else {
-		set_state(sub, PJSIP_EVSUB_STATE_TERMINATED, NULL, NULL);
+		char errmsg[PJ_ERR_MSG_SIZE];
+		pj_str_t reason;
+
+		reason = pj_strerror(status, errmsg, sizeof(errmsg));
+		set_state(sub, PJSIP_EVSUB_STATE_TERMINATED, NULL, NULL, 
+			  &reason);
 	    }
 
 	    return;
@@ -1720,19 +1933,14 @@ static void on_tsx_state_uac( pjsip_evsub *sub, pjsip_transaction *tsx,
 	    (pj_stricmp(&sub_state->sub_state, &STR_ACTIVE)==0 ||
 	     pj_stricmp(&sub_state->sub_state, &STR_PENDING)==0))
 	{
-	    next_refresh = sub_state->expires_param;
+	    int next_refresh = sub_state->expires_param;
+	    unsigned timeout;
 
-	} else {
-	    next_refresh = sub->expires->ivalue;
-	}
+	    update_expires(sub, next_refresh);
 
-	/* Update time */
-	update_expires(sub, next_refresh);
-
-	/* Start UAC refresh timer, only when we're not unsubscribing */
-	if (sub->expires->ivalue != 0) {
-	    unsigned timeout = (next_refresh > TIME_UAC_REFRESH) ?
-		next_refresh - TIME_UAC_REFRESH : next_refresh;
+	    /* Start UAC refresh timer, only when we're not unsubscribing */
+	    timeout = (next_refresh > TIME_UAC_REFRESH) ?
+			next_refresh - TIME_UAC_REFRESH : next_refresh;
 
 	    PJ_LOG(5,(sub->obj_name, "Will refresh in %d seconds", timeout));
 	    set_timer(sub, TIMER_TYPE_UAC_REFRESH, timeout);
@@ -1749,9 +1957,15 @@ static void on_tsx_state_uac( pjsip_evsub *sub, pjsip_transaction *tsx,
 
 	/* Set the state */
 	if (status == PJ_SUCCESS) {
-	    set_state(sub, new_state, new_state_str, event);
+	    set_state(sub, new_state, new_state_str, event, 
+		      &sub_state->reason_param);
 	} else {
-	    set_state(sub, PJSIP_EVSUB_STATE_TERMINATED, NULL, event);
+	    char errmsg[PJ_ERR_MSG_SIZE];
+	    pj_str_t reason;
+
+	    reason = pj_strerror(status, errmsg, sizeof(errmsg));
+	    set_state(sub, PJSIP_EVSUB_STATE_TERMINATED, NULL, event, 
+		      &reason);
 	}
 
 
@@ -1792,6 +2006,7 @@ static void on_tsx_state_uas( pjsip_evsub *sub, pjsip_transaction *tsx,
 	pjsip_msg_body *body = NULL;
 	pjsip_evsub_state old_state;
 	pj_str_t old_state_str;
+	pj_str_t reason = { NULL, 0 };
 	pj_status_t status;
 
 
@@ -1808,7 +2023,8 @@ static void on_tsx_state_uas( pjsip_evsub *sub, pjsip_transaction *tsx,
 	 * or package default expiration time.
 	 */
 	event_hdr = (pjsip_event_hdr*)
-		    pjsip_msg_find_hdr_by_name(msg, &STR_EVENT, NULL);
+		    pjsip_msg_find_hdr_by_names(msg, &STR_EVENT, 
+					        &STR_EVENT, NULL);
 	expires = (pjsip_expires_hdr*)
 		  pjsip_msg_find_hdr(msg, PJSIP_H_EXPIRES, NULL);
 	if (event_hdr && expires) {
@@ -1877,9 +2093,9 @@ static void on_tsx_state_uas( pjsip_evsub *sub, pjsip_transaction *tsx,
 	if (st_code/100==2) {
 	    
 	    if (sub->expires->ivalue == 0) {
-		set_state(sub, sub->state, NULL, event);
+		set_state(sub, sub->state, NULL, event, &reason);
 	    } else  if (sub->state == PJSIP_EVSUB_STATE_NULL) {
-		set_state(sub, sub->state, NULL, event);
+		set_state(sub, sub->state, NULL, event, &reason);
 	    }
 
 	    /* Set UAS timeout timer, when state is not terminated. */
@@ -1913,7 +2129,8 @@ static void on_tsx_state_uas( pjsip_evsub *sub, pjsip_transaction *tsx,
 
 	    if (status != PJ_SUCCESS) {
 		/* Can't authenticate. Terminate session (?) */
-		set_state(sub, PJSIP_EVSUB_STATE_TERMINATED, NULL, NULL);
+		set_state(sub, PJSIP_EVSUB_STATE_TERMINATED, NULL, NULL, 
+			  &tsx->status_text);
 		return;
 	    }
 
@@ -1926,7 +2143,8 @@ static void on_tsx_state_uas( pjsip_evsub *sub, pjsip_transaction *tsx,
 	    (tsx->status_code==481 || tsx->status_code==408 ||
 	     tsx->status_code/100 == 7))
 	{
-	    set_state(sub, PJSIP_EVSUB_STATE_TERMINATED, NULL, event);
+	    set_state(sub, PJSIP_EVSUB_STATE_TERMINATED, NULL, event,
+		      &tsx->status_text);
 	    return;
 	}
 
