@@ -1,6 +1,7 @@
-/* $Id: publishc.c 1292 2007-05-23 14:12:35Z bennylp $ */
+/* $Id$ */
 /* 
- * Copyright (C) 2003-2007 Benny Prijono <benny@prijono.org>
+ * Copyright (C) 2008-2011 Teluu Inc. (http://www.teluu.com)
+ * Copyright (C) 2003-2008 Benny Prijono <benny@prijono.org>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -36,7 +37,7 @@
 
 
 #define REFRESH_TIMER		1
-#define DELAY_BEFORE_REFRESH	5
+#define DELAY_BEFORE_REFRESH	PJSIP_PUBLISHC_DELAY_BEFORE_REFRESH
 #define THIS_FILE		"publishc.c"
 
 
@@ -55,6 +56,16 @@ const pjsip_method pjsip_publish_method =
 };
 
 
+/**
+ * Pending request list.
+ */
+typedef struct pending_publish
+{
+    PJ_DECL_LIST_MEMBER(struct pending_publish);
+    pjsip_tx_data		*tdata;
+} pending_publish;
+
+
 /**
  * SIP client publication structure.
  */
@@ -64,7 +75,10 @@ struct pjsip_publishc
     pjsip_endpoint		*endpt;
     pj_bool_t			 _delete_flag;
     int				 pending_tsx;
+    pj_bool_t			 in_callback;
+    pj_mutex_t			*mutex;
 
+    pjsip_publishc_opt		 opt;
     void			*token;
     pjsip_publishc_cb		*cb;
 
@@ -80,6 +94,9 @@ struct pjsip_publishc
     pjsip_expires_hdr		*expires_hdr;
     pj_uint32_t			 expires;
     pjsip_route_hdr		 route_set;
+    pjsip_hdr			 usr_hdr;
+    pjsip_host_port              via_addr;
+    const void                  *via_tp;
 
     /* Authorization sessions. */
     pjsip_auth_clt_sess		 auth_sess;
@@ -89,35 +106,64 @@ struct pjsip_publishc
     pj_time_val			 last_refresh;
     pj_time_val			 next_refresh;
     pj_timer_entry		 timer;
+
+    /* Pending PUBLISH request */
+    pending_publish		 pending_reqs;
+    pending_publish		 pending_reqs_empty;
 };
 
 
+PJ_DEF(void) pjsip_publishc_opt_default(pjsip_publishc_opt *opt)
+{
+    pj_bzero(opt, sizeof(*opt));
+    opt->queue_request = PJSIP_PUBLISHC_QUEUE_REQUEST;
+}
+
 
 /*
  * Initialize client publication module.
  */
 PJ_DEF(pj_status_t) pjsip_publishc_init_module(pjsip_endpoint *endpt)
 {
+    /* Note:
+	Commented out the capability registration below, since it's
+	wrong to include PUBLISH in Allow header of INVITE requests/
+	responses.
+
+	13.2.1 Creating the Initial INVITE
+	  An Allow header field (Section 20.5) SHOULD be present in the 
+	  INVITE. It indicates what methods can be invoked within a dialog
+
+	20.5 Allow
+	  The Allow header field lists the set of methods supported by the
+	  UA generating the message.
+
+	While the semantic of Allow header in non-dialog requests is unclear,
+	it's probably best not to include PUBLISH in Allow header for now
+	until we can find out how to customize the inclusion of methods in
+	Allow header for in-dialog vs out-dialog requests.
+
     return pjsip_endpt_add_capability( endpt, NULL, PJSIP_H_ALLOW, NULL,
 				       1, &pjsip_publish_method.name);
+     */
+    PJ_UNUSED_ARG(endpt);
+    return PJ_SUCCESS;
 }
 
 
 PJ_DEF(pj_status_t) pjsip_publishc_create( pjsip_endpoint *endpt, 
-					   unsigned options,
+					   const pjsip_publishc_opt *opt,
 					   void *token,
 					   pjsip_publishc_cb *cb,	
 					   pjsip_publishc **p_pubc)
 {
     pj_pool_t *pool;
     pjsip_publishc *pubc;
+    pjsip_publishc_opt default_opt;
     pj_status_t status;
 
     /* Verify arguments. */
     PJ_ASSERT_RETURN(endpt && cb && p_pubc, PJ_EINVAL);
-    PJ_ASSERT_RETURN(options == 0, PJ_EINVAL);
-
-    PJ_UNUSED_ARG(options);
 
     pool = pjsip_endpt_create_pool(endpt, "pubc%p", 1024, 1024);
     PJ_ASSERT_RETURN(pool != NULL, PJ_ENOMEM);
@@ -130,11 +176,29 @@ PJ_DEF(pj_status_t) pjsip_publishc_create( pjsip_endpoint *endpt,
     pubc->cb = cb;
     pubc->expires = PJSIP_PUBC_EXPIRATION_NOT_SPECIFIED;
 
-    status = pjsip_auth_clt_init(&pubc->auth_sess, endpt, pubc->pool, 0);
-    if (status != PJ_SUCCESS)
+    if (!opt) {
+	pjsip_publishc_opt_default(&default_opt);
+	opt = &default_opt;
+    }
+    pj_memcpy(&pubc->opt, opt, sizeof(*opt));
+    pj_list_init(&pubc->pending_reqs);
+    pj_list_init(&pubc->pending_reqs_empty);
+
+    status = pj_mutex_create_recursive(pubc->pool, "pubc%p", &pubc->mutex);
+    if (status != PJ_SUCCESS) {
+	pj_pool_release(pool);
 	return status;
+    }
+
+    status = pjsip_auth_clt_init(&pubc->auth_sess, endpt, pubc->pool, 0);
+    if (status != PJ_SUCCESS) {
+	pj_mutex_destroy(pubc->mutex);
+	pj_pool_release(pool);
+	return status;
+    }
 
     pj_list_init(&pubc->route_set);
+    pj_list_init(&pubc->usr_hdr);
 
     /* Done */
     *p_pubc = pubc;
@@ -146,10 +210,18 @@ PJ_DEF(pj_status_t) pjsip_publishc_destroy(pjsip_publishc *pubc)
 {
     PJ_ASSERT_RETURN(pubc, PJ_EINVAL);
 
-    if (pubc->pending_tsx) {
+    if (pubc->pending_tsx || pubc->in_callback) {
 	pubc->_delete_flag = 1;
 	pubc->cb = NULL;
     } else {
+	/* Cancel existing timer, if any */
+	if (pubc->timer.id != 0) {
+	    pjsip_endpt_cancel_timer(pubc->endpt, &pubc->timer);
+	    pubc->timer.id = 0;
+	}
+
+	if (pubc->mutex)
+	    pj_mutex_destroy(pubc->mutex);
 	pjsip_endpt_release_pool(pubc->endpt, pubc->pool);
     }
 
@@ -164,7 +236,9 @@ PJ_DEF(pj_pool_t*) pjsip_publishc_get_pool(pjsip_publishc *pubc)
 
 static void set_expires( pjsip_publishc *pubc, pj_uint32_t expires)
 {
-    if (expires != pubc->expires) {
+    if (expires != pubc->expires && 
+	expires != PJSIP_PUBC_EXPIRATION_NOT_SPECIFIED) 
+    {
 	pubc->expires_hdr = pjsip_expires_hdr_create(pubc->pool, expires);
     } else {
 	pubc->expires_hdr = NULL;
@@ -259,6 +333,41 @@ PJ_DEF(pj_status_t) pjsip_publishc_set_route_set( pjsip_publishc *pubc,
     return PJ_SUCCESS;
 }
 
+PJ_DEF(pj_status_t) pjsip_publishc_set_headers( pjsip_publishc *pubc,
+						const pjsip_hdr *hdr_list)
+{
+    const pjsip_hdr *h;
+
+    PJ_ASSERT_RETURN(pubc && hdr_list, PJ_EINVAL);
+
+    pj_list_init(&pubc->usr_hdr);
+    h = hdr_list->next;
+    while (h != hdr_list) {
+	pj_list_push_back(&pubc->usr_hdr, pjsip_hdr_clone(pubc->pool, h));
+	h = h->next;
+    }
+
+    return PJ_SUCCESS;
+}
+
+PJ_DEF(pj_status_t) pjsip_publishc_set_via_sent_by(pjsip_publishc *pubc,
+				                   pjsip_host_port *via_addr,
+                                                   pjsip_transport *via_tp)
+{
+    PJ_ASSERT_RETURN(pubc, PJ_EINVAL);
+
+    if (!via_addr)
+        pj_bzero(&pubc->via_addr, sizeof(pubc->via_addr));
+    else {
+        if (pj_strcmp(&pubc->via_addr.host, &via_addr->host))
+            pj_strdup(pubc->pool, &pubc->via_addr.host, &via_addr->host);
+        pubc->via_addr.port = via_addr->port;
+    }
+    pubc->via_tp = via_tp;
+
+    return PJ_SUCCESS;
+}
+
 static pj_status_t create_request(pjsip_publishc *pubc, 
 				  pjsip_tx_data **p_tdata)
 {
@@ -323,6 +432,19 @@ static pj_status_t create_request(pjsip_publishc *pubc,
 	    pjsip_msg_add_hdr(tdata->msg, (pjsip_hdr*)hdr);
     }
 
+    /* Add user headers */
+    if (!pj_list_empty(&pubc->usr_hdr)) {
+	const pjsip_hdr *hdr;
+
+	hdr = pubc->usr_hdr.next;
+	while (hdr != &pubc->usr_hdr) {
+	    pjsip_hdr *new_hdr = (pjsip_hdr*)
+	    			 pjsip_hdr_shallow_clone(tdata->pool, hdr);
+	    pjsip_msg_add_hdr(tdata->msg, new_hdr);
+	    hdr = hdr->next;
+	}
+    }
+
 
     /* Done. */
     *p_tdata = tdata;
@@ -435,15 +557,15 @@ static void pubc_refresh_timer_cb( pj_timer_heap_t *timer_heap,
 
     entry->id = 0;
     status = pjsip_publishc_publish(pubc, 1, &tdata);
-    if (status == PJ_SUCCESS) {
-	status = pjsip_publishc_send(pubc, tdata);
-    } 
-    
     if (status != PJ_SUCCESS) {
 	char errmsg[PJ_ERR_MSG_SIZE];
 	pj_str_t reason = pj_strerror(status, errmsg, sizeof(errmsg));
 	call_callback(pubc, status, 400, &reason, NULL, -1);
+	return;
     }
+
+    status = pjsip_publishc_send(pubc, tdata);
+    /* No need to call callback as it should have been called */
 }
 
 static void tsx_callback(void *token, pjsip_event *event)
@@ -456,6 +578,9 @@ static void tsx_callback(void *token, pjsip_event *event)
     pj_assert(pubc->pending_tsx > 0);
     --pubc->pending_tsx;
 
+    /* Mark that we're in callback to prevent deletion (#1164) */
+    ++pubc->in_callback;
+
     /* If publication data has been deleted by user then remove publication 
      * data from transaction's callback, and don't call callback.
      */
@@ -474,19 +599,14 @@ static void tsx_callback(void *token, pjsip_event *event)
 					    rdata, 
 					    tsx->last_tx,  
 					    &tdata);
-
-	if (status == PJ_SUCCESS) {
-	    status = pjsip_publishc_send(pubc, tdata);
-	} 
-	
 	if (status != PJ_SUCCESS) {
 	    call_callback(pubc, status, tsx->status_code, 
 			  &rdata->msg_info.msg->line.status.reason,
 			  rdata, -1);
+	} else {
+    	    status = pjsip_publishc_send(pubc, tdata);
 	}
 
-	return;
-
     } else {
 	pjsip_rx_data *rdata;
 	pj_int32_t expiration = 0xFFFF;
@@ -513,12 +633,18 @@ static void tsx_callback(void *token, pjsip_event *event)
 	    expires = (pjsip_expires_hdr*)
 	    	      pjsip_msg_find_hdr(msg, PJSIP_H_EXPIRES, NULL);
 
-	    if (expires)
+	    if (pubc->auto_refresh && expires)
 		expiration = expires->ivalue;
 	    
 	    if (pubc->auto_refresh && expiration!=0 && expiration!=0xFFFF) {
 		pj_time_val delay = { 0, 0};
 
+		/* Cancel existing timer, if any */
+		if (pubc->timer.id != 0) {
+		    pjsip_endpt_cancel_timer(pubc->endpt, &pubc->timer);
+		    pubc->timer.id = 0;
+		}
+
 		delay.sec = expiration - DELAY_BEFORE_REFRESH;
 		if (pubc->expires != PJSIP_PUBC_EXPIRATION_NOT_SPECIFIED && 
 		    delay.sec > (pj_int32_t)pubc->expires) 
@@ -556,8 +682,56 @@ static void tsx_callback(void *token, pjsip_event *event)
 		      rdata, expiration);
 
 	--pubc->pending_tsx;
+
+	/* If we have pending request(s), send them now */
+	pj_mutex_lock(pubc->mutex);
+	while (!pj_list_empty(&pubc->pending_reqs)) {
+	    pending_publish *pp = pubc->pending_reqs.next;
+	    pjsip_tx_data *tdata = pp->tdata;
+
+	    /* Remove the request from pending request list,
+	     * and keep the unused entry into pending_reqs_empty pool.
+	     */
+	    pj_list_erase(pp);
+	    pj_list_push_back(&pubc->pending_reqs_empty, pp);
+
+	    /* Add SIP-If-Match if we have etag and the request doesn't have
+	     * one (http://trac.pjsip.org/repos/ticket/996)
+	     */
+	    if (pubc->etag.slen) {
+		const pj_str_t STR_HNAME = { "SIP-If-Match", 12 };
+		pjsip_generic_string_hdr *sim_hdr;
+
+		sim_hdr = (pjsip_generic_string_hdr*)
+			  pjsip_msg_find_hdr_by_name(tdata->msg, &STR_HNAME, NULL);
+		if (!sim_hdr) {
+		    /* Create the header */
+		    sim_hdr = pjsip_generic_string_hdr_create(tdata->pool,
+							      &STR_HNAME,
+							      &pubc->etag);
+		    pjsip_msg_add_hdr(tdata->msg, (pjsip_hdr*)sim_hdr);
+
+		} else {
+		    /* Update */
+		    if (pj_strcmp(&pubc->etag, &sim_hdr->hvalue))
+			pj_strdup(tdata->pool, &sim_hdr->hvalue, &pubc->etag);
+		}
+	    }
+
+	    status = pjsip_publishc_send(pubc, tdata);
+	    if (status == PJ_EPENDING) {
+		pj_assert(!"Not expected");
+		pjsip_tx_data_dec_ref(tdata);
+	    } else if (status == PJ_SUCCESS) {
+		break;
+	    }
+	}
+	pj_mutex_unlock(pubc->mutex);
     }
 
+    /* No longer in callback. */
+    --pubc->in_callback;
+
     /* Delete the record if user destroy pubc during the callback. */
     if (pubc->_delete_flag && pubc->pending_tsx==0) {
 	pjsip_publishc_destroy(pubc);
@@ -572,12 +746,39 @@ PJ_DEF(pj_status_t) pjsip_publishc_send(pjsip_publishc *pubc,
     pjsip_cseq_hdr *cseq_hdr;
     pj_uint32_t cseq;
 
+    PJ_ASSERT_RETURN(pubc && tdata, PJ_EINVAL);
+
     /* Make sure we don't have pending transaction. */
+    pj_mutex_lock(pubc->mutex);
     if (pubc->pending_tsx) {
-	PJ_LOG(4,(THIS_FILE, "Unable to send request, pubc has another "
-			     "transaction pending"));
-	pjsip_tx_data_dec_ref( tdata );
-	return PJSIP_EBUSY;
+	if (pubc->opt.queue_request) {
+	    pending_publish *pp = NULL;
+	    if (pj_list_empty(&pubc->pending_reqs_empty)) {
+		pp = PJ_POOL_ZALLOC_T(pubc->pool, pending_publish);
+	    } else {
+		pp = pubc->pending_reqs_empty.next;
+		pj_list_erase(pp);
+	    }
+	    pp->tdata = tdata;
+	    pj_list_push_back(&pubc->pending_reqs, pp);
+	    pj_mutex_unlock(pubc->mutex);
+	    PJ_LOG(4,(THIS_FILE, "Request is queued, pubc has another "
+				 "transaction pending"));
+	    return PJ_EPENDING;
+	} else {
+	    pjsip_tx_data_dec_ref(tdata);
+	    pj_mutex_unlock(pubc->mutex);
+	    PJ_LOG(4,(THIS_FILE, "Unable to send request, pubc has another "
+				 "transaction pending"));
+	    return PJ_EBUSY;
+	}
+    }
+    pj_mutex_unlock(pubc->mutex);
+
+    /* If via_addr is set, use this address for the Via header. */
+    if (pubc->via_addr.host.slen > 0) {
+        tdata->via_addr = pubc->via_addr;
+        tdata->via_tp = pubc->via_tp;
     }
 
     /* Invalidate message buffer. */
@@ -596,11 +797,13 @@ PJ_DEF(pj_status_t) pjsip_publishc_send(pjsip_publishc *pubc,
     status = pjsip_endpt_send_request(pubc->endpt, tdata, -1, pubc, 
 				      &tsx_callback);
     if (status!=PJ_SUCCESS) {
-	--pubc->pending_tsx;
+	// no need to decrement, callback has been called and it should
+	// already decremented pending_tsx. Decrementing this here may 
+	// cause accessing freed memory location.
+	//--pubc->pending_tsx;
 	PJ_LOG(4,(THIS_FILE, "Error sending request, status=%d", status));
     }
 
     return status;
 }
 
-
