@@ -1,6 +1,7 @@
-/* $Id: sip_endpoint.c 1323 2007-05-30 15:20:56Z bennylp $ */
+/* $Id$ */
 /* 
- * Copyright (C) 2003-2007 Benny Prijono <benny@prijono.org>
+ * Copyright (C) 2008-2011 Teluu Inc. (http://www.teluu.com)
+ * Copyright (C) 2003-2008 Benny Prijono <benny@prijono.org>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -34,11 +35,20 @@
 #include <pj/errno.h>
 #include <pj/lock.h>
 
-#define PJSIP_EX_NO_MEMORY  PJ_NO_MEMORY_EXCEPTION
+#define PJSIP_EX_NO_MEMORY  pj_NO_MEMORY_EXCEPTION()
 #define THIS_FILE	    "sip_endpoint.c"
 
 #define MAX_METHODS   32
 
+
+/* List of SIP endpoint exit callback. */
+typedef struct exit_cb
+{
+    PJ_DECL_LIST_MEMBER		    (struct exit_cb);
+    pjsip_endpt_exit_callback	    func;
+} exit_cb;
+
+
 /**
  * The SIP endpoint.
  */
@@ -85,6 +95,9 @@ struct pjsip_endpoint
 
     /** Additional request headers. */
     pjsip_hdr		 req_hdr;
+
+    /** List of exit callback. */
+    exit_cb		 exit_cb_list;
 };
 
 
@@ -105,6 +118,8 @@ static void endpt_on_rx_msg( pjsip_endpoint*,
 			     pj_status_t, pjsip_rx_data*);
 static pj_status_t endpt_on_tx_msg( pjsip_endpoint *endpt,
 				    pjsip_tx_data *tdata );
+static pj_status_t unload_module(pjsip_endpoint *endpt,
+				 pjsip_module *mod);
 
 /* Defined in sip_parser.c */
 void init_sip_parser(void);
@@ -114,27 +129,6 @@ void deinit_sip_parser(void);
 pj_status_t pjsip_tel_uri_subsys_init(void);
 
 
-/* Specifies whether error subsystem has been registered to pjlib. */
-static int error_subsys_initialized;
-
-/**
- * Defined in sip_errno.c
- *
- * Get error message for the specified error code. This can only get
- * PJSIP specific error message. To get all types of error message,
- * use pj_strerror() instead.
- *
- * @param status    The error code.
- * @param buffer    The buffer where to put the error message.
- * @param bufsize   Size of the buffer.
- *
- * @return	    The error message as NULL terminated string,
- *                  wrapped with pj_str_t.
- */
-PJ_DECL(pj_str_t) pjsip_strerror( pj_status_t status, char *buffer,
-				  pj_size_t bufsize);
-
-
 /*
  * This is the global handler for memory allocation failure, for pools that
  * are created by the endpoint (by default, all pools ARE allocated by 
@@ -258,10 +252,33 @@ PJ_DEF(pj_status_t) pjsip_endpt_unregister_module( pjsip_endpoint *endpt,
 	if (status != PJ_SUCCESS) goto on_return;
     }
 
+    /* Unload module */
+    status = unload_module(endpt, mod);
+
+on_return:
+    pj_rwmutex_unlock_write(endpt->mod_mutex);
+
+    if (status != PJ_SUCCESS) {
+	char errmsg[PJ_ERR_MSG_SIZE];
+
+	pj_strerror(status, errmsg, sizeof(errmsg));
+	PJ_LOG(3,(THIS_FILE, "Module \"%.*s\" can not be unregistered: %s",
+		  (int)mod->name.slen, mod->name.ptr, errmsg));
+    }
+
+    return status;
+}
+
+static pj_status_t unload_module(pjsip_endpoint *endpt,
+				 pjsip_module *mod)
+{
+    pj_status_t status;
+
     /* Try to unload the module. */
     if (mod->unload) {
 	status = (*mod->unload)();
-	if (status != PJ_SUCCESS) goto on_return;
+	if (status != PJ_SUCCESS) 
+	    return status;
     }
 
     /* Module MUST NOT set module ID to -1. */
@@ -282,17 +299,6 @@ PJ_DEF(pj_status_t) pjsip_endpt_unregister_module( pjsip_endpoint *endpt,
     PJ_LOG(4,(THIS_FILE, "Module \"%.*s\" unregistered", 
 	      (int)mod->name.slen, mod->name.ptr));
 
-on_return:
-    pj_rwmutex_unlock_write(endpt->mod_mutex);
-
-    if (status != PJ_SUCCESS) {
-	char errmsg[PJ_ERR_MSG_SIZE];
-
-	pj_strerror(status, errmsg, sizeof(errmsg));
-	PJ_LOG(3,(THIS_FILE, "Module \"%.*s\" can not be unregistered: %s",
-		  (int)mod->name.slen, mod->name.ptr, errmsg));
-    }
-
     return status;
 }
 
@@ -428,11 +434,9 @@ PJ_DEF(pj_status_t) pjsip_endpt_create(pj_pool_factory *pf,
     pj_lock_t *lock = NULL;
 
 
-    if (!error_subsys_initialized) {
-	pj_register_strerror(PJSIP_ERRNO_START, PJ_ERRNO_SPACE_SIZE,
-			     &pjsip_strerror);
-	error_subsys_initialized = 1;
-    }
+    status = pj_register_strerror(PJSIP_ERRNO_START, PJ_ERRNO_SPACE_SIZE,
+				  &pjsip_strerror);
+    pj_assert(status == PJ_SUCCESS);
 
     PJ_LOG(5, (THIS_FILE, "Creating endpoint instance..."));
 
@@ -453,6 +457,9 @@ PJ_DEF(pj_status_t) pjsip_endpt_create(pj_pool_factory *pf,
     /* Init modules list. */
     pj_list_init(&endpt->module_list);
 
+    /* Initialize exit callback list. */
+    pj_list_init(&endpt->exit_cb_list);
+
     /* Create R/W mutex for module manipulation. */
     status = pj_rwmutex_create(endpt->pool, "ept%p", &endpt->mod_mutex);
     if (status != PJ_SUCCESS)
@@ -551,6 +558,7 @@ on_error:
 	pj_mutex_destroy(endpt->mutex);
 	endpt->mutex = NULL;
     }
+    deinit_sip_parser();
     if (endpt->mod_mutex) {
 	pj_rwmutex_destroy(endpt->mod_mutex);
 	endpt->mod_mutex = NULL;
@@ -567,14 +575,25 @@ on_error:
 PJ_DEF(void) pjsip_endpt_destroy(pjsip_endpoint *endpt)
 {
     pjsip_module *mod;
+    exit_cb *ecb;
 
     PJ_LOG(5, (THIS_FILE, "Destroying endpoing instance.."));
 
-    /* Unregister modules. */
+    /* Phase 1: stop all modules */
     mod = endpt->module_list.prev;
     while (mod != &endpt->module_list) {
 	pjsip_module *prev = mod->prev;
-	pjsip_endpt_unregister_module(endpt, mod);
+	if (mod->stop) {
+	    (*mod->stop)();
+	}
+	mod = prev;
+    }
+
+    /* Phase 2: unload modules. */
+    mod = endpt->module_list.prev;
+    while (mod != &endpt->module_list) {
+	pjsip_module *prev = mod->prev;
+	unload_module(endpt, mod);
 	mod = prev;
     }
 
@@ -588,8 +607,18 @@ PJ_DEF(void) pjsip_endpt_destroy(pjsip_endpoint *endpt)
     pj_ioqueue_destroy(endpt->ioqueue);
 
     /* Destroy timer heap */
+#if PJ_TIMER_DEBUG
+    pj_timer_heap_dump(endpt->timer_heap);
+#endif
     pj_timer_heap_destroy(endpt->timer_heap);
 
+    /* Call all registered exit callbacks */
+    ecb = endpt->exit_cb_list.next;
+    while (ecb != &endpt->exit_cb_list) {
+	(*ecb->func)(endpt);
+	ecb = ecb->next;
+    }
+
     /* Delete endpoint mutex. */
     pj_mutex_destroy(endpt->mutex);
 
@@ -711,10 +740,11 @@ PJ_DEF(pj_status_t) pjsip_endpt_handle_events2(pjsip_endpoint *endpt,
     do {
 	c = pj_ioqueue_poll( endpt->ioqueue, &timeout);
 	if (c < 0) {
+	    pj_status_t err = pj_get_netos_error();
 	    pj_thread_sleep(PJ_TIME_VAL_MSEC(timeout));
 	    if (p_count)
 		*p_count = count;
-	    return pj_get_netos_error();
+	    return err;
 	} else if (c == 0) {
 	    break;
 	} else {
@@ -742,6 +772,19 @@ PJ_DEF(pj_status_t) pjsip_endpt_handle_events(pjsip_endpoint *endpt,
 /*
  * Schedule timer.
  */
+#if PJ_TIMER_DEBUG
+PJ_DEF(pj_status_t) pjsip_endpt_schedule_timer_dbg(pjsip_endpoint *endpt,
+						    pj_timer_entry *entry,
+						    const pj_time_val *delay,
+						    const char *src_file,
+						    int src_line)
+{
+    PJ_LOG(6, (THIS_FILE, "pjsip_endpt_schedule_timer(entry=%p, delay=%u.%u)",
+			 entry, delay->sec, delay->msec));
+    return pj_timer_heap_schedule_dbg(endpt->timer_heap, entry, delay,
+                                      src_file, src_line);
+}
+#else
 PJ_DEF(pj_status_t) pjsip_endpt_schedule_timer( pjsip_endpoint *endpt,
 						pj_timer_entry *entry,
 						const pj_time_val *delay )
@@ -750,6 +793,7 @@ PJ_DEF(pj_status_t) pjsip_endpt_schedule_timer( pjsip_endpoint *endpt,
 			 entry, delay->sec, delay->msec));
     return pj_timer_heap_schedule( endpt->timer_heap, entry, delay );
 }
+#endif
 
 /*
  * Cancel the previously registered timer.
@@ -769,15 +813,117 @@ PJ_DEF(pj_timer_heap_t*) pjsip_endpt_get_timer_heap(pjsip_endpoint *endpt)
     return endpt->timer_heap;
 }
 
+/* Init with default */
+PJ_DEF(void) pjsip_process_rdata_param_default(pjsip_process_rdata_param *p)
+{
+    pj_bzero(p, sizeof(*p));
+}
+
+/* Distribute rdata */
+PJ_DEF(pj_status_t) pjsip_endpt_process_rx_data( pjsip_endpoint *endpt,
+                                                 pjsip_rx_data *rdata,
+                                                 pjsip_process_rdata_param *p,
+                                                 pj_bool_t *p_handled)
+{
+    pjsip_msg *msg;
+    pjsip_process_rdata_param def_prm;
+    pjsip_module *mod;
+    pj_bool_t handled = PJ_FALSE;
+    unsigned i;
+    pj_status_t status;
+
+    PJ_ASSERT_RETURN(endpt && rdata, PJ_EINVAL);
+
+    if (p==NULL) {
+	p = &def_prm;
+	pjsip_process_rdata_param_default(p);
+    }
+
+    msg = rdata->msg_info.msg;
+
+    if (p_handled)
+	*p_handled = PJ_FALSE;
+
+    if (!p->silent) {
+	PJ_LOG(5, (THIS_FILE, "Distributing rdata to modules: %s",
+		   pjsip_rx_data_get_info(rdata)));
+	pj_log_push_indent();
+    }
+
+    LOCK_MODULE_ACCESS(endpt);
+
+    /* Find start module */
+    if (p->start_mod) {
+	mod = (pjsip_module*)
+	      pj_list_find_node(&endpt->module_list, p->start_mod);
+	if (!mod) {
+	    status = PJ_ENOTFOUND;
+	    goto on_return;
+	}
+    } else {
+	mod = endpt->module_list.next;
+    }
+
+    /* Start after the specified index */
+    for (i=0; i < p->idx_after_start && mod != &endpt->module_list; ++i) {
+	mod = mod->next;
+    }
+
+    /* Start with the specified priority */
+    while (mod != &endpt->module_list && mod->priority < (int)p->start_prio) {
+	mod = mod->next;
+    }
+
+    if (mod == &endpt->module_list) {
+	status = PJ_ENOTFOUND;
+	goto on_return;
+    }
+
+    /* Distribute */
+    if (msg->type == PJSIP_REQUEST_MSG) {
+	do {
+	    if (mod->on_rx_request)
+		handled = (*mod->on_rx_request)(rdata);
+	    if (handled)
+		break;
+	    mod = mod->next;
+	} while (mod != &endpt->module_list);
+    } else {
+	do {
+	    if (mod->on_rx_response)
+		handled = (*mod->on_rx_response)(rdata);
+	    if (handled)
+		break;
+	    mod = mod->next;
+	} while (mod != &endpt->module_list);
+    }
+
+    status = PJ_SUCCESS;
+
+on_return:
+    if (p_handled)
+	*p_handled = handled;
+
+    UNLOCK_MODULE_ACCESS(endpt);
+    if (!p->silent) {
+	pj_log_pop_indent();
+    }
+    return status;
+}
+
 /*
  * This is the callback that is called by the transport manager when it 
  * receives a message from the network.
  */
 static void endpt_on_rx_msg( pjsip_endpoint *endpt,
-				      pj_status_t status,
-				      pjsip_rx_data *rdata )
+			     pj_status_t status,
+			     pjsip_rx_data *rdata )
 {
     pjsip_msg *msg = rdata->msg_info.msg;
+    pjsip_process_rdata_param proc_prm;
+    pj_bool_t handled = PJ_FALSE;
+
+    PJ_UNUSED_ARG(msg);
 
     if (status != PJ_SUCCESS) {
 	char info[30];
@@ -822,7 +968,9 @@ static void endpt_on_rx_msg( pjsip_endpoint *endpt,
 
     PJ_LOG(5, (THIS_FILE, "Processing incoming message: %s", 
 	       pjsip_rx_data_get_info(rdata)));
+    pj_log_push_indent();
 
+#if defined(PJSIP_CHECK_VIA_SENT_BY) && PJSIP_CHECK_VIA_SENT_BY != 0
     /* For response, check that the value in Via sent-by match the transport.
      * If not matched, silently drop the response.
      * Ref: RFC3261 Section 18.1.2 Receiving Response
@@ -876,65 +1024,32 @@ static void endpt_on_rx_msg( pjsip_endpoint *endpt,
 				 pjsip_rx_data_get_info(rdata),
 				 rdata->pkt_info.src_name, 
 				 rdata->pkt_info.src_port));
+	    pj_log_pop_indent();
 	    return;
 	}
     }
+#endif
 
+    pjsip_process_rdata_param_default(&proc_prm);
+    proc_prm.silent = PJ_TRUE;
 
-    /* Distribute to modules, starting from modules with highest priority */
-    LOCK_MODULE_ACCESS(endpt);
+    pjsip_endpt_process_rx_data(endpt, rdata, &proc_prm, &handled);
 
-    if (msg->type == PJSIP_REQUEST_MSG) {
-	pjsip_module *mod;
-	pj_bool_t handled = PJ_FALSE;
-
-	mod = endpt->module_list.next;
-	while (mod != &endpt->module_list) {
-	    if (mod->on_rx_request)
-		handled = (*mod->on_rx_request)(rdata);
-	    if (handled)
-		break;
-	    mod = mod->next;
-	}
-
-	/* No module is able to handle the request. */
-	if (!handled) {
-	    PJ_TODO(ENDPT_RESPOND_UNHANDLED_REQUEST);
-	    PJ_LOG(4,(THIS_FILE, "Message %s from %s:%d was dropped/unhandled by"
-				 " any modules",
-				 pjsip_rx_data_get_info(rdata),
-				 rdata->pkt_info.src_name,
-				 rdata->pkt_info.src_port));
-	}
-
-    } else {
-	pjsip_module *mod;
-	pj_bool_t handled = PJ_FALSE;
-
-	mod = endpt->module_list.next;
-	while (mod != &endpt->module_list) {
-	    if (mod->on_rx_response)
-		handled = (*mod->on_rx_response)(rdata);
-	    if (handled)
-		break;
-	    mod = mod->next;
-	}
-
-	if (!handled) {
-	    PJ_LOG(4,(THIS_FILE, "Message %s from %s:%d was dropped/unhandled"
-				 " by any modules",
-				 pjsip_rx_data_get_info(rdata),
-				 rdata->pkt_info.src_name,
-				 rdata->pkt_info.src_port));
-	}
+    /* No module is able to handle the message */
+    if (!handled) {
+	PJ_LOG(4,(THIS_FILE, "%s from %s:%d was dropped/unhandled by"
+			     " any modules",
+			     pjsip_rx_data_get_info(rdata),
+			     rdata->pkt_info.src_name,
+			     rdata->pkt_info.src_port));
     }
 
-    UNLOCK_MODULE_ACCESS(endpt);
-
     /* Must clear mod_data before returning rdata to transport, since
      * rdata may be reused.
      */
     pj_bzero(&rdata->endpt_info, sizeof(rdata->endpt_info));
+
+    pj_log_pop_indent();
 }
 
 /*
@@ -1012,6 +1127,15 @@ PJ_DEF(pj_status_t) pjsip_endpt_set_resolver( pjsip_endpoint *endpt,
     return pjsip_resolver_set_resolver(endpt->resolver, resv);
 }
 
+/*
+ * Set DNS external resolver implementation to be used by the SIP resolver.
+ */
+PJ_DEF(pj_status_t) pjsip_endpt_set_ext_resolver(pjsip_endpoint *endpt,
+						 pjsip_ext_resolver *ext_res)
+{
+    return pjsip_resolver_set_ext_resolver(endpt->resolver, ext_res);
+}
+
 /*
  * Get the DNS resolver being used by the SIP resolver.
  */
@@ -1064,6 +1188,22 @@ PJ_DEF(pj_status_t) pjsip_endpt_acquire_transport(pjsip_endpoint *endpt,
 }
 
 
+/*
+ * Find/create transport.
+ */
+PJ_DEF(pj_status_t) pjsip_endpt_acquire_transport2(pjsip_endpoint *endpt,
+						   pjsip_transport_type_e type,
+						   const pj_sockaddr_t *remote,
+						   int addr_len,
+						   const pjsip_tpselector *sel,
+						   pjsip_tx_data *tdata,
+						   pjsip_transport **transport)
+{
+    return pjsip_tpmgr_acquire_transport2(endpt->transport_mgr, type, remote, 
+					  addr_len, sel, tdata, transport);
+}
+
+
 /*
  * Report error.
  */
@@ -1075,7 +1215,7 @@ PJ_DEF(void) pjsip_endpt_log_error(  pjsip_endpoint *endpt,
 {
 #if PJ_LOG_MAX_LEVEL > 0
     char newformat[256];
-    int len;
+    pj_size_t len;
     va_list marker;
 
     va_start(marker, format);
@@ -1145,8 +1285,12 @@ PJ_DEF(void) pjsip_endpt_dump( pjsip_endpoint *endpt, pj_bool_t detail )
     pjsip_tpmgr_dump_transports( endpt->transport_mgr );
 
     /* Timer. */
+#if PJ_TIMER_DEBUG
+    pj_timer_heap_dump(endpt->timer_heap);
+#else
     PJ_LOG(3,(THIS_FILE, " Timer heap has %u entries", 
 			pj_timer_heap_count(endpt->timer_heap)));
+#endif
 
     /* Unlock mutex. */
     pj_mutex_unlock(endpt->mutex);
@@ -1157,3 +1301,20 @@ PJ_DEF(void) pjsip_endpt_dump( pjsip_endpoint *endpt, pj_bool_t detail )
 #endif
 }
 
+
+PJ_DEF(pj_status_t) pjsip_endpt_atexit( pjsip_endpoint *endpt,
+					pjsip_endpt_exit_callback func)
+{
+    exit_cb *new_cb;
+
+    PJ_ASSERT_RETURN(endpt && func, PJ_EINVAL);
+
+    new_cb = PJ_POOL_ZALLOC_T(endpt->pool, exit_cb);
+    new_cb->func = func;
+
+    pj_mutex_lock(endpt->mutex);
+    pj_list_push_back(&endpt->exit_cb_list, new_cb);
+    pj_mutex_unlock(endpt->mutex);
+
+    return PJ_SUCCESS;
+}
