@@ -1,6 +1,7 @@
-/* $Id: sip_msg.c 1266 2007-05-11 15:14:34Z bennylp $ */
+/* $Id$ */
 /* 
- * Copyright (C) 2003-2007 Benny Prijono <benny@prijono.org>
+ * Copyright (C) 2008-2011 Teluu Inc. (http://www.teluu.com)
+ * Copyright (C) 2003-2008 Benny Prijono <benny@prijono.org>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -20,17 +21,68 @@
 #include <pjsip/sip_parser.h>
 #include <pjsip/print_util.h>
 #include <pjsip/sip_errno.h>
+#include <pj/ctype.h>
+#include <pj/guid.h>
 #include <pj/string.h>
 #include <pj/pool.h>
 #include <pj/assert.h>
+#include <pjlib-util/string.h>
+
+PJ_DEF_DATA(const pjsip_method) pjsip_invite_method =
+	{ PJSIP_INVITE_METHOD, { "INVITE",6 }};
+
+PJ_DEF_DATA(const pjsip_method) pjsip_cancel_method =
+	{ PJSIP_CANCEL_METHOD, { "CANCEL",6 }};
+
+PJ_DEF_DATA(const pjsip_method) pjsip_ack_method =
+	{ PJSIP_ACK_METHOD, { "ACK",3}};
+
+PJ_DEF_DATA(const pjsip_method) pjsip_bye_method =
+	{ PJSIP_BYE_METHOD, { "BYE",3}};
+
+PJ_DEF_DATA(const pjsip_method) pjsip_register_method =
+	{ PJSIP_REGISTER_METHOD, { "REGISTER", 8}};
+
+PJ_DEF_DATA(const pjsip_method) pjsip_options_method =
+	{ PJSIP_OPTIONS_METHOD, { "OPTIONS",7}};
+
+
+/** INVITE method constant. */
+PJ_DEF(const pjsip_method*) pjsip_get_invite_method(void)
+{
+    return &pjsip_invite_method;
+}
+
+/** CANCEL method constant. */
+PJ_DEF(const pjsip_method*) pjsip_get_cancel_method(void)
+{
+    return &pjsip_cancel_method;
+}
+
+/** ACK method constant. */
+PJ_DEF(const pjsip_method*) pjsip_get_ack_method(void)
+{
+    return &pjsip_ack_method;
+}
+
+/** BYE method constant. */
+PJ_DEF(const pjsip_method*) pjsip_get_bye_method(void)
+{
+    return &pjsip_bye_method;
+}
+
+/** REGISTER method constant.*/
+PJ_DEF(const pjsip_method*) pjsip_get_register_method(void)
+{
+    return &pjsip_register_method;
+}
+
+/** OPTIONS method constant. */
+PJ_DEF(const pjsip_method*) pjsip_get_options_method(void)
+{
+    return &pjsip_options_method;
+}
 
-const pjsip_method 
-    pjsip_invite_method	    = { PJSIP_INVITE_METHOD,	{ "INVITE",6 }	},
-    pjsip_cancel_method	    = { PJSIP_CANCEL_METHOD,	{ "CANCEL",6 }	},
-    pjsip_ack_method	    = { PJSIP_ACK_METHOD,	{ "ACK",3}	},
-    pjsip_bye_method	    = { PJSIP_BYE_METHOD,	{ "BYE",3}	},
-    pjsip_register_method   = { PJSIP_REGISTER_METHOD,	{ "REGISTER",8}	},
-    pjsip_options_method    = { PJSIP_OPTIONS_METHOD,	{ "OPTIONS",7}	};
 
 static const pj_str_t *method_names[] = 
 {
@@ -42,58 +94,61 @@ static const pj_str_t *method_names[] =
     &pjsip_options_method.name
 };
 
-const pj_str_t pjsip_hdr_names[] = 
+const pjsip_hdr_name_info_t pjsip_hdr_names[] = 
 {
-    { "Accept",		     6 },   // PJSIP_H_ACCEPT,
-    { "Accept-Encoding",    15 },   // PJSIP_H_ACCEPT_ENCODING,
-    { "Accept-Language",    15 },   // PJSIP_H_ACCEPT_LANGUAGE,
-    { "Alert-Info",	    10 },   // PJSIP_H_ALERT_INFO,
-    { "Allow",		     5 },   // PJSIP_H_ALLOW,
-    { "Authentication-Info",19 },   // PJSIP_H_AUTHENTICATION_INFO,
-    { "Authorization",	    13 },   // PJSIP_H_AUTHORIZATION,
-    { "Call-ID",	     7 },   // PJSIP_H_CALL_ID,
-    { "Call-Info",	     9 },   // PJSIP_H_CALL_INFO,
-    { "Contact",	     7 },   // PJSIP_H_CONTACT,
-    { "Content-Disposition",19 },   // PJSIP_H_CONTENT_DISPOSITION,
-    { "Content-Encoding",   16 },   // PJSIP_H_CONTENT_ENCODING,
-    { "Content-Language",   16 },   // PJSIP_H_CONTENT_LANGUAGE,
-    { "Content-Length",	    14 },   // PJSIP_H_CONTENT_LENGTH,
-    { "Content-Type",	    12 },   // PJSIP_H_CONTENT_TYPE,
-    { "CSeq",		     4 },   // PJSIP_H_CSEQ,
-    { "Date",		     4 },   // PJSIP_H_DATE,
-    { "Error-Info",	    10 },   // PJSIP_H_ERROR_INFO,
-    { "Expires",	     7 },   // PJSIP_H_EXPIRES,
-    { "From",		     4 },   // PJSIP_H_FROM,
-    { "In-Reply-To",	    11 },   // PJSIP_H_IN_REPLY_TO,
-    { "Max-Forwards",	    12 },   // PJSIP_H_MAX_FORWARDS,
-    { "MIME-Version",	    12 },   // PJSIP_H_MIME_VERSION,
-    { "Min-Expires",	    11 },   // PJSIP_H_MIN_EXPIRES,
-    { "Organization",	    12 },   // PJSIP_H_ORGANIZATION,
-    { "Priority",	     8 },   // PJSIP_H_PRIORITY,
-    { "Proxy-Authenticate", 18 },   // PJSIP_H_PROXY_AUTHENTICATE,
-    { "Proxy-Authorization",19 },   // PJSIP_H_PROXY_AUTHORIZATION,
-    { "Proxy-Require",	    13 },   // PJSIP_H_PROXY_REQUIRE,
-    { "Record-Route",	    12 },   // PJSIP_H_RECORD_ROUTE,
-    { "Reply-To",	     8 },   // PJSIP_H_REPLY_TO,
-    { "Require",	     7 },   // PJSIP_H_REQUIRE,
-    { "Retry-After",	    11 },   // PJSIP_H_RETRY_AFTER,
-    { "Route",		     5 },   // PJSIP_H_ROUTE,
-    { "Server",		     6 },   // PJSIP_H_SERVER,
-    { "Subject",	     7 },   // PJSIP_H_SUBJECT,
-    { "Supported",	     9 },   // PJSIP_H_SUPPORTED,
-    { "Timestamp",	     9 },   // PJSIP_H_TIMESTAMP,
-    { "To",		     2 },   // PJSIP_H_TO,
-    { "Unsupported",	    11 },   // PJSIP_H_UNSUPPORTED,
-    { "User-Agent",	    10 },   // PJSIP_H_USER_AGENT,
-    { "Via",		     3 },   // PJSIP_H_VIA,
-    { "Warning",	     7 },   // PJSIP_H_WARNING,
-    { "WWW-Authenticate",   16 },   // PJSIP_H_WWW_AUTHENTICATE,
+    { "Accept",		    6,  NULL },   // PJSIP_H_ACCEPT,
+    { "Accept-Encoding",    15, NULL },   // PJSIP_H_ACCEPT_ENCODING,
+    { "Accept-Language",    15, NULL },   // PJSIP_H_ACCEPT_LANGUAGE,
+    { "Alert-Info",	    10, NULL },   // PJSIP_H_ALERT_INFO,
+    { "Allow",		    5,  NULL },   // PJSIP_H_ALLOW,
+    { "Authentication-Info",19, NULL },   // PJSIP_H_AUTHENTICATION_INFO,
+    { "Authorization",	    13, NULL },   // PJSIP_H_AUTHORIZATION,
+    { "Call-ID",	    7,  "i" },    // PJSIP_H_CALL_ID,
+    { "Call-Info",	    9,  NULL },   // PJSIP_H_CALL_INFO,
+    { "Contact",	    7,  "m" },    // PJSIP_H_CONTACT,
+    { "Content-Disposition",19, NULL },   // PJSIP_H_CONTENT_DISPOSITION,
+    { "Content-Encoding",   16, "e" },    // PJSIP_H_CONTENT_ENCODING,
+    { "Content-Language",   16, NULL },   // PJSIP_H_CONTENT_LANGUAGE,
+    { "Content-Length",	    14, "l" },    // PJSIP_H_CONTENT_LENGTH,
+    { "Content-Type",	    12, "c" },    // PJSIP_H_CONTENT_TYPE,
+    { "CSeq",		     4, NULL },   // PJSIP_H_CSEQ,
+    { "Date",		     4, NULL },   // PJSIP_H_DATE,
+    { "Error-Info",	    10, NULL },   // PJSIP_H_ERROR_INFO,
+    { "Expires",	     7, NULL },   // PJSIP_H_EXPIRES,
+    { "From",		     4, "f" },    // PJSIP_H_FROM,
+    { "In-Reply-To",	    11, NULL },   // PJSIP_H_IN_REPLY_TO,
+    { "Max-Forwards",	    12, NULL },   // PJSIP_H_MAX_FORWARDS,
+    { "MIME-Version",	    12, NULL },   // PJSIP_H_MIME_VERSION,
+    { "Min-Expires",	    11, NULL },   // PJSIP_H_MIN_EXPIRES,
+    { "Organization",	    12, NULL },   // PJSIP_H_ORGANIZATION,
+    { "Priority",	     8, NULL },   // PJSIP_H_PRIORITY,
+    { "Proxy-Authenticate", 18, NULL },   // PJSIP_H_PROXY_AUTHENTICATE,
+    { "Proxy-Authorization",19, NULL },   // PJSIP_H_PROXY_AUTHORIZATION,
+    { "Proxy-Require",	    13, NULL },   // PJSIP_H_PROXY_REQUIRE,
+    { "Record-Route",	    12, NULL },   // PJSIP_H_RECORD_ROUTE,
+    { "Reply-To",	     8, NULL },   // PJSIP_H_REPLY_TO,
+    { "Require",	     7, NULL },   // PJSIP_H_REQUIRE,
+    { "Retry-After",	    11, NULL },   // PJSIP_H_RETRY_AFTER,
+    { "Route",		     5, NULL },   // PJSIP_H_ROUTE,
+    { "Server",		     6, NULL },   // PJSIP_H_SERVER,
+    { "Subject",	     7, "s" },    // PJSIP_H_SUBJECT,
+    { "Supported",	     9, "k" },    // PJSIP_H_SUPPORTED,
+    { "Timestamp",	     9, NULL },   // PJSIP_H_TIMESTAMP,
+    { "To",		     2, "t" },    // PJSIP_H_TO,
+    { "Unsupported",	    11, NULL },   // PJSIP_H_UNSUPPORTED,
+    { "User-Agent",	    10, NULL },   // PJSIP_H_USER_AGENT,
+    { "Via",		     3, "v" },    // PJSIP_H_VIA,
+    { "Warning",	     7, NULL },   // PJSIP_H_WARNING,
+    { "WWW-Authenticate",   16, NULL },   // PJSIP_H_WWW_AUTHENTICATE,
 
-    { "_Unknown-Header",    15 },   // PJSIP_H_OTHER,
+    { "_Unknown-Header",    15, NULL },   // PJSIP_H_OTHER,
 };
 
+pj_bool_t pjsip_use_compact_form = PJSIP_ENCODE_SHORT_HNAME;
+
 static pj_str_t status_phrase[710];
-static int print_media_type(char *buf, const pjsip_media_type *media);
+static int print_media_type(char *buf, unsigned len,
+			    const pjsip_media_type *media);
 
 static int init_status_phrase()
 {
@@ -140,7 +195,7 @@ static int init_status_phrase()
     pj_strset2( &status_phrase[481], "Call/Transaction Does Not Exist");
     pj_strset2( &status_phrase[482], "Loop Detected");
     pj_strset2( &status_phrase[483], "Too Many Hops");
-    pj_strset2( &status_phrase[484], "Address Incompleted");
+    pj_strset2( &status_phrase[484], "Address Incomplete");
     pj_strset2( &status_phrase[485], "Ambiguous");
     pj_strset2( &status_phrase[486], "Busy Here");
     pj_strset2( &status_phrase[487], "Request Terminated");
@@ -305,13 +360,27 @@ PJ_DEF(void*)  pjsip_msg_find_hdr_by_name( const pjsip_msg *msg,
 	hdr = msg->hdr.next;
     }
     for (; hdr!=end; hdr = hdr->next) {
-	if (hdr->type < PJSIP_H_OTHER) {
-	    if (pj_stricmp(&pjsip_hdr_names[hdr->type], name) == 0)
-		return (void*)hdr;
-	} else {
-	    if (pj_stricmp(&hdr->name, name) == 0)
-		return (void*)hdr;
-	}
+	if (pj_stricmp(&hdr->name, name) == 0)
+	    return (void*)hdr;
+    }
+    return NULL;
+}
+
+PJ_DEF(void*)  pjsip_msg_find_hdr_by_names( const pjsip_msg *msg, 
+					    const pj_str_t *name, 
+					    const pj_str_t *sname,
+					    const void *start)
+{
+    const pjsip_hdr *hdr=(const pjsip_hdr*)start, *end=&msg->hdr;
+
+    if (hdr == NULL) {
+	hdr = msg->hdr.next;
+    }
+    for (; hdr!=end; hdr = hdr->next) {
+	if (pj_stricmp(&hdr->name, name) == 0)
+	    return (void*)hdr;
+	if (pj_stricmp(&hdr->name, sname) == 0)
+	    return (void*)hdr;
     }
     return NULL;
 }
@@ -330,10 +399,15 @@ PJ_DEF(pj_ssize_t) pjsip_msg_print( const pjsip_msg *msg,
 				    char *buf, pj_size_t size)
 {
     char *p=buf, *end=buf+size;
-    int len;
+    pj_ssize_t len;
     pjsip_hdr *hdr;
     pj_str_t clen_hdr =  { "Content-Length: ", 16};
 
+    if (pjsip_use_compact_form) {
+	clen_hdr.ptr = "l: ";
+	clen_hdr.slen = 3;
+    }
+
     /* Get a wild guess on how many bytes are typically needed.
      * We'll check this later in detail, but this serves as a quick check.
      */
@@ -386,16 +460,18 @@ PJ_DEF(pj_ssize_t) pjsip_msg_print( const pjsip_msg *msg,
 
     /* Print each of the headers. */
     for (hdr=msg->hdr.next; hdr!=&msg->hdr; hdr=hdr->next) {
-	len = (*hdr->vptr->print_on)(hdr, p, end-p);
-	if (len < 1)
-	    return -1;
-	p += len;
-
-	if (p+3 >= end)
+	len = pjsip_hdr_print_on(hdr, p, end-p);
+	if (len < 0)
 	    return -1;
 
-	*p++ = '\r';
-	*p++ = '\n';
+	if (len > 0) {
+	    p += len;
+	    if (p+3 >= end)
+		return -1;
+
+	    *p++ = '\r';
+	    *p++ = '\n';
+	}
     }
 
     /* Process message body. */
@@ -410,15 +486,18 @@ PJ_DEF(pj_ssize_t) pjsip_msg_print( const pjsip_msg *msg,
 	    pj_str_t ctype_hdr = { "Content-Type: ", 14};
 	    const pjsip_media_type *media = &msg->body->content_type;
 
+	    if (pjsip_use_compact_form) {
+		ctype_hdr.ptr = "c: ";
+		ctype_hdr.slen = 3;
+	    }
+
 	    /* Add Content-Type header. */
-	    if ( (end-p) < 24 + media->type.slen + media->subtype.slen + 
-			   media->param.slen) 
-	    {
+	    if ( (end-p) < 24 + media->type.slen + media->subtype.slen) {
 		return -1;
 	    }
 	    pj_memcpy(p, ctype_hdr.ptr, ctype_hdr.slen);
 	    p += ctype_hdr.slen;
-	    p += print_media_type(p, media);
+	    p += print_media_type(p, (unsigned)(end-p), media);
 	    *p++ = '\r';
 	    *p++ = '\n';
 
@@ -456,7 +535,7 @@ PJ_DEF(pj_ssize_t) pjsip_msg_print( const pjsip_msg *msg,
 	 */
 	if (clen_pos) {
 	    char tmp[16];
-	    len = pj_utoa(len, tmp);
+	    len = pj_utoa((unsigned long)len, tmp);
 	    if (len > CLEN_SPACE) len = CLEN_SPACE;
 	    pj_memcpy(clen_pos+CLEN_SPACE-len, tmp, len);
 	}
@@ -520,6 +599,82 @@ PJ_DEF(const pj_str_t*) pjsip_get_status_text(int code)
 	&status_phrase[code] : &status_phrase[0];
 }
 
+///////////////////////////////////////////////////////////////////////////////
+/*
+ * Media type
+ */
+/*
+ * Init media type.
+ */
+PJ_DEF(void) pjsip_media_type_init( pjsip_media_type *mt,
+				    pj_str_t *type,
+				    pj_str_t *subtype)
+{
+    pj_bzero(mt, sizeof(*mt));
+    pj_list_init(&mt->param);
+    if (type)
+	mt->type = *type;
+    if (subtype)
+	mt->subtype = *subtype;
+}
+
+PJ_DEF(void) pjsip_media_type_init2( pjsip_media_type *mt,
+				     char *type,
+				     char *subtype)
+{
+    pj_str_t s_type, s_subtype;
+
+    if (type) {
+	s_type = pj_str(type);
+    } else {
+	s_type.ptr = NULL;
+	s_type.slen = 0;
+    }
+
+    if (subtype) {
+	s_subtype = pj_str(subtype);
+    } else {
+	s_subtype.ptr = NULL;
+	s_subtype.slen = 0;
+    }
+
+    pjsip_media_type_init(mt, &s_type, &s_subtype);
+}
+
+/*
+ * Compare two media types.
+ */
+PJ_DEF(int) pjsip_media_type_cmp( const pjsip_media_type *mt1,
+				  const pjsip_media_type *mt2,
+				  pj_bool_t cmp_param)
+{
+    int rc;
+
+    PJ_ASSERT_RETURN(mt1 && mt2, 1);
+
+    rc = pj_stricmp(&mt1->type, &mt2->type);
+    if (rc) return rc;
+
+    rc = pj_stricmp(&mt1->subtype, &mt2->subtype);
+    if (rc) return rc;
+
+    if (cmp_param) {
+	rc = pjsip_param_cmp(&mt1->param, &mt2->param, (cmp_param==1));
+    }
+
+    return rc;
+}
+
+PJ_DEF(void) pjsip_media_type_cp( pj_pool_t *pool,
+				  pjsip_media_type *dst,
+				  const pjsip_media_type *src)
+{
+    PJ_ASSERT_ON_FAIL(pool && dst && src, return);
+    pj_strdup(pool, &dst->type,    &src->type);
+    pj_strdup(pool, &dst->subtype, &src->subtype);
+    pjsip_param_clone(pool, &dst->param, &src->param);
+}
+
 ///////////////////////////////////////////////////////////////////////////////
 /*
  * Generic pjsip_hdr_names/hvalue header.
@@ -558,11 +713,10 @@ PJ_DEF(void) pjsip_generic_string_hdr_init2(pjsip_generic_string_hdr *hdr,
 }
 
 
-PJ_DEF(pjsip_generic_string_hdr*) 
-pjsip_generic_string_hdr_init( pj_pool_t *pool,
-			       void *mem,
-			       const pj_str_t *hnames,
-			       const pj_str_t *hvalue)
+PJ_DEF(pjsip_generic_string_hdr*) pjsip_generic_string_hdr_init(pj_pool_t *pool,
+					       void *mem,
+					       const pj_str_t *hnames,
+					       const pj_str_t *hvalue)
 {
     pjsip_generic_string_hdr *hdr = (pjsip_generic_string_hdr*) mem;
     pj_str_t dup_hname, dup_hval;
@@ -583,8 +737,7 @@ pjsip_generic_string_hdr_init( pj_pool_t *pool,
     return hdr;
 }
 
-PJ_DEF(pjsip_generic_string_hdr*) 
-pjsip_generic_string_hdr_create( pj_pool_t *pool,
+PJ_DEF(pjsip_generic_string_hdr*) pjsip_generic_string_hdr_create(pj_pool_t *pool,
 				 const pj_str_t *hnames,
 				 const pj_str_t *hvalue)
 {
@@ -596,19 +749,20 @@ static int pjsip_generic_string_hdr_print( pjsip_generic_string_hdr *hdr,
 					   char *buf, pj_size_t size)
 {
     char *p = buf;
+    const pj_str_t *hname = pjsip_use_compact_form? &hdr->sname : &hdr->name;
     
-    if ((pj_ssize_t)size < hdr->name.slen + hdr->hvalue.slen + 5)
+    if ((pj_ssize_t)size < hname->slen + hdr->hvalue.slen + 5)
 	return -1;
 
-    pj_memcpy(p, hdr->name.ptr, hdr->name.slen);
-    p += hdr->name.slen;
+    pj_memcpy(p, hname->ptr, hname->slen);
+    p += hname->slen;
     *p++ = ':';
     *p++ = ' ';
     pj_memcpy(p, hdr->hvalue.ptr, hdr->hvalue.slen);
     p += hdr->hvalue.slen;
     *p = '\0';
 
-    return p - buf;
+    return (int)(p - buf);
 }
 
 static pjsip_generic_string_hdr* pjsip_generic_string_hdr_clone( pj_pool_t *pool, 
@@ -619,7 +773,7 @@ static pjsip_generic_string_hdr* pjsip_generic_string_hdr_clone( pj_pool_t *pool
     hdr = pjsip_generic_string_hdr_create(pool, &rhs->name, &rhs->hvalue);
 
     hdr->type = rhs->type;
-    hdr->sname = hdr->name;
+    pj_strdup(pool, &hdr->sname, &rhs->sname);
     return hdr;
 }
 
@@ -678,18 +832,19 @@ static int pjsip_generic_int_hdr_print( pjsip_generic_int_hdr *hdr,
 					char *buf, pj_size_t size)
 {
     char *p = buf;
+    const pj_str_t *hname = pjsip_use_compact_form? &hdr->sname : &hdr->name;
 
-    if ((pj_ssize_t)size < hdr->name.slen + 15)
+    if ((pj_ssize_t)size < hname->slen + 15)
 	return -1;
 
-    pj_memcpy(p, hdr->name.ptr, hdr->name.slen);
-    p += hdr->name.slen;
+    pj_memcpy(p, hname->ptr, hname->slen);
+    p += hname->slen;
     *p++ = ':';
     *p++ = ' ';
 
     p += pj_utoa(hdr->ivalue, p);
 
-    return p - buf;
+    return (int)(p - buf);
 }
 
 static pjsip_generic_int_hdr* pjsip_generic_int_hdr_clone( pj_pool_t *pool, 
@@ -753,8 +908,9 @@ static int pjsip_generic_array_hdr_print( pjsip_generic_array_hdr *hdr,
 					  char *buf, pj_size_t size)
 {
     char *p = buf, *endbuf = buf+size;
+    const pj_str_t *hname = pjsip_use_compact_form? &hdr->sname : &hdr->name;
 
-    copy_advance(p, hdr->name);
+    copy_advance(p, (*hname));
     *p++ = ':';
     *p++ = ' ';
 
@@ -767,7 +923,7 @@ static int pjsip_generic_array_hdr_print( pjsip_generic_array_hdr *hdr,
 	}
     }
 
-    return p - buf;
+    return (int)(p - buf);
 }
 
 static pjsip_generic_array_hdr* pjsip_generic_array_hdr_clone( pj_pool_t *pool, 
@@ -901,12 +1057,13 @@ static int pjsip_clen_hdr_print( pjsip_clen_hdr *hdr,
 {
     char *p = buf;
     int len;
+    const pj_str_t *hname = pjsip_use_compact_form? &hdr->sname : &hdr->name;
 
-    if ((pj_ssize_t)size < hdr->name.slen + 14)
+    if ((pj_ssize_t)size < hname->slen + 14)
 	return -1;
 
-    pj_memcpy(p, hdr->name.ptr, hdr->name.slen);
-    p += hdr->name.slen;
+    pj_memcpy(p, hname->ptr, hname->slen);
+    p += hname->slen;
     *p++ = ':';
     *p++ = ' ';
 
@@ -914,7 +1071,7 @@ static int pjsip_clen_hdr_print( pjsip_clen_hdr *hdr,
     p += len;
     *p = '\0';
 
-    return p-buf;
+    return (int)(p-buf);
 }
 
 static pjsip_clen_hdr* pjsip_clen_hdr_clone( pj_pool_t *pool, const pjsip_clen_hdr *rhs)
@@ -965,6 +1122,7 @@ static int pjsip_cseq_hdr_print( pjsip_cseq_hdr *hdr, char *buf, pj_size_t size)
 {
     char *p = buf;
     int len;
+    /* CSeq doesn't have compact form */
 
     if ((pj_ssize_t)size < hdr->name.slen + hdr->method.name.slen + 15)
 	return -1;
@@ -983,7 +1141,7 @@ static int pjsip_cseq_hdr_print( pjsip_cseq_hdr *hdr, char *buf, pj_size_t size)
 
     *p = '\0';
 
-    return p-buf;
+    return (int)(p-buf);
 }
 
 static pjsip_cseq_hdr* pjsip_cseq_hdr_clone( pj_pool_t *pool, 
@@ -1042,25 +1200,28 @@ PJ_DEF(pjsip_contact_hdr*) pjsip_contact_hdr_create( pj_pool_t *pool )
 static int pjsip_contact_hdr_print( pjsip_contact_hdr *hdr, char *buf, 
 				    pj_size_t size)
 {
+    const pj_str_t *hname = pjsip_use_compact_form? &hdr->sname : &hdr->name;
+    const pjsip_parser_const_t *pc = pjsip_parser_const();
+
     if (hdr->star) {
 	char *p = buf;
-	if ((pj_ssize_t)size < hdr->name.slen + 6)
+	if ((pj_ssize_t)size < hname->slen + 6)
 	    return -1;
-	pj_memcpy(p, hdr->name.ptr, hdr->name.slen);
-	p += hdr->name.slen;
+	pj_memcpy(p, hname->ptr, hname->slen);
+	p += hname->slen;
 	*p++ = ':';
 	*p++ = ' ';
 	*p++ = '*';
-	return p - buf;
+	return (int)(p - buf);
 
     } else {
 	int printed;
 	char *startbuf = buf;
 	char *endbuf = buf + size;
 
-	copy_advance(buf, hdr->name);
-	*buf++ = ':';
-	*buf++ = ' ';
+	copy_advance(buf, (*hname));
+	copy_advance_char_check(buf, ':');
+	copy_advance_char_check(buf, ' ');
 
 	printed = pjsip_uri_print(PJSIP_URI_IN_CONTACT_HDR, hdr->uri, 
 				  buf, endbuf-buf);
@@ -1070,6 +1231,8 @@ static int pjsip_contact_hdr_print( pjsip_contact_hdr *hdr, char *buf,
 	buf += printed;
 
 	if (hdr->q1000) {
+	    unsigned frac;
+
 	    if (buf+19 >= endbuf)
 		return -1;
 
@@ -1080,9 +1243,14 @@ static int pjsip_contact_hdr_print( pjsip_contact_hdr *hdr, char *buf,
 	    pj_memcpy(buf, ";q=", 3);
 	    printed = pj_utoa(hdr->q1000/1000, buf+3);
 	    buf += printed + 3;
-	    *buf++ = '.';
-	    printed = pj_utoa(hdr->q1000 % 1000, buf);
-	    buf += printed;
+	    frac = hdr->q1000 % 1000;
+	    if (frac != 0) {
+		*buf++ = '.';
+		if ((frac % 100)==0) frac /= 100;
+		if ((frac % 10)==0) frac /= 10;
+		printed = pj_utoa(frac, buf);
+		buf += printed;
+	    }
 	}
 
 	if (hdr->expires >= 0) {
@@ -1094,14 +1262,15 @@ static int pjsip_contact_hdr_print( pjsip_contact_hdr *hdr, char *buf,
 	    buf += printed + 9;
 	}
 
-	printed = pjsip_param_print_on(&hdr->other_param, buf, endbuf-buf,
-				       &pjsip_TOKEN_SPEC, &pjsip_TOKEN_SPEC, 
-				       ';');
+	printed = (int)pjsip_param_print_on(&hdr->other_param, buf, endbuf-buf,
+					    &pc->pjsip_TOKEN_SPEC,
+					    &pc->pjsip_TOKEN_SPEC, 
+					    ';');
 	if (printed < 0)
 	    return printed;
 	buf += printed;
 
-	return buf-startbuf;
+	return (int)(buf-startbuf);
     }
 }
 
@@ -1157,6 +1326,7 @@ PJ_DEF(pjsip_ctype_hdr*) pjsip_ctype_hdr_init( pj_pool_t *pool,
 
     pj_bzero(mem, sizeof(pjsip_ctype_hdr));
     init_hdr(hdr, PJSIP_H_CONTENT_TYPE, &ctype_hdr_vptr);
+    pj_list_init(&hdr->media.param);
     return hdr;
 
 }
@@ -1167,9 +1337,12 @@ PJ_DEF(pjsip_ctype_hdr*) pjsip_ctype_hdr_create( pj_pool_t *pool )
     return pjsip_ctype_hdr_init(pool, mem);
 }
 
-static int print_media_type(char *buf, const pjsip_media_type *media)
+static int print_media_type(char *buf, unsigned len,
+			    const pjsip_media_type *media)
 {
     char *p = buf;
+    pj_ssize_t printed;
+    const pjsip_parser_const_t *pc;
 
     pj_memcpy(p, media->type.ptr, media->type.slen);
     p += media->type.slen;
@@ -1177,12 +1350,23 @@ static int print_media_type(char *buf, const pjsip_media_type *media)
     pj_memcpy(p, media->subtype.ptr, media->subtype.slen);
     p += media->subtype.slen;
 
-    if (media->param.slen) {
-	pj_memcpy(p, media->param.ptr, media->param.slen);
-	p += media->param.slen;
-    }
+    pc = pjsip_parser_const();
+    printed = pjsip_param_print_on(&media->param, p, buf+len-p,
+				   &pc->pjsip_TOKEN_SPEC,
+				   &pc->pjsip_TOKEN_SPEC, ';');
+    if (printed < 0)
+	return -1;
 
-    return p-buf;
+    p += printed;
+
+    return (int)(p-buf);
+}
+
+
+PJ_DEF(int) pjsip_media_type_print(char *buf, unsigned len,
+				   const pjsip_media_type *media)
+{
+    return print_media_type(buf, len, media);
 }
 
 static int pjsip_ctype_hdr_print( pjsip_ctype_hdr *hdr, 
@@ -1190,24 +1374,24 @@ static int pjsip_ctype_hdr_print( pjsip_ctype_hdr *hdr,
 {
     char *p = buf;
     int len;
+    const pj_str_t *hname = pjsip_use_compact_form? &hdr->sname : &hdr->name;
 
-    if ((pj_ssize_t)size < hdr->name.slen + 
-			   hdr->media.type.slen + hdr->media.subtype.slen + 
-			   hdr->media.param.slen + 8)
+    if ((pj_ssize_t)size < hname->slen + 
+			   hdr->media.type.slen + hdr->media.subtype.slen + 8)
     {
 	return -1;
     }
 
-    pj_memcpy(p, hdr->name.ptr, hdr->name.slen);
-    p += hdr->name.slen;
+    pj_memcpy(p, hname->ptr, hname->slen);
+    p += hname->slen;
     *p++ = ':';
     *p++ = ' ';
 
-    len = print_media_type(p, &hdr->media);
+    len = print_media_type(p, (unsigned)(buf+size-p), &hdr->media);
     p += len;
 
     *p = '\0';
-    return p-buf;
+    return (int)(p-buf);
 }
 
 static pjsip_ctype_hdr* pjsip_ctype_hdr_clone( pj_pool_t *pool, 
@@ -1216,7 +1400,7 @@ static pjsip_ctype_hdr* pjsip_ctype_hdr_clone( pj_pool_t *pool,
     pjsip_ctype_hdr *hdr = pjsip_ctype_hdr_create(pool);
     pj_strdup(pool, &hdr->media.type, &rhs->media.type);
     pj_strdup(pool, &hdr->media.subtype, &rhs->media.subtype);
-    pj_strdup(pool, &hdr->media.param, &rhs->media.param);
+    pjsip_param_clone(pool, &hdr->media.param, &rhs->media.param);
     return hdr;
 }
 
@@ -1307,25 +1491,33 @@ PJ_DEF(pjsip_to_hdr*) pjsip_to_hdr_create( pj_pool_t *pool )
 PJ_DEF(pjsip_from_hdr*) pjsip_fromto_hdr_set_from( pjsip_fromto_hdr *hdr )
 {
     hdr->type = PJSIP_H_FROM;
-    hdr->name = hdr->sname = pjsip_hdr_names[PJSIP_H_FROM];
+    hdr->name.ptr = pjsip_hdr_names[PJSIP_H_FROM].name;
+    hdr->name.slen = pjsip_hdr_names[PJSIP_H_FROM].name_len;
+    hdr->sname.ptr = pjsip_hdr_names[PJSIP_H_FROM].sname;
+    hdr->sname.slen = 1;
     return hdr;
 }
 
 PJ_DEF(pjsip_to_hdr*) pjsip_fromto_hdr_set_to( pjsip_fromto_hdr *hdr )
 {
     hdr->type = PJSIP_H_TO;
-    hdr->name = hdr->sname = pjsip_hdr_names[PJSIP_H_TO];
+    hdr->name.ptr = pjsip_hdr_names[PJSIP_H_TO].name;
+    hdr->name.slen = pjsip_hdr_names[PJSIP_H_TO].name_len;
+    hdr->sname.ptr = pjsip_hdr_names[PJSIP_H_TO].sname;
+    hdr->sname.slen = 1;
     return hdr;
 }
 
 static int pjsip_fromto_hdr_print( pjsip_fromto_hdr *hdr, 
 				   char *buf, pj_size_t size)
 {
-    int printed;
+    pj_ssize_t printed;
     char *startbuf = buf;
     char *endbuf = buf + size;
+    const pj_str_t *hname = pjsip_use_compact_form? &hdr->sname : &hdr->name;
+    const pjsip_parser_const_t *pc = pjsip_parser_const();
 
-    copy_advance(buf, hdr->name);
+    copy_advance(buf, (*hname));
     *buf++ = ':';
     *buf++ = ' ';
 
@@ -1336,16 +1528,17 @@ static int pjsip_fromto_hdr_print( pjsip_fromto_hdr *hdr,
 
     buf += printed;
 
-    copy_advance_pair(buf, ";tag=", 5, hdr->tag);
+    copy_advance_pair_escape(buf, ";tag=", 5, hdr->tag,
+			     pc->pjsip_TOKEN_SPEC);
 
     printed = pjsip_param_print_on(&hdr->other_param, buf, endbuf-buf, 
-				   &pjsip_TOKEN_SPEC,
-				   &pjsip_TOKEN_SPEC, ';');
+				   &pc->pjsip_TOKEN_SPEC,
+				   &pc->pjsip_TOKEN_SPEC, ';');
     if (printed < 0)
 	return -1;
     buf += printed;
 
-    return buf-startbuf;
+    return (int)(buf-startbuf);
 }
 
 static pjsip_fromto_hdr* pjsip_fromto_hdr_clone( pj_pool_t *pool, 
@@ -1481,23 +1674,48 @@ PJ_DEF(pjsip_route_hdr*) pjsip_route_hdr_create( pj_pool_t *pool )
 PJ_DEF(pjsip_rr_hdr*) pjsip_routing_hdr_set_rr( pjsip_routing_hdr *hdr )
 {
     hdr->type = PJSIP_H_RECORD_ROUTE;
-    hdr->name = hdr->sname = pjsip_hdr_names[PJSIP_H_RECORD_ROUTE];
+    hdr->name.ptr = pjsip_hdr_names[PJSIP_H_RECORD_ROUTE].name;
+    hdr->name.slen = pjsip_hdr_names[PJSIP_H_RECORD_ROUTE].name_len;
+    hdr->sname = hdr->name;
     return hdr;
 }
 
 PJ_DEF(pjsip_route_hdr*) pjsip_routing_hdr_set_route( pjsip_routing_hdr *hdr )
 {
     hdr->type = PJSIP_H_ROUTE;
-    hdr->name = hdr->sname = pjsip_hdr_names[PJSIP_H_ROUTE];
+    hdr->name.ptr = pjsip_hdr_names[PJSIP_H_ROUTE].name;
+    hdr->name.slen = pjsip_hdr_names[PJSIP_H_ROUTE].name_len;
+    hdr->sname = hdr->name;
     return hdr;
 }
 
 static int pjsip_routing_hdr_print( pjsip_routing_hdr *hdr,
 				    char *buf, pj_size_t size )
 {
-    int printed;
+    pj_ssize_t printed;
     char *startbuf = buf;
     char *endbuf = buf + size;
+    const pjsip_parser_const_t *pc = pjsip_parser_const();
+    pjsip_sip_uri *sip_uri;
+    pjsip_param *p;
+
+    /* Check the proprietary param 'hide', don't print this header 
+     * if it exists in the route URI.
+     */
+    sip_uri = (pjsip_sip_uri*) pjsip_uri_get_uri(hdr->name_addr.uri);
+    p = sip_uri->other_param.next;
+    while (p != &sip_uri->other_param) {
+	const pj_str_t st_hide = {"hide", 4};
+
+	if (pj_stricmp(&p->name, &st_hide) == 0) {
+	    /* Check if param 'hide' is specified without 'lr'. */
+	    pj_assert(sip_uri->lr_param != 0);
+	    return 0;
+	}
+	p = p->next;
+    }
+
+    /* Route and Record-Route don't compact forms */
 
     copy_advance(buf, hdr->name);
     *buf++ = ':';
@@ -1510,13 +1728,13 @@ static int pjsip_routing_hdr_print( pjsip_routing_hdr *hdr,
     buf += printed;
 
     printed = pjsip_param_print_on(&hdr->other_param, buf, endbuf-buf, 
-				   &pjsip_TOKEN_SPEC, 
-				   &pjsip_TOKEN_SPEC, ';');
+				   &pc->pjsip_TOKEN_SPEC, 
+				   &pc->pjsip_TOKEN_SPEC, ';');
     if (printed < 0)
 	return -1;
     buf += printed;
 
-    return buf-startbuf;
+    return (int)(buf-startbuf);
 }
 
 static pjsip_routing_hdr* pjsip_routing_hdr_clone( pj_pool_t *pool,
@@ -1567,6 +1785,22 @@ PJ_DEF(pjsip_require_hdr*) pjsip_require_hdr_create(pj_pool_t *pool)
 /*
  * Retry-After header.
  */
+static int pjsip_retry_after_hdr_print(pjsip_retry_after_hdr *r, 
+				       char *buf, pj_size_t size );
+static pjsip_retry_after_hdr* pjsip_retry_after_hdr_clone(pj_pool_t *pool, 
+							  const pjsip_retry_after_hdr *r);
+static pjsip_retry_after_hdr* 
+pjsip_retry_after_hdr_shallow_clone(pj_pool_t *pool, 
+				    const pjsip_retry_after_hdr *r );
+
+static pjsip_hdr_vptr retry_after_hdr_vptr = 
+{
+    (pjsip_hdr_clone_fptr) &pjsip_retry_after_hdr_clone,
+    (pjsip_hdr_clone_fptr) &pjsip_retry_after_hdr_shallow_clone,
+    (pjsip_hdr_print_fptr) &pjsip_retry_after_hdr_print,
+};
+
+
 PJ_DEF(pjsip_retry_after_hdr*) pjsip_retry_after_hdr_init( pj_pool_t *pool,
 							   void *mem,
 							   int value )
@@ -1575,8 +1809,10 @@ PJ_DEF(pjsip_retry_after_hdr*) pjsip_retry_after_hdr_init( pj_pool_t *pool,
 
     PJ_UNUSED_ARG(pool);
 
-    init_hdr(hdr, PJSIP_H_RETRY_AFTER, &generic_int_hdr_vptr);
+    init_hdr(hdr, PJSIP_H_RETRY_AFTER, &retry_after_hdr_vptr);
     hdr->ivalue = value;
+    hdr->comment.slen = 0;
+    pj_list_init(&hdr->param);
     return hdr;
 }
 
@@ -1588,6 +1824,74 @@ PJ_DEF(pjsip_retry_after_hdr*) pjsip_retry_after_hdr_create(pj_pool_t *pool,
 }
 
 
+static int pjsip_retry_after_hdr_print(pjsip_retry_after_hdr *hdr, 
+			    	       char *buf, pj_size_t size)
+{
+    char *p = buf;
+    char *endbuf = buf + size;
+    const pj_str_t *hname = &hdr->name;
+    const pjsip_parser_const_t *pc = pjsip_parser_const();
+    pj_ssize_t printed;
+    
+    if ((pj_ssize_t)size < hdr->name.slen + 2+11)
+	return -1;
+
+    pj_memcpy(p, hdr->name.ptr, hdr->name.slen);
+    p += hname->slen;
+    *p++ = ':';
+    *p++ = ' ';
+
+    p += pj_utoa(hdr->ivalue, p);
+
+    if (hdr->comment.slen) {
+	pj_bool_t enclosed;
+
+	if (endbuf-p < hdr->comment.slen + 3)
+	    return -1;
+
+	enclosed = (*hdr->comment.ptr == '(');
+	if (!enclosed)
+	    *p++ = '(';
+	pj_memcpy(p, hdr->comment.ptr, hdr->comment.slen);
+	p += hdr->comment.slen;
+	if (!enclosed)
+	    *p++ = ')';
+
+	if (!pj_list_empty(&hdr->param))
+	    *p++ = ' ';
+    }
+
+    printed = pjsip_param_print_on(&hdr->param, p, endbuf-p,
+				   &pc->pjsip_TOKEN_SPEC,
+				   &pc->pjsip_TOKEN_SPEC, 
+				   ';');
+    if (printed < 0)
+	return (int)printed;
+
+    p += printed;
+
+    return (int)(p - buf);
+}
+
+static pjsip_retry_after_hdr* pjsip_retry_after_hdr_clone(pj_pool_t *pool, 
+							  const pjsip_retry_after_hdr *rhs)
+{
+    pjsip_retry_after_hdr *hdr = pjsip_retry_after_hdr_create(pool, rhs->ivalue);
+    pj_strdup(pool, &hdr->comment, &rhs->comment);
+    pjsip_param_clone(pool, &hdr->param, &rhs->param);
+    return hdr;
+}
+
+static pjsip_retry_after_hdr* 
+pjsip_retry_after_hdr_shallow_clone(pj_pool_t *pool, 
+				    const pjsip_retry_after_hdr *rhs)
+{
+    pjsip_retry_after_hdr *hdr = PJ_POOL_ALLOC_T(pool, pjsip_retry_after_hdr);
+    pj_memcpy(hdr, rhs, sizeof(*hdr));
+    pjsip_param_shallow_clone(pool, &hdr->param, &rhs->param);
+    return hdr;
+}
+
 ///////////////////////////////////////////////////////////////////////////////
 /*
  * Supported header.
@@ -1672,30 +1976,45 @@ PJ_DEF(pjsip_via_hdr*) pjsip_via_hdr_create( pj_pool_t *pool )
 static int pjsip_via_hdr_print( pjsip_via_hdr *hdr, 
 				char *buf, pj_size_t size)
 {
-    int printed;
+    pj_ssize_t printed;
     char *startbuf = buf;
     char *endbuf = buf + size;
     pj_str_t sip_ver = { "SIP/2.0/", 8 };
+    const pj_str_t *hname = pjsip_use_compact_form? &hdr->sname : &hdr->name;
+    const pjsip_parser_const_t *pc = pjsip_parser_const();
 
-    if ((pj_ssize_t)size < hdr->name.slen + sip_ver.slen + 
+    if ((pj_ssize_t)size < hname->slen + sip_ver.slen + 
 			   hdr->transport.slen + hdr->sent_by.host.slen + 12)
     {
 	return -1;
     }
 
     /* pjsip_hdr_names */
-    copy_advance(buf, hdr->name);
+    copy_advance(buf, (*hname));
     *buf++ = ':';
     *buf++ = ' ';
 
     /* SIP/2.0/transport host:port */
     pj_memcpy(buf, sip_ver.ptr, sip_ver.slen);
     buf += sip_ver.slen;
-    pj_memcpy(buf, hdr->transport.ptr, hdr->transport.slen);
+    //pj_memcpy(buf, hdr->transport.ptr, hdr->transport.slen);
+    /* Convert transport type to UPPERCASE (some endpoints want that) */
+    {
+	int i;
+	for (i=0; i<hdr->transport.slen; ++i) {
+	    buf[i] = (char)pj_toupper(hdr->transport.ptr[i]);
+	}
+    }
     buf += hdr->transport.slen;
     *buf++ = ' ';
-    pj_memcpy(buf, hdr->sent_by.host.ptr, hdr->sent_by.host.slen);
-    buf += hdr->sent_by.host.slen;
+
+    /* Check if host contains IPv6 */
+    if (pj_memchr(hdr->sent_by.host.ptr, ':', hdr->sent_by.host.slen)) {
+	copy_advance_pair_quote_cond(buf, "", 0, hdr->sent_by.host, '[', ']');
+    } else {
+	copy_advance_check(buf, hdr->sent_by.host);
+    }
+
     if (hdr->sent_by.port != 0) {
 	*buf++ = ':';
 	printed = pj_utoa(hdr->sent_by.port, buf);
@@ -1724,18 +2043,28 @@ static int pjsip_via_hdr_print( pjsip_via_hdr *hdr,
     }
 
 
-    copy_advance_pair(buf, ";maddr=", 7, hdr->maddr_param);
+    if (hdr->maddr_param.slen) {
+	/* Detect IPv6 IP address */
+	if (pj_memchr(hdr->maddr_param.ptr, ':', hdr->maddr_param.slen)) {
+	    copy_advance_pair_quote_cond(buf, ";maddr=", 7, hdr->maddr_param,
+				         '[', ']');
+	} else {
+	    copy_advance_pair(buf, ";maddr=", 7, hdr->maddr_param);
+	}
+    }
+
     copy_advance_pair(buf, ";received=", 10, hdr->recvd_param);
-    copy_advance_pair(buf, ";branch=", 8, hdr->branch_param);
+    copy_advance_pair_escape(buf, ";branch=", 8, hdr->branch_param,
+			     pc->pjsip_TOKEN_SPEC);
     
     printed = pjsip_param_print_on(&hdr->other_param, buf, endbuf-buf, 
-				   &pjsip_TOKEN_SPEC,
-				   &pjsip_TOKEN_SPEC, ';');
+				   &pc->pjsip_TOKEN_SPEC,
+				   &pc->pjsip_TOKEN_SPEC, ';');
     if (printed < 0)
 	return -1;
     buf += printed;
     
-    return buf-startbuf;
+    return (int)(buf-startbuf);
 }
 
 static pjsip_via_hdr* pjsip_via_hdr_clone( pj_pool_t *pool, 
@@ -1785,10 +2114,9 @@ PJ_DEF(pjsip_warning_hdr*) pjsip_warning_hdr_create(  pj_pool_t *pool,
     return pjsip_generic_string_hdr_create(pool, &str_warning, &hvalue);
 }
 
-PJ_DEF(pjsip_warning_hdr*) 
-pjsip_warning_hdr_create_from_status( pj_pool_t *pool,
-				      const pj_str_t *host,
-				      pj_status_t status)
+PJ_DEF(pjsip_warning_hdr*) pjsip_warning_hdr_create_from_status(pj_pool_t *pool,
+						      const pj_str_t *host,
+						      pj_status_t status)
 {
     char errbuf[PJ_ERR_MSG_SIZE];
     pj_str_t text;
@@ -1829,12 +2157,8 @@ PJ_DEF(pj_status_t) pjsip_msg_body_copy( pj_pool_t *pool,
     PJ_ASSERT_RETURN( src_body->clone_data!=NULL, PJ_EINVAL );
 
     /* Duplicate content-type */
-    pj_strdup(pool, &dst_body->content_type.type, 
-		    &src_body->content_type.type);
-    pj_strdup(pool, &dst_body->content_type.subtype, 
-		    &src_body->content_type.subtype);
-    pj_strdup(pool, &dst_body->content_type.param,
-		    &src_body->content_type.param);
+    pjsip_media_type_cp(pool, &dst_body->content_type,
+		        &src_body->content_type);
 
     /* Duplicate data. */
     dst_body->data = (*src_body->clone_data)(pool, src_body->data, 
@@ -1880,11 +2204,11 @@ PJ_DEF(pjsip_msg_body*) pjsip_msg_body_create( pj_pool_t *pool,
 
     pj_strdup(pool, &body->content_type.type, type);
     pj_strdup(pool, &body->content_type.subtype, subtype);
-    body->content_type.param.slen = 0;
+    pj_list_init(&body->content_type.param);
 
     body->data = pj_pool_alloc(pool, text->slen);
     pj_memcpy(body->data, text->ptr, text->slen);
-    body->len = text->slen;
+    body->len = (unsigned)text->slen;
 
     body->clone_data = &pjsip_clone_text_data;
     body->print_body = &pjsip_print_text_body;
