@@ -1,6 +1,7 @@
-/* $Id: sip_ua_layer.c 1307 2007-05-28 11:49:46Z bennylp $ */
+/* $Id$ */
 /* 
- * Copyright (C) 2003-2007 Benny Prijono <benny@prijono.org>
+ * Copyright (C) 2008-2011 Teluu Inc. (http://www.teluu.com)
+ * Copyright (C) 2003-2008 Benny Prijono <benny@prijono.org>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -236,6 +237,7 @@ PJ_DEF(pj_status_t) pjsip_ua_destroy(void)
 /*
  * Create key to identify dialog set.
  */
+/*
 PJ_DEF(void) pjsip_ua_create_dlg_set_key( pj_pool_t *pool,
 					  pj_str_t *set_key,
 					  const pj_str_t *call_id,
@@ -252,6 +254,7 @@ PJ_DEF(void) pjsip_ua_create_dlg_set_key( pj_pool_t *pool,
     pj_memcpy(set_key->ptr + call_id->slen + 1, 
 	      local_tag->ptr, local_tag->slen);
 }
+*/
 
 /*
  * Acquire one dlg_set node to be put in the hash table.
@@ -299,9 +302,10 @@ PJ_DEF(pj_status_t) pjsip_ua_register_dlg( pjsip_user_agent *ua,
 	struct dlg_set *dlg_set;
 
 	dlg_set = (struct dlg_set*)
-		  pj_hash_get( mod_ua.dlg_table, dlg->local.info->tag.ptr, 
-			       dlg->local.info->tag.slen,
-			       &dlg->local.tag_hval);
+		  pj_hash_get_lower( mod_ua.dlg_table,
+                                     dlg->local.info->tag.ptr, 
+			             (unsigned)dlg->local.info->tag.slen,
+			             &dlg->local.tag_hval);
 
 	if (dlg_set) {
 	    /* This is NOT the first dialog in the dialog set. 
@@ -323,9 +327,11 @@ PJ_DEF(pj_status_t) pjsip_ua_register_dlg( pjsip_user_agent *ua,
 	    dlg->dlg_set = dlg_set;
 
 	    /* Register the dialog set in the hash table. */
-	    pj_hash_set_np(mod_ua.dlg_table, 
-			   dlg->local.info->tag.ptr, dlg->local.info->tag.slen,
-			   dlg->local.tag_hval, dlg_set->ht_entry, dlg_set);
+	    pj_hash_set_np_lower(mod_ua.dlg_table, 
+			         dlg->local.info->tag.ptr,
+                                 (unsigned)dlg->local.info->tag.slen,
+			         dlg->local.tag_hval, dlg_set->ht_entry,
+                                 dlg_set);
 	}
 
     } else {
@@ -338,9 +344,10 @@ PJ_DEF(pj_status_t) pjsip_ua_register_dlg( pjsip_user_agent *ua,
 
 	dlg->dlg_set = dlg_set;
 
-	pj_hash_set_np(mod_ua.dlg_table, 
-		       dlg->local.info->tag.ptr, dlg->local.info->tag.slen,
-		       dlg->local.tag_hval, dlg_set->ht_entry, dlg_set);
+	pj_hash_set_np_lower(mod_ua.dlg_table, 
+		             dlg->local.info->tag.ptr,
+                             (unsigned)dlg->local.info->tag.slen,
+		             dlg->local.tag_hval, dlg_set->ht_entry, dlg_set);
     }
 
     /* Unlock user agent. */
@@ -384,8 +391,9 @@ PJ_DEF(pj_status_t) pjsip_ua_unregister_dlg( pjsip_user_agent *ua,
 
     /* If dialog list is empty, remove the dialog set from the hash table. */
     if (pj_list_empty(&dlg_set->dlg_list)) {
-	pj_hash_set(NULL, mod_ua.dlg_table, dlg->local.info->tag.ptr,
-		    dlg->local.info->tag.slen, dlg->local.tag_hval, NULL);
+	pj_hash_set_lower(NULL, mod_ua.dlg_table, dlg->local.info->tag.ptr,
+		          (unsigned)dlg->local.info->tag.slen, 
+			  dlg->local.tag_hval, NULL);
 
 	/* Return dlg_set to free nodes. */
 	pj_list_push_back(&mod_ua.free_dlgset_nodes, dlg_set);
@@ -410,6 +418,24 @@ PJ_DEF(pjsip_dialog*) pjsip_tsx_get_dlg( pjsip_transaction *tsx )
 }
 
 
+/*
+ * Retrieve the current number of dialog-set currently registered
+ * in the hash table. 
+ */
+PJ_DEF(unsigned) pjsip_ua_get_dlg_set_count(void)
+{
+    unsigned count;
+
+    PJ_ASSERT_RETURN(mod_ua.endpt, 0);
+
+    pj_mutex_lock(mod_ua.mutex);
+    count = pj_hash_count(mod_ua.dlg_table);
+    pj_mutex_unlock(mod_ua.mutex);
+
+    return count;
+}
+
+
 /* 
  * Find a dialog.
  */
@@ -428,8 +454,8 @@ PJ_DEF(pjsip_dialog*) pjsip_ua_find_dialog(const pj_str_t *call_id,
 
     /* Lookup the dialog set. */
     dlg_set = (struct dlg_set*)
-    	      pj_hash_get(mod_ua.dlg_table, local_tag->ptr, local_tag->slen,
-			  NULL);
+    	      pj_hash_get_lower(mod_ua.dlg_table, local_tag->ptr,
+                                (unsigned)local_tag->slen, NULL);
     if (dlg_set == NULL) {
 	/* Not found */
 	pj_mutex_unlock(mod_ua.mutex);
@@ -441,7 +467,7 @@ PJ_DEF(pjsip_dialog*) pjsip_ua_find_dialog(const pj_str_t *call_id,
      */
     dlg = dlg_set->dlg_list.next;
     while (dlg != (pjsip_dialog*)&dlg_set->dlg_list) {	
-	if (pj_strcmp(&dlg->remote.info->tag, remote_tag) == 0)
+	if (pj_stricmp(&dlg->remote.info->tag, remote_tag) == 0)
 	    break;
 	dlg = dlg->next;
     }
@@ -511,7 +537,7 @@ static struct dlg_set *find_dlg_set_for_msg( pjsip_rx_data *rdata )
 	    role = PJSIP_ROLE_UAC;
 
 	pjsip_tsx_create_key(rdata->tp_info.pool, &key, role, 
-			     &pjsip_invite_method, rdata);
+			     pjsip_get_invite_method(), rdata);
 
 	/* Lookup the INVITE transaction */
 	tsx = pjsip_tsx_layer_find_tsx(&key, PJ_TRUE);
@@ -519,7 +545,7 @@ static struct dlg_set *find_dlg_set_for_msg( pjsip_rx_data *rdata )
 	/* We should find the dialog attached to the INVITE transaction */
 	if (tsx) {
 	    dlg = (pjsip_dialog*) tsx->mod_data[mod_ua.mod.id];
-	    pj_mutex_unlock(tsx->mutex);
+	    pj_grp_lock_release(tsx->grp_lock);
 
 	    /* Dlg may be NULL on some extreme condition
 	     * (e.g. during debugging where initially there is a dialog)
@@ -542,7 +568,8 @@ static struct dlg_set *find_dlg_set_for_msg( pjsip_rx_data *rdata )
 
 	/* Lookup the dialog set. */
 	dlg_set = (struct dlg_set*)
-		  pj_hash_get(mod_ua.dlg_table, tag->ptr, tag->slen, NULL);
+		  pj_hash_get_lower(mod_ua.dlg_table, tag->ptr, 
+				    (unsigned)tag->slen, NULL);
 	return dlg_set;
     }
 }
@@ -564,6 +591,10 @@ static pj_bool_t mod_ua_on_rx_request(pjsip_rx_data *rdata)
 	return PJ_FALSE;
     }
 
+    /* Incoming REGISTER may have tags in it */
+    if (rdata->msg_info.msg->line.req.method.id == PJSIP_REGISTER_METHOD)
+	return PJ_FALSE;
+
 retry_on_deadlock:
 
     /* Lock user agent before looking up the dialog hash table. */
@@ -599,7 +630,7 @@ retry_on_deadlock:
     dlg = dlg_set->dlg_list.next;
     while (dlg != (pjsip_dialog*)&dlg_set->dlg_list) {
 	
-	if (pj_strcmp(&dlg->remote.info->tag, from_tag) == 0)
+	if (pj_stricmp(&dlg->remote.info->tag, from_tag) == 0)
 	    break;
 
 	dlg = dlg->next;
@@ -736,10 +767,10 @@ retry_on_deadlock:
 
 	/* Get the dialog set. */
 	dlg_set = (struct dlg_set*)
-		  pj_hash_get(mod_ua.dlg_table, 
-			      rdata->msg_info.from->tag.ptr,
-			      rdata->msg_info.from->tag.slen,
-			      NULL);
+		  pj_hash_get_lower(mod_ua.dlg_table, 
+			            rdata->msg_info.from->tag.ptr,
+			            (unsigned)rdata->msg_info.from->tag.slen,
+			            NULL);
 
 	if (!dlg_set) {
 	    /* Unlock dialog hash table. */
@@ -787,7 +818,7 @@ retry_on_deadlock:
 		break;
 
 	    /* Otherwise find the one with matching To tag. */
-	    if (pj_strcmp(to_tag, &dlg->remote.info->tag) == 0)
+	    if (pj_stricmp(to_tag, &dlg->remote.info->tag) == 0)
 		break;
 
 	    dlg = dlg->next;
@@ -812,6 +843,10 @@ retry_on_deadlock:
 	    if (mod_ua.param.on_dlg_forked) {
 		dlg = (*mod_ua.param.on_dlg_forked)(dlg_set->dlg_list.next, 
 						    rdata);
+		if (dlg == NULL) {
+		    pj_mutex_unlock(mod_ua.mutex);
+		    return PJ_TRUE;
+		}
 	    } else {
 		dlg = dlg_set->dlg_list.next;
 
@@ -882,7 +917,7 @@ static void print_dialog( const char *title,
     char userinfo[128];
 
     len = pjsip_hdr_print_on(dlg->remote.info, userinfo, sizeof(userinfo));
-    if (len < 1)
+    if (len < 0)
 	pj_ansi_strcpy(userinfo, "<--uri too long-->");
     else
 	userinfo[len] = '\0';
