@@ -1,6 +1,7 @@
-/* $Id: sip_util.c 1317 2007-05-29 11:51:45Z bennylp $ */
+/* $Id$ */
 /* 
- * Copyright (C) 2003-2007 Benny Prijono <benny@prijono.org>
+ * Copyright (C) 2008-2011 Teluu Inc. (http://www.teluu.com)
+ * Copyright (C) 2003-2008 Benny Prijono <benny@prijono.org>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -49,6 +50,139 @@ static const char *event_str[] =
 static pj_str_t str_TEXT = { "text", 4},
 		str_PLAIN = { "plain", 5 };
 
+/* Add URI to target-set */
+PJ_DEF(pj_status_t) pjsip_target_set_add_uri( pjsip_target_set *tset,
+					      pj_pool_t *pool,
+					      const pjsip_uri *uri,
+					      int q1000)
+{
+    pjsip_target *t, *pos = NULL;
+
+    PJ_ASSERT_RETURN(tset && pool && uri, PJ_EINVAL);
+
+    /* Set q-value to 1 if it is not set */
+    if (q1000 <= 0)
+	q1000 = 1000;
+
+    /* Scan all the elements to see for duplicates, and at the same time
+     * get the position where the new element should be inserted to
+     * based on the q-value.
+     */
+    t = tset->head.next;
+    while (t != &tset->head) {
+	if (pjsip_uri_cmp(PJSIP_URI_IN_REQ_URI, t->uri, uri)==PJ_SUCCESS)
+	    return PJ_EEXISTS;
+	if (pos==NULL && t->q1000 < q1000)
+	    pos = t;
+	t = t->next;
+    }
+
+    /* Create new element */
+    t = PJ_POOL_ZALLOC_T(pool, pjsip_target);
+    t->uri = (pjsip_uri*)pjsip_uri_clone(pool, uri);
+    t->q1000 = q1000;
+
+    /* Insert */
+    if (pos == NULL)
+	pj_list_push_back(&tset->head, t);
+    else
+	pj_list_insert_before(pos, t);
+
+    /* Set current target if this is the first URI */
+    if (tset->current == NULL)
+	tset->current = t;
+
+    return PJ_SUCCESS;
+}
+
+/* Add URI's in the Contact header in the message to target-set */
+PJ_DEF(pj_status_t) pjsip_target_set_add_from_msg( pjsip_target_set *tset,
+						   pj_pool_t *pool,
+						   const pjsip_msg *msg)
+{
+    const pjsip_hdr *hdr;
+    unsigned added = 0;
+
+    PJ_ASSERT_RETURN(tset && pool && msg, PJ_EINVAL);
+
+    /* Scan for Contact headers and add the URI */
+    hdr = msg->hdr.next;
+    while (hdr != &msg->hdr) {
+	if (hdr->type == PJSIP_H_CONTACT) {
+	    const pjsip_contact_hdr *cn_hdr = (const pjsip_contact_hdr*)hdr;
+
+	    if (!cn_hdr->star) {
+		pj_status_t rc;
+		rc = pjsip_target_set_add_uri(tset, pool, cn_hdr->uri, 
+					      cn_hdr->q1000);
+		if (rc == PJ_SUCCESS)
+		    ++added;
+	    }
+	}
+	hdr = hdr->next;
+    }
+
+    return added ? PJ_SUCCESS : PJ_EEXISTS;
+}
+
+
+/* Get next target, if any */
+PJ_DEF(pjsip_target*) pjsip_target_set_get_next(const pjsip_target_set *tset)
+{
+    const pjsip_target *t, *next = NULL;
+
+    t = tset->head.next;
+    while (t != &tset->head) {
+	if (PJSIP_IS_STATUS_IN_CLASS(t->code, 200)) {
+	    /* No more target since one target has been successful */
+	    return NULL;
+	}
+	if (PJSIP_IS_STATUS_IN_CLASS(t->code, 600)) {
+	    /* No more target since one target returned global error */
+	    return NULL;
+	}
+	if (t->code==0 && next==NULL) {
+	    /* This would be the next target as long as we don't find
+	     * targets with 2xx or 6xx status after this.
+	     */
+	    next = t;
+	}
+	t = t->next;
+    }
+
+    return (pjsip_target*)next;
+}
+
+
+/* Set current target */
+PJ_DEF(pj_status_t) pjsip_target_set_set_current( pjsip_target_set *tset,
+						  pjsip_target *target)
+{
+    PJ_ASSERT_RETURN(tset && target, PJ_EINVAL);
+    PJ_ASSERT_RETURN(pj_list_find_node(tset, target) != NULL, PJ_ENOTFOUND);
+
+    tset->current = target;
+
+    return PJ_SUCCESS;
+}
+
+
+/* Assign status to a target */
+PJ_DEF(pj_status_t) pjsip_target_assign_status( pjsip_target *target,
+					        pj_pool_t *pool,
+					        int status_code,
+					        const pj_str_t *reason)
+{
+    PJ_ASSERT_RETURN(target && pool && status_code && reason, PJ_EINVAL);
+
+    target->code = (pjsip_status_code)status_code;
+    pj_strdup(pool, &target->reason, reason);
+
+    return PJ_SUCCESS;
+}
+
+
+
 /*
  * Initialize transmit data (msg) with the headers and optional body.
  * This will just put the headers in the message as it is. Be carefull
@@ -111,7 +245,7 @@ static void init_request_throw( pjsip_endpoint *endpt,
 
     /* Add a blank Via header in the front of the message. */
     via = pjsip_via_hdr_create(tdata->pool);
-    via->rport_param = 0;
+    via->rport_param = pjsip_cfg()->endpt.disable_rport ? -1 : 0;
     pjsip_msg_insert_first_hdr(msg, (pjsip_hdr*)via);
 
     /* Add header params as request headers */
@@ -140,7 +274,7 @@ static void init_request_throw( pjsip_endpoint *endpt,
 	body->content_type.subtype = str_PLAIN;
 	body->data = pj_pool_alloc(tdata->pool, param_text->slen );
 	pj_memcpy(body->data, param_text->ptr, param_text->slen);
-	body->len = param_text->slen;
+	body->len = (unsigned)param_text->slen;
 	body->print_body = &pjsip_print_text_body;
 	msg->body = body;
     }
@@ -173,6 +307,7 @@ PJ_DEF(pj_status_t) pjsip_endpt_create_request(  pjsip_endpoint *endpt,
     pjsip_cid_hdr *call_id;
     pj_str_t tmp;
     pj_status_t status;
+    const pj_str_t STR_CONTACT = { "Contact", 7 };
     PJ_USE_EXCEPTION;
 
     status = pjsip_endpt_create_tdata(endpt, &tdata);
@@ -214,11 +349,11 @@ PJ_DEF(pj_status_t) pjsip_endpt_create_request(  pjsip_endpoint *endpt,
 
 	/* Contact. */
 	if (param_contact) {
-	    contact = pjsip_contact_hdr_create(tdata->pool);
 	    pj_strdup_with_null(tdata->pool, &tmp, param_contact);
-	    contact->uri = pjsip_parse_uri( tdata->pool, tmp.ptr, tmp.slen,
-					    PJSIP_PARSE_URI_AS_NAMEADDR);
-	    if (contact->uri == NULL) {
+	    contact = (pjsip_contact_hdr*)
+		      pjsip_parse_hdr(tdata->pool, &STR_CONTACT, tmp.ptr, 
+				      tmp.slen, NULL);
+	    if (contact == NULL) {
 		status = PJSIP_EINVALIDHDR;
 		goto on_error;
 	    }
@@ -261,8 +396,7 @@ on_error:
     return status;
 }
 
-PJ_DEF(pj_status_t)
-pjsip_endpt_create_request_from_hdr( pjsip_endpoint *endpt,
+PJ_DEF(pj_status_t) pjsip_endpt_create_request_from_hdr( pjsip_endpoint *endpt,
 				     const pjsip_method *method,
 				     const pjsip_uri *param_target,
 				     const pjsip_from_hdr *param_from,
@@ -308,7 +442,12 @@ pjsip_endpt_create_request_from_hdr( pjsip_endpoint *endpt,
 	} else {
 	    contact = NULL;
 	}
-	call_id = (pjsip_cid_hdr*) pjsip_hdr_clone(tdata->pool, param_call_id);
+	call_id = pjsip_cid_hdr_create(tdata->pool);
+	if (param_call_id != NULL && param_call_id->id.slen)
+	    pj_strdup(tdata->pool, &call_id->id, &param_call_id->id);
+	else
+	    pj_create_unique_string(tdata->pool, &call_id->id);
+
 	cseq = pjsip_cseq_hdr_create(tdata->pool);
 	if (param_cseq >= 0)
 	    cseq->cseq = param_cseq;
@@ -346,7 +485,8 @@ PJ_DEF(pj_status_t) pjsip_endpt_create_response( pjsip_endpoint *endpt,
     pjsip_tx_data *tdata;
     pjsip_msg *msg, *req_msg;
     pjsip_hdr *hdr;
-    pjsip_via_hdr *via;
+    pjsip_to_hdr *to_hdr;
+    pjsip_via_hdr *top_via = NULL, *via;
     pjsip_rr_hdr *rr;
     pj_status_t status;
 
@@ -388,7 +528,13 @@ PJ_DEF(pj_status_t) pjsip_endpt_create_response( pjsip_endpoint *endpt,
     /* Copy all the via headers, in order. */
     via = rdata->msg_info.via;
     while (via) {
-	pjsip_msg_add_hdr( msg, (pjsip_hdr*)pjsip_hdr_clone(tdata->pool, via));
+	pjsip_via_hdr *new_via;
+
+	new_via = (pjsip_via_hdr*)pjsip_hdr_clone(tdata->pool, via);
+	if (top_via == NULL)
+	    top_via = new_via;
+
+	pjsip_msg_add_hdr( msg, (pjsip_hdr*)new_via);
 	via = via->next;
 	if (via != (void*)&req_msg->hdr)
 	    via = (pjsip_via_hdr*) 
@@ -419,8 +565,18 @@ PJ_DEF(pj_status_t) pjsip_endpt_create_response( pjsip_endpoint *endpt,
     pjsip_msg_add_hdr( msg, hdr);
 
     /* Copy To header. */
-    hdr = (pjsip_hdr*) pjsip_hdr_clone(tdata->pool, rdata->msg_info.to);
-    pjsip_msg_add_hdr( msg, hdr);
+    to_hdr = (pjsip_to_hdr*) pjsip_hdr_clone(tdata->pool, rdata->msg_info.to);
+    pjsip_msg_add_hdr( msg, (pjsip_hdr*)to_hdr);
+
+    /* Must add To tag in the response (Section 8.2.6.2), except if this is
+     * 100 (Trying) response. Same tag must be created for the same request
+     * (e.g. same tag in provisional and final response). The easiest way
+     * to do this is to derive the tag from Via branch parameter (or to
+     * use it directly).
+     */
+    if (to_hdr->tag.slen==0 && st_code > 100 && top_via) {
+	to_hdr->tag = top_via->branch_param;
+    }
 
     /* Copy CSeq header. */
     hdr = (pjsip_hdr*) pjsip_hdr_clone(tdata->pool, rdata->msg_info.cseq);
@@ -484,7 +640,7 @@ PJ_DEF(pj_status_t) pjsip_endpt_create_ack( pjsip_endpoint *endpt,
 
     /* Create new request message from the headers. */
     status = pjsip_endpt_create_request_from_hdr(endpt, 
-						 &pjsip_ack_method,
+						 pjsip_get_ack_method(),
 						 tdata->msg->line.req.uri,
 						 from_hdr, to_hdr,
 						 NULL, cid_hdr,
@@ -575,7 +731,7 @@ PJ_DEF(pj_status_t) pjsip_endpt_create_cancel( pjsip_endpoint *endpt,
 
     /* Create new request message from the headers. */
     status = pjsip_endpt_create_request_from_hdr(endpt, 
-						 &pjsip_cancel_method,
+						 pjsip_get_cancel_method(),
 						 req_tdata->msg->line.req.uri,
 						 from_hdr, to_hdr,
 						 NULL, cid_hdr,
@@ -615,6 +771,22 @@ PJ_DEF(pj_status_t) pjsip_endpt_create_cancel( pjsip_endpoint *endpt,
 	    break;
     }
 
+    /* Must also copy the saved strict route header, otherwise CANCEL will be
+     * sent with swapped Route and request URI!
+     */
+    if (req_tdata->saved_strict_route) {
+	cancel_tdata->saved_strict_route = (pjsip_route_hdr*)
+	    pjsip_hdr_clone(cancel_tdata->pool, req_tdata->saved_strict_route);
+    }
+
+    /* Copy the destination host name from the original request */
+    pj_strdup(cancel_tdata->pool, &cancel_tdata->dest_info.name,
+	      &req_tdata->dest_info.name);
+
+    /* Finally copy the destination info from the original request */
+    pj_memcpy(&cancel_tdata->dest_info, &req_tdata->dest_info,
+	      sizeof(req_tdata->dest_info));
+
     /* Done.
      * Return the transmit buffer containing the CANCEL request.
      */
@@ -628,6 +800,87 @@ on_missing_hdr:
 }
 
 
+/* Fill-up destination information from a target URI */
+PJ_DEF(pj_status_t) pjsip_get_dest_info(const pjsip_uri *target_uri,
+				 	const pjsip_uri *request_uri,
+				 	pj_pool_t *pool,
+				 	pjsip_host_info *dest_info)
+{
+    /* The target URI must be a SIP/SIPS URL so we can resolve it's address.
+     * Otherwise we're in trouble (i.e. there's no host part in tel: URL).
+     */
+    pj_bzero(dest_info, sizeof(*dest_info));
+
+    /* When request URI uses sips scheme, TLS must always be used regardless
+     * of the target scheme or transport type (see ticket #1740).
+     */
+    if (PJSIP_URI_SCHEME_IS_SIPS(target_uri) || 
+	(pjsip_cfg()->endpt.disable_tls_switch == 0 && request_uri &&
+	 PJSIP_URI_SCHEME_IS_SIPS(request_uri)))
+    {
+	pjsip_uri *uri = (pjsip_uri*) target_uri;
+	const pjsip_sip_uri *url=(const pjsip_sip_uri*)pjsip_uri_get_uri(uri);
+	unsigned flag;
+
+	if (!PJSIP_URI_SCHEME_IS_SIPS(target_uri)) {
+	    PJ_LOG(4,(THIS_FILE, "Automatic switch to TLS transport as "
+				 "request-URI uses ""sips"" scheme."));
+	}
+
+	dest_info->flag |= (PJSIP_TRANSPORT_SECURE | PJSIP_TRANSPORT_RELIABLE);
+	if (url->maddr_param.slen)
+	    pj_strdup(pool, &dest_info->addr.host, &url->maddr_param);
+	else
+	    pj_strdup(pool, &dest_info->addr.host, &url->host);
+        dest_info->addr.port = url->port;
+	dest_info->type = 
+            pjsip_transport_get_type_from_name(&url->transport_param);
+	/* Double-check that the transport parameter match.
+	 * Sample case:     sips:host;transport=tcp
+	 * See https://trac.pjsip.org/repos/ticket/1319
+	 */
+	flag = pjsip_transport_get_flag_from_type(dest_info->type);
+	if ((flag & dest_info->flag) != dest_info->flag) {
+	    pjsip_transport_type_e t;
+
+	    t = pjsip_transport_get_type_from_flag(dest_info->flag);
+	    if (t != PJSIP_TRANSPORT_UNSPECIFIED)
+		dest_info->type = t;
+	}
+
+    } else if (PJSIP_URI_SCHEME_IS_SIP(target_uri)) {
+	pjsip_uri *uri = (pjsip_uri*) target_uri;
+	const pjsip_sip_uri *url=(const pjsip_sip_uri*)pjsip_uri_get_uri(uri);
+	if (url->maddr_param.slen)
+	    pj_strdup(pool, &dest_info->addr.host, &url->maddr_param);
+	else
+	    pj_strdup(pool, &dest_info->addr.host, &url->host);
+	dest_info->addr.port = url->port;
+	dest_info->type = 
+            pjsip_transport_get_type_from_name(&url->transport_param);
+	dest_info->flag = 
+	    pjsip_transport_get_flag_from_type(dest_info->type);
+    } else {
+	/* Should have never reached here; app should have configured route
+	 * set when sending to tel: URI
+        pj_assert(!"Unsupported URI scheme!");
+	 */
+	PJ_TODO(SUPPORT_REQUEST_ADDR_RESOLUTION_FOR_TEL_URI);
+	return PJSIP_ENOROUTESET;
+    }
+
+    /* Handle IPv6 (http://trac.pjsip.org/repos/ticket/861) */
+    if (dest_info->type != PJSIP_TRANSPORT_UNSPECIFIED && 
+	pj_strchr(&dest_info->addr.host, ':'))
+    {
+	dest_info->type = (pjsip_transport_type_e)
+			  ((int)dest_info->type | PJSIP_TRANSPORT_IPV6);
+    }
+
+    return PJ_SUCCESS;
+}
+
+
 /*
  * Find which destination to be used to send the request message, based
  * on the request URI and Route headers in the message. The procedure
@@ -654,37 +907,8 @@ PJ_DEF(pj_status_t) pjsip_get_request_dest(const pjsip_tx_data *tdata,
 	target_uri = tdata->msg->line.req.uri;
     }
 
-
-    /* The target URI must be a SIP/SIPS URL so we can resolve it's address.
-     * Otherwise we're in trouble (i.e. there's no host part in tel: URL).
-     */
-    pj_bzero(dest_info, sizeof(*dest_info));
-
-    if (PJSIP_URI_SCHEME_IS_SIPS(target_uri)) {
-	pjsip_uri *uri = (pjsip_uri*) target_uri;
-	const pjsip_sip_uri *url=(const pjsip_sip_uri*)pjsip_uri_get_uri(uri);
-	dest_info->flag |= (PJSIP_TRANSPORT_SECURE | PJSIP_TRANSPORT_RELIABLE);
-	pj_strdup(tdata->pool, &dest_info->addr.host, &url->host);
-        dest_info->addr.port = url->port;
-	dest_info->type = 
-            pjsip_transport_get_type_from_name(&url->transport_param);
-
-    } else if (PJSIP_URI_SCHEME_IS_SIP(target_uri)) {
-	pjsip_uri *uri = (pjsip_uri*) target_uri;
-	const pjsip_sip_uri *url=(const pjsip_sip_uri*)pjsip_uri_get_uri(uri);
-	pj_strdup(tdata->pool, &dest_info->addr.host, &url->host);
-	dest_info->addr.port = url->port;
-	dest_info->type = 
-            pjsip_transport_get_type_from_name(&url->transport_param);
-	dest_info->flag = 
-	    pjsip_transport_get_flag_from_type(dest_info->type);
-    } else {
-        pj_assert(!"Unsupported URI scheme!");
-	PJ_TODO(SUPPORT_REQUEST_ADDR_RESOLUTION_FOR_TEL_URI);
-	return PJSIP_EINVALIDSCHEME;
-    }
-
-    return PJ_SUCCESS;
+    return pjsip_get_dest_info(target_uri, tdata->msg->line.req.uri,
+			       (pj_pool_t*)tdata->pool, dest_info);
 }
 
 
@@ -701,11 +925,25 @@ PJ_DEF(pj_status_t) pjsip_process_route_set(pjsip_tx_data *tdata,
     const pjsip_uri *new_request_uri, *target_uri;
     const pjsip_name_addr *topmost_route_uri;
     pjsip_route_hdr *first_route_hdr, *last_route_hdr;
+    pj_status_t status;
     
     PJ_ASSERT_RETURN(tdata->msg->type == PJSIP_REQUEST_MSG, 
 		     PJSIP_ENOTREQUESTMSG);
     PJ_ASSERT_RETURN(dest_info != NULL, PJ_EINVAL);
 
+    /* If the request contains strict route, check that the strict route
+     * has been restored to its original values before processing the
+     * route set. The strict route is restored to the original values
+     * with pjsip_restore_strict_route_set(). If caller did not restore
+     * the strict route before calling this function, we need to call it
+     * here, or otherwise the strict-route and Request-URI will be swapped
+     * twice!
+     */
+    if (tdata->saved_strict_route != NULL) {
+	pjsip_restore_strict_route_set(tdata);
+    }
+    PJ_ASSERT_RETURN(tdata->saved_strict_route==NULL, PJ_EBUG);
+
     /* Find the first and last "Route" headers from the message. */
     last_route_hdr = first_route_hdr = (pjsip_route_hdr*)
 	pjsip_msg_find_hdr(tdata->msg, PJSIP_H_ROUTE, NULL);
@@ -740,7 +978,7 @@ PJ_DEF(pj_status_t) pjsip_process_route_set(pjsip_tx_data *tdata,
 	    PJSIP_URI_SCHEME_IS_SIPS(topmost_route_uri))
 	{
 	    const pjsip_sip_uri *url = (const pjsip_sip_uri*)
-		pjsip_uri_get_uri((void*)topmost_route_uri);
+		pjsip_uri_get_uri((const void*)topmost_route_uri);
 	    has_lr_param = url->lr_param;
 	} else {
 	    has_lr_param = 0;
@@ -761,8 +999,9 @@ PJ_DEF(pj_status_t) pjsip_process_route_set(pjsip_tx_data *tdata,
 	    new_request_uri = (const pjsip_uri*) 
 	    		      pjsip_uri_get_uri((pjsip_uri*)topmost_route_uri);
 	    pj_list_erase(first_route_hdr);
+	    tdata->saved_strict_route = first_route_hdr;
 	    if (first_route_hdr == last_route_hdr)
-		last_route_hdr = NULL;
+		first_route_hdr = last_route_hdr = NULL;
 	}
 
 	target_uri = (pjsip_uri*)topmost_route_uri;
@@ -771,34 +1010,11 @@ PJ_DEF(pj_status_t) pjsip_process_route_set(pjsip_tx_data *tdata,
 	target_uri = new_request_uri = tdata->msg->line.req.uri;
     }
 
-    /* The target URI must be a SIP/SIPS URL so we can resolve it's address.
-     * Otherwise we're in trouble (i.e. there's no host part in tel: URL).
-     */
-    pj_bzero(dest_info, sizeof(*dest_info));
-
-    if (PJSIP_URI_SCHEME_IS_SIPS(target_uri)) {
-	pjsip_uri *uri = (pjsip_uri*) target_uri;
-	const pjsip_sip_uri *url=(const pjsip_sip_uri*)pjsip_uri_get_uri(uri);
-	dest_info->flag |= (PJSIP_TRANSPORT_SECURE | PJSIP_TRANSPORT_RELIABLE);
-	pj_strdup(tdata->pool, &dest_info->addr.host, &url->host);
-        dest_info->addr.port = url->port;
-	dest_info->type = 
-            pjsip_transport_get_type_from_name(&url->transport_param);
-
-    } else if (PJSIP_URI_SCHEME_IS_SIP(target_uri)) {
-	pjsip_uri *uri = (pjsip_uri*) target_uri;
-	const pjsip_sip_uri *url=(const pjsip_sip_uri*)pjsip_uri_get_uri(uri);
-	pj_strdup(tdata->pool, &dest_info->addr.host, &url->host);
-	dest_info->addr.port = url->port;
-	dest_info->type = 
-            pjsip_transport_get_type_from_name(&url->transport_param);
-	dest_info->flag = 
-	    pjsip_transport_get_flag_from_type(dest_info->type);
-    } else {
-        pj_assert(!"Unsupported URI scheme!");
-	PJ_TODO(SUPPORT_REQUEST_ADDR_RESOLUTION_FOR_TEL_URI);
-	return PJSIP_EINVALIDSCHEME;
-    }
+    /* Fill up the destination host/port from the URI. */
+    status = pjsip_get_dest_info(target_uri, new_request_uri, tdata->pool,
+			   	 dest_info);
+    if (status != PJ_SUCCESS)
+	return status;
 
     /* If target URI is different than request URI, replace 
      * request URI add put the original URI in the last Route header.
@@ -819,6 +1035,56 @@ PJ_DEF(pj_status_t) pjsip_process_route_set(pjsip_tx_data *tdata,
 }
 
 
+/*
+ * Swap the request URI and strict route back to the original position
+ * before #pjsip_process_route_set() function is called. This function
+ * should only used internally by PJSIP client authentication module.
+ */
+PJ_DEF(void) pjsip_restore_strict_route_set(pjsip_tx_data *tdata)
+{
+    pjsip_route_hdr *first_route_hdr, *last_route_hdr;
+
+    /* Check if we have found strict route before */
+    if (tdata->saved_strict_route == NULL) {
+	/* This request doesn't contain strict route */
+	return;
+    }
+
+    /* Find the first "Route" headers from the message. */
+    first_route_hdr = (pjsip_route_hdr*)
+		      pjsip_msg_find_hdr(tdata->msg, PJSIP_H_ROUTE, NULL);
+
+    if (first_route_hdr == NULL) {
+	/* User has modified message route? We don't expect this! */
+	pj_assert(!"Message route was modified?");
+	tdata->saved_strict_route = NULL;
+	return;
+    }
+
+    /* Find last Route header */
+    last_route_hdr = first_route_hdr;
+    while (last_route_hdr->next != (void*)&tdata->msg->hdr) {
+	pjsip_route_hdr *hdr;
+	hdr = (pjsip_route_hdr*)
+	      pjsip_msg_find_hdr(tdata->msg, PJSIP_H_ROUTE, 
+                                 last_route_hdr->next);
+	if (!hdr)
+	    break;
+	last_route_hdr = hdr;
+    }
+
+    /* Put the last Route header as request URI, delete last Route
+     * header, and insert the saved strict route as the first Route.
+     */
+    tdata->msg->line.req.uri = last_route_hdr->name_addr.uri;
+    pj_list_insert_before(first_route_hdr, tdata->saved_strict_route);
+    pj_list_erase(last_route_hdr);
+
+    /* Reset */
+    tdata->saved_strict_route = NULL;
+}
+
+
 /* Transport callback for sending stateless request. 
  * This is one of the most bizzare function in pjsip, so
  * good luck if you happen to debug this function!!
@@ -855,7 +1121,7 @@ static void stateless_send_transport_cb( void *token,
 	     * (2) Failure (i.e. sent <= 0)
 	     */
 	    cont = (sent > 0) ? PJ_FALSE :
-		   (stateless_data->cur_addr<stateless_data->addr.count-1);
+		   (tdata->dest_info.cur_addr<tdata->dest_info.addr.count-1);
 	    if (stateless_data->app_cb) {
 		(*stateless_data->app_cb)(stateless_data, sent, &cont);
 	    } else {
@@ -882,11 +1148,11 @@ static void stateless_send_transport_cb( void *token,
 	 * first invocation. 
 	 */
 	if (sent != -PJ_EPENDING) {
-	    stateless_data->cur_addr++;
+	    tdata->dest_info.cur_addr++;
 	}
 
 	/* Have next address? */
-	if (stateless_data->cur_addr >= stateless_data->addr.count) {
+	if (tdata->dest_info.cur_addr >= tdata->dest_info.addr.count) {
 	    /* This only happens when a rather buggy application has
 	     * sent 'cont' to PJ_TRUE when the initial value was PJ_FALSE.
 	     * In this case just stop the processing; we don't need to
@@ -898,16 +1164,17 @@ static void stateless_send_transport_cb( void *token,
 	}
 
 	/* Keep current server address information handy. */
-	cur_addr = &stateless_data->addr.entry[stateless_data->cur_addr].addr;
-	cur_addr_type = stateless_data->addr.entry[stateless_data->cur_addr].type;
-	cur_addr_len = stateless_data->addr.entry[stateless_data->cur_addr].addr_len;
+	cur_addr = &tdata->dest_info.addr.entry[tdata->dest_info.cur_addr].addr;
+	cur_addr_type = tdata->dest_info.addr.entry[tdata->dest_info.cur_addr].type;
+	cur_addr_len = tdata->dest_info.addr.entry[tdata->dest_info.cur_addr].addr_len;
 
 	/* Acquire transport. */
-	status = pjsip_endpt_acquire_transport( stateless_data->endpt,
+	status = pjsip_endpt_acquire_transport2(stateless_data->endpt,
 						cur_addr_type,
 						cur_addr,
 						cur_addr_len,
 						&tdata->tp_sel,
+						tdata,
 						&stateless_data->cur_transport);
 	if (status != PJ_SUCCESS) {
 	    sent = -status;
@@ -939,8 +1206,38 @@ static void stateless_send_transport_cb( void *token,
 	}
 
 	via->transport = pj_str(stateless_data->cur_transport->type_name);
-	via->sent_by = stateless_data->cur_transport->local_name;
-	via->rport_param = 0;
+        if (tdata->via_addr.host.slen > 0 &&
+            tdata->via_tp == (void *)stateless_data->cur_transport)
+        {
+            via->sent_by = tdata->via_addr;
+        } else {
+	    via->sent_by = stateless_data->cur_transport->local_name;
+        }
+	via->rport_param = pjsip_cfg()->endpt.disable_rport ? -1 : 0;
+
+	/* Add/remove "alias" param to/from Via header on connection 
+	 * oriented/less transport, if configured.
+	 */
+	if (pjsip_cfg()->endpt.req_has_via_alias &&
+	    tdata->msg->type == PJSIP_REQUEST_MSG)
+	{
+	    const pj_str_t ALIAS_STR = {"alias", 5};
+	    pjsip_param *alias_param;
+	    pj_bool_t is_datagram;
+
+	    alias_param = pjsip_param_find(&via->other_param, &ALIAS_STR);
+	    is_datagram = (stateless_data->cur_transport->flag & 
+			   PJSIP_TRANSPORT_DATAGRAM);
+	    if (!is_datagram && !alias_param) {
+		alias_param = PJ_POOL_ZALLOC_T(tdata->pool, pjsip_param);
+		alias_param->name = ALIAS_STR;
+		pj_list_push_back(&via->other_param, alias_param);
+	    } else if (is_datagram && alias_param) {
+		pj_list_erase(alias_param);
+	    }
+	}
+
+	pjsip_tx_data_invalidate_msg(tdata);
 
 	/* Send message using this transport. */
 	status = pjsip_transport_send( stateless_data->cur_transport,
@@ -974,6 +1271,7 @@ stateless_send_resolver_callback( pj_status_t status,
 				  const struct pjsip_server_addresses *addr)
 {
     pjsip_send_state *stateless_data = (pjsip_send_state*) token;
+    pjsip_tx_data *tdata = stateless_data->tdata;
 
     /* Fail on server resolution. */
     if (status != PJ_SUCCESS) {
@@ -981,16 +1279,69 @@ stateless_send_resolver_callback( pj_status_t status,
 	    pj_bool_t cont = PJ_FALSE;
 	    (*stateless_data->app_cb)(stateless_data, -status, &cont);
 	}
-	pjsip_tx_data_dec_ref(stateless_data->tdata);
+	pjsip_tx_data_dec_ref(tdata);
 	return;
     }
 
     /* Copy server addresses */
-    pj_memcpy( &stateless_data->addr, addr, sizeof(pjsip_server_addresses));
+    if (addr && addr != &tdata->dest_info.addr) {
+	pj_memcpy( &tdata->dest_info.addr, addr, 
+	           sizeof(pjsip_server_addresses));
+    }
+    pj_assert(tdata->dest_info.addr.count != 0);
+
+    /* RFC 3261 section 18.1.1:
+     * If a request is within 200 bytes of the path MTU, or if it is larger
+     * than 1300 bytes and the path MTU is unknown, the request MUST be sent
+     * using an RFC 2914 [43] congestion controlled transport protocol, such
+     * as TCP.
+     */
+    if (pjsip_cfg()->endpt.disable_tcp_switch==0 &&
+	tdata->msg->type == PJSIP_REQUEST_MSG &&
+	tdata->dest_info.addr.count > 0 && 
+	tdata->dest_info.addr.entry[0].type == PJSIP_TRANSPORT_UDP)
+    {
+	int len;
+
+	/* Encode the request */
+	status = pjsip_tx_data_encode(tdata);
+	if (status != PJ_SUCCESS) {
+	    if (stateless_data->app_cb) {
+		pj_bool_t cont = PJ_FALSE;
+		(*stateless_data->app_cb)(stateless_data, -status, &cont);
+	    }
+	    pjsip_tx_data_dec_ref(tdata);
+	    return;
+	}
+
+	/* Check if request message is larger than 1300 bytes. */
+	len = (int)(tdata->buf.cur - tdata->buf.start);
+	if (len >= PJSIP_UDP_SIZE_THRESHOLD) {
+	    int i;
+	    int count = tdata->dest_info.addr.count;
+
+	    PJ_LOG(5,(THIS_FILE, "%s exceeds UDP size threshold (%u), "
+				 "sending with TCP",
+				 pjsip_tx_data_get_info(tdata),
+				 PJSIP_UDP_SIZE_THRESHOLD));
+
+	    /* Insert "TCP version" of resolved UDP addresses at the
+	     * beginning.
+	     */
+	    if (count * 2 > PJSIP_MAX_RESOLVED_ADDRESSES)
+		count = PJSIP_MAX_RESOLVED_ADDRESSES / 2;
+	    for (i = 0; i < count; ++i) {
+		pj_memcpy(&tdata->dest_info.addr.entry[i+count],
+			  &tdata->dest_info.addr.entry[i],
+			  sizeof(tdata->dest_info.addr.entry[0]));
+		tdata->dest_info.addr.entry[i].type = PJSIP_TRANSPORT_TCP;
+	    }
+	    tdata->dest_info.addr.count = count * 2;
+	}
+    }
 
     /* Process the addresses. */
-    stateless_send_transport_cb( stateless_data, stateless_data->tdata,
-				 -PJ_EPENDING);
+    stateless_send_transport_cb( stateless_data, tdata, -PJ_EPENDING);
 }
 
 /*
@@ -1001,11 +1352,10 @@ stateless_send_resolver_callback( pj_status_t status,
  *  - establish transport (#pjsip_endpt_acquire_transport)
  *  - send the message (#pjsip_transport_send)
  */
-PJ_DEF(pj_status_t) 
-pjsip_endpt_send_request_stateless(pjsip_endpoint *endpt, 
+PJ_DEF(pj_status_t) pjsip_endpt_send_request_stateless(pjsip_endpoint *endpt, 
 				   pjsip_tx_data *tdata,
 				   void *token,
-				   pjsip_endpt_callback cb)
+				   pjsip_send_callback cb)
 {
     pjsip_host_info dest_info;
     pjsip_send_state *stateless_data;
@@ -1025,14 +1375,178 @@ pjsip_endpt_send_request_stateless(pjsip_endpoint *endpt,
     stateless_data->tdata = tdata;
     stateless_data->app_cb = cb;
 
+    /* If destination info has not been initialized (this applies for most
+     * all requests except CANCEL), resolve destination host. The processing
+     * then resumed when the resolving callback is called. For CANCEL, the
+     * destination info must have been copied from the original INVITE so
+     * proceed to sending the request directly.
+     */
+    if (tdata->dest_info.addr.count == 0) {
+	/* Copy the destination host name to TX data */
+	pj_strdup(tdata->pool, &tdata->dest_info.name, &dest_info.addr.host);
+
+	pjsip_endpt_resolve( endpt, tdata->pool, &dest_info, stateless_data,
+			     &stateless_send_resolver_callback);
+    } else {
+	PJ_LOG(5,(THIS_FILE, "%s: skipping target resolution because "
+	                     "address is already set",
+			     pjsip_tx_data_get_info(tdata)));
+	stateless_send_resolver_callback(PJ_SUCCESS, stateless_data,
+					 &tdata->dest_info.addr);
+    }
+    return PJ_SUCCESS;
+}
+
+
+/*
+ * Send raw data to a destination.
+ */
+PJ_DEF(pj_status_t) pjsip_endpt_send_raw( pjsip_endpoint *endpt,
+					  pjsip_transport_type_e tp_type,
+					  const pjsip_tpselector *sel,
+					  const void *raw_data,
+					  pj_size_t data_len,
+					  const pj_sockaddr_t *addr,
+					  int addr_len,
+					  void *token,
+					  pjsip_tp_send_callback cb)
+{
+    return pjsip_tpmgr_send_raw(pjsip_endpt_get_tpmgr(endpt), tp_type, sel,
+				NULL, raw_data, data_len, addr, addr_len,
+				token, cb);
+}
+
+
+/* Callback data for sending raw data */
+struct send_raw_data
+{
+    pjsip_endpoint	    *endpt;
+    pjsip_tx_data	    *tdata;
+    pjsip_tpselector	    *sel;
+    void		    *app_token;
+    pjsip_tp_send_callback   app_cb;
+};
+
+
+/* Resolver callback for sending raw data. */
+static void send_raw_resolver_callback( pj_status_t status,
+    					void *token,
+					const pjsip_server_addresses *addr)
+{
+    struct send_raw_data *sraw_data = (struct send_raw_data*) token;
+
+    if (status != PJ_SUCCESS) {
+	if (sraw_data->app_cb) {
+	    (*sraw_data->app_cb)(sraw_data->app_token, sraw_data->tdata,
+				 -status);
+	}
+    } else {
+	pj_size_t data_len;
+
+	pj_assert(addr->count != 0);
+
+	/* Avoid tdata destroyed by pjsip_tpmgr_send_raw(). */
+	pjsip_tx_data_add_ref(sraw_data->tdata);
+
+	data_len = sraw_data->tdata->buf.cur - sraw_data->tdata->buf.start;
+	status = pjsip_tpmgr_send_raw(pjsip_endpt_get_tpmgr(sraw_data->endpt),
+				      addr->entry[0].type,
+				      sraw_data->sel, sraw_data->tdata,
+				      sraw_data->tdata->buf.start, data_len,
+				      &addr->entry[0].addr, 
+				      addr->entry[0].addr_len, 
+				      sraw_data->app_token,
+				      sraw_data->app_cb);
+	if (status == PJ_SUCCESS) {
+	    (*sraw_data->app_cb)(sraw_data->app_token, sraw_data->tdata,
+				 data_len);
+	} else if (status != PJ_EPENDING) {
+	    (*sraw_data->app_cb)(sraw_data->app_token, sraw_data->tdata,
+				 -status);
+	}
+    }
+
+    if (sraw_data->sel) {
+	pjsip_tpselector_dec_ref(sraw_data->sel);
+    }
+    pjsip_tx_data_dec_ref(sraw_data->tdata);
+}
+
+
+/*
+ * Send raw data to the specified destination URI. 
+ */
+PJ_DEF(pj_status_t) pjsip_endpt_send_raw_to_uri(pjsip_endpoint *endpt,
+						const pj_str_t *p_dst_uri,
+						const pjsip_tpselector *sel,
+						const void *raw_data,
+						pj_size_t data_len,
+						void *token,
+						pjsip_tp_send_callback cb)
+{
+    pjsip_tx_data *tdata;
+    struct send_raw_data *sraw_data;
+    pj_str_t dst_uri;
+    pjsip_uri *uri;
+    pjsip_host_info dest_info;
+    pj_status_t status;
+
+    /* Allocate buffer */
+    status = pjsip_endpt_create_tdata(endpt, &tdata);
+    if (status != PJ_SUCCESS)
+	return status;
+
+    pjsip_tx_data_add_ref(tdata);
+
+    /* Duplicate URI since parser requires URI to be NULL terminated */
+    pj_strdup_with_null(tdata->pool, &dst_uri, p_dst_uri);
+
+    /* Parse URI */
+    uri = pjsip_parse_uri(tdata->pool, dst_uri.ptr, dst_uri.slen, 0);
+    if (uri == NULL) {
+	pjsip_tx_data_dec_ref(tdata);
+	return PJSIP_EINVALIDURI;
+    }
+
+    /* Build destination info. */
+    status = pjsip_get_dest_info(uri, NULL, tdata->pool, &dest_info);
+    if (status != PJ_SUCCESS) {
+	pjsip_tx_data_dec_ref(tdata);
+	return status;
+    }
+
+    /* Copy data (note: data_len may be zero!) */
+    tdata->buf.start = (char*) pj_pool_alloc(tdata->pool, data_len+1);
+    tdata->buf.end = tdata->buf.start + data_len + 1;
+    if (data_len)
+	pj_memcpy(tdata->buf.start, raw_data, data_len);
+    tdata->buf.cur = tdata->buf.start + data_len;
+
+    /* Init send_raw_data */
+    sraw_data = PJ_POOL_ZALLOC_T(tdata->pool, struct send_raw_data);
+    sraw_data->endpt = endpt;
+    sraw_data->tdata = tdata;
+    sraw_data->app_token = token;
+    sraw_data->app_cb = cb;
+
+    if (sel) {
+	sraw_data->sel = PJ_POOL_ALLOC_T(tdata->pool, pjsip_tpselector);
+	pj_memcpy(sraw_data->sel, sel, sizeof(pjsip_tpselector));
+	pjsip_tpselector_add_ref(sraw_data->sel);
+    }
+
+    /* Copy the destination host name to TX data */
+    pj_strdup(tdata->pool, &tdata->dest_info.name, &dest_info.addr.host);
+
     /* Resolve destination host.
      * The processing then resumed when the resolving callback is called.
      */
-    pjsip_endpt_resolve( endpt, tdata->pool, &dest_info, stateless_data,
-			 &stateless_send_resolver_callback);
+    pjsip_endpt_resolve( endpt, tdata->pool, &dest_info, sraw_data,
+			 &send_raw_resolver_callback);
     return PJ_SUCCESS;
 }
 
+
 /*
  * Determine which address (and transport) to use to send response message
  * based on the received request. This function follows the specification
@@ -1180,11 +1694,12 @@ static void send_response_resolver_cb( pj_status_t status, void *token,
     /* Only handle the first address resolved. */
 
     /* Acquire transport. */
-    status = pjsip_endpt_acquire_transport( send_state->endpt, 
+    status = pjsip_endpt_acquire_transport2(send_state->endpt, 
 					    addr->entry[0].type,
 					    &addr->entry[0].addr,
 					    addr->entry[0].addr_len,
 					    &send_state->tdata->tp_sel,
+					    send_state->tdata,
 					    &send_state->cur_transport);
     if (status != PJ_SUCCESS) {
 	if (send_state->app_cb) {
@@ -1196,7 +1711,7 @@ static void send_response_resolver_cb( pj_status_t status, void *token,
     }
 
     /* Update address in send_state. */
-    send_state->addr = *addr;
+    pj_memcpy(&send_state->tdata->dest_info.addr, addr, sizeof(*addr));
 
     /* Send response using the transoprt. */
     status = pjsip_transport_send( send_state->cur_transport, 
@@ -1224,7 +1739,7 @@ PJ_DEF(pj_status_t) pjsip_endpt_send_response( pjsip_endpoint *endpt,
 					       pjsip_response_addr *res_addr,
 					       pjsip_tx_data *tdata,
 					       void *token,
-					       pjsip_endpt_callback cb)
+					       pjsip_send_callback cb)
 {
     /* Determine which transports and addresses to send the response,
      * based on Section 18.2.2 of RFC 3261.
@@ -1260,6 +1775,10 @@ PJ_DEF(pj_status_t) pjsip_endpt_send_response( pjsip_endpoint *endpt,
 	    return status;
 	}
     } else {
+	/* Copy the destination host name to TX data */
+	pj_strdup(tdata->pool, &tdata->dest_info.name, 
+		  &res_addr->dst_host.addr.host);
+
 	pjsip_endpt_resolve(endpt, tdata->pool, &res_addr->dst_host, 
 			    send_state, &send_response_resolver_cb);
 	return PJ_SUCCESS;
@@ -1273,7 +1792,7 @@ PJ_DEF(pj_status_t) pjsip_endpt_send_response2( pjsip_endpoint *endpt,
 					        pjsip_rx_data *rdata,
 					        pjsip_tx_data *tdata,
 						void *token,
-						pjsip_endpt_callback cb)
+						pjsip_send_callback cb)
 {
     pjsip_response_addr res_addr;
     pj_status_t status;
@@ -1348,8 +1867,12 @@ PJ_DEF(pj_status_t) pjsip_endpt_respond_stateless( pjsip_endpoint *endpt,
 
     /* Send! */
     status = pjsip_endpt_send_response( endpt, &res_addr, tdata, NULL, NULL );
+    if (status != PJ_SUCCESS) {
+	pjsip_tx_data_dec_ref(tdata);
+	return status;
+    }
 
-    return status;
+    return PJ_SUCCESS;
 }
 
 
