@@ -1,6 +1,7 @@
-/* $Id: pjsua_im.c 1285 2007-05-21 13:48:35Z bennylp $ */
+/* $Id$ */
 /* 
- * Copyright (C) 2003-2007 Benny Prijono <benny@prijono.org>
+ * Copyright (C) 2008-2011 Teluu Inc. (http://www.teluu.com)
+ * Copyright (C) 2003-2008 Benny Prijono <benny@prijono.org>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -67,14 +68,14 @@ static pjsip_module mod_pjsua_im =
 /* MIME constants. */
 static const pj_str_t STR_MIME_APP	   = { "application", 11 };
 static const pj_str_t STR_MIME_ISCOMPOSING = { "im-iscomposing+xml", 18 };
-static const pj_str_t STR_MIME_TEXT	   = { "text", 4 };
-static const pj_str_t STR_MIME_PLAIN	   = { "plain", 5 };
 
 
 /* Check if content type is acceptable */
 #if 0
 static pj_bool_t acceptable_message(const pjsip_media_type *mime)
 {
+    const pj_str_t STR_MIME_TEXT	   = { "text", 4 };
+    const pj_str_t STR_MIME_PLAIN	   = { "plain", 5 };
     return (pj_stricmp(&mime->type, &STR_MIME_TEXT)==0 &&
 	    pj_stricmp(&mime->subtype, &STR_MIME_PLAIN)==0)
 	    ||
@@ -124,9 +125,21 @@ pj_bool_t pjsua_im_accept_pager(pjsip_rx_data *rdata,
 	if (p_accept_hdr)
 	    *p_accept_hdr = pjsua_im_create_accept(rdata->tp_info.pool);
 
+	return PJ_FALSE;
+    }
+#elif 0
+    pjsip_msg *msg;
+
+    msg = rdata->msg_info.msg;
+    if (msg->body == NULL) {
+	/* Create Accept header. */
+	if (p_accept_hdr)
+	    *p_accept_hdr = pjsua_im_create_accept(rdata->tp_info.pool);
+
 	return PJ_FALSE;
     }
 #else
+    /* Ticket #693: allow incoming MESSAGE without message body */
     PJ_UNUSED_ARG(rdata);
     PJ_UNUSED_ARG(p_accept_hdr);
 #endif
@@ -145,15 +158,18 @@ void pjsua_im_process_pager(int call_id, const pj_str_t *from,
     pj_str_t contact;
     pjsip_msg_body *body = rdata->msg_info.msg->body;
 
+#if 0
+    /* Ticket #693: allow incoming MESSAGE without message body */
     /* Body MUST have been checked before */
     pj_assert(body != NULL);
+#endif
 
 
     /* Build remote contact */
     contact_hdr = (pjsip_contact_hdr*)
 		  pjsip_msg_find_hdr(rdata->msg_info.msg, PJSIP_H_CONTACT,
 				     NULL);
-    if (contact_hdr) {
+    if (contact_hdr && contact_hdr->uri) {
 	contact.ptr = (char*) pj_pool_alloc(rdata->tp_info.pool, 
 				    	    PJSIP_MAX_URL_SIZE);
 	contact.slen = pjsip_uri_print(PJSIP_URI_IN_CONTACT_HDR,
@@ -163,7 +179,7 @@ void pjsua_im_process_pager(int call_id, const pj_str_t *from,
 	contact.slen = 0;
     }
 
-    if (pj_stricmp(&body->content_type.type, &STR_MIME_APP)==0 &&
+    if (body && pj_stricmp(&body->content_type.type, &STR_MIME_APP)==0 &&
 	pj_stricmp(&body->content_type.subtype, &STR_MIME_ISCOMPOSING)==0)
     {
 	/* Expecting typing indication */
@@ -183,6 +199,21 @@ void pjsua_im_process_pager(int call_id, const pj_str_t *from,
 					     is_typing);
 	}
 
+	if (pjsua_var.ua_cfg.cb.on_typing2) {
+	    pjsua_acc_id acc_id;
+
+	    if (call_id == PJSUA_INVALID_ID) {
+		acc_id = pjsua_acc_find_for_incoming(rdata);
+	    } else {
+		pjsua_call *call = &pjsua_var.calls[call_id];
+		acc_id = call->acc_id;
+	    }
+
+
+	    (*pjsua_var.ua_cfg.cb.on_typing2)(call_id, from, to, &contact,
+					      is_typing, rdata, acc_id);
+	}
+
     } else {
 	pj_str_t mime_type;
 	char buf[256];
@@ -190,20 +221,27 @@ void pjsua_im_process_pager(int call_id, const pj_str_t *from,
 	pj_str_t text_body;
 	
 	/* Save text body */
-	text_body.ptr = (char*)rdata->msg_info.msg->body->data;
-	text_body.slen = rdata->msg_info.msg->body->len;
+	if (body) {
+	    text_body.ptr = (char*)rdata->msg_info.msg->body->data;
+	    text_body.slen = rdata->msg_info.msg->body->len;
 
-	/* Get mime type */
-	m = &rdata->msg_info.msg->body->content_type;
-	mime_type.ptr = buf;
-	mime_type.slen = pj_ansi_snprintf(buf, sizeof(buf),
-				          "%.*s/%.*s",
-				          (int)m->type.slen,
-					  m->type.ptr,
-					  (int)m->subtype.slen,
-					  m->subtype.ptr);
-	if (mime_type.slen < 1)
-	    mime_type.slen = 0;
+	    /* Get mime type */
+	    m = &rdata->msg_info.msg->body->content_type;
+	    mime_type.ptr = buf;
+	    mime_type.slen = pj_ansi_snprintf(buf, sizeof(buf),
+					      "%.*s/%.*s",
+					      (int)m->type.slen,
+					      m->type.ptr,
+					      (int)m->subtype.slen,
+					      m->subtype.ptr);
+	    if (mime_type.slen < 1)
+		mime_type.slen = 0;
+
+
+	} else {
+	    text_body.ptr = mime_type.ptr = "";
+	    text_body.slen = mime_type.slen = 0;
+	}
 
 	if (pjsua_var.ua_cfg.cb.on_pager) {
 	    (*pjsua_var.ua_cfg.cb.on_pager)(call_id, from, to, &contact, 
@@ -211,8 +249,18 @@ void pjsua_im_process_pager(int call_id, const pj_str_t *from,
 	}
 
 	if (pjsua_var.ua_cfg.cb.on_pager2) {
+	    pjsua_acc_id acc_id;
+
+	    if (call_id == PJSUA_INVALID_ID) {
+		acc_id = pjsua_acc_find_for_incoming(rdata);
+	    } else {
+		pjsua_call *call = &pjsua_var.calls[call_id];
+		acc_id = call->acc_id;
+	    }
+
 	    (*pjsua_var.ua_cfg.cb.on_pager2)(call_id, from, to, &contact, 
-					     &mime_type, &text_body, rdata);
+					     &mime_type, &text_body, rdata,
+					     acc_id);
 	}
     }
 }
@@ -226,7 +274,6 @@ static pj_bool_t im_on_rx_request(pjsip_rx_data *rdata)
     pj_str_t from, to;
     pjsip_accept_hdr *accept_hdr;
     pjsip_msg *msg;
-    pj_status_t status;
 
     msg = rdata->msg_info.msg;
 
@@ -258,8 +305,8 @@ static pj_bool_t im_on_rx_request(pjsip_rx_data *rdata)
     /* Respond with 200 first, so that remote doesn't retransmit in case
      * the UI takes too long to process the message. 
      */
-    status = pjsip_endpt_respond( pjsua_var.endpt, NULL, rdata, 200, NULL,
-				  NULL, NULL, NULL);
+    pjsip_endpt_respond( pjsua_var.endpt, NULL, rdata, 200, NULL,
+			 NULL, NULL, NULL);
 
     /* For the source URI, we use Contact header if present, since
      * Contact header contains the port number information. If this is
@@ -321,6 +368,9 @@ static void im_callback(void *token, pjsip_event *e)
 		pjsua_var.acc[im_data->acc_id].cred_cnt,
 		pjsua_var.acc[im_data->acc_id].cred);
 
+	    pjsip_auth_clt_set_prefs(&auth, 
+				     &pjsua_var.acc[im_data->acc_id].cfg.auth_pref);
+
 	    status = pjsip_auth_clt_reinit_req(&auth, rdata, tsx->last_tx,
 					       &tdata);
 	    if (status == PJ_SUCCESS) {
@@ -329,6 +379,9 @@ static void im_callback(void *token, pjsip_event *e)
 		/* Must duplicate im_data */
 		im_data2 = pjsua_im_data_dup(tdata->pool, im_data);
 
+		/* Increment CSeq */
+		PJSIP_MSG_CSEQ_HDR(tdata->msg)->cseq++;
+
 		/* Re-send request */
 		status = pjsip_endpt_send_request( pjsua_var.endpt, tdata, -1,
 						   im_data2, &im_callback);
@@ -352,7 +405,7 @@ static void im_callback(void *token, pjsip_event *e)
 		      tsx->status_text.ptr));
 	}
 
-	if (pjsua_var.ua_cfg.cb.on_pager_status)
+	if (pjsua_var.ua_cfg.cb.on_pager_status) {
 	    pjsua_var.ua_cfg.cb.on_pager_status(im_data->call_id, 
 					        &im_data->to,
 						&im_data->body,
@@ -360,6 +413,26 @@ static void im_callback(void *token, pjsip_event *e)
 						(pjsip_status_code) 
 						    tsx->status_code,
 						&tsx->status_text);
+	}
+
+	if (pjsua_var.ua_cfg.cb.on_pager_status2) {
+	    pjsip_rx_data *rdata;
+
+	    if (e->body.tsx_state.type == PJSIP_EVENT_RX_MSG)
+		rdata = e->body.tsx_state.src.rdata;
+	    else
+		rdata = NULL;
+
+	    pjsua_var.ua_cfg.cb.on_pager_status2(im_data->call_id, 
+					         &im_data->to,
+					 	 &im_data->body,
+						 im_data->user_data,
+						 (pjsip_status_code) 
+						    tsx->status_code,
+						 &tsx->status_text,
+						 tsx->last_tx,
+						 rdata, im_data->acc_id);
+	}
     }
 }
 
@@ -397,6 +470,9 @@ static void typing_callback(void *token, pjsip_event *e)
 		pjsua_var.acc[im_data->acc_id].cred_cnt,
 		pjsua_var.acc[im_data->acc_id].cred);
 
+	    pjsip_auth_clt_set_prefs(&auth, 
+				     &pjsua_var.acc[im_data->acc_id].cfg.auth_pref);
+
 	    status = pjsip_auth_clt_reinit_req(&auth, rdata, tsx->last_tx,
 					       &tdata);
 	    if (status == PJ_SUCCESS) {
@@ -405,6 +481,9 @@ static void typing_callback(void *token, pjsip_event *e)
 		/* Must duplicate im_data */
 		im_data2 = pjsua_im_data_dup(tdata->pool, im_data);
 
+		/* Increment CSeq */
+		PJSIP_MSG_CSEQ_HDR(tdata->msg)->cseq++;
+
 		/* Re-send request */
 		status = pjsip_endpt_send_request( pjsua_var.endpt, tdata, -1,
 						   im_data2, &typing_callback);
@@ -432,19 +511,22 @@ PJ_DEF(pj_status_t) pjsua_im_send( pjsua_acc_id acc_id,
 {
     pjsip_tx_data *tdata;
     const pj_str_t mime_text_plain = pj_str("text/plain");
-    const pj_str_t STR_CONTACT = { "Contact", 7 };
     pjsip_media_type media_type;
     pjsua_im_data *im_data;
-    pj_str_t contact;
+    pjsua_acc *acc;
     pj_status_t status;
 
     /* To and message body must be specified. */
     PJ_ASSERT_RETURN(to && content, PJ_EINVAL);
 
+    acc = &pjsua_var.acc[acc_id];
+
     /* Create request. */
     status = pjsip_endpt_create_request(pjsua_var.endpt, 
-					&pjsip_message_method, to, 
-					&pjsua_var.acc[acc_id].cfg.id,
+					&pjsip_message_method,
+                                        (msg_data && msg_data->target_uri.slen? 
+                                         &msg_data->target_uri: to),
+					&acc->cfg.id,
 					to, NULL, NULL, -1, NULL, &tdata);
     if (status != PJ_SUCCESS) {
 	pjsua_perror(THIS_FILE, "Unable to create request", status);
@@ -454,10 +536,10 @@ PJ_DEF(pj_status_t) pjsua_im_send( pjsua_acc_id acc_id,
     /* If account is locked to specific transport, then set transport to
      * the request.
      */
-    if (pjsua_var.acc[acc_id].cfg.transport_id != PJSUA_INVALID_ID) {
+    if (acc->cfg.transport_id != PJSUA_INVALID_ID) {
 	pjsip_tpselector tp_sel;
 
-	pjsua_init_tpselector(pjsua_var.acc[acc_id].cfg.transport_id, &tp_sel);
+	pjsua_init_tpselector(acc->cfg.transport_id, &tp_sel);
 	pjsip_tx_data_set_transport(tdata, &tp_sel);
     }
 
@@ -465,17 +547,29 @@ PJ_DEF(pj_status_t) pjsua_im_send( pjsua_acc_id acc_id,
     pjsip_msg_add_hdr( tdata->msg, 
 		       (pjsip_hdr*)pjsua_im_create_accept(tdata->pool));
 
-    /* Add contact. */
-    status = pjsua_acc_create_uac_contact(tdata->pool, &contact, acc_id, to);
-    if (status != PJ_SUCCESS) {
-	pjsua_perror(THIS_FILE, "Unable to generate Contact header", status);
-	pjsip_tx_data_dec_ref(tdata);
-	return status;
+    /* Create suitable Contact header unless a Contact header has been
+     * set in the account.
+     */
+    /* Ticket #1632: According to RFC 3428:
+     * MESSAGE requests do not initiate dialogs.
+     * User Agents MUST NOT insert Contact header fields into MESSAGE requests
+     */
+    /*
+    if (acc->contact.slen) {
+	contact = acc->contact;
+    } else {
+	status = pjsua_acc_create_uac_contact(tdata->pool, &contact, acc_id, to);
+	if (status != PJ_SUCCESS) {
+	    pjsua_perror(THIS_FILE, "Unable to generate Contact header", status);
+	    pjsip_tx_data_dec_ref(tdata);
+	    return status;
+	}
     }
 
     pjsip_msg_add_hdr( tdata->msg, (pjsip_hdr*)
 	pjsip_generic_string_hdr_create(tdata->pool, 
 					&STR_CONTACT, &contact));
+    */
 
     /* Create IM data to keep message details and give it back to
      * application on the callback
@@ -510,7 +604,13 @@ PJ_DEF(pj_status_t) pjsua_im_send( pjsua_acc_id acc_id,
     pjsua_process_msg_data(tdata, msg_data);
 
     /* Add route set */
-    pjsua_set_msg_route_set(tdata, &pjsua_var.acc[acc_id].route_set);
+    pjsua_set_msg_route_set(tdata, &acc->route_set);
+
+    /* If via_addr is set, use this address for the Via header. */
+    if (acc->cfg.allow_via_rewrite && acc->via_addr.host.slen > 0) {
+        tdata->via_addr = acc->via_addr;
+        tdata->via_tp = acc->via_tp;
+    }
 
     /* Send request (statefully) */
     status = pjsip_endpt_send_request( pjsua_var.endpt, tdata, -1, 
@@ -532,15 +632,16 @@ PJ_DEF(pj_status_t) pjsua_im_typing( pjsua_acc_id acc_id,
 				     pj_bool_t is_typing,
 				     const pjsua_msg_data *msg_data)
 {
-    const pj_str_t STR_CONTACT = { "Contact", 7 };
     pjsua_im_data *im_data;
     pjsip_tx_data *tdata;
-    pj_str_t contact;
+    pjsua_acc *acc;
     pj_status_t status;
 
+    acc = &pjsua_var.acc[acc_id];
+
     /* Create request. */
     status = pjsip_endpt_create_request( pjsua_var.endpt, &pjsip_message_method,
-					 to, &pjsua_var.acc[acc_id].cfg.id,
+					 to, &acc->cfg.id,
 					 to, NULL, NULL, -1, NULL, &tdata);
     if (status != PJ_SUCCESS) {
 	pjsua_perror(THIS_FILE, "Unable to create request", status);
@@ -551,10 +652,10 @@ PJ_DEF(pj_status_t) pjsua_im_typing( pjsua_acc_id acc_id,
     /* If account is locked to specific transport, then set transport to
      * the request.
      */
-    if (pjsua_var.acc[acc_id].cfg.transport_id != PJSUA_INVALID_ID) {
+    if (acc->cfg.transport_id != PJSUA_INVALID_ID) {
 	pjsip_tpselector tp_sel;
 
-	pjsua_init_tpselector(pjsua_var.acc[acc_id].cfg.transport_id, &tp_sel);
+	pjsua_init_tpselector(acc->cfg.transport_id, &tp_sel);
 	pjsip_tx_data_set_transport(tdata, &tp_sel);
     }
 
@@ -563,18 +664,29 @@ PJ_DEF(pj_status_t) pjsua_im_typing( pjsua_acc_id acc_id,
 		       (pjsip_hdr*)pjsua_im_create_accept(tdata->pool));
 
 
-    /* Add contact. */
-    status = pjsua_acc_create_uac_contact(tdata->pool, &contact, acc_id, to);
-    if (status != PJ_SUCCESS) {
-	pjsua_perror(THIS_FILE, "Unable to generate Contact header", status);
-	pjsip_tx_data_dec_ref(tdata);
-	return status;
+    /* Create suitable Contact header unless a Contact header has been
+     * set in the account.
+     */
+    /* Ticket #1632: According to RFC 3428:
+     * MESSAGE requests do not initiate dialogs.
+     * User Agents MUST NOT insert Contact header fields into MESSAGE requests
+     */
+    /*
+    if (acc->contact.slen) {
+	contact = acc->contact;
+    } else {
+	status = pjsua_acc_create_uac_contact(tdata->pool, &contact, acc_id, to);
+	if (status != PJ_SUCCESS) {
+	    pjsua_perror(THIS_FILE, "Unable to generate Contact header", status);
+	    pjsip_tx_data_dec_ref(tdata);
+	    return status;
+	}
     }
 
     pjsip_msg_add_hdr( tdata->msg, (pjsip_hdr*)
 	pjsip_generic_string_hdr_create(tdata->pool, 
 					&STR_CONTACT, &contact));
-
+    */
 
     /* Create "application/im-iscomposing+xml" msg body. */
     tdata->msg->body = pjsip_iscomposing_create_body( tdata->pool, is_typing,
@@ -584,7 +696,13 @@ PJ_DEF(pj_status_t) pjsua_im_typing( pjsua_acc_id acc_id,
     pjsua_process_msg_data(tdata, msg_data);
 
     /* Add route set */
-    pjsua_set_msg_route_set(tdata, &pjsua_var.acc[acc_id].route_set);
+    pjsua_set_msg_route_set(tdata, &acc->route_set);
+
+    /* If via_addr is set, use this address for the Via header. */
+    if (acc->cfg.allow_via_rewrite && acc->via_addr.host.slen > 0) {
+        tdata->via_addr = acc->via_addr;
+        tdata->via_tp = acc->via_tp;
+    }
 
     /* Create data to reauthenticate */
     im_data = PJ_POOL_ZALLOC_T(tdata->pool, pjsua_im_data);
